//	Mona1ASIC96 (Converted by RBF2VxD)


// ----------------------------------------------------------------------------
//
//   Copyright Echo Digital Audio Corporation (c) 1998 - 2004
//   All rights reserved
//   www.echoaudio.com
//   
//   Echo Digital Audio does not disclose the source code from which these 
//   firmware images are derived. Permission is hereby granted for the 
//   distribution of these firmware images as part of the Linux kernel or 
//   other GPL project in text or binary 
//   form as required.
//   
//   This file is part of Echo Digital Audio's generic driver library.
//   
//   Echo Digital Audio's generic driver library is free software; 
//   you can redistribute it and/or modify it under the terms of 
//   the GNU General Public License as published by the Free Software Foundation.
//   
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//   
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
//   MA  02111-1307, USA.
//
// *******************************************************


//	The array is 31146 bytes.

u8 Mona1ASIC96[] =
{
	0xff,0x04,0x3c,0x2b,0xf9,0xda,0xff,0xd7,
	0xfd,0xf5,0xb7,0x75,0x7f,0xfd,0xf5,0xd7,
	0xdf,0xde,0xfd,0xed,0xbd,0xf5,0x5b,0xef,
	0xd7,0xdf,0xf6,0xfd,0xf5,0xd7,0x5f,0x7f,
	0xfd,0x6d,0xdd,0xdf,0xff,0x7f,0xf5,0xff,
	0xdd,0xf7,0xdf,0xff,0xff,0xfd,0xf7,0xdf,
	0x7f,0xff,0xff,0xf7,0xff,0xfd,0xf7,0x7f,
	0xff,0x7d,0xff,0xdf,0xf5,0xdf,0x7f,0xff,
	0xfd,0xf7,0xff,0x7d,0xff,0xdd,0xfd,0x93,
	0xff,0xff,0xff,0xff,0xff,0xbf,0xff,0xff,
	0xff,0xff,0xff,0xff,0xfc,0xff,0xdf,0x3f,
	0xff,0xff,0xef,0xff,0xff,0xfc,0xff,0xff,
	0xff,0xff,0xff,0xff,0xcf,0xff,0xff,0xff,
	0x7f,0xfd,0xff,0xfd,0xf7,0xdf,0x7f,0xff,
	0xfd,0xf7,0xdf,0x7f,0xff,0xfd,0xf7,0xdf,
	0x7f,0xff,0xf7,0xdf,0x7f,0xff,0xfd,0xf6,
	0xdf,0x7f,0xff,0xfd,0xf7,0xdf,0x7f,0xff,
	0xfd,0xfe,0xe8,0xff,0xff,0xf9,0xf7,0xdf,
	0x7d,0xfe,0xfd,0xf7,0xdf,0x7f,0xf7,0xf9,
	0x77,0x9f,0x7d,0xde,0x67,0xdf,0x7f,0xf7,
	0xf1,0xf7,0xdf,0x7f,0xff,0xfd,0x77,0x9f,
	0x7f,0xff,0xf1,0x27,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xdf,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xbf,0x3f,0xe8,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xc7,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xef,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xcf,
	0xfa,0xff,0xff,0xbf,0xff,0xfe,0xff,0xaf,
	0xbf,0xff,0xfa,0xfb,0xff,0xbf,0xfe,0xff,
	0xbf,0xff,0xff,0xfa,0xeb,0xff,0xbf,0xfe,
	0xfa,0xfb,0xef,0xbf,0xfd,0xff,0xfb,0xff,
	0xff,0xd6,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xfd,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xdf,0x7f,0xff,0xff,
	0xf7,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xa3,0xee,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xfe,0xff,0xff,
	0xbf,0xff,0xff,0xff,0x7f,0xff,0xff,0xff,
	0xff,0x7f,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xf7,0xff,0xbf,0xfc,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0x7f,0xff,0xff,0xff,
	0xff,0xff,0xfe,0xff,0xf7,0xff,0xff,0xff,
	0xff,0xef,0xff,0xff,0xff,0xea,0xff,0xff,
	0xff,0x3d,0xff,0xdf,0xff,0xff,0xfd,0xff,
	0xdf,0xff,0xff,0xff,0xff,0xff,0xfa,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xef,0xff,0xff,0xff,0xdf,0x17,0xff,
	0xff,0xfe,0xfb,0xed,0xbf,0xff,0xfe,0xfb,
	0xef,0xbf,0xff,0xfe,0xf3,0xef,0xbb,0xbf,
	0xbb,0xce,0xbf,0xff,0xfe,0xf3,0xef,0xbf,
	0xff,0xfe,0xfb,0xaf,0xbf,0xff,0xfe,0xbf,
	0xf8,0xff,0xf7,0xdf,0x7f,0xff,0xfd,0xf7,
	0xdf,0x7f,0xdf,0xfd,0xf7,0xdd,0x7f,0xdf,
	0x7d,0x5f,0x7f,0xff,0xfd,0xf7,0xdf,0x7f,
	0xff,0x7d,0xf7,0xdf,0x73,0xff,0xfd,0xf7,
	0xff,0xc7,0xff,0xe7,0x9f,0x6f,0x7f,0xf9,
	0xe7,0x9b,0x7f,0xee,0xf9,0xe5,0x9b,0x7f,
	0xfe,0xd9,0x97,0x5f,0xfe,0xf9,0xe7,0x9f,
	0x7f,0xee,0xf9,0x67,0x9f,0x5f,0xfe,0xf9,
	0xe7,0xff,0x3f,0xfe,0xff,0xfd,0xf7,0xdf,
	0xbf,0xff,0xfd,0xf7,0xdf,0x7f,0xfe,0xfe,
	0xf7,0xaf,0xbf,0x3f,0xdb,0xef,0x7f,0xfe,
	0xfe,0xf7,0xdf,0x7f,0xff,0x7d,0xf7,0xef,
	0x7f,0xff,0xfd,0x7e,0xf6,0xd9,0xf9,0xe7,
	0x9f,0x7f,0xfe,0xf9,0xe7,0x9f,0x7f,0xfe,
	0xf9,0x67,0x9f,0x7b,0xee,0xe3,0x9f,0x7f,
	0xfe,0xf9,0xe7,0x9f,0x7f,0xfe,0xf9,0xe7,
	0x9f,0x7f,0xfe,0xf9,0xb7,0xaf,0xff,0xbf,
	0xfe,0xfe,0xfb,0xbf,0xbf,0xff,0xfe,0xfb,
	0xef,0xff,0xfe,0xfe,0xef,0xbf,0xff,0xfb,
	0xff,0xef,0xff,0xfe,0xfe,0xfb,0xef,0xbf,
	0xff,0xfe,0xef,0xef,0xbf,0xff,0x9f,0xfc,
	0x7e,0xfb,0xfd,0xf6,0xdf,0x7e,0xbf,0xfd,
	0xf7,0xdf,0x7f,0x3b,0xfd,0xb7,0xdf,0x7e,
	0xed,0xf7,0xcf,0x7f,0xfb,0xfd,0xf7,0xdf,
	0x7f,0xff,0xfd,0xb7,0xdf,0x7f,0xff,0xed,
	0xe8,0xff,0xcb,0x6f,0xbf,0xfd,0xf2,0xdb,
	0x67,0xbf,0xbd,0xf6,0xcb,0x6f,0xbf,0xdc,
	0xf2,0x23,0x9f,0xfd,0xf6,0x8b,0x6d,0xff,
	0xfd,0xf6,0xdb,0x6f,0xbf,0xdc,0xf6,0xdb,
	0xbf,0x5f,0xf7,0xff,0xff,0xdf,0xff,0xff,
	0xff,0xf7,0xff,0xff,0xfd,0xff,0xe7,0xff,
	0xff,0xff,0xfb,0xf7,0xff,0xff,0xfe,0xef,
	0xff,0xf9,0xff,0xff,0xff,0xff,0xff,0xfe,
	0xff,0xff,0xbb,0xfb,0xff,0xff,0xff,0xff,
	0xff,0x7f,0xbf,0xff,0xff,0xef,0x7f,0xdf,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xef,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0x7f,0xcd,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xfe,0xff,0xff,0xff,
	0xff,0xef,0xfb,0xbf,0x5e,0xfb,0xf5,0xff,
	0xff,0xfe,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0x8f,0xfe,0xff,
	0xef,0xff,0xf3,0xff,0xff,0xff,0xfc,0xff,
	0xff,0xff,0xbf,0xff,0xff,0xfb,0xff,0xff,
	0xff,0xff,0xff,0xbf,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xef,0xf4,
	0xff,0xef,0xbf,0xbf,0xfe,0xfb,0xef,0xaf,
	0xff,0xfa,0xfb,0xef,0xbf,0xff,0xfa,0xfb,
	0xbb,0xef,0xfe,0xfb,0xaf,0xbb,0xff,0xfe,
	0xfb,0xef,0x3f,0xff,0xfe,0xfb,0xef,0xff,
	0x8f,0xff,0x2f,0xff,0xff,0xff,0xfb,0xff,
	0xff,0xbf,0xff,0xff,0xef,0xfe,0x9f,0xf6,
	0xcb,0xbf,0xfd,0xff,0xff,0xfd,0xff,0xff,
	0xff,0xff,0xfd,0xff,0xff,0x7e,0xff,0xfd,
	0xfe,0x1f,0xfc,0xff,0xff,0xef,0x3f,0xff,
	0xff,0x7f,0xff,0xff,0xf7,0xfd,0xff,0xff,
	0xf7,0xbb,0xef,0xfb,0xfc,0xfb,0xdf,0xef,
	0xff,0xfd,0xfb,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xef,0xe8,0xff,0xf7,0xff,0xff,
	0xdf,0xfd,0xff,0xff,0xff,0xff,0x7f,0xf7,
	0xff,0xbf,0xf7,0xcf,0xe7,0xfe,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x7f,
	0xff,0x7f,0xff,0xfb,0x57,0xbf,0xff,0xdf,
	0xff,0xff,0xfe,0xff,0xff,0xff,0xbf,0xff,
	0xff,0xff,0xff,0xbf,0xff,0xbf,0xff,0xfe,
	0xbf,0xff,0xfd,0xff,0xff,0xff,0xfb,0xff,
	0xff,0xff,0xff,0xff,0xff,0x3f,0xfa,0xfd,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xfe,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xfd,0xff,0xbf,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xc9,
	0xfb,0xff,0xff,0xff,0xff,0xfd,0xff,0xff,
	0xff,0xff,0xff,0xe7,0xff,0x7f,0xfd,0xf5,
	0x4f,0xcf,0xfd,0xff,0xdf,0xfd,0xff,0xff,
	0xff,0xff,0xf7,0x6f,0xfe,0xff,0xff,0xff,
	0x9f,0xfe,0xff,0xff,0xff,0xff,0x7f,0xf7,
	0xff,0xff,0xff,0xff,0xff,0xd7,0xff,0xdf,
	0xbe,0xfb,0xf5,0xfb,0xff,0xfb,0xbf,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xfb,0xff,
	0xff,0x7f,0xf4,0xfe,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xdf,0xff,0xfd,
	0xdf,0x7f,0xff,0xff,0xfe,0xbf,0xff,0xbd,
	0xff,0xff,0xbf,0xff,0xff,0xff,0xef,0xbf,
	0xff,0x7f,0xff,0x97,0xfb,0xff,0xff,0xff,
	0xff,0xfd,0xff,0xff,0xff,0xdf,0xff,0x7f,
	0xfe,0xf7,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfe,0xff,0xff,0xcf,0x7f,0x7f,0xff,0xfd,
	0xff,0xff,0xff,0xff,0x5e,0xbd,0xff,0xff,
	0xbf,0xff,0xfa,0xff,0xff,0xff,0xff,0xff,
	0xff,0xdf,0xff,0xff,0xff,0xff,0xff,0xfb,
	0xfb,0xff,0xef,0xff,0xff,0xfd,0xff,0xff,
	0xff,0xff,0xfd,0xff,0xff,0xff,0xe6,0xfe,
	0xf3,0xcf,0xff,0xff,0xff,0xf3,0xcf,0xff,
	0xff,0xfc,0xf3,0xcf,0x3b,0xbb,0xef,0xcd,
	0x3f,0xfb,0xfc,0xb3,0xcf,0x3f,0xff,0xfc,
	0xf3,0xcf,0xfb,0xff,0xf8,0xbf,0xff,0x1f,
	0xfd,0xbf,0xf7,0xfe,0xff,0xff,0xbf,0xf7,
	0xfe,0x7f,0xef,0xbf,0xff,0xfe,0xfb,0xff,
	0xbf,0xfe,0xfb,0xef,0xbd,0xfd,0xfe,0xfb,
	0xef,0xbf,0xff,0xfe,0xdf,0xbf,0xff,0xff,
	0xff,0xbb,0xdf,0xfd,0xe3,0xbf,0xff,0xfe,
	0xf8,0xe3,0xbf,0x3f,0xfe,0xf9,0xa7,0x9d,
	0xf9,0xee,0xa7,0x9e,0x7b,0xf6,0xf9,0xe7,
	0x9f,0x7f,0xfe,0xf9,0xe7,0xbf,0xef,0xf7,
	0xff,0xff,0xd3,0xff,0xdf,0x5f,0x7f,0xfd,
	0xe5,0xd7,0x5f,0x7f,0xfd,0xf5,0xd7,0x57,
	0x7b,0xfd,0xf5,0x4f,0x7b,0xfd,0xf5,0x97,
	0x5f,0x7f,0xfd,0xf5,0xd7,0x5f,0xff,0xfd,
	0xff,0xff,0xff,0x7f,0xfe,0x7f,0xef,0xff,
	0xfb,0xef,0xff,0xff,0xff,0xf9,0xff,0x9f,
	0xff,0xbe,0xfb,0xef,0xfe,0xfe,0xfb,0xef,
	0xbf,0xff,0xfe,0xfb,0xef,0xbf,0xff,0xfe,
	0x9f,0xff,0xff,0xff,0x7f,0xd1,0xff,0xff,
	0xff,0xff,0xff,0x8f,0x3f,0xfe,0xf2,0xe3,
	0xff,0xef,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xbf,0xfe,0xff,0x94,0xf5,
	0xff,0xff,0xff,0x7f,0xfe,0xfc,0xf3,0x9f,
	0x3f,0xff,0xff,0xff,0xff,0x7f,0xfe,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xf9,0xff,0xba,
	0xbd,0xff,0xff,0xff,0xbf,0xff,0xfe,0xff,
	0xff,0xbf,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfe,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xf7,0x65,0xff,0xf7,0xdf,0x7f,0xff,0xfd,
	0xf7,0xdf,0x7f,0xfb,0xfd,0xf7,0xdf,0x7f,
	0xff,0xed,0xdf,0x7f,0xff,0xfd,0xf7,0xdf,
	0x7f,0xff,0xfd,0xf7,0xdf,0xff,0xff,0xff,
	0xff,0xff,0x3e,0xff,0xef,0xfe,0xeb,0xcf,
	0x3f,0xfe,0xfa,0xeb,0xaf,0xbf,0xff,0xfe,
	0xfb,0xef,0xbf,0xff,0xfb,0xef,0xbf,0xff,
	0xfe,0xfb,0xef,0xbf,0xff,0xfe,0xeb,0xbf,
	0xff,0xfe,0xfb,0x6f,0xdb,0x2f,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xbf,0x89,0xff,0xff,
	0xff,0xff,0xff,0xff,0xfe,0xff,0xff,0xff,
	0xff,0xff,0xeb,0xff,0xbf,0xfe,0xff,0xff,
	0xbf,0xff,0xff,0xfb,0xff,0xff,0xff,0xfe,
	0xff,0xff,0xb7,0xff,0xff,0xff,0xbd,0xfe,
	0xff,0xff,0xef,0xff,0xf5,0xe6,0xfb,0xef,
	0xbf,0xff,0xff,0xff,0xff,0xff,0xfd,0x7f,
	0xff,0xff,0xf9,0xff,0x9f,0xff,0xff,0xff,
	0xe7,0xff,0xef,0xff,0xf9,0xd6,0xfb,0xff,
	0xe6,0xf9,0x6f,0xbb,0xff,0xe6,0x7b,0xef,
	0xbf,0xff,0xfe,0xfb,0xef,0xbf,0xed,0xde,
	0xfb,0xbf,0xfd,0xde,0xfb,0xef,0xbd,0xff,
	0xfe,0x7b,0x6f,0xbb,0xed,0xdf,0xfb,0xef,
	0xef,0x2b,0xef,0x7f,0xe6,0xfd,0x67,0xdf,
	0x7f,0xf7,0xdd,0xf7,0x9f,0x7f,0xfe,0xf9,
	0x67,0x9f,0xff,0x79,0x67,0x9f,0x7d,0xf6,
	0xf9,0xe7,0x9f,0x7f,0xe6,0xfd,0xff,0x9f,
	0x7f,0xfe,0xff,0xf8,0xff,0xf5,0xf7,0x5f,
	0x7d,0xff,0xdd,0xf7,0xd8,0x7f,0xfd,0xf5,
	0xd7,0x5f,0x7d,0xad,0xd6,0x5f,0x77,0xf5,
	0x75,0xd7,0x5a,0x7f,0xfd,0xf5,0xf7,0xdf,
	0x7f,0xfe,0xf9,0x7e,0x64,0xff,0xfd,0xf7,
	0xff,0x7f,0xff,0xff,0xff,0xdf,0x7f,0xff,
	0xfd,0xf7,0xff,0x7f,0xff,0xfd,0xff,0xff,
	0xff,0xff,0xf7,0xdf,0x7f,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xfb,0x0e,0xff,0xf7,
	0xdf,0xff,0xff,0xfd,0xff,0xff,0xdf,0xff,
	0xfd,0xf7,0xf7,0xff,0x7f,0xff,0xbd,0xff,
	0xff,0xff,0xff,0xef,0xdf,0xff,0xfd,0xff,
	0xff,0xff,0xff,0x7f,0xff,0xfd,0xbf,0xf9,
	0xff,0xf2,0xcb,0x2f,0xbf,0xd8,0xe2,0x8b,
	0x3f,0xbe,0xf8,0x73,0xcf,0x2f,0xbe,0xf8,
	0xcf,0x2d,0xb6,0xfc,0x62,0xcf,0x3f,0xbf,
	0xd8,0xf2,0xcb,0xbf,0xf7,0xf9,0xe7,0xff,
	0xdb,0xff,0x77,0xdf,0x7d,0xf5,0xdd,0x57,
	0x5f,0x7d,0xf5,0xdd,0x57,0x5f,0x7d,0xf5,
	0xdd,0x5f,0x7d,0xf5,0xd5,0x56,0x5f,0x7d,
	0xf5,0xdd,0x57,0x5f,0xfd,0xf1,0xdf,0x7f,
	0xff,0xdf,0xfe,0xff,0xfe,0xba,0xfb,0xaf,
	0xbf,0xff,0xfe,0xef,0xaf,0xf7,0xfe,0xff,
	0xaf,0xad,0xff,0xff,0xbb,0xef,0xff,0xfa,
	0xfb,0xef,0xaf,0xbe,0xff,0xbe,0xed,0x7e,
	0xf3,0x7f,0x3f,0xf5,0xff,0xff,0xf1,0xc7,
	0x1f,0x7b,0xec,0xb1,0x47,0x1e,0x7b,0xf4,
	0xf1,0x67,0x1f,0x7b,0xf6,0xc7,0x1f,0x7f,
	0xf6,0xd1,0x47,0x1f,0x7f,0xfc,0xf1,0x67,
	0x9f,0x7f,0xee,0xf9,0x97,0xff,0xaf,0xff,
	0xfe,0xfb,0xef,0x3f,0xff,0xfc,0xff,0xef,
	0xff,0xff,0xfe,0xfe,0xef,0xdf,0xfe,0xfb,
	0xef,0xef,0xff,0xff,0xff,0xef,0xbf,0xff,
	0xfe,0xe3,0x9f,0x7f,0xff,0x7f,0xfd,0x7f,
	0xff,0xff,0xdf,0xff,0xff,0xf9,0xe7,0xf7,
	0xff,0x7f,0xff,0xf5,0xf7,0xff,0xff,0xf4,
	0xdf,0x7f,0x7f,0xff,0xfd,0xf7,0xff,0xff,
	0xfd,0xf7,0x5f,0xff,0xfd,0xf7,0xff,0xeb,
	0xff,0xef,0xbf,0xff,0xff,0xfb,0xff,0xff,
	0xff,0xff,0xfb,0xff,0xff,0xff,0xff,0xfb,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfb,0xff,0xff,0xff,0xfe,0xff,0xdf,0xff,
	0x47,0xff,0xff,0xfb,0xef,0xbf,0xff,0xfe,
	0xfa,0xeb,0xbd,0xff,0xde,0x7b,0xef,0xbf,
	0xff,0x7e,0xef,0xbf,0xff,0xfe,0x7b,0xef,
	0xbd,0xff,0xfe,0xfb,0xef,0xf7,0xff,0xff,
	0xff,0xff,0xf8,0xff,0xff,0xff,0xff,0xff,
	0xff,0xef,0xff,0xff,0xff,0xbf,0xff,0xff,
	0xff,0xf7,0xff,0xbf,0xff,0xfe,0xff,0xff,
	0xff,0xff,0xff,0xff,0x7f,0xff,0xff,0xfe,
	0xfb,0xff,0xff,0xce,0xff,0x7f,0xfd,0xd5,
	0x97,0x5f,0x7f,0xf5,0xe5,0x97,0x5e,0x7f,
	0xed,0xe5,0x97,0x5f,0xed,0xf4,0xd7,0x5f,
	0x7b,0xf9,0xf5,0xd7,0x5f,0x7f,0xfd,0xf5,
	0x57,0x5f,0x7f,0xfd,0x1f,0xfe,0xff,0xfc,
	0xfb,0x8f,0xff,0xff,0xff,0xff,0xef,0xff,
	0xfe,0xfc,0xf3,0xbf,0xbe,0xfb,0xe3,0x7f,
	0x3f,0xff,0xff,0xf3,0xcf,0xf7,0xfe,0xfc,
	0xf3,0xff,0xff,0xff,0xdc,0xff,0xe3,0xff,
	0x7c,0xfb,0x3d,0xaf,0xbf,0xff,0xfb,0xef,
	0xff,0xff,0xbc,0xf7,0xfd,0xff,0xdd,0xf6,
	0xef,0xff,0xfc,0xfe,0xcf,0xd7,0xf7,0xbf,
	0xf3,0xce,0xef,0xff,0xff,0xf3,0xfe,0xa1,
	0xff,0xdf,0x7f,0xff,0xff,0xff,0xff,0xff,
	0xff,0xfd,0xff,0xbf,0x77,0xff,0xff,0xff,
	0xef,0xfe,0xff,0xff,0xf7,0xff,0xff,0xfd,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0x7f,0xfc,0xff,0xff,0xfd,0xf7,0x5f,0x6f,
	0xff,0xf5,0xd7,0xff,0x7f,0xfd,0xff,0xf7,
	0x5f,0xff,0xff,0xd6,0x7f,0x6f,0x7d,0xf1,
	0xd7,0x5f,0xff,0xfd,0xf5,0xfe,0xff,0xf7,
	0xff,0x7f,0xc7,0xff,0xff,0xbf,0xff,0xfe,
	0xfb,0xef,0x7f,0xef,0x7e,0xbf,0xee,0xff,
	0xff,0xfd,0xeb,0xf3,0xbf,0x7d,0xff,0xde,
	0xdf,0x77,0xfe,0xf6,0xfb,0x9f,0xff,0xbf,
	0xff,0xff,0xff,0x5f,0xee,0xff,0xfd,0xf7,
	0x7f,0xb7,0xff,0x75,0xdf,0xdd,0x7f,0xdd,
	0xff,0xff,0x7f,0xf3,0x9d,0xdf,0xef,0xf7,
	0xdf,0xf3,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xef,0xff,0x9f,0xb3,0xff,0x7f,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xfe,0xf7,0xdf,0xfe,0xff,
	0xdf,0xe7,0xff,0xff,0xff,0xff,0xf3,0xff,
	0x7f,0xff,0xff,0xfd,0xff,0xff,0xd9,0xff,
	0xfe,0xff,0xfb,0xaf,0xdf,0xff,0xff,0xff,
	0xf7,0xff,0xfe,0xfd,0xff,0xff,0xff,0xff,
	0xef,0xff,0xbf,0xff,0xee,0xff,0xff,0x7f,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xaf,
	0xfc,0x3f,0xff,0xbc,0xfd,0xcf,0x9f,0xff,
	0xf5,0xfb,0xcf,0x3b,0x7f,0xbc,0xd3,0xcf,
	0xbf,0xf4,0xe3,0xcf,0x3b,0xfe,0xfc,0xf3,
	0xdf,0x3d,0xff,0xbe,0xff,0xff,0xff,0xff,
	0xfe,0xf3,0xff,0xfb,0xef,0xff,0xff,0xfd,
	0xfd,0xdb,0xdf,0x7f,0xfe,0xfb,0xef,0xaf,
	0x7f,0xfe,0xef,0xff,0xfe,0xfe,0xfb,0xef,
	0xbf,0xff,0xfe,0xfb,0xf7,0xef,0xfd,0xff,
	0xff,0xff,0x9b,0xed,0xdf,0x7f,0xfd,0xfb,
	0xf7,0xff,0xbf,0xfe,0xff,0xef,0xdf,0x7f,
	0xee,0xed,0xbf,0x7f,0xf6,0xfa,0xc7,0xff,
	0x7f,0xfc,0xf9,0xef,0x1f,0xff,0xfe,0xff,
	0xee,0x7f,0xff,0xbf,0xfd,0xfe,0xfd,0xff,
	0xff,0xff,0xef,0xff,0xf5,0xfb,0xff,0xff,
	0xff,0xf3,0xdf,0xef,0xfb,0xf5,0xd7,0x5f,
	0xbf,0xff,0xf5,0xd7,0x5f,0x7f,0xfd,0xf5,
	0xff,0x7f,0xff,0xff,0xff,0xe4,0xfb,0xb7,
	0xff,0xbf,0xff,0xff,0xff,0xff,0xfe,0xff,
	0xfe,0xff,0xdf,0x7f,0xff,0xff,0xef,0xff,
	0xff,0xfe,0xff,0xef,0x9f,0x7b,0xfe,0xfb,
	0xef,0xff,0xff,0xfa,0xff,0xff,0x7f,0xfc,
	0xff,0xaf,0xbf,0x7e,0x7e,0xff,0xfd,0xff,
	0xdf,0xff,0xeb,0xff,0xff,0xff,0xff,0xff,
	0xff,0x9f,0xfc,0xeb,0xfd,0xff,0xcf,0xff,
	0xff,0xff,0xff,0xff,0xff,0xeb,0xff,0x4f,
	0x5a,0xff,0x7f,0xfd,0xf5,0xf3,0xfd,0xef,
	0xff,0xff,0xfe,0x9f,0x7f,0xfd,0xff,0xff,
	0xff,0xfd,0xff,0xe6,0x9f,0xff,0xff,0x7f,
	0xff,0xff,0xff,0xff,0xff,0xff,0x9f,0xff,
	0xaf,0xdb,0xfb,0xff,0xff,0xff,0xfb,0xff,
	0xff,0xff,0xfe,0xff,0xef,0xff,0xff,0xff,
	0xff,0xef,0xff,0xff,0xff,0xff,0x7f,0xff,
	0xff,0xff,0xef,0xff,0xff,0xff,0xfb,0xef,
	0xff,0x7f,0xbf,0xf6,0x7f,0xff,0xfd,0xb7,
	0xff,0xff,0xff,0xed,0xff,0xdf,0x7e,0xff,
	0xfd,0xf7,0xdf,0x7e,0xfd,0xff,0xdf,0xff,
	0xff,0xfd,0xf7,0x5f,0x7f,0xff,0xfd,0xbf,
	0xff,0xfe,0xff,0xef,0xf5,0xff,0xaf,0xbf,
	0xfe,0xfa,0xef,0xbf,0xbf,0xff,0xfb,0xfb,
	0xaf,0xbf,0xfe,0xfa,0xfb,0xbf,0xfe,0xfb,
	0xeb,0xbf,0xbf,0xfe,0xfe,0xfb,0xaf,0xbf,
	0xff,0xfb,0xff,0xbf,0xff,0xab,0xfd,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0x1b,0xf8,
	0xff,0xff,0xff,0xff,0xff,0xff,0xb7,0xff,
	0x7f,0xff,0xff,0xff,0xff,0xff,0xfb,0xff,
	0xff,0x7f,0xfb,0xff,0xf7,0xff,0xfe,0xea,
	0xff,0xbf,0xfe,0x7f,0xfb,0xff,0xff,0x7f,
	0xef,0xff,0xbf,0xff,0xfe,0xfb,0xcf,0x7f,
	0xf8,0xff,0xc7,0xff,0xbf,0xff,0xfe,0x1b,
	0xff,0xf7,0xfe,0x17,0xef,0x7f,0xfc,0xe6,
	0xdf,0xff,0xbf,0xff,0xf3,0x8f,0xff,0xbf,
	0xff,0x6f,0xff,0xff,0xe6,0xdb,0x6e,0xbf,
	0xff,0xfe,0xdb,0xee,0xbf,0xff,0xb6,0xdb,
	0x6e,0xbb,0xed,0xfb,0xef,0xbf,0xed,0xfe,
	0x5b,0xee,0xbd,0xff,0xfe,0xfb,0xbf,0xff,
	0xfe,0x7e,0x3e,0xfa,0xff,0x77,0xdf,0x79,
	0xdf,0xf9,0xe7,0x9f,0x7f,0xfe,0xf9,0x77,
	0xde,0x79,0xe7,0xd9,0xdf,0x7f,0xfe,0xfd,
	0xe7,0xdf,0x79,0xf6,0xf9,0xf7,0x9f,0xff,
	0xf7,0xff,0xe7,0xf7,0x95,0xff,0xdf,0x69,
	0xaf,0xfd,0xe7,0x1f,0x7f,0xfd,0xf1,0xd7,
	0xdb,0x7f,0xff,0xfd,0xd6,0x6d,0xff,0xf1,
	0xf7,0x1f,0x7f,0xff,0xd5,0xd7,0xdf,0x6b,
	0xfd,0x9d,0xf6,0x9f,0xff,0x97,0xf6,0xff,
	0x7f,0xff,0xfd,0xff,0xff,0xff,0xff,0xff,
	0xf7,0xcf,0xff,0xff,0xff,0xf3,0xdf,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xf7,0xdf,
	0x7f,0xff,0x7f,0xf7,0xfd,0xff,0xef,0xf6,
	0xff,0xff,0xfe,0xfd,0xff,0xf7,0xff,0xff,
	0xff,0xef,0xbf,0xff,0xff,0xff,0xf7,0xdf,
	0xfb,0xff,0xff,0xff,0xff,0xff,0xff,0xf7,
	0xdf,0x7f,0xff,0xff,0xf7,0xff,0xdf,0xff,
	0x83,0xfe,0x3f,0xbf,0xf8,0x63,0x1f,0x7d,
	0xb4,0xf8,0xc7,0x8f,0x3f,0xbe,0xf8,0xe3,
	0x8b,0xff,0xfc,0xc3,0x8b,0x7d,0xbc,0xd8,
	0xe2,0x8b,0x3d,0xff,0xf8,0xff,0xff,0x7f,
	0xfe,0x97,0xfd,0xff,0xf5,0xd5,0x77,0xff,
	0xfd,0xf7,0xd5,0x7f,0x7f,0x7d,0xf7,0xd5,
	0x5f,0x5b,0xfd,0xd4,0x1f,0x5f,0xfd,0xf7,
	0xd5,0x57,0xdf,0x7d,0xf5,0xd5,0x7f,0x5f,
	0xfd,0xf7,0xdf,0xed,0xff,0xeb,0xee,0xff,
	0xee,0xaf,0xbf,0x3f,0xff,0xfd,0xf3,0xeb,
	0xbe,0xbb,0xea,0x77,0xbf,0xf7,0xfc,0xfb,
	0xbf,0xee,0xbe,0xdb,0xf2,0xef,0xbf,0xbe,
	0xee,0x3f,0xff,0xff,0x67,0xff,0x5f,0x9f,
	0x79,0xe6,0xf9,0xe7,0x9f,0x7d,0xfe,0xd9,
	0x47,0x9e,0x79,0xf4,0xf9,0x67,0x7d,0xf6,
	0x99,0xe7,0x1f,0x7f,0xfc,0xf1,0x47,0x1f,
	0x7d,0xf4,0xd1,0xe7,0x9e,0xff,0xf8,0xff,
	0xf3,0xcf,0xbf,0xff,0xff,0xf7,0xfb,0x7f,
	0xfe,0xfe,0xf3,0xfb,0x3f,0xff,0xfe,0xdf,
	0x3f,0xbe,0xff,0xf7,0xcf,0xbf,0xff,0xfe,
	0xe7,0xff,0x2f,0xfe,0xfe,0xf7,0x7f,0xcf,
	0xff,0x9f,0xdf,0x7e,0xff,0xff,0xbf,0xdf,
	0xff,0xff,0xfd,0xb7,0xdf,0xff,0xf9,0xfd,
	0xcf,0xfe,0xfd,0xfd,0xbf,0x7f,0xfe,0xfd,
	0xff,0xf7,0xdf,0x7f,0xed,0xb7,0x7f,0xff,
	0x5f,0xfe,0xff,0xff,0xff,0xef,0xff,0xfe,
	0xfb,0xff,0xbf,0xbf,0xff,0xff,0xff,0xff,
	0xbf,0xff,0xff,0xaf,0xff,0xff,0xfb,0xff,
	0xff,0xbf,0xff,0xff,0xff,0xff,0xbf,0xff,
	0xfd,0xff,0xf3,0xff,0xaf,0xbf,0x7e,0xfb,
	0xff,0xff,0xff,0xfe,0x7f,0xef,0xaf,0xff,
	0xfe,0x7a,0xef,0xb7,0x7e,0xff,0xef,0xff,
	0xbf,0xfe,0xfb,0xef,0xad,0xf7,0xfe,0xff,
	0xfd,0xff,0xff,0x9f,0xff,0xff,0xff,0xff,
	0xff,0xbf,0xff,0xfe,0xff,0xff,0xbf,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xef,0xff,
	0xff,0xfe,0xfb,0xff,0xff,0xff,0xff,0xfb,
	0xff,0xff,0xff,0xff,0xdf,0xf4,0xff,0x57,
	0x5f,0x7f,0xf1,0xe4,0x97,0x4f,0x7f,0xfd,
	0xf5,0xd7,0x5f,0x7f,0xfd,0xf5,0x5e,0x7d,
	0xfc,0xb5,0xd7,0x5f,0x7f,0xf1,0xd5,0x57,
	0x5f,0x7f,0xfd,0xf5,0xd7,0xbf,0xe1,0xff,
	0x9f,0xfd,0xf7,0xfe,0xff,0xdf,0x3f,0xdf,
	0xfe,0xf3,0xbf,0xff,0xff,0xef,0xbb,0x3f,
	0xf7,0xfb,0xff,0xff,0xff,0xfe,0xff,0xff,
	0xcf,0xff,0xd6,0xfc,0xf3,0xcf,0x7f,0x7f,
	0xfe,0xdf,0xff,0x7f,0xff,0xff,0xaf,0x3f,
	0xbf,0xf7,0xf7,0xfd,0xbf,0xbf,0xff,0xfb,
	0x3b,0xfd,0xde,0xfb,0xf7,0xdf,0x7f,0xff,
	0xff,0x3f,0xb7,0xff,0xff,0xff,0x3f,0xbf,
	0x9f,0xfa,0xff,0xff,0xff,0xdf,0xff,0x7f,
	0xff,0xef,0xff,0x7f,0xff,0xff,0xff,0xfd,
	0x7f,0xf7,0xfe,0xff,0xff,0xff,0xff,0xff,
	0xfd,0xff,0xff,0xff,0xff,0xcf,0x3f,0xff,
	0xff,0x7f,0xcb,0xff,0xd7,0x5f,0xff,0xff,
	0xff,0xff,0x5f,0xff,0xff,0xf5,0xf7,0x5f,
	0x7f,0xfc,0xfd,0xcf,0xfb,0xff,0x7d,0xfe,
	0x5f,0x7f,0xfc,0xf5,0xc7,0xdf,0xff,0xfd,
	0xf1,0xff,0xff,0x5f,0xfe,0xff,0xf6,0x9a,
	0xff,0xfe,0xfd,0xef,0xf7,0xfe,0xab,0xef,
	0xfd,0xfb,0xf7,0xbf,0xff,0xeb,0xff,0x7f,
	0xef,0xff,0xfb,0xf6,0xbb,0x7b,0xff,0xfb,
	0xfb,0xdf,0xff,0xef,0xff,0xf7,0xff,0xfd,
	0xff,0xff,0xff,0xff,0xff,0xb7,0xff,0xfe,
	0xdf,0xff,0xf7,0xed,0xf7,0xdf,0xfd,0xfd,
	0xff,0xdf,0xff,0xf7,0xfd,0x77,0xdf,0xbf,
	0xfe,0xff,0xf7,0xff,0xff,0xff,0xa3,0xff,
	0xf7,0xff,0xdf,0xff,0xfe,0xff,0xff,0xf7,
	0xff,0xff,0xfd,0xfd,0xfb,0xff,0xff,0xef,
	0xff,0xff,0xff,0xf7,0xfd,0xfd,0xff,0xff,
	0xff,0xdb,0xf3,0xff,0xdf,0xff,0xff,0x3f,
	0xfc,0xff,0xb7,0xdf,0x7f,0xff,0xfb,0xff,
	0xbf,0xfb,0xff,0xff,0xff,0xff,0xfd,0xfb,
	0xff,0xbf,0xff,0xff,0xbf,0xff,0xff,0xff,
	0xff,0xf9,0xbf,0xff,0xff,0xff,0xff,0xff,
	0xbf,0xe0,0xff,0xf3,0xf7,0x5b,0xff,0xfc,
	0xff,0xcf,0xff,0xff,0xbe,0xf3,0xf7,0xf7,
	0x7f,0xff,0xcf,0x7f,0xfd,0xff,0xd7,0xc7,
	0x3f,0x7f,0xfd,0xf3,0xff,0x3f,0xfe,0xff,
	0xff,0xff,0x57,0xff,0x9f,0xff,0xef,0xf7,
	0xdf,0xff,0xdf,0xfd,0xff,0xf7,0xff,0xfe,
	0x7f,0xff,0xff,0x17,0xdd,0xfb,0xff,0x7f,
	0x7f,0xfe,0xf7,0xbf,0x9f,0xff,0x7f,0xef,
	0xff,0xff,0xff,0xbf,0xf9,0xfe,0xfb,0xff,
	0x7f,0x7f,0xdf,0x7f,0xd7,0xfd,0xdb,0xff,
	0xf5,0xef,0xff,0xdf,0xfd,0xdf,0xff,0xff,
	0xfd,0xf2,0xff,0x5d,0xf7,0xff,0xf3,0xff,
	0x3f,0xef,0xff,0xf7,0xff,0x81,0xef,0xdf,
	0x7f,0xef,0xff,0xff,0xff,0xff,0xff,0xbd,
	0xef,0xff,0x3f,0xff,0xf7,0xff,0xff,0x7f,
	0xfd,0x7f,0xdf,0xff,0xbf,0xff,0xff,0xdf,
	0xff,0xff,0xfd,0xff,0xff,0xff,0x57,0xbe,
	0xbf,0xfe,0xeb,0xff,0xff,0xf6,0xff,0xff,
	0xaf,0xff,0xff,0xff,0xfd,0xbf,0xff,0xff,
	0xef,0xff,0xff,0x7f,0x7f,0x7f,0xff,0xff,
	0xbf,0xfe,0xff,0xab,0xff,0xfe,0xff,0x7f,
	0xc6,0xff,0xff,0xff,0xff,0xff,0x97,0x5f,
	0x7e,0xff,0xff,0x3f,0xff,0xfa,0xdf,0xcd,
	0xf7,0x7b,0xfe,0xcf,0x37,0x5f,0xfe,0xf3,
	0xff,0x27,0xff,0x7a,0xff,0xcf,0xbf,0xfe,
	0xff,0xbe,0xf5,0xff,0xe7,0xff,0xff,0xbf,
	0xf5,0xd7,0xff,0xff,0xff,0xf5,0xd7,0x5f,
	0x6e,0xfd,0xfc,0xff,0x7f,0xfd,0xf9,0xf2,
	0x5f,0x7f,0xdd,0xff,0xe7,0xff,0xff,0xff,
	0xf9,0xff,0x7a,0xbd,0xff,0xff,0xef,0xff,
	0xff,0xff,0xf7,0xdf,0xbf,0xff,0xff,0xff,
	0xef,0xf7,0xff,0xfd,0xdf,0xd7,0xff,0xfd,
	0xff,0xff,0xff,0xff,0xff,0xfb,0xdf,0xff,
	0xff,0xff,0x3f,0xf7,0x6e,0xff,0xf7,0xdf,
	0xff,0xff,0xff,0xff,0xff,0x7f,0xfb,0xff,
	0xff,0xdf,0xbe,0xff,0xff,0xff,0x37,0xfe,
	0xff,0xff,0xdf,0xff,0xff,0xff,0xd7,0xff,
	0x7f,0xff,0xff,0xff,0xfd,0x36,0xff,0xff,
	0xfa,0xe3,0xbf,0xff,0xfe,0xfb,0xef,0xaf,
	0xff,0xfe,0xfb,0xfb,0xbd,0xff,0xbe,0x6f,
	0xaf,0xff,0xfe,0xfb,0xeb,0xbf,0xff,0xfe,
	0xfa,0xef,0xaf,0xff,0xfe,0xdb,0x4f,0xdb,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xef,0xfe,0xff,
	0xff,0xff,0xfe,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xb7,
	0x99,0xff,0xff,0xff,0xff,0xbf,0xfe,0xff,
	0xff,0xff,0xff,0x5f,0xfe,0xf9,0x7f,0xf7,
	0xff,0xff,0x7d,0xfd,0xdf,0x7e,0xff,0xff,
	0xbf,0xfe,0xff,0xff,0xfd,0xff,0xdf,0xff,
	0xbf,0x1d,0xfe,0xff,0xfb,0x6f,0x3d,0xfc,
	0xfc,0xc3,0xcf,0xbf,0xff,0xfd,0xc3,0xff,
	0x7b,0xfd,0x74,0x1f,0xab,0xff,0xe5,0xf7,
	0xef,0x3f,0xfd,0xd4,0xfb,0x1f,0xbf,0xff,
	0xd7,0xfb,0xff,0xf6,0xf9,0x6f,0xbe,0xed,
	0xfe,0xfb,0xef,0xbf,0xff,0xfe,0xfb,0xef,
	0xbf,0xed,0xde,0x7b,0xbf,0xff,0xb6,0x7b,
	0xef,0xbf,0xed,0xde,0xfb,0xef,0xbf,0xff,
	0xfe,0xfb,0xef,0xe7,0x95,0xef,0x7d,0xf7,
	0x9d,0xe7,0x9f,0x7f,0xfe,0xf9,0x77,0x9f,
	0x7f,0xfe,0xf9,0xe7,0x9f,0xff,0xf9,0xf7,
	0x9f,0x7f,0xfe,0xdd,0xe7,0x9f,0x7f,0xff,
	0xf9,0xf7,0x9f,0x7f,0x7e,0x7f,0xf0,0xf7,
	0x9d,0xf6,0x9f,0x7f,0xfe,0xf9,0xe7,0xda,
	0x77,0xfc,0xf9,0xd6,0x1f,0x7f,0xfe,0xc6,
	0xdf,0x7e,0xfc,0xf1,0xf7,0x9d,0x7f,0xfe,
	0xfd,0xc7,0xdb,0x7f,0xfc,0xf9,0x7e,0x6c,
	0xff,0xff,0xf7,0xff,0xff,0xff,0xff,0xff,
	0xdf,0xff,0xff,0xff,0xf6,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xf7,0xff,0xff,
	0xff,0xfd,0xff,0xdb,0xff,0xff,0xff,0xf6,
	0x36,0xff,0xff,0xef,0xff,0xdf,0x7f,0xff,
	0xfd,0xb7,0xff,0xff,0xff,0xf5,0xff,0xff,
	0x7f,0xbf,0xff,0xff,0xff,0xff,0xf7,0xff,
	0xdf,0x7f,0xfb,0xff,0x7f,0xff,0xff,0xff,
	0xdd,0xff,0xe9,0xff,0xf2,0x8b,0x7f,0xf5,
	0xd1,0xd7,0x1f,0x2f,0xfe,0xd5,0xc7,0x8f,
	0x7d,0xfc,0xd1,0x1f,0x2f,0xf7,0xf1,0xc3,
	0x8b,0x7d,0xf4,0xf1,0x63,0x1f,0x3f,0xff,
	0xf9,0xe7,0xff,0xc9,0xff,0x57,0x7f,0xfd,
	0xf7,0xdf,0x7f,0xff,0xfd,0xb5,0xdf,0x7f,
	0x5f,0xfd,0xf7,0xdf,0xff,0x7d,0xf5,0xdf,
	0x1f,0x5f,0xfd,0xf7,0xdf,0x57,0xff,0x7d,
	0xb5,0xdf,0x7f,0xff,0x6f,0xfe,0xff,0xfe,
	0x7b,0xff,0xff,0xfa,0xff,0xef,0xcf,0xfe,
	0xff,0xcd,0xff,0xbf,0xfc,0xfa,0xef,0xff,
	0xff,0xf2,0xfe,0xfe,0xdf,0xff,0xff,0xfc,
	0xef,0xef,0xff,0xf2,0xff,0x7f,0xf5,0xff,
	0xf7,0xb9,0xe7,0x9f,0x7f,0xfe,0xf9,0xe7,
	0x9f,0x7f,0xfe,0xb1,0xe7,0x9f,0x7f,0xfe,
	0xe7,0x9f,0x7f,0xe6,0x99,0xe7,0x9f,0x7b,
	0xf4,0xf9,0x47,0x9f,0x7f,0xee,0xf9,0xbb,
	0xff,0xff,0xbf,0xfe,0xf7,0xdf,0x7f,0xfe,
	0xfd,0xfa,0xdf,0xff,0xff,0xfe,0xff,0xdf,
	0xdf,0xfd,0xfb,0xdf,0x3f,0xff,0xfe,0xff,
	0x9f,0xff,0xff,0xfd,0xff,0xff,0x7f,0xff,
	0xff,0xfc,0xff,0xff,0xf7,0xbf,0xff,0xfe,
	0xff,0xef,0xdf,0xff,0xfe,0xff,0xf5,0xff,
	0xff,0xff,0xef,0xdf,0xff,0xfe,0xf9,0xfd,
	0xff,0xff,0x7f,0xff,0xff,0xf7,0xff,0xfd,
	0xf7,0xff,0xef,0xff,0xaf,0xbf,0xfe,0xff,
	0xff,0xff,0xff,0xff,0xfa,0xff,0xff,0xff,
	0xfe,0xff,0xf7,0xf7,0xff,0xfa,0xff,0xef,
	0xff,0xfe,0xff,0xff,0xbf,0x7f,0xff,0xfb,
	0xff,0xdf,0xff,0x37,0xff,0x7d,0xff,0xff,
	0xbf,0xff,0xfe,0xfb,0xef,0xff,0xff,0xfe,
	0x7b,0xff,0xbf,0xff,0xfe,0xef,0xf7,0xff,
	0x7e,0xfb,0xff,0xbf,0xff,0xde,0xff,0xef,
	0xfd,0xff,0xff,0xff,0x3f,0xfb,0xff,0xef,
	0xbe,0xff,0xfe,0xbb,0xff,0xff,0xff,0xf7,
	0xfb,0x7f,0xbf,0x2f,0xff,0xff,0xfd,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xfb,0xff,
	0xff,0xff,0xff,0xff,0x7f,0xff,0xc4,0x7f,
	0x7f,0xfd,0xb4,0xd7,0x5f,0x7d,0xed,0xb5,
	0x57,0x4f,0x7f,0xf5,0xf5,0x56,0x5f,0xff,
	0xd5,0xd7,0x5f,0x7f,0xfd,0xf5,0xd7,0x5f,
	0x3d,0xf9,0xa5,0xd7,0x5e,0x7f,0xfd,0x4f,
	0xfe,0xff,0x7c,0xff,0xce,0xbf,0xff,0xfc,
	0xff,0x4f,0xbd,0xfe,0x5f,0xf3,0xdf,0xb6,
	0xfb,0xff,0xcf,0xff,0xfe,0xfa,0xff,0xff,
	0xff,0xff,0xff,0xf3,0xcf,0xff,0xff,0xfc,
	0xff,0xe5,0xff,0xfc,0xff,0xbe,0xff,0xf9,
	0xfd,0xfd,0x5f,0xff,0xff,0xfe,0xf7,0xbf,
	0xff,0xbf,0xfb,0x2f,0xff,0xbf,0x77,0xff,
	0xff,0xff,0xff,0xff,0xdf,0xbf,0xff,0xff,
	0xfd,0xfe,0x9d,0xff,0xbf,0xfd,0xff,0xfd,
	0xff,0xbf,0xff,0xff,0xff,0xf7,0xff,0xff,
	0xfe,0xff,0x6f,0xef,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xbf,0xff,0xfb,
	0xff,0xbf,0xff,0x3f,0xf8,0x7f,0xfd,0xff,
	0xfe,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xef,0xff,0xff,0xf7,0xff,
	0xef,0x7f,0xf7,0xff,0xf7,0x7f,0xfd,0xdf,
	0xd7,0xff,0xff,0xff,0x7f,0xe4,0xff,0xeb,
	0xff,0xff,0xff,0xff,0xff,0xff,0xeb,0xff,
	0xbf,0xff,0xff,0xff,0xbf,0xff,0xff,0xff,
	0xb9,0xff,0xfe,0xff,0xff,0xff,0xbf,0xe7,
	0xfe,0xff,0x6e,0xbf,0xff,0xff,0x07,0xfe,
	0xff,0xff,0xfb,0xff,0xff,0xfd,0xff,0xff,
	0x7f,0xff,0xff,0xff,0xdf,0xff,0x7f,0xdd,
	0xef,0x7f,0xff,0xfd,0x7f,0xff,0x5f,0xff,
	0xff,0xff,0xdf,0xfd,0xff,0xff,0xff,0x1f,
	0xfa,0xff,0xff,0xff,0xfd,0xff,0xff,0xdf,
	0xff,0xff,0xfe,0x7f,0xff,0xff,0xff,0xff,
	0xff,0xfe,0xf7,0xff,0xf7,0xff,0xdf,0xff,
	0xdf,0xff,0xff,0xdf,0xff,0xed,0xff,0xff,
	0xff,0xd5,0xbf,0xff,0xfb,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xbe,0x7f,0xfb,0xff,
	0xff,0xfe,0xff,0xef,0xff,0xff,0xff,0xfe,
	0xff,0xff,0xff,0x7b,0xfe,0xff,0xff,0xff,
	0xff,0xff,0xef,0xfe,0x3f,0xff,0xff,0xf7,
	0xcf,0xff,0xff,0xff,0xf3,0xcf,0x3f,0xff,
	0xfc,0xf3,0xff,0xbf,0xff,0xff,0xcf,0x2f,
	0xfa,0xff,0xe3,0xc6,0x3f,0xff,0xfe,0xfd,
	0xff,0xff,0xff,0x7f,0xf1,0xff,0xfb,0xff,
	0xfe,0xfe,0xfe,0xff,0xff,0x5f,0xff,0xde,
	0xfb,0xef,0x7f,0xff,0xdf,0xff,0xff,0xef,
	0xfe,0xfb,0xff,0xbf,0xff,0xfd,0xfb,0xf7,
	0xff,0xff,0xff,0xff,0xff,0xab,0xfb,0x8f,
	0xf7,0xff,0x6f,0xe3,0xaf,0xff,0xff,0xf7,
	0xe3,0xdf,0x7f,0xff,0xf5,0xff,0xbf,0xfe,
	0xd7,0xe7,0xff,0xf7,0xff,0xf7,0xf7,0xdf,
	0xbf,0xfe,0xff,0xfe,0xff,0xff,0x9f,0xdd,
	0x7f,0xfd,0x7f,0xff,0x5f,0x7f,0xfd,0xf7,
	0xff,0x5d,0xff,0xff,0xff,0xff,0xff,0xfb,
	0x74,0xdf,0x7f,0xfb,0xff,0xff,0xff,0xff,
	0xff,0xff,0x66,0xff,0xff,0xff,0xdf,0xff,
	0xe5,0xff,0xff,0xff,0xff,0xfb,0x7f,0xbf,
	0xb7,0xfe,0xff,0xff,0xff,0xff,0xff,0x7f,
	0xef,0xff,0xfd,0x7e,0xfc,0xff,0xff,0xd7,
	0x7f,0xef,0xff,0xbf,0xff,0xff,0xef,0xff,
	0xff,0x27,0xfc,0xff,0xe7,0xb7,0xde,0xff,
	0xff,0xaf,0x9f,0x9f,0x3f,0xf3,0xaf,0xf3,
	0xdf,0x7f,0xeb,0x37,0xff,0xff,0xf3,0xad,
	0x37,0xdf,0x74,0xff,0xff,0x37,0xdf,0x7a,
	0xf3,0xfd,0x8f,0x58,0xff,0x7f,0xde,0x75,
	0xff,0xff,0x7f,0xfd,0xf5,0xff,0xcd,0x7f,
	0xdd,0x79,0xfd,0x5b,0xbd,0xfc,0xff,0xcf,
	0x6f,0xbe,0xf5,0x97,0xff,0xff,0xbf,0xf9,
	0xd6,0xfb,0xef,0xaf,0xdd,0xfb,0xff,0xff,
	0xff,0xff,0xff,0xbf,0xff,0xfe,0xf7,0xff,
	0xff,0xff,0x7f,0xfd,0xff,0xff,0xff,0xfb,
	0xff,0xff,0xff,0xff,0xf7,0xdd,0xff,0xff,
	0xff,0xff,0xff,0xff,0x7f,0xcf,0xf6,0x7f,
	0xff,0xff,0xff,0xdf,0x7f,0xfb,0xfd,0xff,
	0xdf,0x7f,0xff,0x7d,0xeb,0xff,0xff,0xff,
	0xf7,0xdf,0xff,0xff,0xff,0xff,0xef,0x7f,
	0xff,0xff,0xff,0xff,0xff,0xff,0x6f,0xf7,
	0xff,0xaf,0xff,0xfe,0xfb,0xeb,0xef,0x3f,
	0xfe,0xfb,0xfb,0xaf,0xbf,0xf7,0xfb,0xef,
	0xfb,0xfe,0xf8,0xfb,0xbf,0xff,0xfe,0x7b,
	0xef,0xaf,0xff,0xfe,0xfb,0xef,0xbf,0xf7,
	0x9b,0xfd,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xef,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xbb,0xff,0xff,0xfd,0xff,0xff,0xff,
	0xff,0xbb,0xf9,0xff,0xff,0xdf,0xff,0xff,
	0xaf,0xbf,0xff,0xfe,0xff,0xff,0xff,0xff,
	0x57,0xff,0xad,0xdf,0xfe,0xfb,0xef,0xf7,
	0xdf,0x7e,0xdf,0xf9,0xff,0xcf,0x7e,0xfb,
	0xff,0xbf,0x7f,0xe5,0xff,0xbf,0xff,0xfd,
	0xf3,0x6f,0xff,0xf9,0xe2,0xf3,0xff,0xbf,
	0xff,0xbf,0xd6,0x5f,0xf7,0xc5,0x8b,0x7f,
	0x7e,0xff,0xe1,0xc7,0x5e,0xbd,0xff,0xe1,
	0x87,0x0f,0x3f,0xfb,0x37,0xff,0xff,0xf6,
	0xfb,0xef,0xbf,0xf7,0xde,0xfb,0xef,0xbf,
	0xff,0xe6,0xdb,0xef,0xfd,0xb7,0xfb,0x6f,
	0xbf,0xe5,0xfe,0xfb,0xef,0xbf,0xff,0xb6,
	0xfb,0xef,0xbf,0xff,0xfe,0xff,0xfa,0xff,
	0xf7,0x9d,0x7f,0xfe,0xdd,0x67,0xdf,0x7d,
	0xfe,0xf9,0x77,0x9e,0x75,0xfe,0xff,0x9f,
	0x7f,0xdf,0x99,0xe7,0x9f,0x7f,0xfe,0xf9,
	0xf7,0xff,0x7d,0xfe,0xf9,0xe7,0xbf,0x97,
	0xff,0xdf,0x7f,0xfc,0xf9,0x77,0x5b,0x7d,
	0xdf,0xb9,0xd6,0xda,0x7f,0xb5,0xf1,0xf7,
	0x7f,0xfc,0xfd,0xd7,0x1f,0x7f,0xfc,0xf1,
	0xc7,0xdf,0x7f,0xdf,0xf1,0xe7,0x9f,0xff,
	0xd7,0xf6,0xff,0xff,0xff,0xff,0xff,0xdb,
	0xff,0xff,0xff,0xf7,0xdf,0x7f,0xbf,0xfd,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xdf,0xff,0xff,0xff,0xff,0xff,
	0xff,0xef,0xf5,0xff,0xff,0xff,0xff,0xfd,
	0xbf,0xff,0xff,0xff,0xdd,0xdf,0x7f,0xff,
	0xfb,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xbf,0xff,0xff,0xff,0xff,
	0xf7,0xdf,0xff,0x83,0xfe,0x2f,0xff,0xf1,
	0xc7,0x8b,0x2f,0xbe,0xf8,0x47,0xcb,0x2f,
	0xbe,0xd8,0xc3,0xff,0xdf,0xf0,0x62,0x8b,
	0x7f,0xfc,0xf1,0x47,0x1f,0x2d,0xff,0xf9,
	0xc7,0x1f,0x7f,0xfc,0xf7,0xec,0x7f,0xf5,
	0xdf,0x7e,0x5f,0x7d,0xf7,0xd5,0x7f,0xdf,
	0x7d,0xf5,0xd5,0x37,0xff,0xbd,0xc7,0x57,
	0x5f,0xfd,0xf7,0xdf,0x7f,0xff,0x7d,0xf5,
	0xcf,0x7f,0xff,0xfd,0xf7,0xff,0xef,0xff,
	0xef,0xfd,0xf3,0xef,0x6e,0xaf,0xef,0xf3,
	0xfd,0xfa,0xff,0xff,0xff,0xea,0xff,0xb7,
	0xbf,0xff,0xeb,0xff,0xff,0xfe,0xeb,0xef,
	0xef,0x3f,0xff,0xfb,0xef,0xff,0xff,0x6f,
	0xff,0x7f,0x9f,0x7b,0xee,0xf1,0xe7,0x1f,
	0x7b,0xfe,0xf1,0xe7,0x9e,0x7b,0xfc,0xf9,
	0xe3,0x19,0xec,0xd9,0xe7,0x86,0x1f,0xfe,
	0xf9,0xc3,0x9f,0x1d,0x7e,0xf8,0xe1,0x86,
	0xbd,0xfb,0xff,0xf6,0x9f,0x7f,0xfe,0xfc,
	0xfa,0xef,0x7f,0xff,0xfe,0xfb,0xef,0xb7,
	0xdf,0x79,0x8f,0xb7,0xff,0x7f,0xf7,0xfd,
	0x77,0xff,0x5d,0xf6,0x9f,0xf6,0xdf,0x7f,
	0xe7,0xfd,0xc9,0xfb,0xb7,0xff,0xfb,0xef,
	0xe7,0xf7,0x7f,0xff,0xef,0xff,0xf7,0xdf,
	0x1f,0x6f,0xfe,0x69,0x97,0xfd,0xff,0xb9,
	0xe5,0x97,0xeb,0xaf,0x99,0xff,0x9f,0x5f,
	0x7e,0xf9,0xe5,0x9f,0xfe,0xff,0xff,0xff,
	0xff,0xff,0xff,0xfe,0xff,0xdf,0xbf,0xff,
	0xfe,0xfb,0xee,0xfb,0xe6,0x6b,0xfe,0xbf,
	0xef,0x9b,0x6f,0xfe,0xff,0xef,0xff,0x6b,
	0xbe,0xf9,0xe6,0x9b,0xff,0xf3,0xff,0xaf,
	0xff,0xff,0xff,0xeb,0xbf,0xff,0xfe,0xff,
	0xef,0xbd,0xf7,0xde,0xff,0xff,0xdf,0xfd,
	0xfb,0xed,0x7f,0xff,0xfd,0xff,0xff,0xaf,
	0xdf,0xfd,0xf7,0xdf,0x7f,0xff,0x8b,0xff,
	0xfb,0xfe,0xfb,0xef,0xff,0xff,0xff,0xfb,
	0xff,0xff,0xff,0xff,0xfb,0xef,0xff,0xff,
	0xff,0xff,0xff,0xff,0xfe,0xff,0xff,0xfa,
	0xff,0xff,0xfb,0xef,0xbf,0xff,0xff,0xff,
	0xf5,0xf7,0xd7,0x4f,0x3f,0xfd,0xb5,0x57,
	0x5f,0x7b,0xf9,0xf5,0xd7,0x5f,0x7f,0xfd,
	0xf5,0x5e,0x7d,0xf1,0xf5,0xd7,0x5e,0x7c,
	0xb9,0xe5,0x97,0x5e,0x7f,0xed,0xf5,0xd7,
	0xbf,0xec,0xfe,0xcf,0xff,0xff,0xff,0xff,
	0xef,0xfd,0xff,0xe9,0xf3,0xff,0xf5,0xff,
	0xfb,0xb3,0x3d,0xff,0xff,0xff,0xcf,0x3f,
	0xff,0x7f,0xff,0xcf,0x3f,0xdf,0xdc,0xf3,
	0xcf,0xff,0x4f,0xfe,0x5f,0xfb,0xff,0xff,
	0xfd,0xef,0xff,0xdf,0xfb,0xeb,0xff,0xff,
	0xdf,0xdf,0xf9,0xbf,0xff,0xfe,0xff,0xff,
	0xef,0xff,0xfe,0xff,0x36,0xff,0xbf,0xff,
	0xff,0xff,0xbf,0x5f,0xfa,0xff,0xff,0xfe,
	0xff,0xff,0xff,0xfd,0xfe,0xff,0xff,0xfe,
	0xf7,0xff,0xfd,0x3b,0xf7,0xf2,0xff,0xfd,
	0xdf,0xfc,0xf3,0xff,0xff,0xff,0xff,0xf3,
	0xcf,0x3f,0xff,0xfc,0xff,0xc5,0xff,0xf7,
	0xff,0xdf,0xff,0xfd,0xff,0x7f,0xff,0xff,
	0xf5,0xc7,0x1f,0xff,0xff,0xff,0xfe,0x7f,
	0xdf,0xf4,0xfd,0xff,0xff,0xf7,0xff,0xe7,
	0xff,0xff,0xff,0xff,0xff,0xff,0xef,0xfe,
	0xff,0xfe,0xbf,0xff,0xbf,0xf6,0xff,0xff,
	0xff,0x9e,0xbf,0xff,0xfe,0xff,0xfe,0xff,
	0xbf,0xdf,0x7f,0xff,0xef,0xff,0xff,0xff,
	0xfe,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xf3,0xdf,0xff,0xff,0xff,0xff,0xff,0xff,
	0xf5,0xf7,0xff,0xdf,0xff,0xff,0xff,0xff,
	0xf7,0xff,0xf7,0x7f,0xff,0xff,0xbe,0xff,
	0xff,0xff,0x73,0xff,0xff,0xf7,0xff,0xff,
	0xff,0xbf,0xff,0xff,0xff,0xff,0xff,0x7e,
	0xff,0xff,0xfd,0xff,0xbf,0xff,0xf9,0xff,
	0xdf,0xff,0xef,0xff,0xdf,0xdf,0xff,0xff,
	0xfd,0xff,0x7f,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0x5f,0xfc,0xff,0xff,0xff,0xff,
	0xff,0xfd,0xf7,0xff,0x7e,0xf7,0xbb,0xb7,
	0xff,0x7f,0xff,0xa7,0xff,0x7a,0xff,0xbf,
	0xb7,0xff,0xff,0xff,0xbd,0xff,0xff,0x7f,
	0xff,0xff,0xff,0xee,0xe0,0xff,0xf7,0xdf,
	0xff,0xff,0xf8,0xe3,0xef,0xff,0xbf,0xff,
	0xe3,0xff,0x3f,0x6f,0xbe,0xf3,0xff,0xff,
	0xf8,0xc3,0x87,0x1f,0x6e,0xf8,0xdf,0xff,
	0xfe,0xff,0xff,0xf3,0xff,0x67,0xff,0x7f,
	0xff,0xbe,0xff,0xef,0xff,0x7e,0xff,0xff,
	0xff,0xff,0xfe,0xff,0xed,0xf7,0xdf,0xff,
	0xff,0xef,0xff,0xfe,0xfd,0xef,0xdf,0xfd,
	0xff,0xff,0xff,0xff,0xdf,0xfe,0x7f,0xfa,
	0xf7,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xaf,0xff,0xec,0xbb,0xef,0xff,0xff,0xff,
	0xdd,0xbe,0xbf,0xf6,0xf7,0xfe,0xff,0xff,
	0xff,0xe7,0xed,0xff,0xff,0xff,0xe7,0xff,
	0x95,0xff,0xff,0x7f,0xff,0xff,0xff,0xff,
	0x7f,0xff,0xf5,0xf3,0xcf,0x3f,0xfe,0xff,
	0xff,0x6f,0xff,0xfc,0xf6,0xfe,0xff,0xff,
	0xbf,0xff,0xff,0xfb,0xfb,0xef,0xbf,0xff,
	0xff,0x37,0xfe,0xff,0xfb,0xfb,0xff,0xfe,
	0xfe,0xff,0xfb,0xf7,0xdf,0x7f,0x7f,0xfd,
	0xff,0xff,0xff,0xfb,0xef,0xbf,0xff,0xff,
	0xff,0xbf,0xff,0xff,0xfb,0xfe,0xff,0xff,
	0xff,0xff,0xff,0xc1,0x9f,0xdf,0xff,0xf9,
	0xe5,0xf7,0xff,0x3f,0xff,0xff,0xff,0xff,
	0x3f,0xff,0xfd,0xff,0xff,0xf9,0xfd,0xef,
	0xbf,0xff,0xd9,0xe7,0xaf,0x7c,0xfe,0xff,
	0xff,0xef,0xff,0xff,0xb0,0xf5,0xf9,0xff,
	0x5f,0x7f,0xde,0xff,0xff,0x9b,0xfc,0xff,
	0xff,0xff,0xfb,0xf7,0xff,0xdf,0x5f,0xf7,
	0xff,0xff,0xff,0xcf,0x38,0xff,0xe5,0xf3,
	0xff,0x7f,0xfe,0xff,0xff,0x7a,0xbd,0xff,
	0xf7,0xff,0x7f,0xff,0xff,0xff,0xef,0x9f,
	0xff,0xfe,0xff,0xef,0xff,0xff,0xff,0xef,
	0xff,0xff,0xfe,0xfb,0xef,0x97,0xff,0x7e,
	0xfb,0xef,0xff,0xff,0xff,0xff,0xf7,0x6e,
	0xff,0xff,0xdf,0xff,0xff,0xff,0xf7,0xdf,
	0x3e,0xfb,0xed,0xf7,0xdf,0xfe,0xff,0xff,
	0xd7,0xfe,0xff,0xed,0xbf,0xff,0xbf,0xfc,
	0xf7,0xb3,0x7f,0xff,0xff,0xff,0xff,0xff,
	0x16,0xdf,0xff,0xfb,0xeb,0xbf,0xff,0xfe,
	0xfa,0xfb,0xa7,0xbb,0xfe,0xfa,0xfb,0xbf,
	0xff,0xfe,0xeb,0xbf,0xbf,0xfe,0xfb,0xf7,
	0xf5,0xff,0xfe,0xfa,0xef,0xbf,0xff,0xfe,
	0xfb,0x3f,0xd8,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xef,0xc6,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xaf,0xfe,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xbf,0x9b,0xff,0x7f,0xf9,0xff,
	0x9f,0xff,0xff,0xfb,0xff,0xfd,0xff,0xff,
	0xff,0xff,0x97,0x3f,0xee,0xef,0xff,0xff,
	0x7f,0xff,0x7f,0xfd,0x9f,0xff,0xff,0xfd,
	0xff,0xdf,0x7f,0xff,0x77,0xfe,0xff,0xc7,
	0xef,0x3f,0xfc,0xfc,0x1b,0xff,0xbb,0xff,
	0xfe,0xfb,0x3f,0x7f,0xff,0xfc,0x6f,0x3e,
	0xfc,0xfe,0xff,0xff,0xeb,0xff,0xff,0xfb,
	0xff,0xbf,0xf5,0xff,0xff,0xff,0xf3,0xff,
	0xef,0xbf,0xff,0xfe,0xfb,0x6f,0xbf,0xf9,
	0xe6,0xdb,0xee,0xbf,0xed,0xfe,0xfb,0xbf,
	0xf7,0xfe,0xdb,0xbf,0xff,0xff,0xff,0xef,
	0x6f,0xfb,0xfe,0xfe,0xfb,0xef,0xff,0xaf,
	0xff,0x7f,0xfe,0xfd,0xe7,0x9f,0x7f,0xdf,
	0xd9,0x77,0xde,0x7f,0xf7,0xf9,0xe7,0x9f,
	0xff,0xdd,0xe7,0xdf,0xf7,0xfe,0xfb,0xef,
	0xbf,0x7f,0xfe,0xfb,0xe7,0x9f,0x7f,0xfe,
	0x5f,0xf8,0xff,0xb1,0xf6,0x9f,0x7f,0xfe,
	0xfd,0x56,0xda,0x6f,0xff,0xdd,0xd7,0x1f,
	0x7f,0xfe,0xf7,0x9d,0x7f,0xff,0xfd,0xf7,
	0xdf,0x7f,0xbf,0xfd,0xf7,0x9f,0x7f,0xfc,
	0xf1,0x7f,0x6b,0xff,0xff,0xf7,0xff,0xff,
	0xff,0xfd,0xf6,0xdf,0x3f,0xff,0xfd,0xff,
	0xff,0xff,0xff,0xf7,0xff,0xff,0xff,0x7f,
	0xff,0xfd,0xf7,0xdf,0xfc,0xff,0xfd,0xff,
	0xff,0xff,0xff,0x76,0xff,0xff,0xef,0xff,
	0xdf,0x7f,0xfd,0xef,0xbf,0x7f,0xff,0xfd,
	0xff,0xff,0xff,0x7f,0xff,0xff,0xdf,0xff,
	0xff,0xfd,0xf7,0xdf,0x7f,0xf7,0xff,0xf7,
	0xdf,0xff,0xff,0xff,0xff,0xe8,0xdf,0xc7,
	0xcb,0x7f,0xfc,0xf1,0xe3,0x8f,0x3d,0xfe,
	0xf8,0xf3,0x8b,0x7d,0xfd,0xf1,0x89,0x3f,
	0xf4,0xd8,0xff,0xff,0xff,0xff,0xff,0xe2,
	0xef,0x7d,0xf7,0xd9,0xe7,0x7f,0xc5,0xff,
	0x7f,0x7f,0xfd,0xf7,0xdf,0x77,0xdb,0x7d,
	0xf5,0xd5,0x57,0x5f,0xfd,0xf7,0xdf,0x5f,
	0xed,0xf1,0xd7,0x7e,0xfb,0xfd,0xb7,0xdf,
	0x56,0x79,0xfc,0xf7,0xdf,0x7f,0xff,0x6f,
	0xfe,0xff,0xeb,0x7a,0xbf,0xfe,0xbb,0xfa,
	0xb3,0xff,0x7f,0xff,0xff,0xab,0xff,0x7f,
	0xff,0xeb,0xaf,0x2e,0xf7,0xef,0x7f,0xff,
	0xff,0xff,0x5f,0xbb,0xff,0xff,0xfe,0xfd,
	0xff,0xf0,0xff,0xff,0xd9,0xe7,0x9f,0x7b,
	0xf4,0xd9,0x67,0x1e,0x7f,0xfc,0xd9,0xe7,
	0x9f,0x7f,0xda,0x67,0x1f,0x7d,0xfe,0xf9,
	0xe7,0x9f,0x7f,0xee,0xd9,0xe7,0x9f,0x7f,
	0xfe,0xf9,0x87,0xff,0x7f,0xff,0xfe,0xf7,
	0xdf,0xbf,0xff,0xfe,0xf3,0xcf,0xbf,0xbf,
	0xff,0xe7,0xff,0xbf,0xff,0xfb,0xcf,0x7f,
	0xff,0xfd,0xf7,0xdf,0xbf,0xff,0xfc,0xe7,
	0xff,0x7f,0xff,0x9f,0xfd,0xff,0xfb,0xff,
	0xff,0xff,0x7f,0xff,0xfd,0xd7,0x5f,0x7f,
	0x7d,0xfd,0xff,0xff,0xff,0xfc,0xdf,0x7f,
	0xff,0xfb,0xef,0xbf,0xff,0x7e,0xff,0xf7,
	0xff,0xfe,0xfb,0xef,0xff,0xef,0xff,0xbf,
	0xbf,0xff,0xf9,0xe7,0xff,0xbf,0xff,0xfd,
	0xf7,0xff,0xff,0xff,0xfb,0xef,0xff,0xff,
	0xfa,0xf7,0x7f,0xff,0xfd,0xf7,0xdf,0xef,
	0x3f,0xfd,0xff,0xff,0xdf,0xff,0x47,0xff,
	0xff,0xff,0xef,0xff,0xff,0xff,0xfb,0xed,
	0xa7,0xb7,0xde,0xf3,0xef,0xff,0xff,0xff,
	0xef,0xf7,0xbf,0xfe,0xff,0xff,0xff,0xff,
	0xdf,0xfb,0xfd,0xff,0xff,0xff,0xff,0xbf,
	0xf9,0xff,0xdf,0xff,0xff,0xfe,0xff,0xff,
	0xff,0xff,0xf7,0xff,0xff,0xff,0xff,0xee,
	0xff,0xff,0xff,0xff,0xfb,0xff,0xbf,0xff,
	0xfe,0xff,0xf6,0x7f,0xff,0xff,0xff,0xef,
	0xff,0x5d,0xff,0x7d,0xf5,0xd5,0xd7,0x5f,
	0x7d,0xfd,0xf5,0xd7,0x5f,0x7f,0xf9,0xc5,
	0x97,0x5f,0xed,0xe5,0x97,0x5f,0x7f,0xfd,
	0xb5,0x57,0x5f,0x6f,0xfd,0xf5,0xd7,0x5f,
	0x7f,0xed,0x3b,0xde,0xff,0xfb,0xf3,0xcf,
	0x3f,0xff,0xff,0xf3,0xcf,0x3f,0xff,0x7f,
	0xf3,0x9f,0xff,0xfb,0xf3,0xbf,0x3f,0xff,
	0xfe,0xff,0xff,0xff,0xbf,0xfc,0xe7,0x8f,
	0xff,0xff,0xfe,0x7f,0xe3,0xff,0xff,0xcf,
	0x7f,0xff,0xfd,0xff,0xfb,0xff,0xff,0x7f,
	0xff,0xff,0xdd,0xff,0xff,0xfe,0xf7,0xf5,
	0x7f,0xff,0xff,0xff,0xff,0xff,0xff,0xeb,
	0xdf,0xff,0xff,0xff,0xff,0xb1,0xff,0xff,
	0xff,0xff,0xfe,0x7b,0xff,0x3f,0xff,0xfc,
	0xf3,0xff,0x3f,0xff,0xff,0xff,0x2c,0xbf,
	0xdf,0xf0,0xbd,0xff,0xff,0x7f,0xfe,0xcd,
	0xf7,0xfb,0xfd,0xff,0xdf,0xff,0x3f,0xec,
	0xff,0xff,0xf7,0xff,0xff,0xef,0xf7,0xfe,
	0xff,0xfb,0xff,0xbf,0xf7,0xff,0xff,0xff,
	0xff,0xfd,0x7f,0xdf,0x9f,0xff,0xff,0xff,
	0x5f,0xfc,0xff,0xff,0xff,0xff,0xff,0xff,
	0xea,0xff,0xff,0xef,0xff,0xef,0xff,0xff,
	0xff,0xff,0xef,0xff,0xff,0xeb,0xff,0xef,
	0xbf,0xff,0xdf,0xbb,0xbf,0xff,0xff,0xf7,
	0xbf,0xff,0x7d,0xfb,0xfb,0xff,0xff,0xff,
	0xff,0x17,0xff,0xff,0xff,0xdf,0xff,0xff,
	0xff,0xff,0xff,0xbf,0xff,0xfd,0xfb,0xff,
	0xff,0xff,0xdf,0xff,0xff,0xf7,0xff,0xff,
	0xaf,0xf7,0x3f,0xdf,0xff,0xff,0xbf,0xff,
	0xff,0xff,0xbf,0xf9,0xff,0xff,0xff,0xff,
	0xff,0xff,0xf7,0xfd,0xff,0xff,0xff,0xbf,
	0xff,0xff,0xfe,0xff,0xde,0xff,0xfe,0xf7,
	0xf7,0xff,0xcf,0xfd,0xf7,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xc5,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xef,0xff,0xbf,
	0xff,0xfb,0xff,0xff,0xdf,0xee,0xff,0xff,
	0xde,0xff,0xfb,0xdb,0xcd,0xff,0x7f,0xee,
	0xff,0xff,0xff,0xff,0xff,0xef,0xfe,0x3f,
	0x7e,0xff,0xff,0xcf,0xff,0xff,0xfc,0xfb,
	0xff,0xff,0x7f,0xff,0xfd,0xdf,0xbf,0xdf,
	0xe8,0xfe,0x2f,0xf7,0xff,0xf7,0xff,0x0f,
	0xbe,0xf8,0xff,0xff,0xff,0xff,0x7f,0xf1,
	0xff,0xf7,0xff,0xfd,0xff,0xfd,0xff,0xdf,
	0xdf,0xff,0xff,0xff,0xff,0xff,0xff,0xfb,
	0xff,0xff,0xef,0xff,0xad,0xff,0xbf,0xef,
	0xaf,0xf7,0xef,0xff,0xf7,0xff,0xff,0xff,
	0xb3,0xff,0x2f,0xbf,0xfe,0xfb,0xf7,0xaf,
	0x7f,0xff,0xfa,0xff,0xff,0xfb,0xfe,0xf7,
	0xff,0xdd,0xbd,0xfb,0xff,0x7f,0xff,0xfb,
	0xff,0xdf,0xbf,0xff,0xfe,0xf7,0xbf,0xfe,
	0xff,0x3f,0xf8,0x7f,0xfd,0xf5,0x8f,0xff,
	0x7f,0xfd,0xff,0xd7,0xf7,0xbf,0xff,0xe7,
	0xff,0xfe,0xfb,0xfe,0xdf,0xff,0xff,0xff,
	0xbf,0xd6,0xff,0x7f,0xec,0xa5,0xf9,0xfe,
	0xff,0xfb,0x7f,0xe9,0xff,0xff,0xff,0x7f,
	0xff,0xef,0xfd,0xff,0xff,0xfb,0xef,0xff,
	0xcf,0xff,0xf7,0xef,0xff,0xbf,0xfe,0xaf,
	0xbf,0xff,0xff,0xff,0xef,0xff,0xee,0xfe,
	0xff,0xef,0xff,0xff,0x77,0xfc,0xff,0xcf,
	0x3f,0x7f,0x7e,0xd3,0xff,0xff,0xff,0xff,
	0xf9,0xab,0xbf,0xfe,0x7f,0xff,0x9f,0xdf,
	0xff,0xfe,0xff,0xff,0xff,0x7f,0xde,0x67,
	0xaf,0xfe,0xff,0xff,0xff,0x8f,0x5b,0xff,
	0xff,0xff,0xff,0xe7,0xfb,0xfe,0xff,0xff,
	0xff,0xcf,0x7f,0xfe,0xf9,0xff,0xfd,0xfd,
	0x7c,0xff,0xff,0x3f,0xff,0xff,0xff,0x9b,
	0x3e,0xe3,0xf5,0xff,0x5f,0xff,0xaf,0xc9,
	0xfb,0xbf,0xff,0xfe,0xfb,0xff,0xb7,0xff,
	0xff,0xff,0xff,0xff,0xff,0xfe,0xff,0xff,
	0xff,0x4f,0xed,0xff,0xff,0xff,0x2f,0xfd,
	0xff,0xbf,0x5f,0xfe,0xff,0xff,0xff,0x7f,
	0x0f,0xf6,0x7f,0xff,0xf5,0xd7,0xff,0x3f,
	0xfb,0xff,0xf7,0xff,0xff,0xff,0xfd,0xff,
	0xff,0x7f,0x7f,0xeb,0xff,0xff,0xff,0xff,
	0xe3,0xff,0x77,0xfd,0xf2,0xff,0xff,0xff,
	0xff,0x6f,0xf4,0xff,0x8f,0xbf,0xff,0xfa,
	0xef,0xad,0xff,0xfe,0xfe,0xff,0xbf,0x3f,
	0xff,0xfb,0xef,0xff,0xf6,0xfb,0xef,0xbf,
	0xff,0xfe,0xfa,0x6f,0xad,0xd7,0xfe,0xfb,
	0xef,0xbf,0xff,0x8b,0xfd,0xff,0xff,0xff,
	0xff,0xff,0xef,0xfe,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xef,0xff,
	0xff,0xff,0xff,0xff,0xff,0xef,0xbf,0xfa,
	0xff,0xff,0xff,0xff,0x5b,0xf9,0xff,0xbf,
	0xff,0xff,0xff,0x7b,0xff,0x7f,0xfe,0xfb,
	0xff,0x97,0xff,0x7f,0xf9,0xfb,0xff,0x56,
	0xfe,0xef,0xe7,0xff,0xff,0xff,0xfd,0xff,
	0xd5,0xff,0xfe,0xff,0xff,0x7f,0xe1,0xff,
	0xbf,0xf8,0xff,0xfb,0xcf,0xbb,0xff,0xf0,
	0x9f,0xff,0x7f,0xfc,0xf3,0xf7,0x4f,0xf7,
	0xa2,0xd6,0x2f,0xfe,0xff,0xfe,0xfb,0x5f,
	0xbd,0xaf,0xff,0xfb,0x2f,0xbf,0xff,0x7c,
	0xff,0xff,0xe6,0xfb,0xef,0xbf,0xff,0xb6,
	0xfb,0xef,0xfd,0xf9,0xfe,0xfb,0xef,0xbf,
	0xf7,0xfb,0xef,0xbd,0xff,0xfe,0xfb,0xef,
	0xbf,0xff,0xfe,0xef,0xef,0xbf,0xff,0xfe,
	0xff,0xf9,0xff,0x77,0x9e,0x7f,0xff,0xf9,
	0x77,0x9f,0x7f,0xf6,0xdf,0xe7,0x9f,0x7f,
	0xfe,0xf9,0x9f,0x7f,0xfe,0xf9,0xe7,0x9f,
	0x7f,0xfe,0xf9,0x67,0xfe,0x7f,0xfe,0xf9,
	0xe7,0xff,0x93,0xff,0xdf,0x6f,0xbd,0xfd,
	0xe7,0xdf,0x7d,0xae,0xd5,0x77,0x1f,0x7f,
	0xfd,0xf1,0xe7,0x6f,0xfe,0xf1,0xe7,0x1f,
	0x7f,0xfe,0xfd,0xc7,0xdf,0x7f,0xff,0xf9,
	0xe7,0x9f,0xff,0xa7,0xf6,0xff,0x6f,0xbf,
	0xfd,0xff,0xff,0xff,0xff,0xfd,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xfd,0xff,0xdf,0xff,0xff,
	0xff,0xff,0xff,0xff,0x6f,0xf6,0xff,0xff,
	0xfe,0xf7,0xff,0xf7,0xff,0x7f,0xfd,0xff,
	0xff,0xff,0xff,0xff,0xff,0xf7,0x7b,0xff,
	0xff,0xf7,0xff,0x7f,0xf7,0xff,0xdf,0xff,
	0xff,0xff,0xfd,0xf7,0xdf,0xff,0x83,0xfe,
	0x2f,0xfe,0xfc,0x73,0x1f,0x2f,0xfe,0xf5,
	0xe3,0xe9,0x7d,0xbd,0xf8,0x47,0x1f,0xff,
	0xfd,0x43,0xdf,0x7d,0xfe,0xfd,0xf3,0x1f,
	0x2f,0xfe,0xfe,0xf7,0xdf,0x7f,0xff,0xb7,
	0xfc,0x7f,0xf5,0xd5,0x57,0xff,0x7d,0xb5,
	0xdf,0x57,0x5f,0xfe,0xf3,0xd7,0x7e,0xff,
	0xfd,0xdf,0x1f,0xfb,0xfd,0xb7,0xdf,0x57,
	0xfd,0x6d,0xd7,0xe7,0x7e,0xff,0xfd,0xf7,
	0xff,0xe2,0xff,0xcf,0xbf,0xff,0xfe,0xbf,
	0xfb,0xfe,0xfb,0xfb,0xfa,0xfd,0xaf,0xff,
	0xff,0x6f,0xff,0xfb,0xfa,0xff,0xdf,0xfc,
	0xff,0xde,0xef,0xcf,0xbf,0xff,0xff,0xbf,
	0xff,0xff,0x37,0xff,0x7f,0x1f,0x7d,0xf4,
	0xb9,0xc7,0x9e,0x7f,0xfc,0xd1,0xe7,0x9f,
	0x79,0xee,0xf9,0xe7,0x7f,0xf6,0xf9,0xe7,
	0x9f,0x7f,0xf4,0xf9,0xe7,0x9f,0x7d,0xfe,
	0xf9,0xe7,0x9f,0x3f,0xf8,0xff,0xfe,0xff,
	0xff,0xff,0xfd,0xfb,0xdf,0xbf,0xff,0xff,
	0xff,0xef,0x7f,0xff,0xfd,0x9d,0xbf,0xff,
	0xf9,0xf7,0xff,0xff,0xff,0xfd,0xfa,0xcf,
	0xff,0xff,0xf9,0xff,0xff,0xd1,0xff,0xf7,
	0xdf,0x7f,0xff,0xef,0xdf,0xff,0x7e,0xfd,
	0xff,0xff,0xff,0xff,0xff,0xff,0xef,0xfd,
	0xfd,0xdf,0x7f,0xff,0x7d,0xff,0xef,0xf7,
	0x7f,0xfe,0xf7,0xdf,0x7f,0xff,0xdf,0xfe,
	0xff,0xff,0xff,0xff,0xff,0xfe,0xff,0xef,
	0xff,0xaf,0xff,0xfb,0xfb,0x9f,0x7f,0xfe,
	0xff,0xaf,0xff,0xff,0xfd,0xff,0xff,0xff,
	0xfe,0xfe,0xfb,0xff,0xff,0xff,0xff,0x7f,
	0xf2,0xff,0xbf,0xf7,0xde,0xfb,0xff,0xbf,
	0xff,0xdf,0x7b,0xfd,0xff,0xff,0xfe,0xff,
	0xff,0xff,0x7f,0xff,0xff,0xff,0xff,0xdf,
	0xfb,0xff,0xbf,0xdf,0xff,0xff,0xff,0xff,
	0xff,0x93,0xff,0xff,0xff,0xff,0xef,0xff,
	0x2f,0xff,0xff,0xff,0xff,0xfd,0xff,0xff,
	0xff,0xff,0xff,0xff,0xcf,0xff,0xfb,0xfd,
	0xff,0xff,0xff,0x6f,0xff,0xff,0xff,0xff,
	0xff,0xff,0xdf,0xf4,0xff,0xd7,0x5f,0x7f,
	0xfd,0xf5,0xd6,0x5f,0x7d,0xfd,0xd5,0xd7,
	0x5f,0x7d,0xfd,0xd5,0x5e,0x7f,0xfd,0xf5,
	0xd3,0x4f,0x3d,0xfd,0x35,0x93,0x1f,0x7f,
	0xfd,0xd5,0xd7,0x9f,0xe8,0xff,0xff,0x3f,
	0xff,0xfc,0xfb,0xdf,0x7f,0xff,0xfb,0xff,
	0xdf,0xff,0xfe,0xfe,0xfb,0x3f,0xfe,0xfb,
	0xe3,0x9f,0x37,0xfe,0xfc,0xfb,0xfb,0xff,
	0xd7,0xf8,0xe3,0x8f,0xf7,0x1f,0xfe,0xff,
	0x6f,0x7d,0xf3,0xdf,0xdf,0x77,0xdf,0xff,
	0xfd,0xcf,0xf7,0xff,0xff,0xff,0x7f,0xff,
	0xff,0xf3,0xcf,0x7b,0xff,0xfb,0xde,0xff,
	0xbf,0xff,0xfd,0xf7,0xdf,0xff,0x9f,0xf8,
	0xff,0xff,0xff,0xff,0xff,0xbf,0xff,0xff,
	0xff,0xff,0xbf,0xff,0xfd,0xbf,0xff,0xfe,
	0xf4,0xff,0x7f,0xff,0xff,0xf7,0xbf,0xff,
	0xff,0xff,0xff,0xdb,0x7d,0xff,0xfd,0xbf,
	0xc3,0xb7,0xf7,0x5f,0x7f,0xff,0xff,0xd6,
	0xff,0xf7,0xff,0xf5,0xff,0xfb,0xdf,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xfb,0x7f,
	0xfd,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfe,0x3f,0xbe,0xff,0xfe,0x77,0xdf,0xff,
	0xff,0xfe,0xbf,0x7f,0x9f,0xff,0xff,0xff,
	0xff,0xfe,0xff,0xff,0xff,0xff,0xff,0xbf,
	0xff,0xef,0xfb,0xfe,0xaf,0xff,0xff,0xff,
	0xfb,0xf7,0x7f,0xf3,0xff,0xff,0xf7,0xff,
	0xf7,0x6f,0xfd,0xff,0xff,0xff,0xfb,0xff,
	0xdf,0x7f,0xfd,0xfd,0xdd,0xfe,0xff,0xfd,
	0x57,0xff,0xdf,0xff,0xff,0xff,0xcf,0x35,
	0xff,0xfc,0xff,0xff,0xbf,0xef,0xff,0xfd,
	0xdd,0xff,0xff,0xff,0xff,0xef,0x67,0xff,
	0x7f,0xbf,0xff,0xde,0xfb,0xaf,0xff,0xff,
	0x7b,0xef,0xbf,0xfe,0xff,0x6b,0xff,0xbf,
	0xfc,0xce,0xff,0xef,0xff,0xbf,0xfd,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xbf,0xfe,
	0xff,0xbf,0xff,0xe7,0x9f,0x7f,0xfe,0xe6,
	0xff,0x7b,0xfe,0xf9,0xff,0x7f,0xff,0xdb,
	0xf7,0x7f,0xdf,0xff,0xff,0xfd,0xff,0xe6,
	0xff,0xf3,0xc7,0x1f,0xff,0xfc,0xf3,0xcf,
	0xff,0xff,0xfc,0xd3,0xcf,0x2f,0xff,0xfc,
	0xca,0x3f,0xfb,0xfc,0xf3,0xcf,0x3f,0xff,
	0xfc,0xff,0xcf,0x3e,0xff,0xfc,0xf3,0xff,
	0x47,0xff,0xbf,0xff,0xfe,0xfb,0xef,0x9f,
	0xff,0xfe,0xff,0xef,0x9f,0xff,0xfe,0xfb,
	0xef,0xff,0xfe,0xef,0xef,0xbe,0xff,0xfe,
	0xfb,0xef,0xff,0xff,0xee,0xfb,0xef,0xbf,
	0xff,0xbf,0xfb,0xff,0x75,0xf7,0xdf,0x7f,
	0xff,0xfa,0xf7,0xbf,0x7b,0xff,0xeb,0xe3,
	0x8f,0x3f,0xfe,0xe1,0xdf,0x3e,0xfe,0xfd,
	0xe6,0x9d,0x73,0xee,0xff,0xf7,0x97,0x7e,
	0xfe,0xfd,0xff,0x81,0xff,0xff,0xff,0xff,
	0xff,0xff,0xdb,0xff,0xff,0xf4,0xff,0xdf,
	0x5f,0x7f,0xfd,0xb5,0x4e,0xff,0xdb,0xf5,
	0xd7,0x7f,0xbf,0xf9,0xf7,0xff,0xff,0xff,
	0xfc,0xf7,0xdf,0xff,0x93,0xfe,0xff,0xdf,
	0xef,0xff,0xff,0xf6,0xfe,0xff,0xef,0xff,
	0xbf,0xfe,0x7f,0x7f,0xf7,0xfb,0xdf,0xff,
	0xff,0xff,0xff,0x7a,0xeb,0xaf,0xfe,0xff,
	0xff,0xef,0xaf,0xff,0xfe,0x7f,0xc1,0xff,
	0xff,0xff,0xff,0xff,0xfe,0xfb,0xff,0xff,
	0xff,0xff,0xff,0x77,0xff,0xfd,0xf7,0x7f,
	0xff,0xff,0xf7,0xdf,0x77,0xd3,0xff,0xb7,
	0xbe,0x77,0xff,0x7d,0xf7,0xdf,0xff,0x98,
	0xf5,0xff,0xff,0x5f,0xff,0xff,0xff,0xe7,
	0xff,0xff,0xff,0xff,0xbf,0xfd,0xf7,0xdf,
	0x1f,0xfd,0xff,0xdf,0x4f,0xbf,0xfd,0xff,
	0xdf,0xf9,0xbf,0xfd,0xf7,0xdb,0x4f,0xff,
	0xba,0xbc,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfb,0xff,0xff,0xff,0xff,0x3b,0xfd,0xd4,
	0x53,0xff,0xf7,0xff,0x5f,0xff,0xfd,0xfd,
	0xf7,0x53,0xff,0x3f,0xfd,0xd4,0xff,0xff,
	0xfd,0xf7,0x6d,0xff,0xf7,0xdf,0x7f,0xff,
	0xfd,0xb7,0xdf,0xff,0xff,0xfd,0xf7,0xef,
	0xb3,0xce,0xfa,0xef,0xff,0x9f,0xfa,0xfb,
	0xef,0x7f,0xcf,0xfa,0xff,0xec,0xb3,0xde,
	0xff,0xfb,0xff,0x3e,0xff,0xff,0xfa,0xeb,
	0xaf,0xbf,0xff,0xfe,0xfb,0xff,0xbf,0xff,
	0xfc,0xef,0xfd,0xf7,0x7e,0xef,0xbf,0xf7,
	0xdf,0xfb,0xef,0x2d,0xdb,0xff,0xfb,0xff,
	0xb7,0xf5,0xdf,0xff,0x3f,0xdb,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0x7f,0xff,
	0xff,0xff,0x7f,0xef,0xbf,0xf7,0x79,0xff,
	0xbf,0xff,0x9e,0x7d,0xff,0x97,0xff,0xff,
	0xff,0xef,0x9d,0xff,0xbe,0xbd,0x99,0xff,
	0xff,0xff,0xff,0xff,0xff,0xfe,0xff,0xff,
	0xfb,0xff,0xfa,0x5f,0x7f,0xff,0xfd,0xcf,
	0xff,0x97,0xfd,0xf7,0x5f,0x7f,0xff,0xf7,
	0xff,0x5f,0xff,0x7d,0xf5,0xf7,0xff,0xc7,
	0xfe,0xff,0xfb,0xef,0xbf,0xff,0xe3,0xff,
	0xff,0xff,0xff,0xf7,0xcf,0xda,0x7b,0xef,
	0xfd,0xde,0xff,0xe5,0xbd,0xf7,0xda,0xbf,
	0xef,0xfd,0xfb,0xda,0x7b,0xaf,0xbd,0xf7,
	0x7f,0xf6,0xff,0x6f,0xbe,0xfd,0xfe,0xfb,
	0xef,0xbf,0xfd,0xff,0x7b,0xef,0xff,0xff,
	0xff,0xdf,0xfe,0xf9,0xfb,0x9f,0x7f,0xff,
	0xff,0xf6,0xdf,0xee,0xff,0xed,0xb7,0xff,
	0x7f,0xfb,0x9f,0xff,0x7f,0xf7,0x7d,0x77,
	0x9f,0x7d,0xfe,0x79,0x77,0x9e,0x7f,0xfe,
	0xff,0x6f,0xff,0xf9,0xff,0xcf,0xbf,0xff,
	0xd7,0xff,0x67,0xbc,0x79,0xfe,0xfb,0x7f,
	0xbf,0xff,0xe6,0xdf,0xf8,0xff,0xdd,0xf7,
	0xdb,0x6d,0xdd,0xf5,0xd7,0x9f,0x6b,0xfd,
	0xf5,0xf7,0xdf,0x7d,0xff,0xf6,0xdf,0x7f,
	0xff,0xdd,0xf6,0xdb,0x7f,0xff,0xf9,0xf7,
	0xdf,0x67,0xbf,0xfd,0x7f,0x68,0xff,0xfd,
	0xf7,0xcf,0xbf,0xff,0xff,0xf7,0xff,0x7f,
	0xff,0xfd,0xf7,0xff,0xff,0xff,0xfd,0xff,
	0xf3,0xff,0xff,0xf6,0xef,0xff,0xff,0xff,
	0xff,0xff,0xbf,0xff,0xfe,0xff,0x76,0xff,
	0xfd,0xf7,0x7f,0xff,0xff,0xff,0xdf,0xff,
	0xff,0xfe,0xfd,0xef,0xff,0xff,0xff,0xbb,
	0xff,0xdf,0xff,0xff,0xf7,0xff,0xff,0xff,
	0xff,0xfd,0xff,0xff,0xff,0xff,0xff,0xff,
	0xeb,0xff,0x73,0xcb,0x2d,0xbe,0xd8,0xe3,
	0x8b,0xf7,0xbf,0xf8,0xe3,0x6f,0xaf,0xbd,
	0xf6,0x6b,0xff,0xbf,0xf6,0xdb,0x6b,0x2d,
	0xbe,0xf6,0xe7,0x6b,0xaf,0xbd,0xf6,0xda,
	0x7f,0xd1,0xff,0x77,0x5f,0x7d,0xf5,0xd5,
	0x5f,0x5f,0xfd,0xfb,0xd7,0x77,0x5f,0x7d,
	0xd5,0xd5,0xdf,0xfd,0xf7,0x55,0x5f,0xdf,
	0x7d,0xd7,0x55,0x7f,0x5f,0x75,0xf5,0x5d,
	0x77,0xfd,0xcf,0xfe,0xbf,0xfe,0xbb,0xef,
	0xff,0xfe,0xfc,0xfe,0xef,0xaf,0xf6,0xfc,
	0xfb,0xfb,0xbf,0xff,0xfb,0xff,0xbf,0xbf,
	0xfe,0xfa,0xef,0xef,0xf7,0xff,0xfb,0xef,
	0xaf,0xff,0xfe,0xff,0xf4,0xff,0xf5,0xf1,
	0xc7,0x9e,0x7f,0xf6,0x99,0x67,0x9f,0x7d,
	0xfc,0xd1,0xc1,0x87,0x1d,0xfe,0xe1,0x9f,
	0x1d,0x74,0xf0,0xe1,0x1e,0x1f,0xee,0xf1,
	0x61,0x07,0x1f,0x7e,0xf8,0xbb,0xff,0xbf,
	0xff,0xfd,0xf6,0xef,0xbf,0xff,0xfe,0xff,
	0xef,0xbf,0xff,0x7f,0xfb,0xf9,0x97,0x7e,
	0xe7,0xfb,0xb7,0xdf,0x7e,0xfa,0xef,0x77,
	0xbf,0x7f,0xfe,0xed,0xa7,0xdf,0x5f,0xfd,
	0x7f,0xf7,0xcf,0x1f,0xdf,0x7d,0xf7,0xdd,
	0xff,0xff,0xfd,0xf5,0x7d,0xd9,0xc5,0x97,
	0x7d,0xf9,0xdf,0x97,0x5d,0x7e,0x71,0x7f,
	0x97,0xff,0x77,0xf1,0xe5,0x17,0x5f,0xfe,
	0xe0,0xff,0xff,0xbf,0xff,0xff,0xfb,0xef,
	0xff,0xff,0xfe,0xfb,0xef,0xff,0xe7,0x9f,
	0x7f,0xb6,0xe7,0xf7,0x7f,0xfe,0xb9,0xe7,
	0xfe,0x7f,0xde,0xff,0xe7,0x9f,0x7b,0xee,
	0xf9,0x7f,0xff,0xff,0x7f,0xfb,0xef,0xf7,
	0xdf,0xff,0xff,0xf7,0xff,0x7f,0x7f,0xcf,
	0x3f,0xff,0xfc,0xcf,0xff,0xff,0xfc,0xf3,
	0xcf,0xff,0xff,0xfc,0xfb,0xcf,0x3f,0xff,
	0xfc,0xf3,0xff,0xfa,0xff,0xef,0xff,0xff,
	0xff,0xf7,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfb,0xef,0xbe,0xfc,0x93,0xff,0xbe,0xfb,
	0xe4,0xcf,0xef,0xff,0xff,0xff,0xfb,0xff,
	0xff,0xfc,0xee,0xfe,0x5d,0xff,0x7f,0xfd,
	0xf5,0x57,0x5f,0x7d,0xf5,0xf5,0xd7,0x5f,
	0x7d,0xfd,0xf5,0xd7,0x5b,0xaf,0xf5,0x97,
	0x5d,0x7d,0xfd,0xf5,0x57,0x5f,0x7f,0xf5,
	0xf5,0xd7,0x5f,0x6f,0xfd,0x7b,0xfe,0xff,
	0xbc,0x73,0xff,0xfb,0xfe,0xf4,0xef,0xff,
	0xff,0xff,0xdf,0xfd,0xf7,0xff,0xdf,0xdf,
	0xbf,0xff,0x7f,0x7f,0xff,0xf7,0xff,0xff,
	0xfe,0xfd,0xff,0xff,0xff,0xff,0x7f,0xe5,
	0xff,0xfd,0xcd,0xf7,0xff,0xff,0xf3,0xff,
	0xff,0xff,0xff,0xff,0xfd,0xf7,0xff,0xbf,
	0xff,0xff,0xff,0xdf,0x7f,0xff,0xff,0xff,
	0xff,0xfe,0xfb,0xef,0xee,0xff,0xff,0xfb,
	0x91,0xff,0xbf,0xff,0xef,0x7f,0xff,0xfd,
	0xff,0xff,0xff,0xff,0xff,0xff,0xbf,0xff,
	0xff,0xed,0xef,0xf7,0xff,0xff,0xff,0xff,
	0xff,0x7f,0xd7,0xfe,0xff,0x7f,0xff,0xff,
	0xff,0xff,0xfd,0xff,0xfd,0xf7,0xd7,0x5f,
	0x56,0xf9,0xf7,0x7f,0x5f,0x7f,0xfc,0xff,
	0xf7,0x57,0xff,0xff,0xff,0xdf,0x7f,0xfd,
	0xfd,0xfe,0x5f,0xff,0xdf,0xf5,0xd7,0x5f,
	0xef,0xef,0xff,0xe0,0xff,0xfb,0xef,0xfb,
	0x7e,0xff,0xff,0xdf,0xff,0xdf,0xfa,0xfb,
	0xf7,0xfe,0xde,0x77,0xf3,0xf7,0xff,0x77,
	0xdb,0xbe,0xff,0xb7,0xfb,0xff,0xbf,0xff,
	0xdd,0xfb,0x7f,0xff,0x47,0xff,0xff,0xff,
	0xff,0xfd,0xff,0xff,0xff,0xdf,0xfe,0xff,
	0xfd,0xef,0xfd,0xff,0xf7,0xdf,0xdf,0xf7,
	0xff,0xfd,0xff,0xdf,0x7f,0xf7,0xff,0x7f,
	0xef,0xbd,0xf7,0xff,0xff,0x3f,0xfb,0xff,
	0xdf,0xff,0xff,0xff,0xff,0xbf,0xeb,0xff,
	0xfd,0xff,0xff,0xeb,0xff,0xff,0xf7,0xff,
	0x6f,0xff,0xff,0xbf,0xdf,0xff,0xf7,0xfd,
	0xf7,0xff,0xff,0xff,0xff,0xff,0xff,0xc7,
	0xff,0xff,0xff,0xfd,0xff,0x9f,0xff,0xff,
	0xfb,0x7f,0xff,0x7f,0xff,0xfe,0xff,0xff,
	0xff,0x7b,0xfb,0xbf,0xff,0xff,0xff,0xf7,
	0xfe,0xff,0xff,0xfd,0xff,0xde,0xff,0xff,
	0xff,0xfe,0xff,0xff,0xfa,0xf7,0xff,0x3f,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xfd,
	0x87,0x1f,0xff,0xf3,0xfb,0xbf,0xff,0xdc,
	0xf1,0x0f,0xef,0xff,0xff,0xff,0xaf,0xff,
	0xff,0x7f,0xf6,0xff,0xff,0xff,0xef,0xfe,
	0xff,0xfb,0xff,0xff,0xef,0xff,0xff,0xff,
	0xff,0xff,0xfb,0xff,0xa7,0xbf,0xff,0xfc,
	0xef,0xbf,0xff,0x7e,0xfe,0xff,0xef,0xff,
	0xff,0xff,0xff,0xab,0xef,0xff,0xff,0xff,
	0x77,0xff,0x17,0xff,0xfe,0xfb,0xff,0xff,
	0xdf,0x7c,0xff,0x7d,0xf7,0xff,0xfd,0xff,
	0xff,0x7f,0x7e,0xdd,0xff,0x7f,0xdf,0xff,
	0xd7,0xff,0xfb,0xff,0xdf,0xf8,0xfe,0xff,
	0xff,0xff,0xff,0xff,0xff,0xf7,0xdf,0xff,
	0xff,0xfd,0xf1,0xff,0xff,0xff,0xbf,0xff,
	0xfe,0xfb,0xff,0xf6,0xdf,0xff,0xff,0xff,
	0xff,0xff,0xef,0xef,0xff,0x7f,0xe9,0xfb,
	0xff,0xff,0xff,0xff,0x7f,0xfb,0xaf,0xbf,
	0xfe,0xef,0xef,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xc7,0x7f,0xff,0xbf,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x77,
	0xfd,0xff,0xff,0xff,0x5f,0xfe,0xff,0xfc,
	0xef,0xff,0x7a,0xff,0xff,0xbf,0xfe,0xfa,
	0xaa,0xbf,0xfe,0xfa,0xeb,0xfb,0xff,0xff,
	0xff,0xcb,0xff,0xbf,0xfe,0xfa,0xf9,0xff,
	0xcf,0x59,0xff,0xff,0xff,0x7f,0xd7,0xff,
	0xf7,0xf3,0x9f,0xd7,0xff,0xff,0x7f,0xf9,
	0xe7,0x9f,0xff,0xf5,0xe7,0x9f,0xff,0xff,
	0xff,0xff,0x9f,0x78,0xfd,0xf9,0xe7,0xcf,
	0xff,0xaf,0xcf,0xfb,0xbf,0xff,0xff,0xff,
	0xff,0xff,0xff,0xaf,0xff,0xdf,0xff,0xeb,
	0xff,0xff,0xff,0xff,0x2f,0xf9,0xff,0xff,
	0x7f,0xff,0xfd,0xef,0xd7,0xff,0xff,0xff,
	0xff,0xff,0x7f,0x0f,0xf6,0x7f,0xfb,0xff,
	0xff,0xff,0x7f,0x9f,0x7f,0xff,0xff,0xff,
	0xdf,0xff,0xff,0xff,0xff,0xff,0xcb,0xff,
	0xff,0x9f,0xf9,0xe7,0xff,0xbe,0xfe,0xff,
	0xff,0xff,0xff,0xff,0x6f,0xe2,0xbf,0xae,
	0xff,0xfe,0xfb,0xef,0xef,0xff,0x76,0xfe,
	0xef,0xff,0xbd,0xec,0xfb,0xef,0xfb,0x7e,
	0xb6,0xef,0xbf,0xbf,0xfe,0xfa,0xef,0xb5,
	0xff,0xfe,0xfb,0xef,0xbf,0xff,0xbf,0xfd,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x77,
	0xeb,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0x7f,0xfd,0xff,0xff,0xff,0xdf,0xff,
	0xaf,0xfe,0xff,0xff,0xff,0xff,0xff,0x9b,
	0xf9,0xff,0xff,0x5f,0x7e,0xf9,0xe5,0xff,
	0xff,0xfd,0xff,0xad,0xff,0xbf,0xff,0xff,
	0xff,0xff,0xdf,0xff,0xeb,0xff,0xb7,0xde,
	0xff,0x7d,0xfd,0xff,0xff,0xfb,0xff,0xff,
	0x7f,0xee,0xff,0xbf,0xfc,0xf1,0x57,0x5f,
	0xfd,0xfd,0xff,0xff,0x5f,0xff,0xbf,0xfe,
	0xfb,0xef,0xf7,0x56,0xff,0x6f,0xbe,0xbf,
	0xe2,0xfe,0xff,0xab,0xff,0xfe,0x9b,0xef,
	0xbf,0xff,0x16,0x9e,0xed,0xf6,0xfb,0xef,
	0xbf,0xff,0xd6,0xff,0xff,0xbf,0xf7,0xe7,
	0xdf,0xee,0xbf,0xff,0xfb,0xbf,0xbf,0xf7,
	0xfe,0xdb,0x6f,0xfe,0xfe,0xfe,0xfb,0xef,
	0xbd,0xff,0xfe,0x37,0xfa,0xde,0x77,0x9d,
	0x7f,0xfe,0xf9,0xe7,0xfd,0xff,0xf7,0xf9,
	0xff,0xff,0x7d,0xfe,0xf9,0x9d,0x7f,0xf6,
	0xf9,0xe7,0x9f,0x77,0xfe,0xdf,0xe7,0x9f,
	0x7f,0xfe,0xf9,0xe7,0xff,0x87,0x7f,0xdf,
	0x77,0xfc,0xf1,0xc7,0x5f,0x6b,0xff,0x7d,
	0xc7,0xdb,0x6f,0xf7,0xf9,0xe7,0x7f,0xfe,
	0xd1,0xe7,0x9f,0x6f,0xff,0xf5,0x76,0x9f,
	0x7f,0xfe,0xf9,0xe7,0x9f,0xff,0x67,0xf6,
	0xdf,0xff,0xff,0xff,0xff,0xff,0x7f,0xff,
	0xfd,0xff,0xdb,0x6f,0xff,0xff,0xff,0xff,
	0xff,0xfd,0xff,0xff,0x6f,0xff,0xff,0xfe,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xef,
	0xf3,0xdf,0xff,0xff,0xff,0xff,0xff,0x7f,
	0xff,0xfb,0xff,0xbf,0xff,0xfe,0xff,0xfd,
	0xf7,0x7b,0xfb,0xff,0xf7,0xdf,0xfe,0xff,
	0xff,0xff,0xdf,0x7f,0xff,0xfd,0xf7,0xdf,
	0xff,0x9b,0xff,0x3f,0xf6,0xd1,0xc7,0x5f,
	0x3d,0xde,0xfe,0xff,0x1f,0xbf,0xff,0xdd,
	0x67,0x9f,0xff,0xd9,0xf2,0x9f,0x27,0xfe,
	0xd8,0xe3,0xef,0x7f,0xfe,0xfd,0xe7,0x9f,
	0x7f,0xff,0x17,0xfd,0x7f,0xf7,0xdf,0x7f,
	0xfb,0xfd,0xf5,0xcf,0xff,0xff,0xfd,0xf7,
	0xcf,0x7f,0xff,0xfd,0xdf,0x57,0xfb,0x6d,
	0xf1,0x55,0x5f,0xdf,0xee,0xb7,0xdf,0x7f,
	0xff,0xed,0xf7,0xff,0xe0,0xff,0xaf,0x7f,
	0xfb,0xef,0xbf,0xab,0xbf,0x9d,0xdf,0xaf,
	0x7f,0xff,0xf5,0xdf,0x7f,0xff,0xff,0xc6,
	0x7f,0xeb,0xff,0xfe,0xfe,0x3b,0xff,0xfe,
	0xf3,0xcf,0xbf,0xff,0xfe,0x1f,0xff,0xdf,
	0x9f,0x7f,0xee,0xf9,0x47,0x9f,0x7f,0xfe,
	0xf9,0xe7,0x9f,0x7f,0xee,0xb9,0xe7,0x7b,
	0xfc,0xb9,0x47,0x9e,0x7d,0xe6,0xf1,0xe7,
	0x9e,0x7f,0xee,0xb9,0xe7,0x9f,0x7f,0xfb,
	0xff,0xfa,0xdf,0x7f,0xfe,0xff,0xfb,0xef,
	0xff,0xff,0xfd,0xff,0xef,0x7f,0xff,0xfd,
	0x9d,0x2f,0xff,0xf9,0xf7,0xef,0xaf,0xbf,
	0xfc,0xf7,0xff,0x7f,0xfe,0xfd,0xff,0xff,
	0xcd,0xff,0xd7,0xff,0xfd,0xf7,0xdf,0xdf,
	0xff,0xff,0xff,0xdf,0xff,0xff,0xff,0xff,
	0xef,0xff,0x7f,0xf9,0xff,0xbf,0x5f,0x7f,
	0xfd,0xe5,0xff,0xff,0xff,0xff,0xef,0xff,
	0xff,0x1f,0xfe,0xff,0xfb,0xf7,0xff,0xff,
	0xff,0xfb,0xeb,0xff,0x7f,0xff,0xfb,0xef,
	0xdf,0xff,0x7f,0xff,0xaf,0xff,0xff,0xff,
	0xef,0xbf,0xbf,0xff,0xfd,0xff,0xff,0xff,
	0xff,0xff,0x7f,0xf0,0xff,0xfe,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xfb,0xef,0xff,0xde,0xff,0xef,
	0xbf,0xdf,0xff,0x7f,0xff,0xbf,0xff,0xfe,
	0xfb,0xef,0xbf,0xff,0xbb,0xff,0xff,0xff,
	0xfb,0xff,0x7f,0xff,0xfd,0xff,0xf3,0xff,
	0xff,0x3f,0xfd,0xff,0xbf,0xff,0xfb,0xe6,
	0x7f,0xff,0xbf,0xdc,0xee,0xff,0xff,0xff,
	0xfb,0xef,0xbf,0xff,0xfe,0x7f,0xfc,0xdf,
	0xd7,0x5f,0x7e,0xf9,0xe5,0x97,0x5f,0x7e,
	0xed,0xf5,0xd7,0x5f,0x7f,0xfd,0xb5,0x5e,
	0x7f,0xfd,0xf5,0xd7,0x1b,0x6d,0xfd,0xf5,
	0xd7,0x5f,0x7f,0xfd,0xf4,0xd7,0xbf,0xe0,
	0xff,0xcf,0xbf,0xff,0xff,0xe3,0xfe,0xff,
	0xff,0xfc,0xeb,0xff,0xdf,0xff,0xdc,0xff,
	0xff,0x2f,0xfc,0xef,0xff,0xfd,0xfe,0xff,
	0xf3,0xcd,0x3f,0xfe,0x7f,0xff,0x8f,0xff,
	0x57,0xfe,0xcf,0xef,0xff,0xfe,0xef,0xf7,
	0xdf,0xff,0x7b,0xff,0xff,0xff,0xff,0xfa,
	0xfb,0xff,0xff,0xfa,0xbb,0xff,0xff,0xff,
	0xff,0x73,0x3a,0x3f,0x7f,0xff,0xff,0xdb,
	0xbf,0x5f,0xfb,0xff,0xfb,0xed,0xff,0x7f,
	0xfd,0xff,0xff,0xbf,0xff,0xfe,0xff,0xff,
	0xbf,0xff,0xff,0x5e,0xdf,0xff,0xff,0xff,
	0xff,0xfb,0xdf,0xfe,0xff,0xf6,0xff,0xff,
	0xaf,0xfd,0xff,0x97,0xff,0xd7,0xff,0xbf,
	0xff,0xef,0xf7,0xfd,0x7f,0xfd,0xff,0xdf,
	0x5b,0xff,0xff,0xff,0xfe,0xff,0xff,0xff,
	0xf7,0xff,0x67,0xfc,0xfd,0xfd,0xff,0xff,
	0xbf,0xff,0xff,0xff,0x7f,0xfa,0xbf,0xfa,
	0xff,0xff,0xff,0xff,0xfd,0xff,0xe7,0xfe,
	0xbb,0xaf,0xf9,0xfe,0xff,0xff,0xff,0xff,
	0xfb,0xef,0xbd,0xff,0xf7,0xbb,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xe2,0x7f,
	0xff,0xff,0xff,0xfb,0xdf,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xfd,
	0xfd,0xf7,0xdf,0x6f,0xff,0xfe,0xf7,0xdf,
	0xff,0xbd,0xfe,0xf7,0x6f,0xbf,0xff,0xba,
	0xff,0xff,0xfd,0xff,0xdf,0xff,0x7f,0xdf,
	0xfd,0xf7,0xff,0xff,0xff,0xeb,0xff,0xff,
	0xef,0xfd,0xff,0xdf,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xf7,0xff,0xf7,0xff,
	0x7f,0xfc,0xff,0xff,0xdf,0xff,0xff,0xfb,
	0xef,0xff,0xfa,0xff,0xff,0xaf,0xbf,0x7f,
	0xff,0xfb,0xbf,0xff,0xff,0xdb,0xe7,0xbf,
	0xfe,0xfb,0xff,0xff,0xff,0xff,0xfe,0xff,
	0xff,0xff,0xe9,0xff,0xff,0xff,0xff,0xff,
	0xff,0xfd,0xc5,0x3f,0xff,0x7c,0xff,0xcf,
	0x3f,0xfe,0xf8,0x8a,0xff,0xff,0xff,0xeb,
	0xff,0x3f,0xff,0xdf,0xff,0x8f,0xef,0xbf,
	0x7a,0xff,0xff,0x07,0xff,0xff,0xff,0xff,
	0xff,0xfb,0xff,0xff,0xf6,0xfb,0xef,0xff,
	0x7f,0xf3,0xef,0xbf,0xff,0xfb,0xff,0xfd,
	0xff,0xf7,0xff,0xfb,0xff,0xf7,0xff,0xfb,
	0xff,0xff,0xff,0xff,0xbf,0xf8,0xff,0xf7,
	0xfe,0x77,0xff,0x7d,0xdf,0xe3,0x9f,0x7f,
	0xff,0x7f,0xdd,0xf7,0xd7,0x7f,0xfd,0xff,
	0xff,0x7d,0xff,0xfd,0x87,0xff,0xff,0xf7,
	0xfd,0xf7,0xdf,0xff,0xff,0xff,0xc1,0xff,
	0xfb,0xff,0xfb,0xff,0xff,0xff,0x5f,0xfb,
	0xfd,0xbf,0xff,0xfe,0xfb,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xef,0xfd,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x15,
	0xfe,0xff,0xdf,0xff,0x7f,0xff,0xff,0xff,
	0xbf,0xf3,0xff,0xff,0xff,0x7f,0x7d,0xfe,
	0xfe,0xff,0xff,0xff,0xff,0xff,0xff,0xef,
	0xff,0xff,0xfb,0xaf,0xff,0xfe,0xff,0xef,
	0xff,0xc7,0xff,0xff,0xfa,0xfe,0xff,0xef,
	0x3f,0xfe,0xff,0xff,0xff,0xff,0xfc,0xff,
	0xff,0xff,0xff,0xeb,0xff,0xff,0xfd,0xff,
	0xfe,0xe4,0xbf,0xfe,0xfa,0xea,0xfb,0xff,
	0xff,0xdf,0xb6,0xf5,0xcf,0x3f,0xff,0xfc,
	0xf3,0xff,0xff,0xff,0x7f,0xfe,0xf5,0xe7,
	0xff,0xff,0xff,0xff,0x9f,0xff,0xff,0xef,
	0xff,0xff,0x77,0xfd,0xf9,0xe7,0x9f,0xff,
	0xff,0xfc,0xff,0x3b,0xbd,0x6f,0xbb,0xed,
	0xb6,0xdb,0xfe,0xfb,0xff,0xff,0xff,0xff,
	0xfb,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0x7f,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0x6b,0x7f,0xed,
	0x35,0xd7,0x5a,0xeb,0xbf,0xdf,0x7f,0xff,
	0xfd,0xb7,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xfb,0xff,0x7f,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xdf,0x2b,0xff,
	0x7f,0xa9,0xdd,0x76,0xdb,0xec,0xff,0xfb,
	0xef,0xbf,0xfe,0xfa,0xef,0xbf,0xff,0xfe,
	0xef,0xbf,0xff,0xde,0xfb,0xef,0xef,0xff,
	0xfe,0xfb,0xef,0xbf,0xff,0xfe,0xfb,0xdf,
	0xd9,0xff,0xde,0xf9,0xe7,0x9f,0x57,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xee,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xbf,0x90,0xe7,0xf7,0xc2,0x5f,0x7f,0xfd,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xfd,
	0xff,0xff,0xef,0xff,0xff,0xff,0xf6,0xfe,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x9f,
	0xff,0xff,0x1d,0x7e,0xff,0xfe,0xfb,0xef,
	0xbf,0xff,0xff,0xff,0xff,0xff,0xfe,0xfb,
	0xff,0xbf,0xff,0xfe,0xef,0xbf,0xff,0xa2,
	0xfb,0xef,0xff,0xff,0xfe,0xfb,0xef,0xbf,
	0xff,0xf3,0xfb,0xbf,0xf5,0xff,0xbd,0xf7,
	0xdc,0x73,0xcf,0x3f,0xbf,0xff,0xe6,0xdb,
	0xee,0xbf,0xff,0xfe,0xfb,0xbf,0xff,0xfe,
	0xfb,0xef,0xbf,0xff,0xb6,0xfb,0xef,0xbf,
	0xff,0xfe,0xfb,0xef,0xff,0x8f,0xff,0xff,
	0xff,0xcb,0x2f,0xff,0xbc,0xf3,0xd9,0x67,
	0xdf,0x7f,0xff,0xf9,0xe7,0x9f,0xff,0xf9,
	0xe7,0x9f,0x7f,0xfe,0xf9,0xe7,0x9f,0x7f,
	0xfe,0xf9,0xe7,0x9f,0x7f,0xfe,0xbf,0xf8,
	0x7f,0xfd,0xf5,0xd7,0x5f,0xff,0xbd,0xd6,
	0x5d,0x7d,0xbf,0xfd,0xc7,0x9f,0x7f,0xfe,
	0xe6,0x9f,0x7f,0xfe,0xe9,0xe7,0x5f,0x7f,
	0xfe,0xf9,0xe7,0x9f,0x7f,0xfc,0xf9,0x7f,
	0x64,0xff,0x6b,0xaf,0xbd,0xf6,0xda,0x7f,
	0xf7,0xff,0x7f,0xbf,0xfd,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0x3e,0xff,0xff,0xff,0xfb,0xef,0xbf,
	0xff,0xf7,0xff,0xff,0xfe,0xfb,0xff,0xff,
	0xdf,0x7f,0xbf,0xf7,0xdf,0x7f,0xff,0xfd,
	0xf7,0xff,0x7f,0xff,0xfd,0xf7,0xdf,0xff,
	0xff,0xfd,0xbf,0xe8,0xff,0x7f,0xff,0xfd,
	0xf7,0xff,0xff,0x8f,0x3f,0xff,0xf8,0xe3,
	0x9f,0x7d,0xf7,0xfd,0xdf,0x7f,0xf6,0xfd,
	0xf7,0xdf,0x2d,0xff,0xfd,0xe7,0x9f,0x7f,
	0xf6,0xd9,0xe7,0xff,0xd1,0xff,0x7f,0xfb,
	0xe5,0xd7,0xdf,0x7f,0x5f,0xfd,0xf5,0xdd,
	0x77,0xfb,0xfd,0xf7,0xdf,0xff,0xfd,0xf7,
	0xdf,0x7f,0xff,0x7d,0xf5,0xdf,0x7e,0xff,
	0xed,0xf7,0xdf,0x7f,0xff,0xef,0xfe,0xf7,
	0xf7,0xda,0x6b,0xae,0xb9,0xf6,0xff,0xeb,
	0xbf,0xbb,0xee,0xb7,0xff,0xff,0xf7,0x3f,
	0xff,0xfe,0xf3,0xff,0xff,0xef,0xfe,0xf3,
	0xdf,0x7f,0xff,0x7f,0xff,0xff,0x7f,0xf3,
	0xff,0xbf,0xd0,0x46,0x1b,0x6d,0xb4,0xb1,
	0x47,0x1f,0x7b,0xe4,0xf9,0xe7,0x9f,0x7f,
	0xfe,0xe7,0x8e,0x7f,0xfe,0xf9,0x67,0x8f,
	0x7f,0xee,0xb9,0xe7,0x9e,0x7f,0xee,0xf9,
	0x93,0xff,0xaf,0x9f,0xfe,0xfa,0xeb,0xaf,
	0xff,0xfe,0xfb,0xeb,0xbf,0xff,0xff,0xf7,
	0xdf,0xdf,0xfd,0xf7,0x9d,0x7f,0xff,0xff,
	0xfe,0xfd,0x7f,0xff,0xfd,0xf7,0x9f,0x7f,
	0xff,0x1f,0xfd,0x7f,0xf5,0xd5,0x57,0x5f,
	0x7d,0xf5,0xd5,0x5f,0x5f,0x7d,0xf7,0xff,
	0xbf,0xff,0xfe,0xef,0xff,0xfb,0xff,0xfb,
	0xff,0x77,0xff,0xff,0xff,0xef,0xff,0xff,
	0xff,0xff,0xff,0xe8,0xff,0xeb,0xfd,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xf7,0xff,0xfd,0xff,
	0xff,0xff,0xff,0xdf,0xff,0xdf,0xff,0xff,
	0xfd,0xff,0xdf,0xff,0x73,0xfb,0x5f,0xef,
	0xff,0xff,0xff,0xff,0x7f,0xff,0xff,0xff,
	0xff,0xff,0xef,0xbf,0xff,0xfe,0xef,0xbf,
	0xfe,0xfe,0xfb,0xef,0xff,0xfe,0xfe,0xfb,
	0xef,0xbf,0xff,0xfe,0xfb,0xbf,0xfa,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xef,0xff,0xef,0x7f,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xf2,0xff,0xff,0xff,0xfb,
	0xef,0xbf,0xff,0xff,0xff,0xef,0xff,0x4f,
	0xff,0x7f,0xf5,0xf1,0xc7,0x5e,0x7f,0xfd,
	0xf5,0xd7,0x5f,0x3b,0xfd,0xd4,0xd7,0x5f,
	0xed,0xe5,0x57,0x5f,0x6f,0xf5,0xf5,0xd7,
	0x5f,0x7b,0xed,0xb5,0xd7,0x5f,0x7f,0xed,
	0xaf,0xfe,0xcf,0xfb,0xef,0xaf,0xbf,0xfa,
	0xfa,0xff,0xcf,0xff,0xff,0xfc,0xef,0x0f,
	0xf7,0xff,0xe3,0x4d,0x3f,0xfa,0xf8,0xc3,
	0xcf,0x37,0xfe,0xfc,0xff,0xcf,0xff,0xff,
	0xfc,0xff,0xe2,0xff,0xff,0xff,0xfb,0xff,
	0xff,0xff,0xff,0xb7,0xef,0x7f,0xfd,0xff,
	0xdf,0xff,0xdf,0xf6,0xbb,0x3f,0xbf,0xfd,
	0xf7,0x3b,0x7f,0xff,0xf3,0xff,0x3f,0xff,
	0xff,0xfc,0xff,0xa5,0xff,0xfb,0xff,0xff,
	0xfb,0xef,0xbf,0xff,0xff,0xfb,0xff,0xbf,
	0xf7,0xff,0xfd,0xff,0x6d,0xf7,0xfb,0xf7,
	0xdf,0x6f,0xff,0xff,0xf7,0xf7,0xdf,0x7f,
	0xff,0xfd,0xbf,0x7f,0xde,0xfd,0xff,0x7d,
	0x75,0xd7,0x7f,0xff,0xfd,0xbf,0xb7,0x5f,
	0x7f,0xfd,0xff,0xfe,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xdf,0xff,0xff,0xff,
	0xff,0xff,0xfb,0xff,0xff,0xb7,0xef,0xff,
	0xff,0x6f,0xff,0xee,0xff,0xff,0xfe,0xff,
	0xfe,0xf5,0xef,0xff,0xff,0xff,0xff,0xfb,
	0xff,0xbf,0xff,0x7e,0xff,0xff,0xa7,0xff,
	0x7e,0xff,0xef,0xff,0xff,0xbf,0x7f,0x1b,
	0xff,0xdf,0xf7,0xdf,0xfd,0xf7,0xdd,0xff,
	0xff,0xff,0xff,0xdf,0xfd,0xff,0xff,0xff,
	0xdd,0xef,0xff,0xff,0xfd,0xdf,0xdf,0xed,
	0xff,0xfd,0xfb,0xef,0xff,0xff,0xff,0xf7,
	0x5f,0xf9,0xf7,0xff,0xdf,0xff,0xff,0xfd,
	0xff,0xff,0xf5,0xff,0xff,0xff,0x7d,0xff,
	0xfd,0xfd,0xfe,0x7f,0xff,0xff,0xe7,0xff,
	0xff,0xf7,0xdf,0xff,0xff,0xff,0xed,0x7f,
	0xff,0xff,0xd8,0xff,0xfd,0xff,0xbf,0xff,
	0xfe,0xfb,0xff,0xff,0xf7,0xfe,0xff,0xff,
	0xfd,0xff,0xff,0xfb,0xfb,0xf7,0xbf,0xfc,
	0xef,0xbf,0xff,0xbf,0xff,0xfb,0xff,0xff,
	0x7f,0xff,0xff,0x3f,0xfe,0xaf,0xfe,0xf4,
	0x7f,0xdf,0xff,0xf7,0xfd,0xf1,0xcf,0x3f,
	0xef,0xff,0xff,0x5f,0xb7,0xff,0xf3,0xf7,
	0x7f,0xfd,0x7f,0xf3,0x8d,0xff,0xff,0xf5,
	0xff,0xcf,0xef,0xff,0xbf,0xf2,0xff,0xff,
	0xef,0xff,0xef,0xfb,0xff,0xbf,0xbd,0xff,
	0xfe,0xfb,0xfe,0xff,0xff,0xfd,0xff,0xcf,
	0xfe,0xff,0xf7,0xff,0xbf,0xff,0xfe,0x7f,
	0xdb,0xf7,0x7f,0xfb,0xfe,0xfe,0xa7,0xff,
	0xf7,0x7f,0xfc,0xfa,0xdf,0xff,0xfb,0xfd,
	0xfd,0xf7,0xdf,0xdf,0x7f,0xfa,0xdd,0xbf,
	0x9c,0x7f,0xcb,0x75,0xff,0xed,0xf1,0xdf,
	0xff,0xdf,0xff,0xf7,0xdd,0xff,0xff,0xdf,
	0xf9,0xff,0x7f,0xf3,0xdf,0xff,0xfb,0xff,
	0xef,0xbf,0xff,0x3f,0xfd,0xff,0xdf,0xff,
	0xfe,0xf6,0xff,0x7f,0xff,0xff,0xef,0xdf,
	0xef,0xff,0xff,0xef,0xbf,0xf7,0xff,0xfb,
	0xff,0xef,0xbf,0xff,0xdf,0xff,0xfe,0xef,
	0xbf,0xff,0xff,0xff,0xff,0x7f,0xff,0x77,
	0xff,0xff,0xdf,0xff,0xff,0xfd,0xff,0xfe,
	0xbf,0xfe,0xff,0xbf,0xff,0xff,0xff,0xaf,
	0xff,0xff,0x6f,0xfc,0xf9,0xfd,0xbf,0xfe,
	0x7a,0xfe,0xaf,0xf7,0xff,0xff,0xff,0xfc,
	0xff,0xff,0xfa,0xbe,0xbf,0xfe,0xfa,0xfe,
	0xff,0x27,0xff,0x7f,0xf3,0xfd,0xaf,0xfe,
	0xff,0xeb,0xff,0x0f,0x59,0x5f,0xef,0xff,
	0xf5,0xf3,0xff,0x3f,0xff,0xff,0xff,0xff,
	0xf7,0xff,0xff,0xe7,0xff,0xff,0xf9,0xe7,
	0xff,0xff,0xbf,0xfc,0xff,0xcd,0xff,0xff,
	0xf9,0xd7,0x9f,0xff,0xaf,0xd5,0xfb,0xff,
	0xff,0xff,0xfb,0xdf,0xbf,0xff,0xfd,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x7f,
	0xff,0xff,0xff,0xff,0xff,0x7f,0xcf,0xf6,
	0xff,0xff,0xfd,0xb7,0xff,0x7f,0xfb,0xff,
	0xf7,0xdf,0x7f,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xf7,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x6f,
	0xf4,0xff,0xbf,0xbf,0xfe,0xfe,0xef,0xef,
	0xff,0xfe,0xfa,0xfb,0xef,0xff,0xfe,0xfb,
	0xef,0xfb,0xfe,0xfb,0xef,0xbf,0xff,0xfe,
	0xfa,0xef,0xbf,0xff,0xfe,0xfb,0xef,0xbf,
	0xff,0x8b,0xfd,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xdb,0xf8,0xff,0xf7,0xff,0xff,
	0xff,0xfd,0xff,0x9f,0xff,0xfb,0xaf,0xff,
	0xff,0x7a,0xff,0xff,0xbf,0xff,0xff,0xef,
	0xff,0xff,0xfa,0xff,0xff,0x9f,0xff,0xff,
	0xff,0xf9,0xff,0x79,0x6d,0xff,0x7f,0xff,
	0xfe,0xff,0xdf,0xff,0xff,0xfd,0x9b,0x7f,
	0xff,0xfc,0xf6,0x5f,0xef,0xf7,0xfe,0xfb,
	0xef,0xbe,0xff,0xf4,0xfb,0xcf,0x3f,0xfd,
	0xfe,0xfb,0xff,0xbf,0xef,0x3f,0xfb,0xff,
	0xfe,0xdb,0x6e,0xbb,0xff,0xf6,0xfb,0x6f,
	0xbc,0xf5,0xfe,0x7b,0xef,0xbf,0xff,0xfb,
	0xef,0xbf,0xff,0xfe,0x7b,0x6f,0xbb,0xff,
	0xde,0xfb,0xef,0xbf,0xff,0xfe,0x7f,0xfa,
	0xff,0xe7,0xdf,0x7f,0xfe,0xf9,0xe7,0x9d,
	0x7f,0xf7,0x59,0xe7,0x9f,0x7d,0xfe,0xf9,
	0x9d,0x7f,0xfe,0xf9,0xe7,0x9f,0x7f,0xf7,
	0xf9,0xe7,0x9f,0x7f,0xfe,0xf9,0xe7,0xff,
	0x87,0xff,0x1f,0x7f,0xff,0xf5,0xc7,0x5a,
	0x7f,0xbc,0x9d,0xd6,0x5d,0x7f,0xf6,0xf1,
	0xe7,0x7f,0xfe,0xf9,0xe7,0x9f,0x7f,0xbe,
	0x7d,0xe7,0x9f,0x7f,0xfe,0xf9,0xc7,0x9f,
	0xff,0xb5,0xf6,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xfe,0xf7,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xbf,0xfd,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0x6f,0xf7,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xf7,0xff,0xff,
	0x7f,0xff,0xff,0xf7,0x7b,0xff,0xfd,0xf7,
	0xdf,0x7f,0xfd,0xff,0xf7,0xdf,0x7f,0xff,
	0xfd,0xff,0xdf,0xff,0x8f,0xfe,0x7d,0xb4,
	0xfc,0xe2,0x1f,0x3f,0xf6,0xf1,0x62,0x8b,
	0x3f,0xbe,0xfc,0xe3,0x9f,0xf5,0xfc,0xe2,
	0x8d,0x7f,0xff,0xf1,0xe2,0x5f,0x7f,0xfd,
	0xfd,0xe7,0x9f,0x7f,0xfe,0xb7,0xfc,0xef,
	0xf7,0xd5,0x57,0xff,0x7d,0xf7,0xdf,0x57,
	0x5f,0xfd,0xf5,0xc5,0x1f,0xfb,0xfd,0xc7,
	0x17,0x7f,0xfc,0xf7,0xdf,0x5f,0xff,0xfd,
	0xf7,0xdf,0x7f,0xff,0xed,0xf7,0xff,0xe2,
	0xff,0xdf,0xbe,0xff,0xff,0x77,0xcb,0x7f,
	0xff,0xfa,0x6f,0xab,0xf7,0xf7,0xca,0xff,
	0xbf,0xb7,0xde,0xfb,0xff,0xfe,0xfe,0xde,
	0x7f,0xff,0xff,0xf7,0xdf,0x77,0xff,0xff,
	0x2f,0xff,0xff,0x9f,0x7d,0xee,0xf9,0x47,
	0x9f,0x7f,0xfc,0xf9,0x47,0x9f,0x7f,0xf6,
	0xb9,0xe7,0x7d,0x64,0x99,0xe7,0x9f,0x7f,
	0xee,0xf9,0xe7,0x9f,0x7f,0xee,0xf9,0xe7,
	0x9e,0xff,0xfa,0xff,0xff,0xdb,0xbf,0xff,
	0xfd,0xfb,0xdf,0xff,0xff,0xfe,0xfb,0x8b,
	0xbf,0xff,0xfd,0xed,0x7f,0xff,0xf8,0xf7,
	0xff,0xaf,0xff,0xff,0xe7,0xdf,0x7f,0xfe,
	0xff,0xf7,0xff,0xc9,0xff,0x7f,0xdf,0xfe,
	0xfd,0xdf,0xf7,0xff,0xfd,0xff,0xfd,0xdf,
	0x5f,0xff,0xfd,0xef,0x7f,0xff,0x5b,0xf7,
	0xbf,0xff,0xfd,0xfd,0xdf,0x7f,0xff,0xfe,
	0xff,0xff,0xff,0xff,0x1f,0xfe,0xff,0xff,
	0xef,0xaf,0x7f,0xff,0xfb,0xf7,0xaf,0xbf,
	0xfe,0xfb,0xff,0xef,0xff,0x7f,0xfb,0xff,
	0xbf,0xff,0xff,0xff,0xaf,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xfd,0xff,0xf0,0xff,
	0xff,0xbf,0x7f,0xff,0xff,0xff,0xff,0xdf,
	0x7f,0xff,0xff,0xff,0x7e,0xfb,0xef,0xdf,
	0xfe,0xf3,0xed,0xbf,0xff,0xff,0xff,0xff,
	0xff,0xff,0xfe,0xfb,0xef,0xbf,0xff,0x9f,
	0xff,0xfd,0xff,0xff,0xef,0xbf,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xbf,0xff,0xff,0xf7,0xff,
	0x7f,0xff,0xfd,0xf7,0xdf,0xff,0xff,0xfe,
	0xdf,0xf5,0xff,0xd7,0x5f,0x3f,0xfd,0xb5,
	0xd3,0x5f,0x7f,0xfd,0xf5,0x97,0x5f,0x7f,
	0xfd,0xd5,0x5e,0x7e,0xf9,0xb5,0xd7,0x1f,
	0x7d,0xfd,0xd4,0xd7,0x5f,0x7b,0xfd,0xf5,
	0xd7,0xbe,0xe8,0xff,0xff,0x3f,0xff,0xfa,
	0xf3,0xcf,0x3f,0xef,0xff,0xff,0x7f,0x3f,
	0xf6,0xef,0xdf,0x3f,0xfe,0xef,0xbf,0x8f,
	0xbf,0xfe,0x7f,0xe7,0x9f,0x3f,0xfe,0xfb,
	0xff,0xef,0xff,0x4f,0xfe,0xfb,0xff,0x9c,
	0xff,0xeb,0x3f,0xff,0xfe,0xff,0xff,0xf7,
	0x7f,0xff,0xfe,0xfb,0x6f,0xff,0xfe,0xfb,
	0xd7,0xff,0xfd,0xff,0xef,0xaf,0x3f,0xff,
	0xff,0xff,0xef,0xff,0x9e,0xf8,0xff,0xff,
	0xff,0xdf,0x7f,0xff,0xff,0xf7,0xfd,0xff,
	0xff,0xff,0x57,0xff,0xff,0xf7,0xf4,0xff,
	0xff,0xff,0xfd,0xff,0xff,0xff,0x7f,0xff,
	0xf7,0xff,0xf7,0xf7,0xdb,0xff,0xd3,0xfe,
	0xff,0xdf,0x67,0xf9,0xef,0x9f,0xff,0x7e,
	0xfc,0xf1,0xf7,0xff,0xff,0x77,0xff,0xfe,
	0x7f,0xbc,0xff,0xff,0xff,0xfb,0xff,0xff,
	0xff,0xef,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfe,0xef,0xef,0xf7,0xeb,0xff,0xbf,0xef,
	0xff,0xf7,0xef,0xff,0xf9,0xff,0xfd,0xff,
	0x7f,0xff,0xfb,0xff,0xef,0xff,0xdf,0xff,
	0xfb,0xef,0xff,0xff,0xfe,0xff,0xef,0xff,
	0x6f,0xf7,0xff,0xdf,0xff,0xff,0xff,0xff,
	0xbf,0xff,0xff,0xff,0xff,0xff,0xff,0xfd,
	0xff,0xff,0xf9,0xf7,0xdf,0xef,0x7f,0xff,
	0xfd,0xff,0xef,0xff,0xff,0xfd,0xfe,0xdf,
	0x7f,0xff,0x8b,0xff,0xff,0xff,0x7f,0xff,
	0x7d,0xfb,0xff,0x57,0xff,0xdf,0xff,0xfb,
	0xfd,0xdf,0xdf,0xef,0xb7,0xdf,0xff,0xff,
	0xff,0x6d,0xff,0xff,0xff,0xff,0xff,0xff,
	0xdf,0xff,0xff,0x7f,0xfd,0xff,0xff,0xbf,
	0xfe,0xff,0xeb,0xbf,0xff,0x7f,0xff,0xbf,
	0x77,0xff,0xff,0xff,0xff,0x9f,0xff,0xff,
	0xff,0xff,0xbf,0xfb,0xff,0xff,0xbf,0xff,
	0xfe,0xee,0xfd,0x7f,0xff,0xe9,0xff,0xf3,
	0xc7,0xbf,0xfe,0xfa,0xf3,0xff,0xff,0xff,
	0xfd,0xf7,0xff,0x3f,0xfb,0xf5,0xfb,0x3f,
	0xfe,0xfc,0xfe,0xcf,0x3f,0xff,0xfc,0xf3,
	0x5e,0x7f,0xaf,0xf8,0xff,0xff,0x37,0xff,
	0xbf,0x7f,0xfe,0xff,0xff,0xbf,0xff,0xff,
	0xff,0xdf,0xff,0xde,0x7f,0xf9,0xdd,0xff,
	0xff,0xd7,0xd6,0xf7,0xff,0xfe,0xfb,0xef,
	0xbf,0xdf,0xfd,0xef,0xbf,0xff,0xff,0xff,
	0xdb,0x6e,0x75,0xff,0xfd,0xff,0xff,0x71,
	0xff,0xff,0xfd,0xfd,0xdf,0x5f,0x7f,0xff,
	0xfd,0xff,0xf7,0xff,0xf6,0xe7,0xd7,0x8f,
	0x7f,0x7f,0xf9,0x5f,0x2f,0xef,0x77,0xff,
	0xdf,0x17,0xef,0xdf,0x7f,0xff,0xbf,0xff,
	0xef,0xff,0xf7,0xff,0xbf,0xff,0xfe,0xff,
	0xff,0xff,0xff,0xff,0xff,0xeb,0xfd,0x5f,
	0xff,0xff,0xf7,0xdf,0xff,0xff,0xfd,0xff,
	0xff,0xfb,0xe6,0xbe,0xff,0xde,0xed,0xff,
	0xff,0xfe,0xdf,0xef,0xff,0xff,0xff,0xdf,
	0xdf,0xff,0xff,0xfe,0xff,0xff,0xff,0xfe,
	0xfb,0x5f,0xff,0xbf,0xbf,0xfe,0xff,0x77,
	0xff,0xfe,0xfb,0xff,0xd7,0xff,0xff,0x3a,
	0xf8,0xad,0xb7,0xce,0xff,0xf9,0xfd,0xf7,
	0xff,0xfa,0xff,0xfb,0xff,0xff,0xf8,0xf9,
	0xaf,0xfe,0xff,0xff,0xac,0x8f,0xcf,0xff,
	0xff,0xab,0xaf,0xfe,0xff,0x92,0xf5,0xff,
	0xff,0x5f,0x7f,0xbd,0x75,0xff,0xcf,0x2f,
	0xff,0xff,0xe7,0xff,0xff,0xff,0xff,0x5f,
	0xef,0xff,0xf9,0xff,0xff,0x77,0xfd,0x7c,
	0xff,0xff,0x7f,0xfe,0xf9,0xff,0xda,0xbd,
	0xff,0xff,0xcf,0x7f,0xff,0xff,0xff,0xef,
	0xff,0xff,0xfd,0xff,0xff,0xd4,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0x9f,0xf7,
	0x64,0xff,0xf7,0xdf,0xfe,0xff,0xff,0xf7,
	0xdf,0xfe,0xff,0xff,0xff,0xff,0xbf,0xfe,
	0xff,0xff,0xff,0xff,0xff,0xff,0xdf,0x7f,
	0xff,0xfd,0xf7,0xff,0xff,0xff,0xff,0xff,
	0xf9,0x6e,0xfe,0xff,0xfa,0xfb,0xbf,0xff,
	0xfe,0xfe,0xfb,0xbf,0xff,0xfe,0xfb,0xef,
	0xbf,0xff,0xfe,0xef,0xbf,0xff,0xfe,0xfb,
	0xeb,0xef,0xbf,0xfe,0xfe,0xef,0xbf,0xff,
	0xfe,0xfb,0x17,0xd8,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xfe,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0x8b,0xe7,0xff,0xff,
	0xff,0xff,0x9f,0xff,0xff,0xff,0xb7,0xdf,
	0x7f,0xf8,0x7b,0xfd,0xbf,0xff,0xef,0xf7,
	0xff,0xff,0xff,0xaf,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0x7e,0xff,
	0xfb,0xff,0x3f,0xff,0xfd,0xff,0xff,0x7f,
	0xf1,0xfd,0xc7,0xef,0xab,0xff,0x7e,0xef,
	0x7e,0xff,0xfe,0xcb,0x6f,0xff,0xff,0xfe,
	0xff,0xef,0xbf,0xff,0xfe,0xcb,0x7f,0xe0,
	0xf9,0x6f,0xbe,0xff,0xfe,0xfb,0xef,0xbf,
	0xed,0xfe,0xfb,0xef,0xbf,0xf9,0xfe,0xfb,
	0xbf,0xff,0xb6,0xfb,0xef,0xbf,0xe5,0xfe,
	0x9b,0x6f,0xbe,0xff,0xfe,0xfb,0xef,0xff,
	0x81,0xef,0x7d,0xf7,0xd9,0xe7,0x9f,0x7f,
	0xfe,0xf9,0xe7,0x9f,0x7f,0xfe,0xd9,0xe7,
	0x9f,0xdf,0xf9,0xe7,0x9f,0x7f,0xfe,0x9d,
	0xe7,0xdf,0x79,0xe6,0xf9,0xe7,0x9f,0x7f,
	0xfe,0x3f,0xf9,0xdf,0xdd,0xd7,0x9d,0x7f,
	0xfc,0xf5,0xd7,0x1f,0x7f,0xfc,0xf1,0x67,
	0x9b,0x7f,0xfe,0xe7,0x1f,0x7f,0xfe,0xf9,
	0xf7,0x5b,0x6f,0xbf,0xf5,0xe7,0x9f,0x7f,
	0xfe,0xf9,0x7e,0x64,0xff,0xff,0xf7,0xff,
	0xff,0xff,0xfd,0xf7,0xff,0xff,0xff,0xff,
	0xff,0xdb,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xdb,0xef,0xbf,0xfd,0xff,0xff,
	0xff,0xff,0xff,0xfb,0x76,0xff,0xff,0xf7,
	0xff,0xdf,0xff,0xfb,0xdf,0xff,0xff,0xff,
	0xff,0xff,0xb7,0xdf,0x7f,0xbf,0xf7,0xff,
	0x7f,0xff,0xfd,0xbf,0xff,0xff,0xfd,0xff,
	0xf7,0xdf,0x7f,0xff,0xfd,0x7f,0xfb,0xff,
	0xf3,0x8b,0x7d,0xfd,0xf1,0xe2,0x8b,0x7d,
	0xf4,0xd5,0xc7,0x89,0x35,0xff,0xfd,0xdf,
	0x2f,0xfc,0xfc,0x67,0x8b,0x3f,0xbf,0xd8,
	0x62,0xdf,0x3d,0xff,0xf9,0xe7,0xff,0xd5,
	0xff,0x5f,0xdf,0xfd,0xf7,0xdf,0x77,0x5f,
	0xfd,0xf7,0xdf,0x7f,0x5f,0x7c,0xb3,0xdf,
	0xff,0x7d,0xb1,0xc6,0x7f,0x5f,0x7d,0xf5,
	0xdd,0x77,0xff,0xfd,0xf1,0xdf,0x7f,0xff,
	0x4f,0xfe,0xff,0xde,0xf3,0xff,0xfc,0xbf,
	0xfe,0xf7,0xbf,0xfe,0xfb,0xdd,0xff,0xef,
	0xff,0xf7,0x7f,0xdf,0xbf,0xf7,0xff,0x6e,
	0xef,0xbc,0xff,0xee,0xff,0xed,0xfd,0xf7,
	0xff,0x7f,0xf2,0xff,0xfd,0xf9,0xe7,0x9f,
	0x7b,0xec,0xf9,0xe7,0x9f,0x7f,0xfe,0xb9,
	0xc7,0x9f,0x7f,0xfa,0xe7,0x9f,0x7d,0xee,
	0xf1,0x47,0x9f,0x7b,0xee,0xf9,0x67,0x9f,
	0x7b,0xee,0xf9,0x83,0xff,0x3f,0xbf,0xfe,
	0xf7,0xdf,0xbf,0xff,0xfe,0xff,0xdf,0x7f,
	0xbf,0xfc,0xf3,0xdf,0xff,0xf9,0xfa,0xcf,
	0x7f,0xfe,0xfc,0xff,0xcb,0xaf,0xff,0xfd,
	0xf3,0xdf,0x7f,0xfe,0x5f,0xfc,0xff,0xe9,
	0xbd,0x3f,0xff,0xfc,0xef,0xbd,0x7f,0xff,
	0xfc,0xf7,0xf5,0xb7,0xff,0xfe,0xff,0x57,
	0x7f,0xfe,0xff,0xa7,0xdf,0xde,0x7a,0xef,
	0xef,0x9f,0xff,0xfe,0xff,0xff,0xef,0xff,
	0xef,0xbf,0xff,0xff,0xff,0xef,0xbf,0xff,
	0xff,0xff,0xdf,0x7f,0xff,0xfe,0xff,0xf7,
	0xff,0xff,0xfb,0xff,0xff,0xff,0xff,0xfb,
	0xef,0xff,0xff,0xfe,0xff,0xff,0xff,0x2b,
	0xff,0xdf,0xfe,0xff,0xff,0xff,0xff,0x7f,
	0xff,0xff,0xff,0xff,0xff,0xef,0xbd,0xff,
	0xfe,0xef,0xff,0xdf,0xfe,0xfb,0xfb,0xf7,
	0xbf,0xff,0xff,0xef,0xb7,0xff,0xfe,0xfb,
	0xbf,0xf8,0xff,0xef,0xff,0xff,0xef,0xff,
	0xff,0xff,0xfd,0xff,0xff,0xff,0xfe,0xff,
	0xff,0xff,0xff,0xfd,0xff,0xff,0xdf,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xf7,
	0xff,0xff,0xca,0xff,0x7f,0xfd,0xf5,0xd7,
	0x5f,0x7d,0xfd,0xf5,0xd7,0x5f,0x7f,0xfd,
	0xf5,0xd7,0x4f,0xed,0xf5,0xc7,0x5f,0x7f,
	0xfd,0xf5,0xd7,0x5f,0x7b,0xfd,0xf5,0xd7,
	0x5f,0x7f,0x9d,0x8f,0xfe,0xff,0xfc,0x7b,
	0xef,0xff,0xf7,0xfb,0xf3,0xdf,0x7f,0xff,
	0xee,0xf3,0x8d,0x3e,0xfa,0xc3,0xcf,0x3b,
	0xfa,0xff,0x7f,0xce,0xfd,0xff,0xff,0xa3,
	0x8f,0xfe,0xff,0xff,0xff,0xe4,0xff,0xfd,
	0xff,0x77,0xff,0xfb,0xff,0xcb,0xbf,0xbf,
	0xfe,0xff,0xcb,0xdf,0x7d,0xff,0xf2,0xef,
	0x6d,0xdf,0xfe,0xfb,0x4f,0xeb,0xff,0xff,
	0xf3,0xcf,0xf7,0xff,0xee,0x7f,0xa1,0xff,
	0xbf,0x7f,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xb7,0xdf,0xef,0xbd,0xf6,0x67,
	0x7f,0xba,0xf7,0xfb,0xdb,0x7d,0xff,0xfd,
	0xf7,0x7f,0xff,0xfd,0x7f,0xff,0x7f,0x77,
	0x78,0xff,0xff,0xff,0xff,0xfe,0x7f,0xff,
	0xf5,0xff,0xff,0xff,0xff,0xff,0xfe,0xff,
	0xef,0xff,0xff,0xf3,0xfe,0xff,0xf5,0xf7,
	0xff,0xff,0xff,0x7f,0xff,0xff,0xff,0xff,
	0x7b,0xe6,0xfb,0xff,0xff,0xff,0xff,0xff,
	0x5f,0x5f,0xdf,0xbf,0xff,0xff,0xff,0xef,
	0xbf,0xff,0xff,0xfd,0xff,0xdf,0xfe,0x6f,
	0xfd,0xb6,0xff,0xfe,0xfb,0xef,0xff,0xff,
	0xfe,0xd7,0x46,0xfe,0xff,0xfe,0xff,0xff,
	0xff,0xfd,0xbf,0xff,0x7f,0xff,0xff,0xff,
	0xff,0xf7,0xdf,0xdd,0x9d,0xbf,0xdf,0x7d,
	0xbf,0xdd,0xff,0x9f,0xfe,0xf7,0xf9,0xf7,
	0xdd,0x7e,0xfd,0x1f,0xe8,0xfe,0xff,0xff,
	0xff,0xfd,0xff,0xff,0xff,0x7f,0xff,0xff,
	0xdf,0xde,0x8f,0x3d,0xfc,0xa2,0xff,0x3f,
	0xfe,0x58,0xc3,0xff,0x39,0xbe,0xff,0xe3,
	0x8f,0x2d,0xfe,0xf8,0xf7,0xc5,0xff,0xff,
	0xff,0xfd,0xbf,0x9f,0xfb,0xfb,0xeb,0xff,
	0xff,0xff,0xff,0xfb,0xff,0xbf,0xfe,0xff,
	0xb7,0xff,0xff,0xff,0xff,0x77,0xff,0xff,
	0xfb,0xef,0xff,0xff,0xfe,0xff,0x0b,0xfe,
	0x7f,0xbf,0x7f,0xf7,0xd7,0x2f,0x7f,0xff,
	0xd7,0xee,0xff,0xfd,0xfd,0xdf,0xf7,0xbf,
	0x7f,0xff,0xff,0xf3,0xff,0xf7,0xf7,0xfb,
	0xff,0xef,0xff,0xff,0xfe,0xff,0xef,0xff,
	0xf7,0xff,0xef,0xff,0xff,0xfe,0xfb,0xf9,
	0xff,0x7f,0x7f,0xff,0xff,0xdf,0xff,0xff,
	0xff,0xfc,0xff,0xfb,0xff,0xff,0xfb,0xff,
	0xfe,0x7f,0xff,0xff,0xff,0xff,0xff,0xff,
	0x77,0x83,0xff,0xbf,0xbb,0xee,0xf7,0xeb,
	0x2e,0xfb,0xff,0xff,0xff,0xff,0xff,0xfd,
	0xfb,0xef,0xbf,0xfe,0xfa,0xad,0xbf,0xff,
	0xfe,0xfb,0xef,0x2f,0xff,0xf6,0xfb,0x6f,
	0xb7,0xef,0x9f,0xfd,0xdf,0xfc,0xf7,0xbf,
	0x5f,0xff,0xff,0xff,0xff,0xff,0x7f,0xfd,
	0xbf,0x83,0x17,0xde,0x76,0xd7,0x1f,0x3f,
	0xfc,0xf1,0xdf,0x1f,0xff,0xfd,0xf1,0xc7,
	0x1e,0x7f,0xec,0xff,0xe6,0xff,0xaf,0xd7,
	0xff,0xff,0xef,0xff,0xff,0xfe,0x7f,0xff,
	0xbf,0xff,0xfe,0xff,0xff,0xbf,0xff,0xef,
	0xff,0xbf,0xff,0x3e,0xff,0xef,0xf7,0xf7,
	0xff,0xff,0xcf,0xbf,0xff,0x07,0xfd,0xff,
	0xaf,0xff,0xdf,0xff,0xff,0x7c,0xbf,0xfe,
	0xff,0xf3,0xff,0xff,0xdd,0x7f,0x3f,0xf7,
	0xff,0x7f,0xdf,0x7d,0x37,0xff,0x7f,0xff,
	0xfc,0xf7,0xdd,0x7f,0xdf,0xfd,0xcf,0x5a,
	0xff,0x7f,0xfd,0xff,0xfe,0xff,0xef,0xf3,
	0xf9,0xff,0xff,0xff,0xf3,0xef,0x3e,0xfb,
	0xbb,0xcf,0xff,0xfb,0xee,0xbb,0xfc,0xff,
	0xfb,0xef,0xbf,0xef,0xfe,0xfb,0xee,0xaf,
	0xd5,0xfb,0xbf,0xff,0xfe,0xff,0xef,0xbf,
	0x59,0xfb,0xff,0xef,0xff,0x7f,0x2f,0xff,
	0xf7,0xdf,0xff,0xed,0xf5,0xf7,0xdf,0xff,
	0x3f,0xf5,0xff,0x4b,0x7f,0x3f,0xf5,0xf7,
	0x7f,0x3f,0xf6,0x7f,0xfb,0xed,0xff,0xdf,
	0x7e,0xcb,0xfa,0xff,0xdf,0xfe,0xff,0xfe,
	0xfb,0xef,0xff,0xfe,0xfb,0xac,0xbf,0xcf,
	0xfe,0xf7,0xac,0xff,0x9f,0x7a,0xfa,0xaf,
	0xbf,0xff,0xff,0xe4,0xff,0xaf,0xbf,0xfe,
	0xfb,0xeb,0xef,0xd7,0xeb,0xfb,0xfb,0xff,
	0xdf,0xec,0x7f,0xef,0xda,0x7e,0x7b,0xff,
	0xb7,0xfb,0xfe,0x7a,0xff,0xff,0xf7,0xff,
	0x7b,0xff,0x77,0xfb,0x8b,0xfd,0xf6,0xff,
	0xff,0xff,0xff,0xff,0xbf,0x6a,0xff,0xff,
	0xff,0xbf,0x5b,0xff,0x7b,0xb3,0xf7,0xae,
	0xf9,0xef,0xcd,0xf7,0xff,0xfb,0xff,0xbf,
	0xff,0xdf,0xfb,0xef,0xd7,0x5b,0xf9,0xff,
	0xff,0xff,0x7f,0xfa,0xff,0xff,0xf5,0x7f,
	0xeb,0xff,0xff,0xb7,0xd7,0xdf,0xdf,0xf6,
	0x5f,0xd9,0xff,0x7d,0xf5,0xff,0xdf,0xff,
	0xff,0xfd,0xd7,0xdf,0xff,0xfd,0xbb,0x60,
	0xff,0xbf,0xff,0xfe,0xb7,0xef,0xff,0xaf,
	0xff,0xd7,0xff,0xff,0xef,0xbe,0xf6,0xde,
	0xef,0xbd,0x5f,0xde,0x7b,0xaf,0xfd,0xfb,
	0xde,0xff,0xef,0xbd,0xf6,0xde,0x7b,0xff,
	0x47,0x9c,0xfd,0xb6,0xdb,0xef,0xbf,0xf9,
	0xf6,0xff,0xef,0xbd,0xff,0xff,0xfb,0x7f,
	0xfb,0xff,0x9f,0xff,0xff,0xed,0xff,0xff,
	0x6f,0xfe,0xed,0xff,0xdf,0xff,0xff,0xff,
	0xff,0xbf,0xfb,0x5e,0x77,0xdf,0x77,0xfe,
	0xfd,0xe7,0xfd,0x7d,0xfe,0xd9,0xef,0x9f,
	0xff,0xe6,0xdf,0xff,0xfd,0xff,0xdb,0xff,
	0xff,0x7f,0xf7,0x9b,0xff,0xbf,0xf1,0xff,
	0xfb,0xef,0xff,0x99,0xff,0xdd,0x77,0xff,
	0xf1,0xf6,0x5f,0x6f,0x9f,0xb1,0xd6,0xdd,
	0x7f,0xfe,0xfd,0xf7,0x78,0xb7,0xfd,0xf7,
	0xd9,0x6f,0xaf,0xdd,0xf7,0xdf,0x6f,0xbf,
	0xfd,0xf7,0xdb,0xef,0xa7,0xf6,0xff,0x7f,
	0xff,0xff,0xf3,0xdf,0xef,0xff,0xfc,0xf7,
	0xff,0x7f,0xff,0xfd,0xff,0x5f,0xff,0xfe,
	0xff,0xdb,0xef,0xff,0xfd,0xff,0xff,0x3f,
	0xbf,0xfd,0xf7,0xcf,0x6f,0xef,0xf6,0xff,
	0x7f,0xff,0xff,0xdf,0x7f,0xff,0xff,0xfd,
	0xef,0xff,0xdf,0x7d,0xfb,0xff,0x7f,0xff,
	0xff,0xff,0xbf,0xff,0xff,0xfd,0xff,0xff,
	0x7f,0xff,0xfd,0xf7,0xdf,0xff,0xfd,0xab,
	0xff,0x2f,0xfe,0xf8,0xc7,0x8f,0x3f,0xfe,
	0xfe,0xc3,0x8b,0xbf,0xff,0xfc,0xda,0x6b,
	0x9f,0xf6,0xff,0x6b,0xbf,0xbd,0xf6,0xe3,
	0x6b,0xbd,0xff,0xf6,0xdb,0x6b,0xbf,0xfd,
	0x5f,0xfd,0x7f,0xf5,0xdd,0x7f,0x5f,0xfd,
	0xf5,0xed,0x33,0x5f,0xfd,0xf1,0xcd,0x77,
	0x5d,0xfd,0xd7,0x77,0x5f,0x75,0xf5,0xdd,
	0x57,0x5f,0xf5,0xf1,0x55,0x57,0xdf,0x75,
	0xd5,0xff,0xe9,0xff,0xdf,0xbf,0xff,0xfd,
	0xf7,0xcf,0xae,0xff,0xee,0xf6,0xef,0xbd,
	0xff,0xfe,0xfe,0xb7,0xff,0xff,0xfe,0xef,
	0xaf,0xff,0xff,0xfe,0xef,0xff,0xff,0xfe,
	0xfa,0xef,0xbf,0x23,0xff,0xff,0x9f,0x79,
	0xfe,0xf1,0xc7,0x0f,0x7d,0xec,0xd9,0x67,
	0x1f,0x1f,0x7e,0xf0,0xe1,0x1f,0xfc,0xf1,
	0x41,0x07,0x1f,0xe6,0xf1,0x61,0x1b,0x1f,
	0x74,0xf0,0x41,0x87,0x3f,0xf9,0xff,0xf3,
	0xef,0x7f,0xfe,0xfe,0xfb,0xed,0x3f,0xff,
	0xfe,0xfb,0xcf,0xac,0xff,0xfe,0xeb,0x7f,
	0xfe,0xfe,0xff,0xef,0xbf,0xff,0xfe,0xff,
	0xef,0xff,0xff,0xfe,0xff,0xef,0xc1,0xff,
	0xdf,0xdf,0xff,0xef,0xf5,0xdf,0xdf,0x7f,
	0xeb,0xfd,0xdf,0xdf,0x3e,0xdf,0x76,0x7b,
	0x37,0xff,0x77,0xf3,0xed,0x37,0xfd,0x77,
	0xfb,0x5f,0x37,0xdf,0x7e,0xf3,0xed,0x2f,
	0xfe,0xff,0xf8,0xeb,0xff,0xff,0xfe,0xda,
	0xfa,0xef,0xff,0xfe,0xfe,0xfb,0xee,0xf9,
	0x67,0x7b,0xbe,0xff,0xe7,0x9f,0x7b,0xbe,
	0xff,0xe7,0xfb,0x7f,0xfe,0xb9,0xe7,0x9f,
	0x3f,0xf1,0xff,0xe6,0xdf,0xff,0x7f,0xff,
	0xfd,0xd6,0xdf,0xff,0xff,0xb7,0xf7,0xe4,
	0x93,0x4f,0xfe,0xc4,0xff,0x4f,0x3c,0xf9,
	0x64,0xff,0x4f,0xfe,0xf7,0xc4,0x93,0x4f,
	0x3c,0xb9,0x9f,0xff,0xff,0xff,0xff,0xef,
	0xff,0xff,0xff,0xfb,0xcf,0xff,0xff,0xfe,
	0xff,0xfe,0xfb,0xbf,0xfc,0x7f,0x7b,0xef,
	0xbf,0xfc,0x7f,0xfb,0xfb,0xbf,0xff,0xf3,
	0xcb,0xef,0xff,0x0f,0xfd,0xff,0xd3,0x4f,
	0x7f,0xf5,0xf5,0xd7,0x4f,0x7d,0xfd,0xb5,
	0x57,0x5f,0x75,0xd4,0xf5,0x1a,0x7c,0xdd,
	0xd5,0xd7,0x5b,0x7f,0xdd,0xf5,0xd7,0x5f,
	0x7b,0xbd,0x71,0xd7,0xff,0xe7,0xff,0xcf,
	0xbd,0xff,0xfe,0xf3,0x8f,0xff,0xff,0xff,
	0xe3,0xbe,0xff,0xff,0xfc,0x33,0x3f,0xff,
	0xff,0xf3,0xc6,0x3e,0xff,0xff,0xf3,0xdf,
	0x15,0xfb,0xec,0xf3,0xcf,0xff,0x27,0xfe,
	0xff,0xbf,0xff,0xf6,0xfb,0xed,0xef,0xf7,
	0xff,0xff,0xf7,0xbf,0xff,0xfe,0xfb,0xaf,
	0xef,0xfe,0xfe,0xef,0xd7,0xff,0xfe,0xfd,
	0xd7,0xbd,0xff,0xff,0xfd,0xee,0xdf,0x1f,
	0xf9,0xff,0xfc,0xf7,0xff,0x3f,0xff,0xfc,
	0xdf,0xff,0x3f,0xff,0xb7,0xff,0xcf,0x3d,
	0xbf,0xf2,0x77,0x3d,0xf5,0xbc,0x73,0xff,
	0x35,0xff,0xff,0xf3,0xcf,0x3f,0xff,0xfc,
	0xfe,0x97,0xff,0xd7,0xff,0xbf,0xff,0xff,
	0xff,0xfe,0xff,0xff,0xf1,0xff,0xff,0xef,
	0xff,0xed,0xff,0xff,0xbf,0xf5,0xf7,0x5f,
	0xff,0xff,0xfc,0xff,0xef,0x7f,0xfc,0x74,
	0x9f,0xfe,0xc7,0xfe,0xbf,0xbe,0xff,0xff,
	0xfe,0xfb,0xfb,0xff,0xfe,0xdf,0xff,0xff,
	0xff,0xfe,0xbb,0x7f,0xff,0xff,0x67,0xef,
	0x5d,0x37,0xff,0x7d,0xe7,0xff,0xff,0xf6,
	0x6f,0xaf,0xff,0x6f,0xc1,0xff,0xff,0xfd,
	0xff,0xff,0xe7,0x7f,0xff,0xfd,0xf7,0xef,
	0xff,0x8f,0x3d,0xfa,0x9f,0xc9,0xed,0xfe,
	0xed,0x7f,0xc7,0xef,0xd7,0xdc,0x7f,0xce,
	0x3d,0xfb,0xdd,0x73,0xfe,0x99,0xff,0xff,
	0xfd,0x7f,0xff,0xff,0xff,0xdf,0xcf,0x77,
	0xff,0xfd,0xbd,0xfe,0xfe,0xdf,0xaf,0xbe,
	0xff,0xfa,0xf7,0xaf,0xb7,0xff,0xdb,0xff,
	0xaf,0xbf,0xfe,0xf3,0xeb,0xff,0x5f,0xfd,
	0xff,0xe7,0xbf,0xff,0xff,0xf7,0xff,0xff,
	0xfb,0xff,0xff,0xbf,0xff,0xdf,0xff,0xef,
	0xff,0xff,0xda,0xee,0xef,0xf7,0xff,0xdf,
	0xff,0x77,0xbd,0xf7,0xdf,0x6f,0xff,0xbf,
	0xee,0xff,0xff,0xdf,0x7f,0xff,0xfc,0xfd,
	0xe6,0xff,0xcf,0xfc,0xf3,0x7f,0xff,0xff,
	0xbf,0xfb,0xdf,0xff,0x9f,0xfd,0xfb,0x1f,
	0xff,0xff,0xf3,0xff,0xef,0xe7,0xdf,0xff,
	0xf7,0x6f,0xff,0xff,0xdf,0xfb,0xf7,0xe5,
	0xff,0x7f,0xff,0xff,0xe7,0x9f,0xfd,0xbf,
	0xff,0xfe,0xf7,0xff,0xff,0xff,0xff,0x8f,
	0xff,0xed,0xff,0x4b,0xff,0xbf,0xff,0xff,
	0xff,0xff,0x37,0xfb,0xff,0xfb,0xdf,0xfd,
	0xb5,0xde,0x6a,0xbf,0xbf,0xff,0xfe,0xff,
	0xff,0xbf,0xbf,0xfe,0xe9,0x27,0x8f,0xde,
	0xfa,0xef,0x74,0xd7,0xee,0xff,0xff,0xbf,
	0xff,0x7e,0xfa,0xff,0x4b,0xff,0xcf,0xff,
	0xff,0xef,0xb5,0xd7,0xfe,0xff,0xf5,0xf7,
	0xff,0x5f,0x9e,0xdd,0xfe,0x6f,0xff,0x7d,
	0xf7,0xdf,0x7f,0xff,0xff,0xf3,0xff,0x7d,
	0xff,0xbd,0xf7,0xfe,0xfb,0x3f,0xfe,0x7f,
	0xdb,0xff,0xbf,0xff,0xf5,0xff,0xff,0xab,
	0xaf,0xff,0xff,0xff,0xf3,0xfd,0xf7,0xfb,
	0xaf,0xdf,0x7f,0xff,0xfc,0xff,0xbf,0xf7,
	0xdf,0xfd,0xeb,0xcd,0xff,0xff,0x7f,0xf1,
	0xbd,0xff,0xef,0xff,0x7b,0xff,0xdb,0xff,
	0xdf,0x7f,0xff,0xff,0xdf,0x3f,0xff,0xf9,
	0xb7,0x3f,0xfe,0xfc,0xeb,0xb7,0x5f,0xff,
	0xf6,0xeb,0xaf,0xbf,0xfa,0xfa,0xf3,0xf7,
	0x96,0xd1,0xdf,0xbe,0xf9,0xff,0xdb,0x7f,
	0xfe,0xfb,0xe6,0xff,0x7f,0x7e,0xf9,0xe7,
	0x97,0xff,0xf9,0xe7,0x97,0x7f,0xfe,0xfd,
	0xe6,0xdb,0x5f,0xbe,0xf9,0xe5,0x9f,0x7f,
	0xf7,0x1f,0xb4,0xcf,0xff,0xdf,0x7c,0xff,
	0xff,0xff,0xff,0x37,0xff,0xff,0xff,0xdf,
	0x7f,0xff,0xfd,0xde,0xfa,0xff,0xfd,0xf7,
	0xdf,0xff,0xfe,0xfd,0xef,0xdf,0x7f,0xff,
	0xfd,0xf7,0x3f,0x67,0xff,0xff,0x7f,0xd7,
	0xfd,0xff,0xff,0xff,0x3f,0xfa,0xff,0xff,
	0xff,0xfe,0xfb,0xec,0xff,0xe6,0xff,0xef,
	0xbb,0xff,0xfe,0xff,0xec,0xff,0xff,0xfe,
	0xfb,0xef,0xbf,0xbf,0x76,0xff,0xff,0xfe,
	0xff,0xff,0xbf,0xff,0xfe,0xbf,0xfc,0xbf,
	0xff,0x2f,0xbf,0x7c,0xde,0x6b,0xff,0x35,
	0xde,0x6b,0x67,0xff,0xff,0xdc,0xff,0x7f,
	0xff,0x7d,0xde,0xd3,0xff,0xbd,0xd3,0xfe,
	0xed,0xff,0xff,0x7f,0xfb,0xfd,0xff,0xed,
	0x7f,0xfb,0x7f,0xfb,0xe9,0xb7,0xff,0xfb,
	0xad,0xf3,0xfe,0xfa,0xfb,0xff,0xa7,0xff,
	0x7f,0xfb,0xec,0xf3,0x9e,0xfb,0xbf,0x9f,
	0xff,0xbf,0xfe,0xfb,0xef,0xef,0x3f,0xff,
	0x5b,0xf9,0xef,0xff,0x76,0x7b,0x7d,0xbf,
	0x6a,0xdb,0xef,0xfd,0xff,0x6e,0xfb,0xef,
	0x7f,0xfb,0x56,0x7f,0xed,0xfd,0xdf,0xeb,
	0x9f,0xf4,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xab,0xff,0xff,0xff,0xdb,0xeb,
	0xff,0xfd,0xdb,0xbf,0xef,0xbf,0x27,0xdb,
	0xff,0xff,0xd7,0xff,0xfa,0x6b,0xef,0xbf,
	0x7f,0x6f,0xe0,0xf9,0xff,0xff,0xff,0xfd,
	0xff,0xff,0x7f,0xdf,0xff,0xff,0xff,0xd7,
	0xdf,0x7f,0xfd,0xd3,0x5f,0x7d,0xfd,0xf7,
	0xd7,0xff,0x7d,0xfd,0xdf,0xf7,0xdf,0x7f,
	0xfd,0xf5,0xfd,0x25,0xdf,0xff,0xfe,0xf3,
	0xdf,0xff,0xff,0xf3,0x57,0xf6,0xff,0xff,
	0xbe,0xf6,0xde,0x6b,0xbf,0xf6,0x7a,0x69,
	0xaf,0xbd,0xf6,0xff,0x6b,0xff,0xbf,0xf7,
	0xde,0x6b,0xaf,0xfd,0xdb,0xfd,0xe7,0xfb,
	0xef,0xbf,0xff,0xfe,0xdb,0xee,0xbf,0xf9,
	0xfe,0xfb,0xff,0xff,0xff,0xf7,0xff,0xbf,
	0xff,0xb7,0xff,0xff,0xbf,0xfd,0xb7,0xfb,
	0x7f,0xfe,0xff,0xff,0xff,0xbf,0xea,0x7f,
	0xdf,0x7d,0xfe,0xf9,0x67,0x9f,0x79,0xfe,
	0x9d,0xe7,0x9f,0xff,0xf7,0xfb,0x7f,0xfd,
	0xff,0xf9,0x6f,0xff,0xff,0xe7,0x59,0x6f,
	0x9f,0xff,0xe6,0xdf,0xef,0xbf,0xff,0x6f,
	0xfe,0x7d,0xf7,0xf9,0xc7,0x5b,0x63,0xfd,
	0xf1,0xe7,0x5f,0x7f,0xbe,0xdd,0xf7,0xdf,
	0xed,0xfd,0xc7,0xdf,0x69,0xaf,0xfd,0x47,
	0xdb,0x7d,0xbc,0xbd,0xf6,0xd8,0x6b,0xff,
	0x9f,0xdb,0xff,0xff,0xff,0xff,0xcf,0x7f,
	0xff,0xff,0xff,0xff,0xff,0xbf,0xfd,0xf7,
	0xff,0x3f,0xfd,0xff,0xff,0xff,0xff,0xff,
	0xff,0xfb,0xff,0xbf,0xfd,0xf7,0xdf,0x7f,
	0xff,0xbf,0xd9,0xff,0xff,0xff,0xfd,0xdf,
	0x7f,0xff,0xff,0xff,0xff,0xff,0x7f,0xfb,
	0xf7,0xff,0xff,0xfd,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xfb,0xef,0xbf,
	0xff,0xfe,0xff,0x2f,0xfe,0xbf,0xf8,0x57,
	0x1f,0x3f,0xbe,0xdc,0xd7,0x9d,0x3f,0xf6,
	0xf9,0xdb,0x6f,0xaf,0x7d,0xdb,0x8f,0xaf,
	0xfd,0xf6,0xdb,0x8b,0xaf,0x9d,0xf8,0xda,
	0x6f,0xbf,0xbd,0xd6,0xff,0xf2,0xff,0xd5,
	0x7f,0xff,0x7d,0xf5,0xdd,0x7f,0xff,0xff,
	0xf5,0xdf,0x57,0x7f,0x75,0xf5,0x5f,0x7f,
	0x7c,0xd5,0xdd,0x77,0x5f,0xfc,0xd5,0xc5,
	0x5f,0x5d,0x7d,0xd7,0x5d,0xff,0xaf,0xff,
	0xbf,0xf7,0xef,0xbf,0xdf,0xbf,0xff,0xff,
	0xff,0xcb,0x7f,0xfb,0xff,0xfb,0xfb,0xdf,
	0xfe,0xab,0xff,0xbf,0xff,0xfe,0xf7,0xef,
	0x2f,0xff,0xfe,0xff,0xef,0xaf,0xff,0xae,
	0xfd,0xff,0x79,0xfe,0xb9,0xc7,0x9f,0x7f,
	0xfe,0xb9,0x41,0x9f,0x7f,0x7c,0xd8,0xc1,
	0x87,0x77,0xd8,0xe7,0x07,0x1f,0x76,0xf8,
	0xe7,0x07,0x7d,0x7e,0xf0,0xc1,0x07,0x1f,
	0xfe,0xe3,0xff,0xfb,0x7f,0xfe,0xfd,0xfb,
	0xeb,0x7f,0xff,0x7f,0xfb,0xdf,0xbf,0xdf,
	0x7e,0xfb,0xbd,0xdf,0xfe,0xfb,0xed,0xb7,
	0x9f,0xfc,0xfa,0xdd,0xaf,0xdf,0x7e,0xfb,
	0xed,0xf7,0x67,0xff,0xdf,0xff,0xef,0xbf,
	0xd7,0xdf,0xff,0xeb,0xff,0xdd,0xff,0x7f,
	0x5d,0x7c,0xd9,0x65,0x5f,0xf6,0xdf,0x65,
	0x17,0x5f,0xf4,0xdf,0xe5,0xff,0x5d,0x74,
	0xd9,0xe5,0x97,0xbf,0xf8,0xff,0xef,0xff,
	0x7f,0xff,0xfb,0xeb,0xff,0xef,0xf3,0xfb,
	0xf7,0xff,0xb9,0xe7,0x9f,0xed,0xb9,0xff,
	0x9e,0x7b,0xee,0x79,0xff,0x9e,0xf7,0xef,
	0xf9,0xe7,0x9e,0x7b,0xfe,0xc6,0xff,0xff,
	0xff,0xff,0xff,0xfd,0xff,0xff,0xff,0x9f,
	0xff,0xff,0xde,0x73,0xcf,0x3f,0xdf,0xf3,
	0xed,0x37,0xdf,0x7c,0xf3,0xef,0x3f,0xff,
	0xfe,0x73,0xcf,0x37,0xff,0xfc,0x7f,0xfe,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xcb,0xff,0xf6,0xb7,0xf2,0xcb,0xff,0xff,
	0xf2,0xff,0x2d,0xad,0xfc,0xf2,0xff,0xff,
	0xff,0xd7,0xf2,0xcb,0x4f,0xff,0xff,0xf7,
	0xf6,0xff,0x5f,0x7f,0xfd,0xf5,0xd7,0x5f,
	0x7f,0xdd,0xf5,0xd7,0x5f,0x6f,0xbd,0xf5,
	0xc7,0x72,0xed,0x65,0x97,0x5d,0x76,0xf5,
	0xf5,0x57,0x5f,0x6f,0xad,0xf5,0x96,0x4e,
	0xfe,0x8f,0xff,0x3f,0xff,0xf5,0xeb,0xcf,
	0x3f,0xfe,0xba,0xfb,0x8f,0x3f,0xde,0xfc,
	0xf3,0xcf,0xff,0xfc,0xff,0xcd,0x3f,0xff,
	0xfc,0xf3,0xcf,0xbf,0xd6,0xfc,0xb3,0xcf,
	0x3f,0xff,0xfb,0xf9,0xff,0xff,0xfa,0xff,
	0xff,0xf7,0xfc,0xf7,0xdf,0xef,0xff,0xfb,
	0xfe,0xfb,0xed,0xbf,0xff,0xfb,0xff,0xbf,
	0xff,0xfe,0x73,0xed,0xb7,0xfd,0xff,0xfd,
	0xef,0xff,0xff,0x7e,0x63,0xff,0xf3,0xff,
	0xff,0xfe,0xfc,0xff,0xff,0xfd,0xff,0xfc,
	0xeb,0xcf,0x3f,0xf7,0xfc,0xcb,0xff,0xf5,
	0xdc,0xf3,0xcf,0x7f,0xff,0xfc,0xff,0x4f,
	0x3f,0xdf,0xfc,0x72,0x57,0x77,0xde,0x5f,
	0xff,0xff,0x7f,0xff,0xff,0xff,0x7f,0xe1,
	0xdf,0xff,0xff,0x3f,0xfd,0xf5,0x7f,0xfd,
	0xff,0xd6,0x5f,0xce,0xbb,0xff,0xff,0xff,
	0xf7,0xff,0xff,0x9f,0x5b,0xfe,0xbe,0xf8,
	0xfe,0xfa,0xfe,0xff,0xfd,0xff,0xff,0x7f,
	0xff,0x7d,0xff,0xff,0xff,0xbf,0xfb,0xed,
	0x67,0xff,0x6f,0x7e,0xd9,0x7f,0xff,0xfd,
	0xff,0xff,0x7f,0xfb,0x6b,0xff,0xfa,0x37,
	0x8b,0xff,0xff,0xf3,0xff,0xfb,0xff,0xff,
	0xff,0x7f,0xfb,0xdf,0xff,0x7f,0xf7,0xfc,
	0xf3,0x37,0xbd,0xfb,0x77,0xff,0xfd,0xf7,
	0xdf,0xa3,0xfb,0xf5,0xf7,0xff,0xa7,0xcf,
	0xfd,0xbf,0xb6,0xff,0xe7,0xff,0xff,0xff,
	0xef,0xbf,0xff,0xff,0xfd,0xff,0xf7,0xfe,
	0x6b,0xef,0xbf,0xb6,0xff,0xef,0xbf,0xfd,
	0xff,0xff,0xff,0xee,0xbd,0xf7,0xeb,0xff,
	0xff,0xff,0x7d,0x72,0xff,0xbf,0x7f,0xff,
	0xff,0xff,0xff,0xff,0xff,0xd7,0xbf,0xff,
	0xb7,0xfd,0xbf,0xbd,0xdf,0xfb,0x7b,0x9b,
	0xb7,0xff,0xff,0xff,0xdf,0xff,0xff,0xfb,
	0xff,0xfb,0xbe,0x7f,0x8f,0xff,0x8f,0xef,
	0xff,0xff,0xf3,0xff,0xfb,0xff,0xe8,0xf3,
	0xcf,0xdf,0x3f,0xff,0xfe,0xcf,0xef,0x7c,
	0xff,0xff,0xff,0xff,0xff,0x7f,0xcb,0xff,
	0xff,0xff,0xff,0xfd,0xff,0x5f,0xfc,0xff,
	0xfd,0xbf,0xff,0x9f,0xff,0xff,0xff,0xbf,
	0x9f,0x7f,0xfb,0xff,0xfe,0xff,0xff,0xff,
	0x37,0xe7,0xfb,0x7f,0xff,0xf9,0xfe,0x4f,
	0xfb,0xff,0xff,0xfe,0xfb,0xdf,0xdf,0xe1,
	0xff,0xff,0xff,0xff,0xff,0xfa,0xcf,0xbd,
	0xff,0xeb,0xfa,0xdf,0xbf,0xdf,0xee,0xfa,
	0xaf,0xdb,0xbd,0xff,0xef,0xbe,0xbf,0x7c,
	0xff,0xdf,0xff,0xff,0xfe,0xff,0xff,0xff,
	0x57,0xff,0xff,0xff,0xff,0xff,0x9f,0xb7,
	0xff,0xfd,0xff,0xdf,0xff,0xff,0xfe,0xfb,
	0xdf,0xbf,0xfd,0xff,0xff,0x7f,0xff,0xfc,
	0xf7,0xdf,0xff,0xff,0xfd,0xf7,0xff,0xff,
	0xef,0x0f,0xfb,0xff,0xff,0xbf,0xff,0x7f,
	0xff,0x7f,0xf3,0xff,0x7f,0xff,0xff,0xfd,
	0xff,0xff,0xfe,0xf7,0xff,0xfe,0xff,0xda,
	0xef,0xbf,0x7e,0xff,0xff,0xf7,0xcd,0xff,
	0xf7,0xdf,0xff,0x4b,0xfb,0xff,0xf9,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xcf,0xfc,
	0xfe,0xfd,0xf7,0xdf,0xff,0xcd,0xef,0xdf,
	0x7f,0xff,0xad,0xff,0xdf,0xfc,0xfe,0xfd,
	0xf7,0xdf,0x7f,0xff,0x53,0xf6,0xff,0x9f,
	0xff,0xff,0xff,0xf3,0xff,0xff,0xff,0xff,
	0xf2,0xff,0xef,0xbf,0xff,0xfe,0x6f,0xfd,
	0xff,0xfe,0xfb,0xee,0xff,0xff,0xd6,0xff,
	0xec,0xbf,0xff,0x7e,0xfb,0xeb,0xf3,0xff,
	0xef,0xff,0xff,0xff,0xfb,0xef,0xbf,0xff,
	0xff,0xfb,0xff,0xda,0x6b,0x67,0xbf,0xf6,
	0xff,0xe7,0xbd,0xfc,0xff,0xff,0x66,0xff,
	0x7f,0xf6,0xd3,0xcf,0xfd,0xf7,0xdf,0x0f,
	0xed,0xdf,0xfe,0xff,0xff,0xb7,0xdf,0x7e,
	0xfb,0xfd,0xb7,0xff,0xbf,0x9f,0xfe,0xfb,
	0xff,0xfe,0xff,0xfb,0xbd,0xbf,0xff,0xfd,
	0xfb,0xff,0xb7,0x9f,0x7e,0xfa,0xff,0xff,
	0xbf,0xf8,0xff,0xfb,0xbf,0xff,0xff,0xfa,
	0xeb,0xaf,0xbf,0xfe,0xfe,0xef,0xb5,0xd3,
	0x5f,0xfb,0xb7,0xfa,0xde,0x8f,0xad,0xbc,
	0x3a,0x5e,0xff,0x6f,0xf7,0xdf,0xde,0x2f,
	0xbd,0xfe,0x4b,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xaf,
	0xad,0x6e,0xda,0xef,0xdd,0xff,0xfb,0x79,
	0xa7,0xdd,0xff,0xfa,0xff,0xef,0xdf,0xf6,
	0xbb,0xf9,0xf5,0x36,0xfe,0xff,0xff,0xf7,
	0xff,0xff,0xff,0xff,0xbf,0xff,0xfe,0xff,
	0xfd,0xfd,0xf7,0xf7,0x3f,0x7d,0xdf,0xd7,
	0x57,0x5f,0x7f,0xff,0xf7,0xff,0x7d,0x7f,
	0xf5,0xd7,0x5f,0xff,0x2f,0xf1,0xff,0xff,
	0xff,0xff,0xff,0xfb,0xef,0xbf,0xfb,0xe6,
	0xff,0xff,0x7b,0xef,0xbd,0xf7,0x6b,0xff,
	0xbf,0xf6,0xda,0x6b,0xff,0x96,0xf7,0xff,
	0x7b,0xaf,0xbd,0xf6,0xde,0xff,0xc9,0xff,
	0xbf,0xfd,0xff,0xff,0x6f,0xbf,0xff,0xe6,
	0x7b,0xef,0xbf,0xff,0xf7,0xdf,0x7f,0xff,
	0xb7,0xfb,0x7f,0xfb,0xff,0xff,0xfb,0x7f,
	0xbe,0xff,0xf7,0xff,0x7f,0xfb,0x9f,0xbf,
	0xfe,0xff,0x59,0xff,0xff,0x79,0xd7,0xdd,
	0x77,0xdf,0x7d,0xfe,0xf9,0xff,0xbd,0xf5,
	0xdf,0x7f,0x9e,0xff,0xf6,0x9f,0x7f,0xde,
	0xff,0xf6,0xf9,0xef,0xfd,0xff,0xfe,0xdb,
	0x7d,0xe7,0xff,0xd7,0xdd,0x6f,0xff,0x7d,
	0xf7,0xdd,0x69,0x8f,0xf5,0xc7,0xdf,0x6f,
	0x9f,0xfd,0xdf,0x7f,0xbc,0x9d,0xf6,0xda,
	0x6b,0xff,0xdd,0xc7,0xda,0x6f,0xaf,0xbd,
	0xf6,0xf9,0x89,0xfd,0xf7,0xff,0x6f,0xff,
	0xfd,0xff,0xdf,0x7f,0xff,0xfd,0xff,0xdf,
	0xbf,0xbf,0xff,0xdf,0xff,0xff,0xfe,0xff,
	0xdf,0x7f,0xff,0xff,0xff,0xdf,0x6f,0xff,
	0xff,0xf7,0xef,0x3b,0xfd,0xf7,0xff,0x7f,
	0xff,0xfd,0xff,0xdf,0xff,0xfe,0xfb,0xff,
	0xdf,0xff,0xff,0xff,0xff,0xfe,0xff,0xff,
	0xff,0xbf,0x7f,0xff,0xff,0xff,0xdf,0xff,
	0xfe,0xff,0xf7,0xff,0xff,0xeb,0x7f,0x8f,
	0xff,0xbf,0xfe,0xe3,0x8b,0x3f,0xbe,0xd8,
	0xe2,0x9f,0xbf,0xbd,0xf6,0xdb,0xaf,0xfd,
	0xf9,0xda,0x6f,0xbf,0xfd,0xf8,0xda,0x9f,
	0xaf,0xfd,0xf6,0xda,0x6b,0xff,0x3f,0xff,
	0xdf,0xfd,0xf7,0xe5,0x5f,0x5f,0xfd,0xf5,
	0xd5,0x5f,0xfb,0x7d,0xf5,0x57,0x57,0x7f,
	0xf5,0xdf,0x5f,0x7d,0xfd,0xf5,0xd5,0x57,
	0xfd,0xfd,0xd5,0xd7,0x57,0x5d,0xe5,0xbf,
	0xf9,0xff,0xf2,0xff,0xef,0xff,0xec,0xfe,
	0xaf,0xee,0xba,0xec,0xf7,0xff,0xbf,0xff,
	0xff,0xff,0x7d,0xff,0xfe,0xfb,0xef,0xbf,
	0xff,0xde,0xf7,0xef,0xbf,0xff,0xff,0xff,
	0xff,0xda,0xff,0xd7,0xe7,0x9f,0x7d,0xfc,
	0xb1,0xc7,0x9f,0x7d,0xf6,0xf9,0x67,0x87,
	0x1f,0x76,0xe8,0x87,0x7f,0x7e,0xf0,0xc1,
	0x07,0x79,0x76,0xf8,0xe7,0x87,0x1d,0x7e,
	0xf8,0xe1,0x57,0xfe,0xff,0xfe,0xff,0xeb,
	0xbf,0xff,0xfc,0xfb,0xef,0xbf,0xff,0xff,
	0xfb,0xeb,0xbf,0xff,0xfb,0xff,0xaf,0xff,
	0xfe,0xfb,0xff,0xef,0xff,0xff,0xfa,0xef,
	0xbf,0xff,0xfe,0xff,0xf2,0xff,0xbd,0xff,
	0x5f,0xff,0xed,0xb7,0xdf,0xde,0xfb,0xef,
	0xff,0xd7,0x6d,0x37,0xdd,0xde,0xed,0xff,
	0xdd,0x76,0xdb,0x6d,0x7b,0xdf,0xfe,0xdf,
	0xcd,0x37,0xdf,0x7c,0xfb,0x9b,0xff,0xff,
	0xff,0xff,0xfe,0xef,0x7f,0xff,0xfe,0xff,
	0xef,0xff,0xff,0x9f,0x7b,0xfe,0xd9,0x9e,
	0xff,0xef,0xb9,0xe7,0x9e,0xff,0xff,0xf9,
	0xff,0x9e,0x7b,0xee,0xb9,0xe7,0xef,0xfd,
	0xff,0xff,0xbf,0xdf,0xdf,0xff,0xfb,0xfd,
	0xff,0xff,0xff,0xef,0x37,0xf9,0x64,0x93,
	0x37,0xf9,0xfe,0x13,0x4f,0x3c,0xd9,0xff,
	0x93,0xef,0x3f,0xf1,0xc4,0x13,0x4f,0xfe,
	0xe0,0xff,0xbf,0xff,0xfe,0xfb,0xef,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfe,0xff,0xff,0xf3,0xff,0xff,0x4f,0xff,
	0xdf,0xf6,0xbf,0x2f,0x3f,0xfd,0xff,0xfb,
	0xfb,0x5b,0xfd,0xff,0xb5,0xd7,0x5e,0x7b,
	0xed,0xf5,0xd7,0x5f,0x7f,0xfd,0xf5,0xd7,
	0x5f,0x6e,0xbd,0xd7,0x5f,0x7d,0xfd,0x55,
	0xd7,0x5f,0x77,0xed,0xf5,0xd7,0x5b,0x7d,
	0xbd,0x65,0x3f,0xfb,0xff,0xf3,0xff,0xfe,
	0xff,0xfc,0xf3,0xef,0xff,0xff,0xfc,0xbf,
	0xcf,0x37,0xef,0xfc,0xcf,0xf5,0xfe,0xfc,
	0xf3,0xcf,0xff,0x9f,0xfc,0xbf,0xc5,0x3f,
	0xff,0xfc,0xf3,0xff,0x87,0xff,0xff,0xff,
	0xff,0xf7,0xff,0xff,0xff,0xdf,0xff,0xff,
	0xfb,0xef,0xff,0xff,0xff,0xff,0xfb,0xff,
	0xbe,0xfe,0xef,0xbf,0xff,0xff,0xfd,0xff,
	0xee,0xff,0xfe,0xff,0xee,0x57,0xfe,0x3f,
	0xff,0xff,0xff,0xcf,0x3f,0xff,0xdd,0x7f,
	0xcf,0xff,0xbf,0xfc,0xf3,0xcf,0xbf,0xfc,
	0xff,0xcf,0x3f,0xf7,0xfc,0xff,0xcf,0xff,
	0xff,0xfc,0xf2,0xc7,0x17,0xf7,0xf9,0xf4,
	0xff,0xfd,0xfd,0xfe,0xff,0xff,0xf5,0xff,
	0xdf,0xff,0xfd,0xff,0xff,0xff,0xff,0xff,
	0xdf,0xff,0xff,0xff,0xff,0xfe,0xfb,0x7f,
	0xdf,0xff,0xff,0x7f,0x7f,0xfb,0xfd,0xfd,
	0xa8,0xff,0x7f,0xff,0xeb,0xff,0xff,0xbf,
	0xff,0xff,0xf7,0xfb,0xfb,0xff,0xff,0xff,
	0xff,0x4d,0xaf,0xff,0xfe,0xff,0xff,0xff,
	0xff,0xef,0xfa,0xff,0xff,0xfd,0xdf,0xbf,
	0xdf,0x9f,0xfd,0xff,0x7f,0xff,0xbf,0xff,
	0x7f,0xe7,0xff,0xff,0xff,0xf7,0x7f,0xbf,
	0xdf,0x7d,0x3e,0xf7,0xbf,0x7e,0xf7,0xdd,
	0x7f,0xff,0x3d,0xff,0xeb,0xff,0xdf,0x7d,
	0xff,0xfd,0x7f,0xeb,0xdf,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0x77,0xff,0x7d,
	0xef,0x3f,0xfd,0xec,0xfb,0xff,0xff,0xfc,
	0xcb,0x6f,0xef,0xff,0xdf,0xff,0xff,0x3b,
	0xfd,0xfc,0x23,0xff,0x6e,0xff,0xff,0xfb,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xef,0x7f,0xff,0xf7,0xf7,0xbd,0xfb,0xff,
	0xf3,0xff,0xfa,0xfd,0xff,0xfb,0xff,0xff,
	0xff,0xfd,0xf7,0xff,0xff,0x27,0xf8,0xff,
	0xf8,0xe3,0xdf,0xff,0xff,0xfc,0xff,0xf7,
	0xdf,0xff,0xfc,0xff,0xff,0xff,0xff,0xfe,
	0x87,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfc,0xf9,0xd7,0x9f,0xff,0xf5,0xbf,0xd3,
	0xff,0xef,0xbf,0xff,0xfd,0xff,0xdf,0xff,
	0xff,0xff,0xff,0xb7,0xfe,0xff,0xff,0xff,
	0xff,0xff,0xfd,0xff,0xff,0xff,0xff,0xff,
	0xff,0xb7,0xde,0xff,0xfe,0xfd,0xeb,0xfb,
	0x2f,0xfe,0xff,0xff,0xff,0xff,0xff,0x7f,
	0xfd,0xf7,0xfd,0xaf,0xff,0xfd,0xff,0xed,
	0xff,0xff,0xff,0xfd,0xab,0xff,0xff,0xfa,
	0xdd,0xfe,0xff,0xfd,0xff,0xff,0xff,0xff,
	0xf7,0x7f,0xe4,0xff,0xff,0xff,0xfe,0xff,
	0xff,0xff,0xff,0xff,0x7f,0xfd,0xff,0xff,
	0xff,0xff,0xfd,0xfb,0xfd,0xff,0xff,0xff,
	0xdf,0xff,0xff,0xff,0xff,0xdf,0x5e,0xff,
	0xff,0xf5,0xff,0x87,0xff,0xff,0xff,0xf7,
	0xff,0xff,0xff,0xfe,0xff,0xff,0xff,0xff,
	0xde,0xff,0xfe,0xf7,0xfd,0xff,0xff,0xff,
	0xff,0x7f,0xff,0xff,0xff,0xff,0xfe,0xfe,
	0xff,0xff,0xff,0xff,0xdf,0xf0,0xe7,0x97,
	0x7f,0x7e,0xf3,0xe5,0xe7,0xff,0xff,0xff,
	0xcf,0xff,0xdf,0x77,0xff,0xfd,0xdf,0xff,
	0xdf,0xfd,0xff,0xdf,0xff,0xfe,0xcd,0xff,
	0xdd,0x7f,0xff,0xfd,0xf7,0x3f,0x68,0x7d,
	0xfe,0xfc,0xf3,0x5d,0x2f,0xdf,0xff,0xff,
	0xff,0x7f,0xfd,0xff,0xbe,0xfb,0x6f,0xf6,
	0xff,0xff,0xee,0xf7,0xe9,0xff,0x9f,0xff,
	0xff,0x6f,0xff,0xfd,0xf7,0xdf,0xbf,0x0e,
	0xef,0xff,0xfd,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0x7f,0x9f,0x7f,0xde,
	0xff,0xf5,0xff,0xf2,0x7f,0xff,0xf7,0x7f,
	0x76,0xff,0xff,0x3d,0xf5,0xd2,0x73,0xff,
	0x7d,0xdb,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xfb,0xff,
	0xe7,0xce,0xef,0xff,0xbf,0xff,0xff,0xfb,
	0xff,0xbf,0xff,0x7f,0xfb,0xac,0xbf,0xce,
	0xfe,0xbf,0xc9,0xff,0xff,0xfe,0xfb,0xef,
	0xbf,0xff,0xfe,0xfb,0xff,0xff,0xff,0x7e,
	0x7b,0xff,0xbd,0xcf,0xfb,0xef,0xff,0xd7,
	0x6e,0xb3,0xef,0xf7,0xff,0xd6,0x37,0xef,
	0xfd,0xf7,0xeb,0x1f,0xf6,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xdb,0xfb,0xef,0xfd,0xde,0xff,0xff,
	0xbf,0xb7,0xda,0xff,0xbf,0xff,0xff,0x7e,
	0x79,0xef,0xff,0x7e,0xef,0xe7,0xff,0xff,
	0xff,0xf9,0xfd,0xf7,0xdf,0x7f,0xff,0xff,
	0xff,0xdf,0xd7,0x5f,0x7f,0xfd,0xd7,0xff,
	0x7d,0xfd,0xf5,0xdf,0xff,0x7d,0xfd,0xdf,
	0xdf,0xdf,0x7f,0xff,0xf5,0xfd,0xb1,0xff,
	0xff,0xfc,0xc3,0xdf,0x7f,0xff,0xfd,0xff,
	0xff,0xff,0xff,0xbf,0xf6,0xda,0x6b,0x9f,
	0xf6,0xff,0x6b,0xef,0xb1,0xf6,0xff,0x6b,
	0xff,0xbf,0xf7,0xde,0x7b,0xaf,0xfd,0xdf,
	0xfd,0xff,0xfb,0xef,0xbf,0xff,0xfe,0xfb,
	0xff,0xff,0xfd,0xff,0xfb,0x7f,0xfb,0xed,
	0xb7,0xff,0xbf,0xff,0xb7,0xff,0xff,0xbf,
	0xff,0xb7,0xfb,0xff,0xff,0xff,0xff,0xff,
	0xff,0xe4,0xff,0x9f,0x7f,0xfe,0xf9,0xe7,
	0x9f,0xff,0xff,0x1f,0xff,0x9f,0xff,0xe7,
	0xdb,0x7f,0xf6,0xf7,0xf9,0x6f,0xff,0xf9,
	0xff,0xf9,0x6f,0x9f,0xff,0xfe,0xff,0x6f,
	0xbe,0xf9,0x07,0xfe,0x7f,0xfe,0xf9,0xc7,
	0x1f,0x7f,0xfc,0xfd,0xf6,0xdb,0x7f,0xbc,
	0xfd,0xf7,0xdd,0xeb,0x7d,0xc7,0xdf,0x63,
	0xbf,0xfd,0xc7,0xdf,0x77,0xfc,0xfd,0xf6,
	0xdb,0x6b,0xff,0x5f,0xd9,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xfe,0xfb,0xff,
	0xbf,0xfd,0xff,0xff,0x7f,0xff,0xff,0xff,
	0x7f,0xbf,0xfd,0xff,0xdf,0xff,0xff,0xfd,
	0xfe,0xef,0x7f,0xff,0xbf,0xd7,0xff,0x7f,
	0xff,0xfd,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xfb,0xff,0xff,0x7f,0xef,0xff,
	0xff,0xff,0xfd,0xfd,0xff,0xdf,0xff,0xff,
	0xf7,0xff,0xff,0xff,0xfd,0xff,0xcf,0xfa,
	0xf7,0xd1,0xc7,0x5f,0x7d,0xfc,0xf5,0xff,
	0xef,0xaf,0xff,0xf9,0xda,0x6f,0xbf,0xfd,
	0xdb,0x9f,0xaf,0xbd,0xf6,0xdb,0x9f,0xbf,
	0xfd,0xf9,0xdb,0x6f,0xbf,0xfd,0xf6,0xdf,
	0xf7,0xff,0xdf,0x7f,0xff,0xfd,0xf7,0xdf,
	0x7f,0xdf,0x7e,0xfb,0xdf,0x57,0x7f,0xf5,
	0xf5,0x77,0xfb,0x7d,0xd5,0xd5,0x77,0xff,
	0x7d,0x97,0xdf,0x57,0xdd,0x7d,0xd7,0x57,
	0xff,0xb7,0xff,0x7f,0xf3,0xdd,0xbf,0xff,
	0xfc,0xff,0xdf,0xfa,0xe9,0x7f,0xfb,0xff,
	0xfb,0xef,0xff,0xfe,0xb7,0xfb,0xbf,0xff,
	0xfe,0xef,0xef,0x7f,0xff,0xfe,0xfb,0xeb,
	0xbf,0xff,0x7e,0xfc,0xff,0x7f,0xfe,0xf9,
	0xe7,0x9e,0x7f,0xfe,0xd1,0xc7,0x9f,0x7f,
	0x7e,0xd8,0x61,0x87,0x7d,0xf8,0xc7,0x07,
	0x1f,0x7c,0xf8,0xc7,0x87,0x7f,0x74,0xf0,
	0x41,0x07,0x1f,0xfe,0xea,0xff,0x9f,0xff,
	0xff,0xff,0xe7,0xff,0x7f,0xff,0xfc,0xfa,
	0xff,0xbf,0x9f,0x7e,0xfa,0xa5,0xdf,0xfd,
	0xfb,0xfd,0xb7,0xdf,0xfd,0xfa,0xfd,0xff,
	0xdf,0x7e,0xfb,0xed,0xf7,0x7f,0xff,0xff,
	0xfd,0xf7,0xdf,0x7f,0xff,0xfd,0xff,0xfd,
	0xf7,0xff,0x7f,0x5f,0x74,0xd1,0x65,0x5d,
	0xfe,0xdf,0xe5,0x97,0x5d,0xfe,0xd7,0xe5,
	0x7f,0x5f,0x76,0xf1,0x65,0x97,0xff,0xf9,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x9f,
	0x2f,0xbf,0xfe,0xef,0xaf,0xf9,0xe6,0x9b,
	0xbf,0x79,0xfe,0x9b,0x6b,0xae,0x79,0xfe,
	0x9b,0xef,0xbf,0xb9,0xe6,0x9b,0x6b,0xfe,
	0xd3,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0x7f,0xf9,0xfd,0xff,0xdf,0xf7,0xdf,
	0x7f,0xff,0xf7,0xff,0x7f,0xf7,0x7d,0xf7,
	0xff,0x7f,0xff,0xdf,0xf7,0xdd,0x7f,0xf7,
	0xfd,0x9d,0xf6,0xff,0xf3,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xf2,
	0xcf,0x2f,0xff,0xf2,0xff,0x4b,0xbf,0xfc,
	0xf2,0x7f,0x2f,0xff,0xff,0xf3,0xfb,0x2f,
	0xff,0xff,0xff,0xd1,0xff,0x5f,0x3f,0xfd,
	0xf5,0xd3,0x5f,0x7f,0xf5,0xb5,0x57,0x5f,
	0x77,0xf5,0x75,0x57,0x77,0xfd,0xf5,0x97,
	0x5d,0x73,0xfd,0x75,0xd7,0x5f,0x7f,0xdd,
	0xf5,0xd6,0x4f,0xff,0xa2,0xff,0xff,0xfe,
	0xef,0xf7,0xdf,0xbf,0xff,0xfc,0xf3,0xff,
	0xff,0xfe,0xfe,0xf7,0xef,0xff,0xff,0xab,
	0xc7,0xff,0xff,0xff,0xe7,0x7f,0xff,0xca,
	0xff,0xff,0xff,0xff,0xdf,0x7b,0xf9,0xff,
	0xff,0xff,0xef,0xb7,0xff,0xfd,0xfc,0x6f,
	0xff,0xff,0xff,0xff,0xf3,0xef,0xf7,0x7f,
	0xdf,0x5b,0xbf,0xff,0xff,0xf3,0xef,0xbf,
	0xff,0xbf,0xff,0xeb,0xfe,0xbf,0x7f,0xe1,
	0xff,0xff,0xff,0xff,0xfb,0xff,0xff,0xbf,
	0x7f,0xff,0xff,0xff,0xdf,0xfd,0xff,0xfd,
	0xfb,0xef,0xff,0xff,0xff,0xff,0xff,0xf7,
	0xbf,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0x45,0xff,0xff,0xff,0xff,0xff,0xfd,0xff,
	0xff,0xff,0xff,0xff,0xff,0x77,0xfd,0xf5,
	0xd7,0x3b,0xfd,0xbf,0xdd,0x5f,0x6f,0xbd,
	0xff,0xd7,0xfd,0x7f,0xbd,0xe7,0x96,0xff,
	0xfe,0xbe,0xf8,0xff,0xff,0xff,0xff,0xff,
	0xef,0xff,0xff,0xff,0xff,0xff,0xff,0x66,
	0x9f,0x7f,0xfe,0x67,0xff,0xbf,0xf7,0xd7,
	0x67,0xff,0x7f,0xe6,0xff,0xdf,0xdf,0xff,
	0xfe,0xbf,0xf6,0xc9,0xff,0xff,0xfd,0xff,
	0xff,0xff,0xff,0xff,0xfe,0xdf,0xef,0xff,
	0xfd,0xf7,0xdf,0x7f,0xf7,0xf7,0xbf,0x57,
	0xf7,0xff,0xff,0xff,0xff,0xff,0xda,0xf7,
	0xfd,0xfd,0xff,0xff,0x5f,0xfe,0xff,0xff,
	0xff,0x7f,0xfb,0xfd,0xff,0xff,0xff,0xff,
	0xff,0xfb,0xf7,0xff,0xff,0x3f,0xff,0xff,
	0x6d,0xff,0xfd,0x77,0xdf,0xff,0xff,0xf7,
	0xff,0xd3,0xfe,0xff,0xff,0xdd,0x76,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xf7,
	0xff,0xff,0xff,0xff,0xff,0xff,0xef,0xfd,
	0xff,0x7f,0xfb,0xbf,0xff,0xff,0xff,0xef,
	0xff,0xff,0xcf,0xff,0xef,0xff,0xfe,0xbf,
	0xff,0xcf,0x2f,0xbf,0xfe,0xf3,0xcb,0xff,
	0xff,0xfc,0xf7,0xcf,0x3f,0xff,0xdc,0xf3,
	0x07,0x7f,0xfc,0xd3,0xce,0x3f,0xff,0xf4,
	0xf3,0xff,0x3f,0xf7,0xfc,0xf3,0xcf,0xff,
	0x9f,0xfc,0xff,0xbe,0xfb,0xf7,0xbf,0xff,
	0xee,0xff,0xed,0xbf,0x7f,0xee,0xfb,0xef,
	0xb7,0xff,0xfb,0xb7,0xbf,0xdf,0xfe,0xba,
	0xb5,0xbf,0xff,0xff,0xfb,0xef,0xbf,0xff,
	0xfe,0xef,0xee,0xff,0xf7,0xdf,0xff,0xfe,
	0xfd,0xe7,0xff,0x7f,0xfc,0xff,0xff,0xdf,
	0x5f,0xff,0xf9,0xdf,0xdf,0xbd,0x78,0xf7,
	0xdd,0xff,0xfd,0xf8,0xfd,0xdf,0x3f,0xfe,
	0xfd,0xe3,0xff,0x5f,0xfe,0xff,0xff,0xff,
	0xe3,0xdd,0x7f,0x7f,0xfd,0xf7,0xff,0xff,
	0xf6,0xff,0xff,0xdf,0xbf,0xff,0xbf,0xdf,
	0xff,0xef,0xff,0xff,0xff,0xff,0xbf,0xff,
	0xfe,0xfb,0xff,0xff,0x5f,0xf8,0xff,0xbf,
	0xff,0xdf,0x7f,0xfb,0x6a,0xff,0xaf,0xff,
	0xff,0xff,0x7f,0xff,0x3f,0xff,0xff,0xff,
	0xfd,0xfe,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xfb,0x1d,0xff,
	0xff,0xff,0xff,0x8f,0x3f,0xfe,0xbf,0xff,
	0x3f,0xff,0xff,0xff,0xfd,0xf7,0xdf,0xdf,
	0xcd,0xff,0xdf,0x77,0xdf,0xfd,0xef,0xdf,
	0xff,0xff,0xfd,0xf7,0xdf,0x77,0xff,0x43,
	0xd6,0xff,0xff,0xff,0xff,0xfc,0xf3,0x5f,
	0xff,0xff,0xff,0xff,0xff,0xf5,0xdb,0x4f,
	0x7f,0xf7,0xff,0x7f,0xbf,0xfd,0xf7,0xff,
	0x6f,0xd7,0xff,0xf6,0xdf,0x6f,0xff,0xfd,
	0x6b,0xf2,0xfe,0xff,0xff,0x59,0xfe,0xff,
	0xff,0xbf,0xff,0xff,0xfb,0xff,0xd2,0x7f,
	0x67,0xbf,0xd6,0xff,0xff,0x9d,0xf7,0xd4,
	0xff,0xff,0xfd,0xff,0xdf,0x53,0xff,0xfd,
	0xf5,0xdf,0xa7,0xfd,0xdf,0x7f,0xdf,0xe8,
	0xf7,0xdf,0x7f,0xfb,0xfd,0xd7,0xff,0xb7,
	0xff,0xfa,0xfb,0xbd,0xff,0xff,0xef,0xef,
	0xa7,0xff,0xff,0xeb,0xff,0xbf,0x9e,0x7a,
	0xea,0xef,0xff,0xff,0xfd,0xff,0xeb,0xaf,
	0xbd,0xfe,0xfa,0xeb,0xef,0xbf,0xfe,0xfa,
	0x6f,0xb7,0xfb,0x7f,0xdb,0xbd,0xfa,0xde,
	0x7f,0xcf,0xbc,0xfa,0xfe,0xaf,0xef,0xff,
	0xf7,0xfe,0x3f,0xbf,0xfe,0x64,0xff,0xff,
	0xff,0xff,0xfb,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0x89,0x6f,0xde,0xf5,0xdd,
	0xff,0xfe,0x7b,0xf5,0xdd,0xff,0xbe,0xfd,
	0xef,0x9b,0xf7,0xbe,0xfc,0xf7,0x06,0xfe,
	0xff,0xff,0xff,0xf5,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0x7d,0xfd,0xf7,0xf7,0x37,
	0x7d,0x9f,0xd7,0x7f,0x7f,0x7d,0xdf,0xdf,
	0xff,0xfd,0xfd,0xfd,0xdf,0x5f,0xdf,0x2f,
	0xf9,0xff,0xef,0xbf,0xaf,0xfe,0xfb,0xef,
	0xff,0xff,0xd6,0xfb,0x7f,0x79,0xaf,0xbd,
	0xf7,0x7b,0xff,0xbf,0xf6,0xde,0x7b,0xff,
	0xb7,0xf6,0x7f,0x69,0xef,0xbd,0xf6,0xde,
	0xff,0xd9,0xff,0xbf,0xfd,0xb6,0xfb,0xef,
	0xbf,0xff,0xfe,0xfb,0xef,0xbf,0xff,0xe7,
	0xff,0xff,0xff,0xb7,0xfb,0xff,0xff,0xff,
	0xff,0x7b,0x7f,0xbf,0xff,0xf7,0xff,0x7f,
	0xfb,0xff,0x0f,0xfe,0xff,0x7d,0x77,0xde,
	0x7f,0xf7,0xfd,0x67,0xdf,0x7f,0xf7,0xf9,
	0x7f,0xbf,0xfd,0xff,0x7f,0x9e,0xff,0xfe,
	0x9f,0x7f,0x9e,0xff,0xd6,0xf9,0xef,0xfd,
	0xff,0xf6,0xdb,0xff,0xe0,0xff,0xf6,0xda,
	0x6b,0xff,0x7d,0xf7,0x5f,0x63,0xaf,0x7d,
	0xc7,0xdb,0x7d,0xf7,0xfd,0xda,0x7f,0xbc,
	0xfd,0xf6,0xdf,0x7f,0xfc,0xdd,0xc7,0xdf,
	0x7f,0xff,0x7d,0x77,0xff,0xa5,0xfd,0xf6,
	0xdf,0xff,0xff,0xff,0xf7,0xff,0x7f,0xff,
	0xfd,0xff,0xef,0xff,0xff,0xff,0xfe,0xff,
	0xff,0xfc,0xf6,0xdf,0xff,0xff,0xff,0xff,
	0xdf,0x7f,0xff,0xff,0xff,0xff,0x5b,0xfc,
	0xef,0xdf,0xff,0xff,0xff,0xf7,0xff,0xff,
	0xfe,0xfb,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfe,0xff,0xf7,0xdf,0xdf,0xff,0xff,0xff,
	0xff,0x7f,0x7f,0xff,0xff,0xff,0xff,0xff,
	0xae,0xff,0xcb,0x2f,0xbf,0xf8,0xe3,0x8f,
	0x3f,0xbe,0xf8,0xe2,0x9f,0xbf,0xbd,0xf6,
	0xdb,0xaf,0xfd,0xf9,0xdb,0x6b,0xaf,0xfd,
	0xf9,0xda,0x9f,0xaf,0xbd,0xf6,0xda,0x6b,
	0xff,0x07,0xff,0x5f,0x7d,0xf5,0xdd,0x57,
	0x7f,0xfd,0xf5,0xd7,0x57,0xfb,0x7d,0xf7,
	0x55,0x5f,0x7f,0xf5,0xdf,0x5f,0x5d,0x7d,
	0xf5,0xdf,0x57,0xfd,0x7d,0xd5,0xdd,0x57,
	0x5d,0xb5,0x7f,0xf8,0xff,0xff,0xff,0x2e,
	0xff,0xff,0xbb,0xef,0xaf,0xb6,0xed,0xaf,
	0xef,0xff,0xff,0xfe,0xef,0x7f,0xff,0xfe,
	0xfb,0xef,0x7d,0xbf,0xff,0xaf,0xff,0xaf,
	0xff,0xff,0xfe,0xef,0xd2,0xff,0xff,0xe7,
	0x1f,0x7f,0xe6,0x99,0x67,0x9e,0x7d,0xf6,
	0xf9,0xc7,0x87,0x1f,0x76,0xd8,0x87,0x7f,
	0x76,0xf0,0xc1,0x87,0x7f,0x7c,0xf8,0xe7,
	0x07,0x1f,0x7e,0xf0,0xe1,0x4f,0xfe,0xff,
	0xfe,0xfb,0xef,0xbf,0xff,0xfe,0xfa,0xef,
	0x3f,0xff,0xfd,0xfa,0xed,0xb7,0x5f,0xfe,
	0xfd,0xbf,0xdf,0x7f,0xff,0x9d,0xbf,0xdf,
	0xff,0xfb,0xed,0xb7,0xdf,0x7e,0x7f,0xf6,
	0xff,0xdd,0x77,0xff,0x7d,0xf5,0xdd,0x57,
	0xff,0x7d,0xf7,0xef,0xd7,0xc5,0x17,0x5f,
	0xde,0xe5,0x7f,0x5f,0x7e,0xf9,0xe5,0xff,
	0x5d,0xfe,0xf7,0xe5,0x17,0x5f,0x76,0xf9,
	0x93,0xff,0xbf,0xff,0xfe,0xfb,0xff,0xbf,
	0xff,0xff,0xfb,0xdf,0xff,0xff,0x9f,0x7b,
	0xee,0xd9,0x9f,0xff,0xef,0xb9,0xe7,0x9e,
	0xff,0xff,0xf9,0xff,0x9e,0x7b,0xee,0xf9,
	0xe7,0x5f,0xfd,0x7f,0xff,0xfd,0xff,0x7f,
	0x7f,0xff,0xff,0xff,0xff,0xfe,0xef,0x3f,
	0xf7,0xdc,0xf3,0x3f,0xff,0xde,0x73,0xcf,
	0x3d,0xff,0xfe,0xf3,0xef,0x3d,0xff,0xdc,
	0xf3,0xcf,0xff,0xe8,0xff,0xbf,0xff,0xff,
	0xfb,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0x2f,0xbf,0xfc,0x64,0x6f,0xff,0xfb,0xf2,
	0x8b,0xff,0xff,0xfb,0xf2,0xff,0x2f,0xbf,
	0xfd,0xf4,0x9b,0xff,0x1b,0xfd,0xff,0xf5,
	0xd7,0x5f,0x7f,0xfd,0xf5,0xd7,0x5f,0x7e,
	0xfd,0x75,0x57,0x5d,0x76,0xb9,0x96,0x5f,
	0x73,0xb9,0xf5,0x57,0x5f,0x77,0xf4,0x55,
	0x57,0x5d,0x2d,0xbc,0xf5,0x3b,0xf8,0xff,
	0xf3,0xcf,0x3f,0xff,0xff,0xff,0x8f,0xff,
	0xdf,0x7c,0xff,0xff,0xfe,0xff,0xff,0xfb,
	0xff,0xf6,0xff,0xff,0x7d,0xfd,0xde,0xff,
	0x6f,0xff,0xef,0xff,0xf7,0xfe,0xbf,0x9b,
	0xff,0xf7,0x3f,0x5f,0xff,0xff,0xff,0x39,
	0xbf,0xbf,0xf5,0xfe,0xef,0xbf,0xff,0xee,
	0xfb,0xff,0x5f,0x7f,0xfb,0xfd,0xbf,0xff,
	0xff,0xfb,0xed,0xf7,0xdf,0xfe,0xbe,0xef,
	0xa7,0xfe,0xff,0xfc,0xff,0xed,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xdf,0xff,0x7f,
	0xff,0xbf,0xbf,0xff,0xff,0xff,0xff,0xff,
	0x7f,0xff,0xff,0x7f,0x7f,0xfd,0xfd,0xff,
	0xff,0xff,0xf3,0xff,0xf5,0xff,0xdf,0xdf,
	0xff,0xff,0xd7,0xdf,0x7f,0xfd,0xff,0xd7,
	0x5f,0x7f,0xbd,0xc3,0xef,0x7f,0xfd,0xf5,
	0xdd,0xff,0xf7,0xbd,0xff,0xd7,0xff,0xff,
	0xf9,0xe7,0xcf,0x87,0xff,0xaf,0xfb,0xff,
	0xb7,0xff,0xff,0xfb,0xff,0xfb,0xef,0xfb,
	0x6f,0xbe,0xd9,0xe6,0xaf,0xf7,0xff,0xe7,
	0x77,0xff,0xff,0xff,0xff,0xf7,0x7f,0xbe,
	0x7f,0xf6,0xf5,0x6f,0xed,0xfc,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0x7f,0xfe,0xcd,
	0xff,0xff,0xaf,0xbf,0xff,0x77,0xff,0xff,
	0xf7,0x7a,0xdf,0x7d,0xff,0xdd,0xfb,0xfb,
	0xe7,0xff,0xfc,0x7b,0xdf,0x7f,0xe2,0xff,
	0x7f,0xfb,0xff,0xff,0xff,0xfe,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xdf,0xef,0x77,
	0xff,0xbf,0xff,0xff,0x7f,0xf7,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xdf,0x43,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xfd,0xff,0xef,0xf7,0xdf,0xff,
	0xfa,0xff,0xfe,0x7e,0xff,0xfb,0xf7,0xff,
	0xff,0xfe,0xfd,0xab,0xff,0xff,0xfb,0xef,
	0xbb,0xfa,0xff,0xfc,0xf3,0xcf,0x3f,0xff,
	0xfc,0xf3,0xcf,0x3f,0xff,0xff,0xff,0xcd,
	0xff,0xff,0x7c,0xcf,0x3f,0xff,0xfc,0xf3,
	0xcf,0x3f,0xff,0xfc,0xf3,0xcf,0x3f,0xff,
	0xfc,0xff,0xc4,0xff,0xee,0xbf,0xff,0xfe,
	0xfb,0xe7,0xbf,0xff,0xfe,0xfb,0xfb,0xff,
	0x5f,0xeb,0xff,0xff,0x6f,0xfe,0xed,0xb7,
	0xd7,0x7e,0xfb,0xed,0xb7,0xdf,0x76,0xfb,
	0xf9,0xb7,0xd7,0x5d,0xfe,0x3f,0xfe,0xf1,
	0xe7,0xdf,0xbf,0xfe,0xf5,0xf7,0xdf,0xef,
	0xff,0xff,0xdf,0xff,0xff,0xf7,0xff,0x7d,
	0xff,0xfd,0xf7,0xdf,0x7f,0xdf,0x75,0xe7,
	0xdd,0xf7,0xdf,0xfd,0x7f,0xe0,0xff,0xf7,
	0xdf,0x3f,0xf7,0xfd,0xf7,0xff,0x7f,0xff,
	0xff,0xff,0xff,0xff,0xff,0xcf,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0x7f,0x97,0xff,
	0xbf,0x3f,0xff,0xfb,0xef,0xbf,0xff,0xff,
	0xfb,0xff,0xff,0xfe,0xff,0xff,0xff,0xdf,
	0xff,0xff,0xff,0xfd,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xdf,
	0xf5,0xff,0xff,0xff,0x3f,0xff,0xcf,0xbf,
	0xca,0xff,0xfe,0xe3,0xff,0x3f,0xfe,0xff,
	0xcf,0xbf,0xff,0xfe,0xe7,0xef,0xff,0xff,
	0xff,0xfb,0xef,0xbf,0xff,0xff,0xff,0xff,
	0x3f,0x64,0xfd,0xff,0xff,0xff,0xfd,0xff,
	0xff,0x75,0xff,0xff,0x3f,0xff,0xff,0xf3,
	0x9f,0x7f,0xf6,0xff,0xff,0xff,0xf3,0xff,
	0xff,0x9f,0xff,0xff,0xff,0xff,0xff,0x7f,
	0xfe,0xbf,0x0e,0xef,0xff,0xff,0xff,0xff,
	0xff,0xff,0xfe,0xff,0xff,0xff,0xff,0xfe,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x7f,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xfd,0xd9,0xff,0xfd,0xf7,
	0xdf,0x7f,0xff,0xf5,0xf7,0xdf,0x7f,0xff,
	0xed,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xfe,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xbf,0x9b,0xff,0xbb,
	0xfe,0xfa,0xfb,0xef,0xbf,0xff,0xfe,0xeb,
	0xaf,0xbf,0xff,0xfb,0xef,0xbf,0xff,0xfb,
	0xef,0xbf,0xdf,0xfe,0xfb,0xef,0xbf,0xff,
	0x5e,0xfb,0xef,0xbf,0xff,0xfe,0x8f,0xf6,
	0xef,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x7f,
	0xff,0xff,0xff,0xbf,0x7b,0xff,0xff,0xff,
	0xff,0xff,0xea,0xff,0xff,0xff,0xff,0xef,
	0xe5,0xff,0xff,0xff,0xeb,0xef,0xbf,0xfe,
	0xff,0xff,0xff,0xff,0xff,0xff,0xfb,0xfd,
	0xf7,0x7b,0xff,0xfd,0xf7,0xd7,0x7d,0xff,
	0xa9,0xf7,0xdf,0x56,0xff,0xad,0xe7,0xdf,
	0xff,0xb9,0xfc,0xff,0xfe,0xdb,0x3f,0xfe,
	0xfd,0xff,0xff,0xef,0xbf,0xfc,0xff,0x1b,
	0xff,0xff,0xf5,0x5f,0xff,0xff,0xe5,0xff,
	0xff,0x7f,0xff,0xff,0xa7,0xfe,0x7f,0xff,
	0xf5,0xff,0xff,0xf8,0xff,0xdb,0xee,0xbd,
	0xff,0xd6,0xdb,0xee,0xbf,0xed,0xf6,0xfb,
	0xef,0xbf,0xff,0xfe,0xef,0xbf,0xff,0xfe,
	0xfb,0xef,0xbf,0xf7,0xfe,0x7b,0xef,0xbf,
	0xf7,0xfe,0xfb,0xff,0xe6,0xff,0xdf,0x79,
	0xff,0xf9,0x67,0x9d,0x7d,0xf6,0x9d,0x77,
	0x9d,0x7d,0xfe,0xf9,0xe7,0x7f,0xfe,0xf9,
	0xe7,0x9f,0x7f,0xfe,0xf9,0xe7,0x9f,0x7f,
	0xfe,0xf9,0xe7,0x9f,0xff,0x27,0xfe,0x6f,
	0xbf,0xbd,0xd6,0x5b,0x77,0x9d,0x75,0xf7,
	0xdf,0x6d,0xf5,0xf9,0xc7,0x1f,0xbf,0xf1,
	0xc7,0x1f,0x7f,0xf8,0xf1,0xc7,0x1f,0x7f,
	0xfc,0xe1,0xc7,0x1f,0x7f,0xfc,0x1f,0xd8,
	0x3f,0xff,0xfc,0xf7,0xdb,0xff,0xff,0xfc,
	0xff,0xdf,0x3f,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xbf,
	0xc9,0xff,0xfd,0xf7,0xf7,0xbf,0xff,0xff,
	0xf7,0xff,0xbf,0xff,0xfd,0xff,0xfd,0xff,
	0xff,0xef,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0x4f,0xfe,0xff,0xfc,0xe3,0x8f,0x3f,
	0xfe,0xf8,0xe2,0xcb,0x3d,0xbe,0xf8,0xf7,
	0x8b,0x7f,0xfe,0xe7,0x9f,0x7d,0xfe,0xd9,
	0xe7,0x9f,0x7f,0xfe,0xd9,0x67,0x9f,0x7f,
	0xf6,0xf9,0x5f,0xf5,0xff,0xd5,0x57,0xdf,
	0x7d,0xf7,0xd7,0x56,0x5f,0x7d,0xf7,0xd5,
	0x7f,0x5f,0xec,0xf7,0x7f,0xfb,0xfd,0xb7,
	0xdf,0x7f,0xff,0xfd,0xb7,0xdf,0x7f,0xff,
	0xfd,0xb7,0xdf,0xff,0x97,0xff,0xff,0xff,
	0xda,0xeb,0xaf,0x3d,0xff,0xff,0xff,0xaf,
	0xef,0xff,0x5f,0xf2,0xbf,0xfc,0xeb,0xef,
	0xdf,0x7f,0xff,0xdd,0x77,0xdf,0x7f,0xff,
	0xfd,0x37,0xbf,0x7e,0xfb,0xbf,0xfc,0xff,
	0x7d,0xf4,0xf1,0x67,0x9f,0x7d,0xee,0xf9,
	0xc7,0x9f,0x79,0xfe,0xd1,0xe7,0x9f,0xff,
	0xf9,0xe7,0x9f,0x7f,0xfe,0xf9,0xe7,0x9f,
	0x7f,0xfe,0xf9,0xe7,0x9f,0x7f,0xfe,0xeb,
	0xff,0xef,0xff,0xff,0xfe,0xfb,0xeb,0xbf,
	0xff,0xfe,0xf2,0xef,0x7f,0xfe,0xff,0xff,
	0xf7,0xff,0xfd,0xe7,0xff,0x7f,0xff,0xf9,
	0xff,0x9f,0x7f,0xff,0xf9,0xf7,0xff,0xff,
	0x1f,0xff,0x5f,0xfd,0xf5,0xd7,0x77,0x5f,
	0x7d,0xf7,0xd7,0x17,0xdf,0xfd,0xff,0xff,
	0xff,0xff,0xff,0xef,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xbf,0xff,
	0xff,0xbf,0xf9,0xff,0xff,0xff,0xbf,0xff,
	0xfe,0xff,0xef,0xbf,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xfd,0xff,0xff,0xff,0xff,
	0xdf,0xff,0xff,0xff,0xff,0xdf,0xff,0xff,
	0xff,0xff,0xff,0xcf,0xff,0xdf,0x7f,0xff,
	0xfd,0xf7,0xff,0xdf,0xff,0xfd,0xef,0xff,
	0xff,0xfb,0xef,0xbf,0xff,0xfb,0xef,0xbf,
	0xff,0xfe,0xfb,0xef,0xbf,0xff,0xfe,0xfb,
	0xef,0xbf,0xff,0xfe,0xef,0xfe,0xff,0xff,
	0xff,0xbf,0xff,0xf7,0xff,0xef,0xff,0xff,
	0xff,0xff,0xff,0xbf,0xfb,0xfc,0xef,0xff,
	0xff,0xfe,0xbf,0xef,0xbf,0xff,0xff,0xff,
	0xff,0xbe,0xff,0xef,0xff,0xff,0xd4,0xff,
	0x5f,0x3f,0xfd,0xb5,0xd7,0x5f,0x7f,0xfd,
	0xe5,0x57,0x5f,0x7f,0xfd,0xa5,0x97,0x7b,
	0xed,0xf5,0xc7,0x5f,0x7f,0xfc,0xf5,0xd7,
	0x5f,0x7f,0xf5,0xd5,0xc7,0x5e,0xfb,0x82,
	0xff,0x3f,0xff,0xff,0xff,0xbf,0xff,0xf7,
	0xdf,0xfb,0xff,0x3d,0xff,0x78,0xef,0xbf,
	0xff,0xfb,0xef,0x7f,0xff,0xff,0xff,0xff,
	0x7f,0xff,0xfe,0xfc,0xff,0xbf,0xff,0xff,
	0x9b,0xf8,0xff,0xbc,0xff,0xfb,0xff,0xbf,
	0xff,0xfe,0xde,0xff,0xff,0xfc,0xfd,0xfe,
	0xff,0xbf,0xff,0xff,0xf5,0xbf,0x9f,0xff,
	0xfb,0xef,0xbf,0xff,0xfd,0xff,0xef,0xfe,
	0xff,0x7f,0xe8,0xff,0xff,0xff,0xdf,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x5f,
	0xff,0xff,0xff,0xfb,0xff,0xff,0xff,0xff,
	0xff,0xff,0x7f,0xbf,0xff,0xbf,0xff,0xcf,
	0x7f,0xff,0xff,0x1d,0xff,0xff,0x7f,0xfd,
	0xff,0xff,0xff,0x7f,0xfd,0xf1,0xc7,0x7f,
	0xd7,0xff,0xb7,0xff,0xfb,0xff,0xff,0xff,
	0xff,0xff,0xfb,0xef,0xff,0xfb,0xff,0xff,
	0xff,0xff,0xff,0xff,0x3f,0xfb,0xf7,0xff,
	0xef,0xff,0xff,0xff,0xff,0xfe,0xef,0xbf,
	0xff,0xee,0xff,0xeb,0xff,0x7f,0xfc,0xfc,
	0xff,0xff,0xdf,0xff,0xff,0xfb,0xef,0xff,
	0xff,0xfb,0xef,0xef,0xbf,0xfe,0x51,0xff,
	0xff,0xdf,0xf7,0xff,0xef,0xff,0xff,0xf7,
	0xff,0xff,0xff,0xff,0xf7,0xff,0x7f,0x77,
	0xff,0xff,0x6f,0xff,0xff,0xff,0xfc,0xff,
	0xff,0xff,0xff,0xdf,0xff,0xff,0xff,0x4f,
	0xfe,0xf7,0xbf,0xff,0xff,0xff,0xef,0xff,
	0xff,0xff,0xff,0xd5,0xff,0xff,0xdf,0x7f,
	0xe7,0xfd,0xdf,0xff,0xbf,0xff,0xfe,0xfb,
	0xfb,0xff,0xfd,0xff,0xdf,0xfe,0xff,0xff,
	0xff,0xf7,0xfd,0xff,0xff,0xfe,0xff,0xff,
	0xdf,0xfe,0xff,0xb9,0xff,0xfe,0xfe,0xfb,
	0xef,0xef,0xfb,0xf7,0xef,0xff,0xf7,0xdf,
	0xdf,0xff,0xff,0xfe,0xfb,0xfb,0xff,0xb7,
	0xdf,0xbe,0x93,0xef,0xcb,0x6f,0xff,0xff,
	0xf3,0xcf,0x3f,0xff,0xf8,0xff,0xcf,0x3f,
	0xff,0xfc,0xf3,0x2f,0xff,0xff,0xe3,0xcf,
	0x3f,0xff,0xfc,0xf3,0xcf,0xff,0xff,0xfc,
	0xf3,0xcf,0xff,0xbf,0x7d,0xff,0xfe,0xfb,
	0xff,0x9f,0xff,0x7e,0xf5,0xbf,0xff,0xf7,
	0xee,0xfb,0xb7,0xdf,0xee,0xf9,0xff,0x7f,
	0xff,0xfe,0xfb,0xef,0xdf,0x7e,0xfb,0xff,
	0xb7,0xdf,0x7e,0xfb,0xef,0xe7,0x7f,0xc7,
	0x7f,0xff,0xff,0xbb,0xe3,0x7e,0xfb,0xcf,
	0xf7,0xc7,0x9f,0xff,0xfc,0xf7,0xfe,0xff,
	0xf7,0xff,0xe3,0x8e,0x77,0xdb,0x77,0xdf,
	0xff,0xdd,0x75,0xc7,0x9d,0xfb,0x2f,0xee,
	0x7f,0xff,0xff,0xff,0xd7,0x7f,0xff,0xff,
	0xfe,0xff,0xbf,0xfe,0xfd,0xff,0xff,0xff,
	0xbf,0xff,0xff,0x7e,0xfa,0xed,0xff,0xff,
	0xff,0xfd,0xdb,0xff,0xff,0xff,0xff,0xdf,
	0xeb,0xff,0xfa,0xff,0xff,0xbf,0xff,0xfd,
	0xff,0xff,0xff,0xff,0xff,0x33,0xef,0xff,
	0xef,0xff,0xfe,0xff,0xfb,0xfd,0xaf,0xfe,
	0xfe,0xfb,0xff,0xff,0xff,0xfc,0xff,0xff,
	0xff,0x1b,0xff,0xff,0xff,0xe7,0xff,0xff,
	0xff,0xf9,0xfd,0xf7,0xff,0x3f,0xff,0xfd,
	0xff,0xbf,0xef,0xe7,0xef,0xbf,0xff,0xff,
	0xe7,0xff,0xff,0xf7,0xff,0xff,0xef,0xbf,
	0xfc,0xff,0xc3,0xd6,0xff,0xcf,0x3f,0xff,
	0xff,0xd7,0xff,0xff,0xdf,0xff,0xf3,0xfd,
	0xf6,0xff,0xff,0xff,0x1f,0xff,0xff,0xff,
	0xff,0x3f,0xff,0xff,0xbf,0x9f,0xff,0xff,
	0xff,0xf3,0xff,0xeb,0xf1,0xfe,0xff,0xff,
	0xff,0xff,0x9b,0xb5,0x7f,0xff,0xff,0xff,
	0xff,0xdf,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0x7f,0xfe,0xff,0xff,0xff,0xfd,0xff,
	0xff,0xff,0xff,0xff,0xdf,0x87,0xfd,0xdf,
	0xff,0xff,0xff,0xf7,0x8d,0xff,0xff,0xff,
	0xff,0xdf,0xa7,0xfe,0xff,0xff,0xff,0xff,
	0xff,0xff,0xdf,0x7f,0xff,0xfd,0xff,0xef,
	0xff,0xff,0xff,0xff,0xff,0xff,0x7b,0xfd,
	0xff,0xeb,0xbf,0xff,0xfe,0xd8,0xeb,0xbf,
	0xff,0xfe,0xfb,0xfb,0xbf,0xff,0xfe,0xfb,
	0xb4,0xff,0xfe,0xfb,0xe9,0xa5,0xb7,0xea,
	0x7b,0xef,0xbf,0xff,0xfe,0xfb,0xef,0xff,
	0x61,0xff,0xff,0xff,0xff,0xff,0xff,0xbf,
	0xff,0xff,0xff,0xff,0xff,0xef,0xfd,0xff,
	0xff,0xaf,0xfe,0xff,0xff,0xef,0xaf,0xfd,
	0x66,0xff,0xbb,0xff,0xff,0xff,0xff,0xff,
	0xff,0x76,0xfe,0xff,0xff,0xa7,0x9e,0xfa,
	0x5b,0xff,0xf7,0x7f,0xfe,0xf8,0xfb,0xfd,
	0x9f,0x7e,0x7f,0xfd,0xff,0x7f,0x7f,0xaf,
	0xbf,0xfc,0x7f,0xde,0xfd,0xe7,0xdf,0x7f,
	0xeb,0xfd,0xdf,0xfb,0xff,0xef,0x7f,0xfd,
	0xf5,0x9b,0xea,0x7f,0xff,0xfc,0xc3,0xff,
	0x6b,0xff,0xe7,0xdf,0xa9,0xff,0xfe,0xdf,
	0x6b,0xff,0xbd,0xff,0xdf,0xfe,0xff,0xf5,
	0xff,0xdf,0xff,0xbf,0xc3,0xff,0xbf,0xff,
	0xde,0x7b,0xef,0xbd,0xff,0xfe,0xfb,0xef,
	0xbf,0xed,0xff,0x7b,0xef,0xfd,0xfe,0xfb,
	0xef,0xbd,0xe5,0xde,0xdb,0xee,0xbd,0xff,
	0xfe,0xfb,0xef,0xbd,0xff,0x8d,0xfe,0xff,
	0xdd,0xe7,0x9f,0x7f,0xf7,0xdd,0xe7,0x9f,
	0x7f,0xfe,0x99,0xff,0x9f,0x7f,0x7e,0xe7,
	0x9f,0x7f,0xfe,0x99,0x67,0x9f,0x7d,0xfe,
	0xf9,0xe7,0x9f,0x7f,0xf6,0xf9,0xff,0xe3,
	0xff,0xf6,0x1d,0x7f,0xfc,0x7d,0x77,0x1f,
	0x7f,0xfe,0xf9,0xd6,0xdf,0x7f,0xfc,0xf1,
	0x9f,0x7f,0xfe,0xf1,0xf7,0x5f,0x6d,0xf5,
	0xf1,0xc7,0x1f,0x7f,0xfc,0xd1,0xc7,0xff,
	0x8d,0xfd,0xf3,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xf6,0xdf,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x6f,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xfb,0xfd,0xdf,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0x7f,0xff,0xed,0xbf,0xff,
	0xff,0xff,0xff,0xde,0x7f,0xff,0xff,0xff,
	0x7f,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xa5,0xff,0xcb,0x7d,0xf4,
	0xf5,0x62,0x8f,0x7f,0xf5,0xf1,0xc7,0x8b,
	0xbd,0xfd,0xd8,0xe7,0x7f,0xff,0xfd,0xe7,
	0x8b,0x2f,0xfe,0xd8,0x67,0x9f,0x7d,0xfe,
	0x79,0x62,0x9f,0xff,0x7d,0xff,0x7f,0xfd,
	0xf7,0xdf,0x57,0x7f,0xfd,0xf7,0xdf,0x7f,
	0xdf,0xfd,0xb5,0xc7,0x7f,0xff,0xf7,0xdf,
	0x7f,0x5b,0x75,0xf5,0xdd,0x7f,0xff,0xfd,
	0xb7,0xdf,0x16,0xff,0xfd,0xff,0xfb,0xff,
	0xfb,0xdf,0xff,0xbf,0xed,0x7b,0xff,0xff,
	0xfa,0xdd,0xfb,0xef,0x36,0xff,0xdd,0xff,
	0xff,0xf7,0xfd,0x2e,0xbb,0xbd,0xff,0xfd,
	0xf7,0xbf,0x7e,0xff,0xfd,0xf7,0xff,0xc9,
	0xff,0xff,0xe7,0x9f,0x7f,0xfc,0x99,0xe7,
	0x9f,0x7f,0xfe,0xb9,0x67,0x9f,0x7d,0xfe,
	0xf9,0x9f,0x7f,0xfe,0xf1,0xc7,0x1f,0x7b,
	0xfe,0xf9,0xe5,0x9f,0x7f,0xfe,0xf9,0xe7,
	0x5f,0xfe,0xbf,0xfe,0xff,0xff,0xbf,0xbf,
	0xfc,0xf7,0xdf,0x7f,0xbf,0xfe,0xfb,0x8f,
	0x7f,0x7e,0xf7,0xff,0x7f,0xfe,0xfe,0xfb,
	0xeb,0x7f,0xfe,0xfd,0xff,0xdf,0x2f,0xfe,
	0xff,0xff,0xf5,0xff,0xf7,0x7f,0xff,0xfd,
	0xfd,0xe5,0x3f,0xf7,0xfd,0xf7,0xfd,0xf7,
	0x7f,0xff,0xff,0xbf,0xff,0xff,0xff,0xf7,
	0xdf,0x5f,0xff,0xff,0xff,0xfd,0xff,0x7f,
	0xfd,0xff,0xff,0x97,0xff,0xbf,0xfe,0xff,
	0xff,0xbf,0xff,0xfe,0xff,0xf7,0xdf,0xbf,
	0xfe,0xfe,0xfb,0xff,0xdf,0xff,0xff,0xff,
	0xff,0xfe,0xfb,0xef,0xff,0x7f,0xfb,0xff,
	0xf7,0xff,0xff,0xff,0xff,0xfc,0xff,0xff,
	0xff,0xff,0xff,0xff,0xfb,0x7f,0xff,0xff,
	0xff,0xff,0xbd,0xdf,0xfe,0xfb,0xbf,0xff,
	0xfe,0xfb,0xff,0xff,0xff,0xff,0xfb,0xef,
	0xbf,0xff,0xfe,0xfb,0xef,0xff,0xe7,0xff,
	0xff,0xff,0xed,0xbb,0x7f,0xbf,0xfd,0xef,
	0xfb,0xff,0xbf,0x4f,0xff,0xff,0xff,0xfe,
	0xff,0xff,0xff,0x8b,0xff,0xef,0xff,0xef,
	0xfb,0xfb,0xff,0xbf,0xff,0xbf,0xff,0x5f,
	0xfd,0xff,0xf5,0xd7,0x5f,0x7b,0xfd,0xf5,
	0xd7,0x5f,0x7a,0xfd,0xd5,0xd7,0x5f,0x7d,
	0xbd,0x57,0x5f,0x7b,0xb5,0xf5,0xd7,0x5f,
	0x7b,0xdd,0xf5,0xd7,0x5f,0x7f,0xfd,0xf5,
	0x2f,0xf9,0xff,0xf3,0xbf,0xbf,0xff,0xff,
	0xff,0xcf,0x3f,0xff,0xfc,0xff,0xf7,0xff,
	0xff,0x7f,0x8f,0x3f,0xfe,0xff,0xff,0xfd,
	0xf7,0xff,0xff,0xf3,0xbf,0xbf,0xdd,0xfb,
	0xef,0xbf,0x97,0xff,0xd5,0xff,0xfe,0xf9,
	0xff,0xff,0x6f,0xff,0xfe,0xfb,0xfb,0xff,
	0xef,0xff,0xff,0xce,0x7f,0xff,0xff,0xfb,
	0xfb,0xff,0xff,0xff,0xf7,0xfd,0xfb,0xff,
	0xff,0xfd,0xef,0x96,0xfe,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xfe,0xef,0xbf,0xfd,
	0xff,0xff,0xff,0xff,0x97,0xbd,0xf5,0xf7,
	0xff,0xbf,0xff,0xfd,0xf6,0x77,0xff,0xff,
	0x77,0xff,0xff,0xff,0xdf,0xf1,0xff,0xff,
	0xff,0xff,0x7f,0xff,0xff,0xff,0xfe,0xff,
	0xff,0xff,0xc7,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xf5,0xd7,0x1f,0xff,0xff,0xfe,
	0xff,0xfc,0xfe,0xff,0xff,0xff,0xa3,0xff,
	0xff,0xff,0xff,0xff,0xdf,0xfe,0xef,0xef,
	0xff,0xff,0xfe,0xfb,0xff,0xef,0xfd,0xdf,
	0xff,0xff,0xff,0xbb,0xbf,0xfe,0xdd,0xfe,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfc,0xff,0xff,0x7f,0xff,0xfe,0xff,0xbf,
	0xff,0x3f,0xea,0xff,0xff,0xed,0xff,0xff,
	0x7b,0xff,0xff,0xff,0x7b,0xdb,0xef,0xff,
	0xfe,0xfa,0xdf,0xbf,0xff,0xff,0xff,0xff,
	0xff,0xe3,0xff,0x7f,0xff,0xdf,0xff,0xcf,
	0xfc,0xff,0xdf,0xef,0xff,0xff,0xff,0xff,
	0xff,0xff,0xfb,0xff,0xbf,0xff,0xfb,0xaf,
	0xff,0xdf,0xff,0xdf,0xff,0xff,0xf7,0xbf,
	0xdf,0xff,0x1f,0xff,0xff,0xff,0xff,0xff,
	0xfe,0xfb,0xff,0xbf,0xbe,0xff,0xff,0xef,
	0xbf,0xfe,0xfe,0xbf,0xf7,0xff,0xff,0x7f,
	0xff,0xef,0xff,0xff,0xf7,0xfd,0xbf,0xfe,
	0xfb,0xeb,0xef,0x7b,0xe9,0xff,0xbd,0xff,
	0xff,0x3f,0xff,0xff,0xff,0xcf,0x3f,0xff,
	0xfc,0xf3,0xcf,0xdf,0x7f,0xde,0xc6,0xdf,
	0x7f,0xfc,0xf3,0xcf,0x3f,0xdf,0xff,0xf3,
	0xcf,0x7f,0xff,0xfc,0xbf,0xd5,0xff,0xef,
	0xff,0xff,0xff,0xef,0xff,0xff,0x7f,0xbd,
	0xfb,0xef,0xdf,0xfe,0xfb,0xff,0xff,0xff,
	0xfe,0xbf,0xef,0xbf,0xff,0xfe,0xef,0xff,
	0xcf,0x7e,0xfb,0xfb,0xe7,0xff,0xaf,0xfe,
	0xff,0xff,0x7f,0xeb,0xdf,0xff,0xef,0xf7,
	0xff,0x9f,0x7f,0x7f,0xd7,0x97,0xff,0xfd,
	0xff,0xf7,0xab,0x7f,0xfb,0x7d,0xe3,0xdf,
	0xff,0xff,0xf7,0xdf,0x7f,0xff,0xff,0xff,
	0xe4,0xff,0xff,0xff,0x7f,0xff,0xdf,0xff,
	0xff,0xff,0xff,0xfc,0xf7,0xff,0x7f,0xff,
	0xff,0xfb,0xff,0xff,0xfd,0xff,0xff,0x7f,
	0xff,0xff,0xdf,0xff,0xff,0xff,0xff,0xff,
	0xff,0xb9,0xff,0xff,0xff,0xdf,0xeb,0xff,
	0xff,0xff,0xfb,0xef,0xb7,0xbe,0xff,0xff,
	0xfb,0xff,0xff,0xff,0xff,0xef,0xff,0xff,
	0xff,0xbb,0xbf,0xff,0xff,0xff,0xef,0xff,
	0xff,0xff,0xf7,0xf1,0xe7,0xf7,0xff,0xfa,
	0xeb,0xe5,0xb7,0xde,0xff,0xf9,0xaf,0xe3,
	0xbf,0x7f,0xeb,0xfb,0xff,0xf7,0xff,0xfb,
	0x9f,0xff,0xff,0xff,0xfd,0xff,0xff,0xff,
	0xff,0xff,0xff,0x3f,0x63,0x7d,0xfe,0xf9,
	0xd7,0xcf,0x7f,0xfd,0xfc,0xff,0xcf,0x7f,
	0xdd,0xff,0xff,0x9f,0x7f,0xf5,0xbf,0xff,
	0xff,0xff,0xff,0xff,0xff,0xef,0xff,0xff,
	0xff,0xff,0xff,0xff,0xbf,0x3e,0xef,0xff,
	0xfd,0xff,0xef,0x7f,0xff,0xfd,0xf7,0xff,
	0xff,0xff,0xff,0xff,0xdf,0xff,0xff,0x3f,
	0xff,0x9f,0x53,0x4f,0xfd,0xf7,0xff,0x4b,
	0xff,0xff,0xff,0xff,0xff,0xff,0x7d,0xd9,
	0xff,0xff,0xff,0xdf,0xfe,0xff,0xff,0xff,
	0xff,0x7f,0xff,0xfd,0xff,0xff,0xff,0xff,
	0xff,0xdc,0x7f,0x9a,0x78,0xe2,0x9f,0xff,
	0xff,0xfa,0xff,0xff,0xff,0xff,0xff,0xbf,
	0xd9,0xff,0xff,0xfe,0xfb,0xeb,0xbf,0xff,
	0xfe,0xfb,0xef,0xaf,0xbf,0xff,0xfb,0xef,
	0xbf,0xef,0x7b,0xeb,0xaf,0xbf,0xfe,0xfa,
	0xeb,0xbf,0xff,0xfe,0xfb,0xef,0xbf,0xff,
	0xfe,0xdf,0xf6,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xfb,0xff,0xfd,0xf7,
	0xff,0xff,0xff,0xff,0xfb,0xff,0xff,0xff,
	0xff,0xff,0x6f,0xe2,0xff,0xdf,0x7f,0xeb,
	0xff,0x97,0x5f,0xfe,0xfe,0xed,0xff,0xff,
	0x7f,0xeb,0xf9,0xbf,0x7b,0xfe,0xfb,0xbd,
	0xff,0xff,0x7e,0xff,0xaf,0xf4,0x7e,0xea,
	0xf9,0xff,0xdf,0xff,0xb9,0xff,0xff,0xfd,
	0xd7,0xef,0x7f,0xfc,0xf1,0xf3,0xdf,0xbe,
	0xff,0xff,0xdf,0xdf,0xbf,0xf8,0x5f,0xef,
	0xef,0xf8,0xf3,0xfb,0xeb,0x3f,0xad,0xe6,
	0xdf,0x7f,0xbf,0xff,0xd7,0x3b,0x7c,0xfe,
	0xfb,0xef,0xbd,0xff,0xfe,0xfb,0xef,0xbf,
	0xff,0xfe,0xdb,0xef,0xbd,0xff,0xfe,0xef,
	0xbf,0xff,0xfe,0xdb,0x6e,0xbf,0xed,0xde,
	0x7b,0xef,0xbd,0xf7,0xfe,0xdb,0xbe,0xe0,
	0xfb,0x9f,0x7f,0xfe,0xfd,0xe7,0x9f,0x7f,
	0xfe,0xf9,0xf7,0x9f,0x75,0xfe,0xf9,0xe7,
	0x7f,0xfe,0xf9,0x67,0x9f,0x7d,0xd6,0x99,
	0xe7,0x9f,0x7f,0xfe,0xf9,0xe7,0x9f,0xf9,
	0x1f,0xfe,0x7f,0xfc,0xb1,0xf6,0x1f,0x7f,
	0xfc,0xf9,0xc7,0xda,0x7f,0xdd,0xf1,0xc7,
	0x9f,0xbf,0xf1,0xf7,0x5f,0x67,0x8d,0xdd,
	0xf7,0x9f,0x7f,0xfe,0xf1,0xc7,0x9f,0x7f,
	0xfc,0xdf,0xd9,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xdf,0x6f,0xff,
	0xfd,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xbf,0xc1,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xfd,0xff,0xff,0xff,
	0xff,0xff,0xff,0xdf,0xef,0xff,0x7f,0xff,
	0xfe,0xfd,0xff,0xff,0xdf,0x7f,0xff,0xff,
	0xff,0xdf,0xff,0xff,0xbf,0xfa,0xff,0xd1,
	0xc7,0x8f,0x7f,0xf4,0xd1,0xc7,0x5f,0x2f,
	0xbe,0xf8,0xe7,0x4f,0x7f,0xf6,0xe7,0xdb,
	0x3f,0xbe,0xf8,0x62,0x8b,0x7f,0xf4,0xdd,
	0xe7,0x9f,0x7d,0x9f,0xf8,0x5f,0xf7,0xbf,
	0xdf,0x7f,0x7f,0xfd,0xf7,0xdf,0x7f,0xff,
	0x7d,0xf7,0xd5,0x7f,0x7f,0xfc,0xf7,0x7f,
	0x7f,0xf5,0xf5,0xd5,0x57,0x5d,0xf5,0xb7,
	0xdf,0x7f,0xff,0xfd,0xf7,0xc5,0xff,0x9f,
	0xff,0x7f,0xfb,0xef,0xfb,0xbf,0xfe,0xfa,
	0xfd,0x7f,0xeb,0xbe,0xf7,0xfd,0x7b,0xff,
	0xff,0xeb,0x7e,0xaf,0xbf,0xfe,0xfe,0x7e,
	0xdf,0xff,0xfb,0xdd,0xf7,0xff,0xfd,0xfa,
	0x1f,0xfd,0xff,0x7f,0xee,0x99,0xe7,0x9f,
	0x7f,0xfe,0xf9,0xc7,0x9e,0x79,0xfe,0xd9,
	0xe7,0x9e,0xff,0xd9,0xc7,0x1f,0x7f,0xec,
	0xb1,0xe5,0x9f,0x7f,0xfe,0xf9,0xe7,0x9f,
	0x7f,0xfe,0xe5,0xff,0xdf,0x7f,0xff,0xfe,
	0xf7,0xdf,0x7f,0xff,0xfd,0xf3,0xff,0x7f,
	0xfe,0xfc,0xe7,0xf7,0xff,0xff,0xfb,0xfb,
	0xef,0xff,0xfe,0xe7,0x9f,0x7f,0xfe,0xf9,
	0xf7,0xcb,0xff,0x2f,0xff,0xff,0xff,0xff,
	0xdd,0xbf,0xff,0xff,0xff,0xef,0x7f,0xff,
	0xfd,0xff,0xe7,0xff,0xff,0xff,0xdf,0x5f,
	0x7f,0xfd,0xf5,0xd7,0xfd,0xff,0xff,0xff,
	0xff,0xbf,0x5f,0xfe,0x3f,0xf8,0xff,0xf7,
	0xdf,0xbf,0xff,0xff,0xf7,0xdf,0xff,0xff,
	0xfc,0xfb,0xbf,0xbf,0xff,0xfb,0xbd,0xff,
	0xff,0xfe,0xff,0xff,0xff,0xfb,0xff,0xef,
	0xbf,0xff,0xfe,0xfb,0xef,0xff,0xd5,0xff,
	0xff,0xfe,0x7b,0xff,0xbf,0xff,0xfe,0xfb,
	0xef,0xef,0xdf,0xff,0xff,0xed,0xff,0xff,
	0xff,0xff,0xfd,0xff,0xff,0xff,0xff,0xbf,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x9f,
	0xfe,0xff,0xf7,0xff,0xff,0xff,0xf6,0xff,
	0xef,0xff,0xfb,0xff,0xff,0xdf,0xff,0xff,
	0x7f,0xff,0xbe,0xef,0xff,0xff,0xff,0x93,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0x7f,0xd1,0xff,0x5f,0x7f,0xfd,0xf5,0x97,
	0x5f,0x7e,0xf9,0xf1,0x57,0x5f,0x7d,0xfd,
	0xe5,0x97,0x7f,0xf9,0xe1,0xd7,0x1f,0x7f,
	0xfc,0xb5,0x97,0x5f,0x7e,0xfd,0xf5,0xd7,
	0x5f,0xff,0xaa,0xff,0x3f,0xfe,0x58,0xf3,
	0xff,0x3f,0xff,0xf8,0xff,0xcf,0xf9,0xfe,
	0xfb,0xff,0xff,0xff,0xff,0xe3,0xc7,0xff,
	0xfe,0xfb,0xfd,0xdf,0x3f,0xfe,0xff,0xff,
	0x8f,0xff,0xff,0x3b,0xf9,0xff,0xfd,0xf7,
	0xcf,0xff,0xff,0xfd,0x77,0xfb,0xf7,0xff,
	0xff,0xbe,0xfe,0xfb,0xff,0xff,0xf7,0xae,
	0xbf,0xbf,0xff,0xff,0xbf,0x7b,0xff,0xff,
	0xff,0xdf,0xff,0xff,0x7f,0xe7,0xff,0xfb,
	0xef,0xff,0xff,0xff,0xfb,0xef,0xff,0xff,
	0xda,0xff,0xff,0xff,0xff,0xff,0xf1,0xdf,
	0xfe,0xfb,0xff,0xff,0xe7,0xff,0xff,0xf6,
	0xf7,0xff,0x7f,0xfd,0xff,0xff,0x77,0xff,
	0xff,0xff,0xff,0xf5,0xfb,0xff,0xef,0xff,
	0xff,0xb7,0x1f,0xf6,0xff,0xff,0xfd,0xbf,
	0xff,0xf5,0xff,0xdf,0xff,0xfd,0xe7,0xfd,
	0xef,0xff,0xff,0xfe,0xff,0xef,0xff,0xbf,
	0xf8,0xbf,0xbf,0xff,0xa6,0xff,0xff,0xbf,
	0xfe,0xfe,0xf7,0xee,0xfd,0x7f,0xff,0xff,
	0xff,0xff,0xbb,0xfd,0xff,0xfb,0x7a,0xd7,
	0xff,0xff,0xff,0xff,0xfd,0xfb,0xfb,0xff,
	0xff,0xdf,0xff,0x7f,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xeb,0xbf,0xbf,0xff,0xff,
	0xff,0x7f,0xe7,0x7f,0x7f,0xe7,0xff,0xfe,
	0xf7,0xbf,0xf3,0xcf,0x7e,0xff,0xfc,0x77,
	0xcf,0xff,0x2b,0xfe,0xff,0x7f,0xff,0xff,
	0xff,0xff,0xfd,0xff,0xff,0xff,0xff,0xff,
	0xff,0x7f,0x7f,0xbf,0xe7,0xde,0xcb,0xff,
	0xfd,0xff,0xfd,0xff,0xdf,0xff,0xf4,0xfe,
	0x4e,0xff,0xff,0xff,0xf5,0xff,0xff,0x7f,
	0xfb,0xff,0xbf,0xff,0xff,0xff,0xff,0xf7,
	0xff,0xff,0xff,0xfd,0xff,0xff,0xef,0xef,
	0xed,0xff,0xff,0xfe,0xfd,0xff,0xff,0xdf,
	0xff,0xff,0xfd,0xff,0xfd,0xb3,0xfd,0xef,
	0x37,0xff,0x7c,0xf3,0xcf,0x7f,0xbf,0xff,
	0xf3,0xc7,0x3f,0xff,0xfc,0xf3,0xe7,0xff,
	0xfc,0xf7,0xcf,0xff,0x5f,0xfd,0x75,0xfb,
	0xff,0xff,0xff,0xff,0xff,0xff,0xf7,0xfd,
	0x7f,0xff,0xfb,0xef,0x5b,0xff,0xfe,0xef,
	0xff,0xb7,0x7f,0x3e,0x6c,0x6f,0xbf,0xff,
	0xfe,0xee,0xbf,0xff,0x7e,0xff,0xbf,0xff,
	0xfe,0x7f,0xff,0xfb,0xff,0xff,0xff,0xff,
	0xe7,0xff,0xeb,0x9f,0x77,0xfd,0xbf,0xf7,
	0xfe,0xbb,0xfe,0xf9,0xef,0x7f,0x7f,0xfe,
	0xf1,0xff,0x7d,0xbb,0xff,0xe3,0xff,0xff,
	0xff,0xe7,0xeb,0xaf,0xbf,0xfe,0xfa,0xeb,
	0xff,0x76,0xfe,0x5f,0xff,0xfd,0xf7,0xff,
	0x7f,0xef,0xff,0xf5,0xcf,0x3f,0xff,0xff,
	0xfb,0xef,0xbf,0xff,0xfe,0xdf,0x7f,0xff,
	0xff,0xff,0xff,0x7f,0xff,0xfd,0xf7,0xdf,
	0x7f,0xff,0xd7,0xfb,0xff,0xff,0xf3,0xbf,
	0xff,0xeb,0xeb,0xff,0xff,0x7e,0xff,0xfd,
	0xff,0xfe,0xfe,0xff,0xff,0xff,0xff,0xfe,
	0xfd,0xff,0xff,0xff,0xff,0xfd,0xaf,0xdf,
	0xff,0xfe,0xfd,0xff,0x1f,0xff,0xff,0xff,
	0xff,0xf3,0xdf,0x3f,0xff,0xad,0xff,0xff,
	0xff,0xff,0xff,0xf3,0xcf,0xfc,0xcb,0xbf,
	0xfe,0xff,0xf8,0xfb,0xb7,0x9e,0xff,0xff,
	0xff,0xeb,0xcf,0xf7,0xff,0xb3,0xd6,0xff,
	0xff,0xff,0xdf,0xff,0xfe,0xcd,0x77,0xfd,
	0xff,0xff,0xff,0x7f,0xde,0x7f,0xff,0xff,
	0xff,0xc9,0xd7,0xcf,0xff,0xff,0xf5,0x7f,
	0xf7,0xdf,0xef,0xff,0xbe,0xff,0xeb,0xf0,
	0xfe,0xef,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xfe,0xff,0xef,0xff,0xff,0xff,0xff,
	0xff,0x6b,0xff,0xfd,0xf7,0x9f,0xff,0xfd,
	0xf7,0xf7,0xff,0xff,0xfb,0x2f,0xbf,0xdf,
	0x93,0xfd,0xdf,0x7e,0xff,0xfd,0xff,0xdf,
	0xff,0xff,0xfd,0xf7,0xdf,0xfe,0xff,0xfd,
	0xf7,0xff,0x9f,0xf9,0xf3,0x9f,0x7f,0xfa,
	0xff,0xff,0xff,0xff,0xff,0xfb,0xff,0xf9,
	0xff,0x1b,0xfd,0xff,0xfb,0xaf,0xbf,0xff,
	0xfb,0xfb,0xbf,0x3f,0xff,0xfa,0xfb,0xbf,
	0xbf,0xff,0xfe,0xbf,0xbf,0xfe,0xb2,0xeb,
	0xaf,0xff,0x5e,0xfb,0xff,0xff,0xfb,0x7f,
	0xff,0xff,0xff,0x65,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xea,0xbf,0xb5,0xee,
	0x5b,0xaf,0xbd,0xff,0x46,0xfe,0xff,0xff,
	0xff,0xff,0x7f,0xf9,0xef,0xff,0xff,0xfe,
	0xff,0xff,0xf7,0xff,0xff,0xff,0xf7,0xf7,
	0xff,0x7f,0xff,0xfd,0x7f,0xd6,0xff,0xff,
	0xff,0xff,0xdf,0xfe,0xef,0x5f,0xfa,0xff,
	0xff,0xbf,0xff,0xff,0xc7,0x7f,0x3c,0xff,
	0xe7,0xcb,0xff,0xff,0xfd,0xff,0xff,0xfd,
	0xbf,0xfe,0xfb,0xeb,0xaf,0xff,0xbc,0xf2,
	0xff,0xef,0xff,0xd7,0xff,0x7f,0xbe,0xd7,
	0xff,0xbf,0xf9,0xb6,0xdb,0xee,0xbf,0xff,
	0xfe,0x5b,0xee,0xbf,0xff,0xde,0xfb,0x6f,
	0xfe,0xfe,0xfb,0xef,0xbf,0xff,0xfe,0xfb,
	0xef,0xff,0xed,0xf7,0xdf,0x7e,0xfb,0xf7,
	0xed,0xfe,0xff,0xd9,0x77,0x9f,0x79,0xfe,
	0xf9,0xe7,0x9f,0x7d,0xe7,0xf9,0xe7,0x9f,
	0x79,0xfe,0xe7,0x9f,0x7f,0xfe,0xd9,0x67,
	0x9f,0x7f,0xfe,0xdf,0x7f,0xfd,0xfd,0xff,
	0xff,0x7f,0xe3,0xff,0x57,0xdf,0x7d,0xfd,
	0xf1,0xd6,0x9f,0x7f,0xdd,0xbd,0xd6,0x1f,
	0x7f,0xfd,0xf5,0x1b,0x7f,0xff,0xfd,0x76,
	0xdb,0x77,0xbe,0xe9,0x77,0xdf,0x77,0xf7,
	0xfd,0xf7,0xff,0xb1,0xfd,0xff,0xff,0xff,
	0xff,0xff,0xf6,0xff,0xff,0xff,0xfd,0xf7,
	0xff,0xff,0xff,0xff,0xff,0x7f,0xff,0xfd,
	0xfe,0xdb,0xff,0xbf,0xfd,0xff,0xff,0xff,
	0xff,0xff,0xf7,0xff,0xfb,0xfd,0xff,0xff,
	0xff,0xff,0xff,0xef,0xff,0xdf,0xff,0xfd,
	0xf7,0xff,0xff,0xff,0xff,0xff,0xff,0xfd,
	0xf7,0xff,0xbf,0xff,0x7f,0xfb,0xfd,0xff,
	0xff,0xff,0xff,0xdf,0xff,0xff,0xaf,0x7f,
	0x8b,0x2d,0xbf,0xdc,0xc7,0x8f,0x7d,0xb4,
	0xd8,0xe3,0x8b,0x2f,0xbe,0xd8,0x72,0x77,
	0xbe,0xf8,0xe3,0x8b,0x2f,0xfe,0xd1,0xd3,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x65,
	0xff,0xdf,0x7d,0xf5,0xd5,0x7e,0xdf,0xfd,
	0xf7,0xd5,0x57,0xdf,0x3d,0xf1,0xd5,0x57,
	0xff,0xf7,0x55,0x77,0xdd,0x65,0x95,0xdf,
	0x37,0xef,0xbe,0xfb,0xef,0xbf,0xff,0xfe,
	0x7f,0xf8,0xff,0xf3,0xff,0xff,0xfb,0xfb,
	0xab,0xff,0xbe,0xff,0xfb,0x72,0xcb,0xef,
	0xfb,0xee,0xfd,0xac,0xff,0xee,0xfb,0xfb,
	0xff,0xfe,0xfe,0xfe,0xff,0xb7,0xff,0x7f,
	0xff,0xff,0xcd,0xff,0xff,0xe7,0x9f,0x7f,
	0xfe,0xf1,0xe7,0x9e,0x79,0xfc,0xf1,0x47,
	0x1f,0x7b,0xf6,0xf9,0x1e,0x79,0xec,0xb9,
	0x67,0x9e,0x7f,0xfc,0xf1,0xe7,0x9f,0x7d,
	0xfe,0xf9,0xe7,0x1f,0xfe,0xbf,0xfe,0xf3,
	0xef,0x7f,0xff,0xfe,0xf7,0xfb,0x3f,0xff,
	0xfe,0xe7,0xef,0xef,0xff,0xf7,0xef,0xbf,
	0xbf,0xfe,0xfb,0xdf,0x3f,0xff,0xff,0xff,
	0xff,0xff,0xff,0xf9,0x7f,0xf6,0xff,0xfd,
	0x9f,0x7f,0xff,0xeb,0xf7,0xff,0xde,0x7f,
	0xf9,0xff,0x7f,0x7f,0x7f,0xff,0x7b,0xdf,
	0xff,0xfd,0xfd,0xf7,0xff,0x7a,0xeb,0xff,
	0xff,0xff,0xff,0xff,0xf7,0xff,0xa7,0xff,
	0xbf,0xff,0xfe,0xfb,0xff,0xbf,0xff,0xfd,
	0xff,0xff,0xbf,0xff,0xff,0xff,0xff,0xff,
	0xfd,0xfb,0xef,0xbf,0xff,0xff,0xff,0xef,
	0xff,0xbf,0xff,0xff,0xfb,0xff,0xff,0x9f,
	0xfc,0xff,0xef,0xa7,0xdf,0xfe,0xff,0xed,
	0xff,0xff,0xde,0xfa,0xef,0xff,0xff,0xfe,
	0xfb,0xff,0xf7,0x7e,0xfb,0xef,0xbf,0xff,
	0xdf,0xff,0xff,0xfd,0xff,0xdf,0xff,0xf9,
	0xff,0xe1,0xff,0xff,0xff,0xff,0xff,0x7f,
	0xff,0xff,0xff,0xff,0xef,0xbf,0xff,0xfe,
	0xff,0xff,0xff,0xbe,0xff,0xff,0xbb,0x2f,
	0xfe,0xbf,0xff,0x4b,0xef,0xbf,0xfc,0xfe,
	0xcb,0xff,0x6f,0xfd,0xff,0xb5,0xd3,0x5f,
	0x7f,0xfd,0xf5,0xd7,0x5e,0x3f,0xfd,0xf5,
	0xd7,0x5f,0x7f,0xb9,0xd7,0x5b,0x7f,0xd4,
	0x71,0xd7,0x5f,0x7b,0xbd,0xd5,0x56,0x1b,
	0x6d,0xb4,0xe1,0x6f,0xf8,0xff,0xb3,0x4f,
	0x3f,0xfe,0xfe,0xff,0xaf,0xfd,0xff,0x7f,
	0xfb,0xff,0xff,0xff,0xff,0xef,0xf7,0xdf,
	0xfb,0xef,0xdf,0xff,0xff,0xfc,0xff,0x7f,
	0xff,0xff,0xf5,0xff,0xfd,0x8f,0xff,0xf7,
	0x3f,0x5f,0xfd,0xfe,0xff,0xff,0xf7,0xff,
	0xff,0xff,0xfb,0xbf,0xff,0xfe,0xed,0xef,
	0xff,0xff,0xff,0xcf,0xf7,0xdf,0x57,0xfb,
	0xff,0xff,0x7f,0xfb,0xff,0xfb,0xd6,0xfe,
	0xff,0xfe,0xff,0xff,0xbb,0xfe,0xff,0xfb,
	0xff,0xfd,0x7f,0xff,0xdf,0x7f,0xff,0xbf,
	0x7d,0x7f,0xff,0xfd,0xff,0xff,0xff,0xef,
	0xff,0xfb,0xff,0xfd,0xfb,0xfe,0xff,0x4f,
	0xf5,0xff,0xff,0xd7,0x5f,0xff,0xff,0xf5,
	0xff,0x7f,0xff,0xff,0xff,0xd3,0x5f,0x7f,
	0xfc,0xff,0x1f,0xff,0xfd,0xff,0xf7,0xff,
	0xff,0xff,0xfd,0xf7,0xdb,0x47,0xff,0xfd,
	0xff,0xaf,0xff,0xff,0xbf,0xfe,0xfa,0xff,
	0xaf,0xfb,0xff,0xbf,0xff,0xff,0xbf,0xee,
	0xfb,0xb7,0xdf,0x9f,0xfe,0xfb,0xff,0xfb,
	0xfd,0xff,0xff,0x7f,0xff,0xbe,0xb7,0xef,
	0x7f,0xff,0x7f,0xfd,0xff,0xff,0x7f,0xff,
	0xfd,0xff,0xff,0xff,0xf7,0xef,0xff,0xff,
	0xdf,0xf7,0xff,0x7f,0xff,0xff,0xf3,0xa7,
	0xff,0x7f,0x7f,0xff,0xdf,0x7f,0x7f,0xdf,
	0x3f,0xef,0xfd,0xff,0xe7,0xff,0xff,0xff,
	0xd7,0x7f,0xff,0xbf,0xfd,0xff,0xff,0xff,
	0xff,0xff,0xfd,0xfd,0xbf,0xfb,0xb7,0xbe,
	0xbf,0x7f,0xff,0xff,0xff,0xfe,0xeb,0xff,
	0xef,0xfb,0xfa,0xd3,0xbd,0x37,0xff,0xff,
	0xff,0xff,0xff,0xff,0xfb,0xff,0xf7,0xff,
	0xff,0xfe,0xff,0xff,0xff,0x7f,0xbe,0xff,
	0xdf,0x7f,0xff,0xff,0xff,0xff,0xdf,0xe7,
	0xff,0xff,0x6f,0xff,0x7f,0xff,0xbb,0xfa,
	0xff,0xff,0xdb,0xff,0xff,0xff,0xf8,0xf7,
	0x8f,0xff,0xff,0xfd,0xf3,0xff,0x7f,0xff,
	0xfe,0x87,0xdf,0x7f,0xff,0xf3,0xff,0xff,
	0xbf,0xff,0xf9,0xf6,0xdf,0x7f,0xff,0xbd,
	0xc9,0xff,0xff,0xdf,0xeb,0xff,0xff,0xbf,
	0xbf,0xff,0xfb,0xff,0xdf,0xff,0xfe,0xff,
	0xef,0xf7,0xff,0xfd,0xff,0xff,0x9f,0xff,
	0xdf,0xff,0xfe,0xdf,0xef,0xff,0xdf,0xff,
	0xff,0x7d,0xbe,0xff,0xff,0xe7,0xff,0xfd,
	0xba,0xfe,0xff,0x6b,0xf7,0xff,0xff,0x78,
	0xdf,0xfc,0xf7,0xff,0xed,0xff,0xff,0xff,
	0xff,0xff,0xff,0xfb,0x7e,0xff,0xfd,0xad,
	0xb3,0xfe,0xff,0xe6,0xfb,0xff,0xff,0x7f,
	0xfb,0xfd,0x77,0xff,0x7f,0xff,0xed,0xff,
	0xd7,0xff,0xfe,0xff,0xfe,0x3f,0xff,0xff,
	0xff,0xff,0xff,0xbf,0xf5,0x56,0xff,0xff,
	0xff,0xfd,0xe7,0xff,0xb9,0xef,0xff,0xff,
	0xfb,0xfd,0xf7,0xbf,0xfe,0xf7,0xfb,0xaf,
	0xff,0xfe,0xfb,0xff,0xff,0xdf,0x7f,0xfd,
	0xff,0xff,0xff,0xff,0xff,0xef,0xaf,0xff,
	0xff,0xff,0xef,0xbf,0xff,0xf7,0xf0,0xcf,
	0x9f,0xff,0xff,0xdf,0xff,0xff,0xff,0xff,
	0xff,0x7f,0xff,0xdf,0x7f,0xf9,0xf9,0xce,
	0xff,0xdf,0xe5,0xf7,0xdf,0xfc,0xff,0x7f,
	0xff,0x9f,0xff,0xff,0xff,0xe3,0xef,0x6e,
	0xf5,0xff,0xf5,0xff,0x9f,0xfe,0xff,0xf9,
	0xff,0xff,0xff,0xfb,0xff,0xff,0xcf,0xf7,
	0xff,0x3e,0xff,0xfe,0xff,0x7f,0xe7,0xff,
	0xff,0xff,0xff,0xfe,0xff,0xff,0x3f,0xff,
	0x4e,0xaf,0xff,0xff,0xff,0xef,0xb7,0xff,
	0xfe,0xfb,0xef,0xbf,0xdf,0xff,0xf7,0xdf,
	0xff,0xff,0xff,0xf7,0x77,0xfe,0xfd,0xff,
	0x7f,0xfe,0xf3,0xfb,0xff,0xff,0xff,0xff,
	0xff,0xff,0xdb,0xff,0xff,0xf7,0xdf,0x3f,
	0xfd,0xfd,0xd7,0xdf,0x7f,0xff,0xfe,0xff,
	0xff,0xff,0xff,0xff,0xcf,0xbf,0xff,0xff,
	0xff,0xff,0xf3,0xce,0xff,0xff,0xff,0xe7,
	0xff,0xfb,0x37,0xc6,0xff,0xff,0xfe,0xfa,
	0xf3,0xed,0x3f,0xff,0xfa,0xf3,0x8f,0xf7,
	0xfe,0xfb,0xef,0xbf,0xef,0xff,0xeb,0xbd,
	0xff,0xfe,0xfb,0xef,0xff,0xff,0xff,0xfb,
	0xff,0xff,0xfb,0xbf,0x07,0xb6,0xff,0xff,
	0xff,0xff,0xef,0xfe,0xff,0xff,0xff,0xff,
	0xbf,0xfb,0xff,0xff,0xff,0xff,0xff,0x7f,
	0xe7,0xfe,0xff,0xff,0xff,0xff,0xfe,0x6b,
	0xff,0xbf,0xb7,0xca,0x7b,0xaf,0xe1,0xff,
	0xdf,0xff,0xff,0x7f,0xff,0xff,0xff,0xff,
	0xef,0xbf,0xdd,0x7f,0xff,0xff,0xff,0xff,
	0xff,0x5f,0xe7,0x9f,0x7f,0xff,0xaf,0xff,
	0xff,0xff,0xeb,0xdf,0xff,0xff,0xfd,0xa1,
	0xff,0xff,0xff,0xdb,0x3f,0xfb,0xff,0xf3,
	0xfb,0x3f,0xbe,0xe8,0xf1,0xf7,0xcf,0x3f,
	0xdf,0xcf,0xff,0x7b,0xff,0xfd,0xf7,0x4f,
	0xfd,0xff,0xff,0xd3,0xfb,0xff,0xff,0xff,
	0x2b,0x7d,0xfe,0xfb,0x6f,0xbd,0xff,0xb6,
	0xdb,0x6e,0xbb,0xfd,0xf6,0xfb,0xef,0xbf,
	0xff,0xfe,0xff,0xbf,0xfd,0xfe,0xfb,0xef,
	0xbf,0xff,0xb7,0xdf,0xee,0xfd,0xed,0xf7,
	0xdf,0xde,0xe9,0xfb,0x9f,0x7f,0xdf,0xf9,
	0x67,0x9e,0x7f,0xff,0x79,0x77,0x9d,0x7f,
	0xfe,0xf9,0xe7,0xff,0xfe,0x59,0xe7,0x9f,
	0x7f,0xfe,0xf9,0x7f,0xff,0x7f,0xfe,0x9f,
	0xff,0xfd,0xf9,0x67,0xfe,0x7f,0xac,0xbd,
	0xd6,0x5a,0x7f,0xbd,0xfd,0xd6,0xdb,0x77,
	0xf8,0xf1,0xe7,0x9f,0xbf,0xfd,0x57,0x1f,
	0x7f,0xfc,0xf1,0xe7,0xdf,0x77,0xfb,0xf9,
	0xb7,0xda,0x7e,0xff,0x1f,0xdb,0xff,0xff,
	0xff,0xf7,0xdf,0xff,0xff,0xfc,0xf3,0xdb,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xdf,0x7f,0xff,
	0xff,0xf7,0xdf,0x7f,0xff,0xff,0xd1,0xff,
	0xff,0xff,0xf7,0xdf,0xff,0xff,0xf7,0xdf,
	0xbf,0xff,0xff,0xff,0xff,0xf7,0xdf,0xef,
	0xfd,0xff,0xff,0xff,0xff,0xff,0xb7,0xff,
	0xfe,0xff,0xdd,0xdf,0xff,0xfd,0xff,0x6f,
	0xfa,0xff,0xf9,0xe2,0xcf,0x3f,0xff,0xf8,
	0xe2,0x8b,0x3f,0xff,0x71,0x53,0x1f,0x7d,
	0xf4,0x7f,0xdb,0x7d,0xf4,0x51,0xc2,0x1f,
	0xfd,0xff,0xff,0xc7,0xfd,0xf7,0xff,0x7f,
	0x7f,0xf1,0xff,0xdf,0x57,0x5f,0x7d,0xf5,
	0xd7,0x57,0x7f,0x7d,0xf7,0xdf,0x1e,0xff,
	0xfd,0xf7,0x7f,0xdf,0xfd,0xf7,0xdf,0x17,
	0xff,0xfd,0xfb,0xef,0x7f,0xeb,0xfe,0xfb,
	0xee,0x7f,0x83,0xff,0xff,0xff,0xfa,0xfb,
	0xff,0x3f,0xbf,0xfd,0xf3,0xef,0x7d,0xff,
	0xee,0xaf,0xff,0xff,0xff,0xff,0xbf,0xff,
	0xbe,0xea,0xbf,0xff,0xe7,0xff,0xfd,0xff,
	0xff,0xff,0xd7,0xcf,0xfc,0xff,0x3b,0xfc,
	0xd1,0xc7,0x9f,0x7d,0xf6,0xf9,0x67,0x9f,
	0x7f,0xf6,0xf9,0xe7,0x9e,0xff,0xf9,0xe7,
	0x9f,0x7f,0xf4,0xb9,0xe7,0x9f,0x3f,0xfe,
	0xf9,0xe7,0x1f,0x7f,0x7e,0xea,0xff,0xdf,
	0xf7,0xff,0xff,0xfb,0xeb,0xbf,0xbf,0xfe,
	0xfb,0xdd,0x3f,0xff,0xfd,0xf7,0x77,0xbe,
	0xff,0xf7,0xdf,0x7f,0xff,0xfd,0xf7,0xff,
	0x75,0xfe,0xfd,0xff,0xff,0xff,0x6f,0xff,
	0xff,0xfd,0xff,0xfd,0xd7,0x5f,0x7f,0xff,
	0xf7,0xf7,0xff,0xff,0xf9,0xff,0xff,0xff,
	0xf7,0xfd,0xff,0xff,0xff,0xff,0xff,0xf7,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xfd,
	0xf9,0xff,0x77,0xaf,0xff,0xfe,0xfb,0xef,
	0xbf,0xbf,0xfe,0xfa,0xf7,0xef,0x7f,0xff,
	0xfd,0xfd,0xff,0xfe,0xfd,0xf7,0xdf,0x7f,
	0xff,0xfd,0xff,0xff,0x7f,0xff,0xfe,0xff,
	0xff,0xc7,0xff,0xfb,0xdf,0x7e,0xff,0xfd,
	0xff,0xff,0xff,0xff,0xf7,0xfe,0x7e,0xfb,
	0xef,0xbf,0xff,0xff,0xff,0xbf,0xff,0xfe,
	0xfb,0xef,0xef,0xff,0xfb,0x7b,0xfb,0xff,
	0xff,0xff,0x7f,0xfe,0xff,0xff,0xef,0xff,
	0x2f,0xff,0xfb,0xff,0xff,0xff,0xbe,0xfc,
	0xff,0xbf,0xff,0x7f,0xff,0xff,0x2f,0xfe,
	0xff,0xef,0xff,0x4f,0xbf,0xfd,0xff,0xfb,
	0xef,0xff,0xfd,0xbf,0xd2,0xff,0x5f,0x7f,
	0xfd,0xd5,0xd7,0x5f,0x7f,0xf5,0xd5,0x97,
	0x5b,0x7e,0xf1,0xe5,0x17,0x7f,0xfd,0x45,
	0x97,0x5f,0x7d,0xfd,0xb5,0x17,0x5d,0x7e,
	0xb1,0xd5,0x17,0x5f,0xfd,0x83,0xff,0x3f,
	0xfe,0xff,0xf3,0x87,0xff,0xdf,0xff,0xef,
	0xef,0x3f,0xdf,0xfe,0xd3,0x8f,0xff,0xfc,
	0xf3,0xcf,0x3f,0xff,0x7c,0xe7,0xe7,0xef,
	0xdf,0xf9,0x77,0xf7,0x6f,0xf7,0x5c,0x78,
	0x7f,0xff,0xff,0xf7,0xad,0xef,0xff,0x7f,
	0xff,0xff,0xff,0xbf,0xff,0xff,0xff,0xff,
	0xff,0xd7,0xff,0xdf,0xef,0x7f,0xef,0xff,
	0xfd,0xbf,0xfd,0x77,0xff,0x7f,0xff,0x7f,
	0xee,0xff,0xf7,0xff,0x7f,0xff,0xfb,0xff,
	0xfb,0xfe,0xff,0xbd,0x6b,0xbf,0xbf,0xf6,
	0xda,0xab,0xd7,0x7f,0xfa,0xeb,0xa7,0xdf,
	0xdf,0xbb,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0x6e,0xff,0xff,0x7f,0xff,0xf5,0xff,
	0x1f,0x7f,0xff,0xff,0xff,0xff,0xef,0xff,
	0xff,0xff,0xbb,0xff,0xff,0xff,0xfe,0xef,
	0xfd,0xfe,0xc7,0x5f,0xfb,0xfb,0xfd,0xd7,
	0x7b,0xf7,0xef,0xf9,0xff,0xff,0xdf,0xbf,
	0xff,0xff,0xff,0xef,0xff,0xff,0xf7,0xdf,
	0xff,0xff,0xff,0x7f,0xff,0xfd,0xff,0xff,
	0xff,0xfb,0xfe,0xff,0xff,0xeb,0xff,0xbf,
	0xff,0xfd,0xfe,0x7f,0xc9,0xff,0xff,0xff,
	0x2f,0xbf,0xff,0xf7,0xff,0xf5,0xff,0xff,
	0xff,0xff,0xff,0xbf,0xff,0x67,0xfb,0xff,
	0xff,0xff,0xff,0xff,0xff,0xfd,0xf6,0x7f,
	0xfb,0xff,0xfd,0xff,0xff,0xdf,0xfe,0xff,
	0xff,0xbf,0xdd,0xff,0xff,0x7d,0xf7,0x7d,
	0xff,0xff,0xff,0xf7,0xdf,0x7f,0xbf,0xff,
	0x7d,0xfb,0xef,0xfd,0xff,0xdf,0xff,0xfd,
	0xfd,0xfd,0xdf,0xff,0xfb,0xbb,0x7f,0xf5,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xab,0xaf,0xff,0xff,0xff,0xfb,0xff,
	0xbe,0xff,0xff,0xef,0xbf,0xff,0xfe,0xff,
	0x7f,0xff,0xfb,0xff,0xff,0xff,0xfe,0xff,
	0x92,0xff,0x8f,0xff,0xff,0xdf,0xf1,0xdd,
	0x3f,0xef,0xfd,0xf3,0xcf,0x3f,0xf7,0xff,
	0xf3,0x2f,0xff,0xff,0xf3,0xcf,0xff,0xff,
	0xfd,0xe3,0xcf,0x3e,0xff,0xfc,0x35,0xdd,
	0xff,0x3f,0xfc,0xff,0xfb,0xff,0xff,0xff,
	0xfe,0xfb,0xf9,0xef,0x5f,0xff,0xfe,0xfb,
	0xff,0xbf,0xff,0xfb,0xfd,0xbf,0xff,0xfe,
	0xff,0xef,0xff,0x7e,0xfd,0xbb,0xef,0xbf,
	0xfd,0xf6,0xbf,0xe8,0xff,0xff,0xff,0xff,
	0xff,0xf5,0xdf,0xaf,0xff,0xfd,0xff,0xe3,
	0x9f,0xff,0xee,0xf1,0x0e,0xff,0xf5,0xfd,
	0xe6,0xbf,0xfe,0xff,0xff,0xff,0x9e,0x7f,
	0xff,0xff,0xff,0xfd,0x77,0xfe,0xfd,0xfb,
	0xef,0xe7,0xff,0xef,0x7f,0xfd,0xff,0xfe,
	0x7f,0xff,0xfc,0xe5,0xcf,0xfb,0xbd,0xff,
	0xff,0xfe,0xfb,0xfd,0xf7,0xff,0xfe,0xff,
	0xdd,0xf4,0xff,0xff,0xff,0x8f,0xfa,0xff,
	0xdf,0xbf,0xbf,0xff,0xff,0xef,0xff,0xff,
	0xff,0xff,0xeb,0xb7,0xef,0xff,0xfe,0xf7,
	0xff,0xff,0xff,0xee,0xf3,0xbf,0xff,0xff,
	0xff,0xf3,0xff,0xff,0xff,0xff,0x3f,0x0f,
	0xff,0xff,0xeb,0xcf,0x9f,0xff,0x7f,0xdf,
	0xff,0xf7,0xff,0xfa,0xff,0xff,0x37,0xcf,
	0xec,0xe7,0xef,0xff,0xff,0xf3,0xcf,0xf7,
	0xff,0x7f,0xff,0xe3,0xef,0xdf,0x7a,0xff,
	0x93,0xd6,0xff,0x9f,0xff,0xff,0xfc,0xff,
	0xfd,0x3e,0xff,0xff,0xd7,0x5f,0xff,0xff,
	0x79,0xf3,0x3f,0xff,0xff,0xff,0xff,0x7f,
	0xdd,0xff,0xff,0xfd,0x3c,0xff,0x7f,0xf3,
	0xfd,0xea,0xf5,0xfe,0xff,0xff,0xff,0xfe,
	0xff,0xff,0xb7,0xff,0xfd,0xff,0xff,0xff,
	0xff,0xfd,0x9f,0xf7,0xfe,0x2f,0x3f,0xf5,
	0x9f,0xff,0xff,0xff,0xff,0xdf,0x7f,0xff,
	0xff,0xff,0xd7,0xb7,0xfd,0xff,0xff,0xff,
	0xf5,0xff,0xff,0x3f,0xfb,0xff,0xff,0xdf,
	0x7f,0xff,0xff,0xf7,0x39,0xff,0xff,0xe7,
	0x89,0x67,0xfa,0xff,0xff,0xff,0x3f,0xff,
	0xf9,0xff,0xff,0xff,0x1b,0xfd,0xff,0xff,
	0xbf,0xbf,0xff,0xfb,0xef,0xed,0xff,0xfe,
	0xfb,0xeb,0xef,0xff,0xfe,0xfe,0xa6,0xff,
	0xfe,0xfa,0xeb,0xaf,0xff,0xfe,0xfb,0xef,
	0xaf,0xba,0xfe,0xfb,0xef,0xff,0x62,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xef,0xfe,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xbf,
	0xfe,0xff,0x7f,0xbd,0xff,0xfd,0xff,0xff,
	0xff,0xff,0xdb,0xf7,0xff,0xff,0xdf,0x46,
	0x9e,0xff,0xff,0xff,0xff,0x7f,0xf9,0x2f,
	0xff,0xdf,0xfa,0xff,0xff,0xff,0xbf,0xfa,
	0x7e,0xfd,0x5f,0xf6,0x5f,0xff,0xfd,0xdf,
	0xff,0xeb,0xe5,0xbf,0xf7,0xff,0xeb,0xfd,
	0x1f,0xf9,0xfd,0xff,0xbf,0xff,0xff,0x57,
	0x4f,0xf9,0xff,0xd5,0xf3,0xef,0xff,0xff,
	0xd4,0xff,0xa9,0xff,0xf5,0xfa,0xea,0xef,
	0xff,0xfd,0x53,0x5f,0xbd,0xb8,0xff,0xc3,
	0xdf,0xbf,0xd1,0xff,0xff,0xff,0xfe,0xfb,
	0xef,0xbf,0xff,0xfe,0xfb,0xef,0xbf,0xff,
	0xfe,0xfb,0xef,0xff,0xf6,0x7b,0xef,0xbf,
	0xff,0xfe,0xfb,0xef,0xbf,0xff,0xfe,0xdb,
	0xef,0xbf,0xff,0x3d,0xfe,0xff,0xfb,0xe7,
	0x9f,0x7f,0xfe,0xf9,0x67,0x9f,0x7f,0xfe,
	0xfd,0xe7,0x9f,0x7f,0xfe,0xf7,0x9d,0x7f,
	0xe6,0xfd,0xe7,0x9f,0x7f,0xfe,0xf9,0xe7,
	0x9f,0x77,0xfe,0xf9,0x7f,0xe1,0xff,0xf7,
	0x9f,0x6f,0xfd,0xf1,0xe7,0x5f,0x7d,0xfc,
	0xf9,0xf6,0x5f,0x7f,0xfe,0xf5,0xda,0x7f,
	0xfc,0xfd,0xb7,0x5f,0x7f,0xfc,0xf9,0xc7,
	0xdf,0x6b,0xfd,0xf9,0xc7,0xff,0xb9,0xfd,
	0xff,0xff,0x3f,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xf3,0xdf,0xff,0xff,0xff,0xf3,
	0xff,0xff,0xff,0xff,0xdf,0xff,0xff,0xff,
	0xff,0xff,0x7f,0xff,0xff,0xff,0xff,0x3b,
	0xfd,0xff,0xf7,0x5f,0xff,0xff,0xff,0xf7,
	0xff,0xff,0xff,0xdd,0xdf,0xff,0x7f,0xff,
	0xdf,0xfe,0xff,0xff,0xff,0xbf,0xff,0xff,
	0xff,0xfd,0xff,0x7f,0xff,0xff,0xfd,0xff,
	0xff,0xa8,0x7f,0xff,0x7f,0xf6,0xf8,0x47,
	0x1f,0x2f,0xf6,0xf1,0x57,0x8f,0x3f,0xf6,
	0xf4,0x63,0x2f,0xfe,0xd1,0xf2,0xcf,0x3f,
	0xfe,0xf4,0xd7,0x5f,0x7f,0xfe,0xf8,0xc7,
	0x1f,0xff,0x05,0xff,0xff,0xfd,0xf7,0xd7,
	0x7f,0xff,0x7d,0xf5,0xdf,0x7f,0x7f,0xfd,
	0xf7,0xc7,0x5f,0xff,0xf5,0xdf,0x57,0x79,
	0xfd,0xf5,0xc7,0x7f,0xff,0xfd,0x97,0xdd,
	0x7f,0xff,0xfd,0xbf,0xfa,0xff,0x7f,0xff,
	0x3f,0xf7,0xef,0xef,0xdb,0x7f,0xf3,0xef,
	0xbb,0xef,0xb7,0xf7,0xfe,0xef,0xff,0xff,
	0xfe,0xbb,0xef,0xbf,0xf3,0xef,0xbf,0xbf,
	0xbf,0xff,0xeb,0xbf,0xff,0xd3,0xff,0xff,
	0xe7,0x9e,0x7d,0xee,0xf9,0xc7,0x9f,0x7f,
	0xfe,0x99,0xe7,0x9e,0x7d,0xe6,0xf9,0x9e,
	0x7b,0xfc,0xd9,0x67,0x9e,0x7d,0xfe,0xf9,
	0xe7,0x9f,0x79,0xfe,0xb9,0xe7,0xef,0xfe,
	0xff,0xfd,0xf7,0xef,0x7f,0xff,0xfd,0xfb,
	0x9f,0x7f,0xff,0xfc,0xfb,0x8f,0xff,0x7f,
	0xf2,0xdf,0xef,0xff,0xfd,0xfb,0xef,0x7f,
	0xff,0xf9,0xfb,0xef,0x7f,0xff,0xfd,0x7f,
	0xf1,0xff,0xcf,0x7f,0xdf,0xff,0xfb,0xff,
	0xdf,0xff,0xff,0xfb,0xed,0xff,0x7f,0xff,
	0xff,0x9f,0xff,0xff,0xfd,0xef,0xf7,0x7f,
	0xff,0xfb,0xff,0xdf,0xdf,0xff,0xff,0xff,
	0xff,0x8b,0xff,0xff,0xff,0xfd,0xeb,0xff,
	0x7f,0xff,0xfb,0xff,0xff,0xbf,0xfe,0xfa,
	0xfb,0xbf,0xdf,0xfa,0xf7,0xbf,0xff,0xfe,
	0xfa,0xfb,0xff,0xff,0xff,0xfa,0xeb,0xdf,
	0x7f,0xff,0x1f,0xfc,0xff,0xff,0xff,0xf7,
	0xff,0xfb,0xef,0xff,0xff,0xfe,0x7b,0xfb,
	0xff,0xdf,0xfe,0xff,0xef,0xff,0xfe,0xff,
	0xfb,0xfd,0xdf,0xfe,0xfb,0xef,0xf7,0xdf,
	0xff,0xfb,0xef,0xff,0xeb,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0x2f,0xff,0xf3,0xff,
	0x7f,0xff,0xfe,0xff,0xff,0xe7,0xff,0xfb,
	0xff,0xff,0x4f,0xff,0xf7,0xef,0xff,0xff,
	0x3f,0xfd,0xff,0xbf,0xff,0x4f,0xff,0xff,
	0xf5,0xd7,0x5f,0x7f,0xfd,0x75,0xd7,0x4e,
	0x7b,0xf5,0xb5,0xd3,0x4f,0x3f,0x7d,0x97,
	0x5e,0x3d,0xfd,0x74,0xd3,0x5f,0x7f,0xfd,
	0xe5,0x07,0x5d,0x7e,0xf1,0xe5,0xaf,0xfa,
	0xff,0xff,0xcf,0xff,0xf7,0xff,0xf3,0x8f,
	0xff,0xff,0xda,0xff,0xff,0xb7,0xfb,0xff,
	0xff,0x3e,0xff,0xfc,0x73,0xad,0xbf,0xff,
	0xfc,0xe3,0xbf,0xff,0xff,0xfa,0xe7,0xfe,
	0x85,0xff,0xff,0xbf,0xff,0xff,0xfe,0xfe,
	0x3f,0xf7,0xff,0x7f,0xff,0xff,0xfb,0xff,
	0xff,0xfe,0xff,0xfe,0xf3,0xf7,0xff,0xd7,
	0xbf,0x7b,0xef,0xff,0xff,0x7f,0xff,0xf7,
	0xff,0x15,0xfe,0xef,0x7f,0xeb,0xff,0xff,
	0xbf,0xfe,0xff,0xff,0xbf,0xff,0xff,0xff,
	0xef,0xff,0xb7,0xff,0x6f,0xff,0xff,0xfe,
	0xfd,0xef,0xbf,0xfd,0xbe,0xef,0xff,0xdf,
	0xff,0xff,0xff,0xf7,0xff,0x7f,0xff,0x5f,
	0xff,0xff,0xfe,0xf6,0xff,0xff,0xff,0xb1,
	0xfe,0xff,0x6f,0xfd,0xfb,0xff,0x7f,0xfd,
	0xf4,0xd7,0xff,0xff,0xff,0xff,0xc7,0xff,
	0xff,0xff,0xff,0xff,0x2f,0xff,0xff,0xff,
	0xdf,0xfa,0xff,0xff,0xff,0xfe,0xff,0xff,
	0xef,0xfd,0xff,0xff,0xd7,0xff,0xff,0xdf,
	0xea,0x5e,0xbf,0xfe,0xff,0xbf,0xfd,0xbf,
	0xbf,0xff,0xff,0xff,0xff,0x17,0xb9,0xff,
	0xec,0xf7,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xef,0xff,0xbf,0xff,0x7f,0xb7,
	0xff,0xff,0xff,0xff,0xff,0xcf,0x7d,0xfb,
	0xff,0xff,0x9f,0xff,0xff,0xff,0xbe,0xcb,
	0xfe,0xfb,0xff,0xff,0xbf,0xff,0xdf,0x7f,
	0xff,0xfd,0xff,0xff,0xff,0xff,0xef,0xff,
	0xf3,0xdf,0xbd,0xff,0xdf,0x7f,0xff,0xfd,
	0xfc,0xdf,0xff,0x3f,0xff,0x77,0xdf,0xff,
	0x67,0xff,0xdf,0xff,0xfb,0xdf,0xfe,0xef,
	0xff,0xff,0xff,0x7f,0xff,0xff,0xff,0xff,
	0xfe,0xfa,0xfb,0xff,0xff,0xff,0xbf,0xff,
	0xff,0xdd,0xff,0xff,0xff,0xb6,0xff,0xfe,
	0xff,0xff,0xf3,0xff,0xf7,0xff,0xcf,0x3f,
	0xff,0xfc,0xf3,0xcf,0x3f,0xf7,0xff,0xd7,
	0xc7,0x1f,0xfe,0xde,0xcf,0x3f,0xff,0xfc,
	0xff,0xcf,0x3e,0xff,0xfc,0xe1,0xcf,0x37,
	0xff,0xfc,0x7f,0xcf,0xff,0xff,0xff,0xef,
	0xfe,0xfb,0xbf,0xbf,0xff,0xfe,0xfb,0xff,
	0x7f,0xff,0xfe,0xaf,0xff,0xff,0xfe,0xfb,
	0xd7,0xfb,0xff,0xfe,0xfb,0xb7,0x7f,0xff,
	0xfe,0xfb,0xef,0xff,0x4f,0xfe,0xff,0xff,
	0xf7,0xf7,0xdf,0x7f,0xfd,0xf5,0xe3,0x9f,
	0xff,0xee,0xdf,0xe7,0x7f,0xcf,0xf7,0xf7,
	0x0d,0xef,0xfd,0xff,0xf7,0x9f,0xff,0xff,
	0xff,0xf5,0x9f,0x7f,0xfe,0xff,0xf4,0xff,
	0xff,0xff,0x7f,0xff,0xbf,0xfe,0xff,0x7f,
	0xde,0xfd,0xf7,0xbb,0x3f,0xff,0xff,0xff,
	0xff,0xff,0xfd,0xff,0xff,0xff,0xde,0xfd,
	0xf7,0xff,0xff,0xff,0xfd,0xf7,0x7f,0xb1,
	0xff,0xff,0xfd,0xff,0xfb,0xbf,0xff,0xfe,
	0xff,0xfd,0xeb,0xaf,0xff,0xff,0xfd,0xff,
	0xff,0xfb,0x6f,0xef,0xff,0xfe,0xff,0xff,
	0xf3,0xbf,0xff,0xff,0xff,0xf3,0xcf,0xff,
	0x7f,0xf0,0xff,0xff,0xff,0xff,0xff,0xcf,
	0xf7,0xcf,0x3f,0xf9,0xff,0xbf,0xde,0xfa,
	0xf9,0xe5,0xd8,0xff,0xff,0xff,0xf7,0xff,
	0xf4,0xea,0xff,0xf7,0xfd,0xfc,0xff,0xcf,
	0xff,0x3f,0x6d,0xfd,0xf3,0xff,0xff,0xff,
	0x7f,0xfd,0x7f,0xff,0xfd,0xff,0xff,0xf5,
	0xd7,0xcf,0x3f,0xef,0xfc,0xff,0xfd,0xdf,
	0xf5,0x97,0x5f,0xff,0xdf,0xef,0x17,0x5f,
	0xff,0xf7,0xbf,0x36,0xef,0xd3,0x7e,0xfb,
	0xff,0xff,0xff,0xfd,0xff,0xff,0xff,0xff,
	0xfe,0xf7,0xff,0x7f,0x79,0xff,0xff,0xdf,
	0xff,0xff,0xff,0xfd,0xd7,0xff,0x7f,0xff,
	0x77,0xde,0x7f,0xff,0xbd,0xd9,0x9f,0x6b,
	0xbe,0xdf,0x7f,0xff,0xff,0xf7,0xdf,0x7f,
	0xff,0xfd,0xff,0xdf,0xff,0xff,0xfb,0xdf,
	0x7f,0xff,0xff,0xff,0xcf,0x33,0xfe,0xff,
	0xfb,0xcf,0x73,0xfe,0xfd,0xbf,0xcd,0xef,
	0xff,0x4d,0xaa,0xeb,0xef,0xff,0xfe,0xfe,
	0xfb,0xaf,0x3f,0xff,0xfb,0xfb,0xbf,0xef,
	0xfb,0xeb,0xa7,0xff,0xfe,0xfb,0xeb,0xaf,
	0xff,0xde,0xfb,0xcb,0xae,0x9f,0xfe,0x6f,
	0xf6,0xff,0x5f,0x9a,0xfd,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xef,0xff,0xff,0xdb,0xff,0xff,0xff,
	0xff,0xfd,0xff,0xee,0xff,0xff,0xff,0xf7,
	0xef,0xe1,0xff,0xf5,0xdf,0xff,0xef,0xff,
	0xdf,0xff,0xff,0xff,0xff,0xff,0xfe,0xfb,
	0xff,0xf7,0x53,0xff,0xff,0xfd,0xff,0x7f,
	0xff,0xff,0xff,0x5f,0x76,0xfb,0xff,0xfd,
	0xf7,0xe7,0xad,0xff,0xaf,0xbf,0xfc,0x6f,
	0xfc,0xff,0xfd,0xff,0xff,0xbf,0xff,0xe3,
	0x93,0xff,0x7f,0xbf,0xf6,0xef,0xbf,0xff,
	0xd4,0xf7,0xef,0xff,0xff,0x95,0xb7,0xef,
	0xaf,0xff,0xbf,0x1f,0x7c,0xfe,0xef,0xaf,
	0xbf,0xed,0xb6,0xfb,0x6f,0xbe,0xff,0xfe,
	0xdb,0xee,0xbd,0xed,0xfe,0xff,0xbf,0xff,
	0xfe,0xfb,0xef,0xbf,0xff,0xf6,0xfb,0xef,
	0xbf,0xff,0xf6,0xdb,0xff,0xec,0xfb,0xbf,
	0xfd,0xff,0x9d,0x67,0x9f,0x7f,0xf6,0xf9,
	0x77,0x9f,0x7d,0xfe,0xf9,0xe7,0xff,0xff,
	0xfd,0xe7,0x9f,0x7f,0xfe,0xf9,0xe7,0x9d,
	0x7f,0xfe,0xf9,0xe7,0x9d,0xf7,0x57,0xfe,
	0x7f,0xf7,0xed,0xf7,0x5f,0x77,0xfc,0x95,
	0xd6,0xdf,0x77,0xdd,0xf9,0xd7,0x1f,0xbf,
	0xbd,0xf6,0xdb,0x7f,0xfe,0xb1,0xf6,0x5b,
	0x7f,0xfc,0xb1,0xf6,0xdb,0x6f,0xfd,0x1f,
	0xd9,0xff,0xdf,0xff,0xff,0xff,0xff,0xff,
	0xff,0xf7,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xf7,0xdb,0xff,0xff,0xff,0xf7,
	0xdb,0xff,0xff,0xff,0xf7,0xdb,0x6f,0xff,
	0xbf,0xdd,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xef,0xff,0xff,0xff,0xff,0xfd,
	0xff,0xff,0xef,0xf7,0x7f,0xff,0x7f,0xff,
	0xf7,0x7f,0xff,0xff,0xff,0xf7,0xbf,0xff,
	0xfe,0xff,0x0f,0xfa,0xbf,0xfe,0x77,0x8f,
	0x2f,0xf7,0xf5,0xf7,0xdf,0x2f,0xfe,0xd8,
	0xc7,0x8b,0x7d,0xfd,0x7f,0xcf,0x7d,0xff,
	0xf1,0xc7,0x8f,0x3d,0xde,0xf4,0xc7,0x9f,
	0x3f,0xfe,0xf8,0x5f,0xf4,0xbf,0x55,0x7f,
	0x7f,0x6d,0xf5,0xdf,0x7f,0x7f,0x7d,0xf5,
	0xd7,0x7f,0x5f,0xfd,0xf7,0x3b,0xdf,0xfd,
	0xd5,0xdf,0x7f,0xdf,0xf5,0xf5,0xc7,0x7f,
	0xff,0xf5,0xd5,0xdd,0xff,0x97,0xff,0xff,
	0xfb,0xef,0xeb,0xff,0xff,0xbb,0xfe,0xff,
	0xff,0x3f,0xff,0xed,0xfb,0xff,0xff,0xff,
	0xfb,0xff,0xff,0xfe,0xef,0xfb,0xef,0xbf,
	0xfb,0xed,0xbf,0xef,0xbe,0xff,0x1f,0xbd,
	0xff,0x7f,0x7e,0xd9,0xe7,0x9f,0x7f,0xf4,
	0xd9,0xe7,0x9e,0x7d,0xfe,0x99,0xe7,0x9f,
	0xff,0xf1,0x67,0x9f,0x7f,0xee,0xb1,0xc7,
	0x9e,0x7d,0xfe,0xb9,0xc7,0x9e,0x79,0xfe,
	0xee,0xff,0xff,0xbf,0xbb,0xfc,0xfb,0xdf,
	0xff,0xff,0xff,0xfb,0xeb,0x77,0xbe,0xff,
	0xf7,0xaf,0xff,0xfd,0xff,0xdf,0x7f,0xff,
	0xff,0xfb,0x8f,0x7f,0xff,0xfe,0xfb,0xef,
	0xff,0x17,0xff,0xff,0xfb,0x6d,0xc5,0x5f,
	0xff,0xfe,0xf7,0xd7,0x5f,0x5f,0xfd,0xff,
	0xdd,0xbf,0x7f,0xfd,0xcf,0x5f,0xff,0xff,
	0xff,0xdf,0x5f,0x7f,0xff,0xff,0xdd,0x5f,
	0xdf,0xfd,0xff,0xfb,0xff,0xfa,0xef,0xff,
	0xff,0xfe,0xff,0xff,0xff,0xff,0xfe,0xff,
	0xff,0xff,0xff,0xff,0xe9,0xff,0xff,0xff,
	0xf7,0xdf,0xff,0xff,0xfe,0xfb,0xdf,0xbf,
	0xff,0xfe,0xfb,0xff,0xc9,0xff,0xd7,0x7f,
	0xff,0xfb,0xf7,0xff,0xfe,0xff,0xfd,0xf7,
	0xff,0xfb,0xfb,0xff,0xbf,0x5f,0xfb,0xfb,
	0xf7,0xff,0xfe,0xfb,0xff,0xfd,0xdf,0xfe,
	0x7b,0xff,0xfd,0xdf,0xff,0x5d,0xfe,0xff,
	0xff,0xf6,0xbf,0xff,0xff,0xfb,0xff,0xff,
	0xff,0xee,0xfb,0xff,0xff,0xff,0x7f,0xf6,
	0xff,0xef,0xff,0xfb,0xef,0xcb,0xff,0xfe,
	0xfb,0xef,0xff,0x4f,0xbe,0xfc,0xff,0xd0,
	0xff,0x5f,0x75,0xfd,0xf5,0xd7,0x5f,0x7f,
	0xfd,0xf5,0xd7,0x5f,0x7f,0xfd,0xf4,0x97,
	0x5f,0xfd,0xf5,0xd6,0x5f,0x3b,0xbd,0xb5,
	0xd7,0x5f,0x7e,0xe9,0x45,0x97,0x5f,0xfc,
	0x9a,0xff,0x3f,0xbe,0xdc,0xff,0xcd,0x3f,
	0xfe,0xfc,0xf3,0xff,0xff,0xef,0xdf,0xef,
	0xcf,0xff,0x6d,0xf3,0xcf,0x3b,0xfe,0xf8,
	0x7f,0xff,0x3f,0xdf,0xfc,0xff,0xff,0xdf,
	0xfe,0x5f,0xf8,0xff,0xdf,0xff,0xff,0x3f,
	0xff,0xdf,0xf3,0xcf,0xff,0xff,0xef,0xff,
	0xff,0xff,0xfb,0xfc,0xcf,0x3f,0xdf,0x7e,
	0xfb,0xff,0xff,0xf7,0xfe,0xf5,0xfd,0xfd,
	0xf7,0xff,0x7f,0xee,0xff,0xf3,0xcf,0xff,
	0xff,0xff,0xeb,0xff,0xff,0xff,0xff,0xdf,
	0xff,0xff,0xfd,0xfa,0xfb,0xff,0xff,0xff,
	0xf7,0xdf,0xff,0xfd,0xff,0xef,0xff,0xff,
	0xff,0xff,0xff,0xff,0x67,0xdf,0x7f,0xff,
	0xfd,0xf5,0xc7,0xfb,0xef,0xfd,0xff,0xc7,
	0x1f,0xff,0xfb,0xe5,0xff,0x5f,0xfd,0xf9,
	0xff,0xff,0xfe,0xff,0xf5,0xff,0xff,0xff,
	0xff,0xff,0xff,0xfd,0xff,0xbf,0xfa,0xff,
	0xfe,0xff,0xae,0xbf,0xff,0xff,0xff,0xff,
	0xbf,0xff,0xfe,0xff,0x5f,0xff,0xbf,0xff,
	0xff,0xfb,0xff,0xff,0xff,0xab,0xfb,0xff,
	0xaf,0xff,0xf7,0xfb,0xff,0x9f,0xff,0xd7,
	0xff,0xff,0xdf,0xff,0xff,0xef,0xff,0xdf,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xf7,0xf7,0x7f,0xff,0xff,0x3f,0xff,0xfc,
	0xff,0xcf,0xff,0xff,0xff,0xff,0xff,0xef,
	0x7f,0xfe,0xff,0xf7,0xff,0xfd,0xff,0xff,
	0xfd,0xff,0xbf,0xff,0xff,0xff,0xff,0xff,
	0xff,0x9f,0xff,0xff,0x7b,0xff,0x7f,0xf6,
	0xff,0x7d,0x9f,0xfe,0x77,0xff,0xff,0xdf,
	0x7f,0x7f,0xf3,0xff,0xff,0x7f,0xff,0xef,
	0xff,0xff,0xff,0xff,0xff,0xef,0xff,0x7f,
	0xfe,0xef,0xf7,0xfe,0xff,0xff,0xff,0xff,
	0xdf,0xff,0xfd,0xbf,0xff,0xff,0xff,0xff,
	0xff,0xff,0xeb,0x83,0xff,0x8f,0xff,0xdf,
	0xfd,0xfb,0xdf,0xff,0xff,0xf8,0xe3,0xff,
	0x7f,0xff,0xfc,0xfd,0x1f,0xeb,0xb8,0xff,
	0xff,0xff,0xff,0xff,0xe3,0xff,0xff,0xff,
	0xfc,0xff,0xff,0x7f,0xbf,0xfc,0xff,0xfd,
	0xff,0xef,0xdf,0xff,0xfb,0xff,0xbf,0xff,
	0xfe,0xff,0xfb,0xe7,0xff,0x7f,0xf4,0xbf,
	0xff,0xff,0xff,0xff,0xff,0xff,0xf6,0xff,
	0xff,0xe7,0xff,0xff,0xff,0xfb,0xeb,0xff,
	0xff,0xfe,0xff,0xff,0x72,0xff,0x2f,0xfb,
	0xfd,0xff,0xff,0xbf,0xff,0xfd,0xfb,0xfe,
	0xfe,0xfd,0xf7,0xef,0xbf,0xbb,0xfe,0xff,
	0xeb,0xff,0xff,0xff,0xff,0xcb,0xff,0x4f,
	0xfe,0xff,0xfe,0x7f,0xff,0xd7,0xff,0xef,
	0xfd,0xef,0xff,0xff,0xff,0xfd,0xff,0xdf,
	0xff,0xff,0xff,0xd5,0x7f,0xff,0xfd,0xe7,
	0xff,0x7f,0xfe,0xff,0x6f,0xbf,0x77,0xff,
	0x5f,0xf8,0xff,0xff,0xff,0xff,0xff,0xdf,
	0xff,0xaf,0xff,0xff,0xff,0xef,0xf3,0xff,
	0x3f,0xff,0xff,0xff,0xfe,0xf7,0xfc,0xf3,
	0xbf,0xff,0xff,0xfb,0xbf,0xff,0xff,0xff,
	0xfb,0xff,0x01,0xff,0xff,0xf8,0xff,0xff,
	0xff,0x7a,0xfe,0xff,0xff,0xff,0xff,0xff,
	0xff,0xbf,0xfe,0xff,0xe3,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xef,
	0xff,0xff,0xff,0x23,0xd6,0xff,0xcf,0xff,
	0xff,0xff,0xd7,0xfd,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xf9,0x67,0x3f,0xff,0xdf,
	0x7f,0xff,0xff,0xf3,0xff,0xe7,0xff,0xff,
	0xf7,0xff,0xe7,0xff,0xea,0xf2,0xfe,0xff,
	0xff,0xff,0xff,0xfb,0xff,0xbf,0xff,0xfe,
	0xfb,0xef,0xbf,0xff,0xfe,0xfb,0xbf,0xff,
	0xff,0xf9,0xe7,0xbf,0x7f,0x4e,0xfb,0xef,
	0xda,0x7f,0xae,0x9d,0xe7,0xd7,0x97,0xfd,
	0xff,0xff,0xff,0xff,0xb7,0xff,0x7f,0xfb,
	0xed,0xb7,0xdf,0x7f,0xfb,0xed,0xb7,0x7f,
	0xfd,0xfd,0xb7,0xdf,0x7e,0xfb,0x6c,0xa7,
	0x5d,0xf7,0xfe,0x6d,0xef,0x9c,0xfe,0xfb,
	0xfd,0xff,0xef,0xbf,0xff,0xfe,0xfa,0xef,
	0xaf,0xbf,0xfe,0xfa,0xe3,0xaf,0xbf,0xfe,
	0xfa,0xae,0xbf,0x7e,0xfa,0xe9,0xa7,0xbf,
	0xf6,0x0a,0x69,0xbf,0x9f,0xf6,0xfb,0xeb,
	0xff,0x61,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xfd,0xb6,
	0xfd,0x5f,0x9f,0x69,0xff,0xfd,0x67,0xdf,
	0x7f,0xdf,0x7e,0x9e,0xff,0xff,0xff,0xff,
	0xff,0xff,0xaf,0xff,0xff,0xff,0xff,0xef,
	0xff,0xff,0xff,0xbf,0xff,0xff,0xdf,0x73,
	0xef,0xbd,0xff,0xf7,0xff,0xff,0xfd,0xf7,
	0xdf,0x7f,0xff,0x9f,0xfb,0xfd,0xef,0xbf,
	0xff,0xfe,0xfb,0x4f,0xbd,0xff,0xfe,0xfb,
	0x6f,0xbe,0xff,0xfe,0xfb,0xbf,0xff,0xfe,
	0x9b,0x2f,0xee,0xf9,0xf6,0xfe,0xff,0xef,
	0xff,0xff,0xfe,0xfb,0xff,0xdb,0xff,0xbf,
	0xff,0xfe,0xfb,0xef,0xbf,0xff,0xfe,0xfb,
	0xef,0xbf,0xf7,0xfe,0xfb,0xef,0xff,0xfe,
	0xfb,0xef,0xbd,0xff,0xde,0x7b,0xef,0xbf,
	0xff,0xb7,0xfb,0xff,0xbf,0xff,0xff,0xfe,
	0xff,0xf9,0xe7,0x9f,0x7f,0xff,0xf9,0xf7,
	0xdf,0x7f,0xff,0xdd,0xf7,0xdf,0x7f,0xff,
	0xf7,0xdf,0x7f,0xfe,0xf9,0xe7,0x9f,0x7f,
	0xfe,0xf9,0x7f,0x9e,0xff,0xff,0xf9,0x7f,
	0xe2,0xff,0xe7,0x9f,0x7f,0xbe,0xfd,0xe7,
	0xdf,0x6f,0xbf,0xfd,0xf7,0xdd,0x6f,0xff,
	0xfd,0xdf,0x6b,0xff,0xfd,0xf7,0x5f,0x7f,
	0xbf,0xf5,0xd6,0xdf,0x7f,0xad,0xfd,0xd6,
	0xff,0x95,0xfd,0xff,0xff,0xff,0xff,0xfc,
	0xff,0xff,0x6f,0xbf,0xfd,0xff,0xff,0x6f,
	0xff,0xff,0xd7,0x7f,0xff,0xff,0xff,0xff,
	0xff,0xbf,0xfd,0xf6,0xff,0xff,0xff,0xfd,
	0xfe,0xff,0x3b,0xfd,0xff,0xf7,0xdf,0x7f,
	0xf7,0xff,0xf7,0xff,0xfe,0xfb,0xff,0xff,
	0xff,0xfe,0xff,0xdf,0x7f,0xff,0xff,0xff,
	0xff,0xff,0xff,0xf7,0xdf,0xff,0xff,0xff,
	0xf7,0xff,0xff,0xff,0xe3,0xff,0xdf,0x7f,
	0xff,0xfd,0x62,0x1f,0x7f,0xfe,0xf9,0xe7,
	0xdf,0x3d,0xfe,0xf8,0xe3,0x37,0xfe,0xf8,
	0x63,0x9f,0x3d,0xfe,0xf9,0xe3,0xcf,0xff,
	0xb7,0xf9,0x7f,0x9f,0xff,0x1d,0xff,0xff,
	0xfd,0xf7,0xdf,0x77,0xff,0xfd,0xf7,0xdf,
	0x7f,0xff,0xfd,0xf5,0xdd,0x5f,0x7f,0xf7,
	0xdd,0x5f,0xfd,0xf5,0xb5,0x5f,0x76,0x7f,
	0xfd,0xf7,0xd7,0x7f,0x7f,0xfd,0x3f,0xf3,
	0xff,0xff,0xff,0xff,0xbf,0xfc,0xef,0xdf,
	0x7f,0xff,0xfd,0xff,0xcf,0x3f,0xff,0xfe,
	0xcd,0xbf,0xff,0xfa,0xef,0xaf,0xff,0xfe,
	0xfc,0x7b,0xff,0x6f,0xf7,0xff,0xff,0xff,
	0xc5,0xff,0xff,0xe7,0x9f,0x7f,0xfc,0xf9,
	0xe7,0x9f,0x7f,0xfe,0xf9,0x67,0x1f,0x7f,
	0xe6,0xd9,0x1f,0x7b,0xf6,0xf9,0x67,0x9f,
	0x7f,0xfc,0xd9,0xe7,0x9f,0x7d,0xfe,0x99,
	0xe7,0x7f,0xfe,0xff,0xfd,0xff,0xdf,0xbf,
	0xff,0xff,0xfb,0xcf,0xbf,0xbf,0xfe,0xf7,
	0xdf,0x7f,0xff,0xff,0xff,0xff,0xff,0xfe,
	0xff,0xef,0xff,0xff,0xfe,0xfa,0xff,0xbf,
	0xff,0xff,0xfb,0xf5,0xff,0xef,0xff,0xff,
	0xfe,0xff,0xbf,0xdf,0xdf,0x7f,0xff,0xf5,
	0xff,0xff,0xff,0xff,0xfb,0xff,0xff,0xff,
	0xf7,0xff,0x7f,0xff,0xff,0xfd,0xd7,0xff,
	0x7f,0xff,0xf7,0xff,0xaf,0xff,0xff,0xfe,
	0xfb,0xef,0xef,0xff,0xff,0xfe,0xf3,0xef,
	0xff,0xff,0xfd,0xf7,0xdf,0xff,0xff,0xff,
	0xff,0xbf,0xff,0xff,0xfb,0xff,0xbf,0xff,
	0xff,0xff,0xef,0xff,0xff,0x3f,0xfc,0xff,
	0xff,0xff,0xff,0xff,0xfb,0xff,0xb7,0xb7,
	0xde,0xfb,0xef,0xaf,0xbf,0xfe,0xfa,0xbf,
	0xff,0xfe,0xfb,0xed,0xbf,0xdf,0xfe,0x7b,
	0xef,0xbf,0xff,0xde,0xfb,0xed,0xff,0xe5,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xf4,0xdf,0x4f,0xff,0xff,
	0xff,0xff,0x4f,0xff,0xfd,0xf4,0xdb,0xff,
	0x47,0xff,0xff,0xf5,0xd7,0x5f,0x7f,0xfd,
	0xf5,0xd7,0x5f,0x3f,0xfd,0xf5,0xd3,0x5f,
	0x7f,0xbd,0xc7,0x5f,0x77,0xfd,0x75,0xd7,
	0x5f,0x7f,0xfd,0x75,0xd7,0x5e,0x77,0xbd,
	0xf5,0x7f,0xf8,0xff,0xe3,0x8f,0x3f,0xfe,
	0xfc,0xff,0xcf,0x3f,0xff,0xfe,0xf3,0xcf,
	0x3f,0xff,0xfc,0xff,0xff,0xfd,0xf8,0xf2,
	0xcf,0x3f,0xff,0xbc,0xe3,0xff,0x2f,0xff,
	0xff,0xf2,0xff,0x8d,0xff,0xf7,0xdf,0x7f,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xd7,
	0xff,0xef,0xdf,0x7f,0xee,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0x7f,0xd7,0xfe,0xff,
	0xff,0xff,0xff,0xfe,0xe7,0xfe,0x7f,0xff,
	0xfd,0xf7,0xcf,0xff,0xff,0xfc,0xf3,0xdf,
	0xff,0xff,0xfc,0xf3,0xcf,0xbf,0xff,0xff,
	0xcf,0x3f,0xff,0xfc,0xd3,0xcf,0xff,0xff,
	0xff,0xf3,0xff,0x3d,0x7f,0x7f,0xf6,0xff,
	0xff,0xff,0xff,0x7f,0xfd,0xff,0xd7,0x5f,
	0x7f,0xfd,0xf5,0xc7,0xdf,0x7f,0xfd,0xd7,
	0x5f,0x7f,0xfd,0xf5,0xd7,0x5f,0xff,0xfd,
	0xf5,0xc7,0x1f,0x7f,0xfc,0xff,0xff,0xb7,
	0xff,0xff,0xff,0xff,0xff,0xdb,0xff,0x7f,
	0xed,0xb5,0xd7,0x5f,0xbf,0xff,0xf7,0xd7,
	0x7a,0xfd,0xf5,0xd5,0x5f,0x7f,0xfd,0xf5,
	0xfb,0xaf,0x7f,0xbf,0xfd,0xf7,0xff,0xff,
	0x7f,0xfd,0xff,0xff,0xef,0xcf,0xff,0xff,
	0xfb,0xff,0xff,0xff,0xff,0xff,0xff,0xcf,
	0xff,0x7f,0xff,0x7f,0xff,0xff,0xdf,0xff,
	0x3f,0xff,0xfe,0xfb,0xff,0xff,0xff,0xff,
	0xfd,0xff,0xee,0xff,0xff,0xff,0xbf,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xbf,0xfe,0xff,0x73,0xff,0xbf,0xf7,0xff,
	0xff,0xff,0xff,0xff,0xff,0x7f,0xdf,0xff,
	0xff,0xf3,0xff,0x17,0xff,0xff,0xcf,0xdf,
	0xf7,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xfe,0xff,0xff,0xff,
	0xff,0xff,0xdf,0xff,0x3f,0xf9,0xff,0xff,
	0xff,0xff,0xff,0xff,0xf7,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xfc,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xdf,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xd7,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x7f,
	0xfe,0xff,0xff,0xff,0xeb,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xbf,
	0xfb,0xff,0xff,0xfb,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xe5,0xff,0xff,0xff,0x7f,0xff,0xef,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xfc,0xfb,0xfd,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xb1,0xff,0xff,0xff,0xff,0xfd,
	0xff,0xff,0xfe,0xff,0xef,0xff,0xff,0xff,
	0xff,0xef,0xaf,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0x3f,0xdc,0xff,0xf9,0xe7,
	0x9f,0x7f,0xfe,0xf9,0x67,0x9f,0x7f,0xfe,
	0xf9,0xe7,0x9f,0x7e,0xfe,0xe7,0x9f,0x7f,
	0xfe,0xf9,0xe7,0x9f,0x7f,0xfe,0xf9,0xe7,
	0x9f,0x7f,0xfe,0xf9,0x7f,0xea,0xff,0xdf,
	0x7b,0xff,0xfd,0xf7,0xdf,0x7f,0xf9,0xfd,
	0xf7,0xdf,0x7f,0xff,0xfd,0xf7,0x7f,0xff,
	0xfd,0xf7,0xdf,0x7f,0xff,0xfd,0xd7,0xdf,
	0x7f,0xff,0xfd,0xf7,0xdf,0xff,0x07,0xff,
	0x7f,0xf7,0xfd,0xf7,0xdf,0x7f,0xff,0xfd,
	0xf7,0xdf,0x77,0xff,0xfd,0xf7,0xdf,0xff,
	0x7d,0xf7,0xdf,0x7f,0xff,0xfd,0xf7,0xdf,
	0x7f,0xff,0xfd,0xf7,0xdf,0x7f,0xff,0x3f,
	0xbb,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xfd,0xff,0xff,0xff,0xff,0xff,0xfe,
	0xff,0x7f,0xfb,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xfb,0xff,
	0xff,0xc5,0xff,0xdf,0x7f,0xff,0xfd,0xf7,
	0xdf,0x7f,0xf3,0xdd,0xf7,0xdd,0x7f,0xff,
	0xfd,0xf7,0x7f,0xff,0xfd,0xf7,0xdf,0x7f,
	0xff,0xfd,0xb7,0xdf,0x7f,0xff,0xfd,0xf7,
	0xdf,0xff,0xdd,0xfe,0x3f,0xff,0xfc,0xf3,
	0xcf,0x3f,0xff,0xfc,0xf3,0xcf,0x2f,0xff,
	0xfc,0xf3,0xcf,0x7f,0xfc,0xf2,0xcf,0x3f,
	0xff,0xfc,0xf3,0xcf,0x3f,0xff,0xfc,0xf3,
	0xcf,0x3f,0xff,0xff,0x74,0xff,0xfd,0xf7,
	0xdf,0x7f,0xff,0xfd,0xf7,0xdb,0x6f,0xff,
	0xfd,0xf7,0xdf,0x7f,0xff,0xf7,0xdf,0x7f,
	0xff,0xfd,0xf7,0xdf,0x7f,0xff,0xfd,0xf7,
	0xdf,0x7f,0xf7,0xfd,0xff,0xa7,0xff,0x9f,
	0x7e,0xfa,0xe9,0xa7,0x9f,0x7e,0xfa,0xe9,
	0xa7,0x9f,0x7e,0xfa,0xe9,0xa7,0x7f,0xfa,
	0xe9,0xa7,0x9f,0x7e,0xfa,0xe9,0xa7,0x9f,
	0x7e,0xfa,0xe9,0xa7,0x9f,0xfe,0xbf,0x7c,
	0xff,0xfe,0xfb,0xef,0xbf,0xff,0xfe,0xfb,
	0xef,0xbf,0xff,0xfe,0xfb,0xef,0xbf,0xff,
	0xfb,0xef,0xbf,0xf7,0xfe,0xfb,0xef,0xb7,
	0xff,0xfe,0xfb,0xef,0xbf,0xfd,0xfe,0xff,
	0xe0,0xfb,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xf7,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfe,0x3f,0xff,0xbf,0xff,0xfe,0xfb,0xef,
	0xbf,0xff,0xbe,0xfb,0xee,0xbf,0xff,0xfe,
	0xfa,0xef,0xcd,0xfe,0xfb,0xef,0xbf,0xfe,
	0xfe,0xfb,0xef,0xbf,0xff,0xfe,0xfb,0xef,
	0xbf,0xff,0x3f,0xfb,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xfe,0xff,0xff,0xbf,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xc3,0xf7,0xbf,0xff,
	0xfe,0xfb,0xef,0xbf,0xff,0xfe,0xfb,0xef,
	0xbf,0xff,0xfe,0xfb,0xef,0xfe,0xfe,0xfb,
	0xef,0xbd,0xff,0xfe,0xfb,0xef,0xbf,0xff,
	0xee,0xfb,0xef,0xbf,0xff,0xaf,0x9e,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0x7f,0xff,
	0xff,0xff,0xef,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xf3,
	0xff,0xff,0xef,0xff,0xff,0xff,0xfb,0xff,
	0xff,0xff,0xff,0xfb,0xff,0xff,0xff,0x7f,
	0xf9,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0x83,0xbf,0xff,0x3f,0xff,0xff,0xff,0xcf,
	0xff,0xff,0xff,0xff,0xcf,0xff,0xff,0xff,
	0xff,0xef,0xff,0xff,0xfd,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfb,0x9d,0x7c,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xe0,0xff,0xff,0xff,0xfb,
	0xff,0xbf,0xff,0xfe,0xfb,0xef,0xbf,0xff,
	0xfe,0xfb,0xff,0xbf,0xfe,0xfb,0xef,0xbf,
	0xff,0xfe,0xfb,0xef,0xbf,0xff,0xfe,0xfb,
	0xef,0xbf,0xff,0xff,0x0f,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xbf,0xf9,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xfe,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xdb,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xdf,0xfe,0xff,0xdf,0x7f,0xff,0xfd,0xf7,
	0xdf,0x7f,0xff,0xfd,0xf7,0xdf,0x7f,0xff,
	0xfd,0xf7,0x7f,0xff,0xfd,0xf7,0xdf,0x7f,
	0xff,0xfd,0xf7,0xdf,0x7f,0xff,0xfd,0xf7,
	0x3f,0x7f,0xf1,0xfe,0xf7,0xdf,0x3f,0xff,
	0xfd,0xf7,0xdf,0x7f,0xff,0xfd,0xf7,0xdf,
	0x7f,0xff,0xfc,0xdd,0x7f,0xff,0xfd,0xf7,
	0xdf,0x7f,0xff,0xfd,0xf7,0xdf,0x7f,0xff,
	0xfd,0xf7,0xf7,0xab,0xff,0xff,0xfd,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xef,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0x9f,0xfd,0x7f,0xff,
	0xfd,0xff,0xd7,0x7f,0xff,0xfd,0xf7,0xdf,
	0x7f,0xff,0xfd,0xf7,0xf7,0xff,0xfd,0xf7,
	0xdf,0x7f,0xff,0xfd,0xf7,0xdf,0x7f,0xff,
	0xfd,0xf7,0xdf,0x7f,0x7f,0xff,0xef,0xff,
	0x7e,0xd7,0xdf,0xde,0xfd,0xf5,0xd7,0x5f,
	0x7f,0xfd,0xf5,0xd7,0x5f,0xdf,0xfb,0xdd,
	0x5f,0x7f,0xfd,0xf5,0xd7,0x5f,0x7f,0xfd,
	0xf5,0xd7,0x5f,0x7f,0xfd,0xf5,0x7d,0x6a,
	0xff,0xff
};
