//	Mona1ASIC48 (Converted by RBF2VxD)


// ----------------------------------------------------------------------------
//
//   Copyright Echo Digital Audio Corporation (c) 1998 - 2004
//   All rights reserved
//   www.echoaudio.com
//   
//   Echo Digital Audio does not disclose the source code from which these 
//   firmware images are derived. Permission is hereby granted for the 
//   distribution of these firmware images as part of the Linux kernel or 
//   other GPL project in text or binary 
//   form as required.
//   
//   This file is part of Echo Digital Audio's generic driver library.
//   
//   Echo Digital Audio's generic driver library is free software; 
//   you can redistribute it and/or modify it under the terms of 
//   the GNU General Public License as published by the Free Software Foundation.
//   
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//   
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
//   MA  02111-1307, USA.
//
// *******************************************************


//	The array is 31146 bytes.

u8 Mona1ASIC48[] =
{
	0xff,0x04,0x3c,0x2b,0xf9,0xda,0xff,0xdf,
	0xfd,0xf5,0xb7,0x77,0x7f,0xfd,0x75,0xdd,
	0xdf,0xd6,0xdd,0x6f,0xfd,0xf5,0x7b,0xef,
	0xd7,0xdf,0xde,0xdd,0xe9,0xd7,0x5f,0x7f,
	0xfd,0xed,0xdd,0xdf,0x7f,0xff,0xf3,0xff,
	0xfd,0xf7,0xdf,0xff,0xff,0xfd,0xf7,0xdf,
	0x7f,0xff,0xdf,0xf7,0xf7,0xfd,0xf7,0xff,
	0xff,0x7d,0xff,0xff,0xf7,0xff,0x7d,0xff,
	0xfd,0xf7,0xff,0x7d,0xff,0xdd,0xff,0xa7,
	0xff,0xff,0xfc,0xff,0xff,0x3f,0xff,0xff,
	0xff,0xff,0xfd,0xff,0xfe,0x7f,0xff,0xff,
	0xff,0xf3,0xff,0xff,0xff,0xfc,0xdf,0xff,
	0xff,0xff,0xff,0xff,0xdf,0xff,0xff,0xff,
	0x1f,0xfd,0xff,0xfd,0xf7,0xdf,0x7f,0xff,
	0xfd,0xf7,0xdf,0x7f,0xff,0xfd,0xf7,0xdb,
	0x1f,0xff,0xf7,0xdf,0x7f,0xff,0xfd,0xf7,
	0xdf,0x7f,0xff,0xfd,0xf7,0xdf,0x7f,0xff,
	0xfd,0xff,0xe0,0xff,0xff,0xf9,0xf7,0xdf,
	0x7d,0xfe,0xfd,0xf7,0x9f,0x7f,0xf7,0xf9,
	0x47,0x9f,0x7c,0xde,0x67,0xdf,0x7f,0xf7,
	0xf9,0x77,0xdf,0x7f,0xff,0xfd,0x77,0x9f,
	0x7f,0xff,0xfd,0x1f,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0x7f,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0x3f,0xe8,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xbf,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xd9,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xbf,0xff,0xfd,0xf7,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0x7f,0x7f,
	0xfa,0xff,0xff,0xbf,0xff,0xfe,0xff,0xef,
	0xbf,0xff,0xff,0xfb,0xff,0xff,0xff,0xff,
	0xbf,0xff,0xff,0xfe,0xfb,0xff,0xff,0xff,
	0xfe,0xfb,0xef,0xbf,0xff,0xff,0xfb,0xef,
	0x7b,0xd6,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0x8f,0xfe,0xff,0xff,0xff,0xfc,
	0xff,0xff,0xff,0xff,0xff,0xff,0xdf,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xfc,0xff,0xff,0xff,0xff,0xfd,
	0xff,0xcf,0xff,0x1f,0xfd,0xfb,0xff,0xff,
	0xf7,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0x7f,0xff,0xff,0xff,
	0xff,0xff,0xfe,0xf7,0xff,0xff,0xff,0xff,
	0xff,0xff,0x7f,0xff,0xff,0xe6,0xff,0xff,
	0xff,0xff,0xff,0xfc,0xff,0xcf,0xff,0xff,
	0xff,0xf7,0xff,0xff,0xff,0xff,0xf9,0xbf,
	0xff,0xfd,0xff,0xff,0xff,0xff,0xff,0x7f,
	0xff,0xfd,0xbf,0xf7,0xdf,0xdf,0x4f,0xff,
	0xff,0xfe,0xfb,0xef,0xbf,0xff,0x7e,0xfb,
	0xef,0xbf,0xff,0xfa,0xfb,0xef,0xbf,0xff,
	0xfb,0xef,0xbf,0xfe,0xfe,0xfb,0xef,0xbf,
	0xff,0xfe,0xfb,0xef,0xbf,0xff,0xfe,0xbd,
	0xfb,0xff,0xf7,0xdf,0x7f,0x7f,0xfd,0xf7,
	0xdd,0x7f,0xdf,0xfd,0xf6,0xdd,0x7f,0xff,
	0xfd,0xdd,0x7f,0x9f,0xfd,0xf7,0xdf,0x7f,
	0xff,0xfd,0xf7,0xdf,0x7f,0xff,0xfd,0xf7,
	0xff,0xcd,0xff,0xe7,0x9f,0x6f,0x7e,0xfd,
	0x67,0xdf,0x7f,0xf6,0xf9,0xf7,0x9e,0x7f,
	0xbe,0xf9,0x8d,0x7f,0xbf,0xfd,0xe3,0x9f,
	0x7f,0xfe,0xf9,0xe7,0x9f,0x7f,0xfe,0xf9,
	0xe7,0xff,0x9f,0xfe,0xff,0xfd,0xf7,0xdf,
	0xb7,0xff,0xfd,0xf7,0xdf,0x7f,0xff,0xfe,
	0xd7,0xef,0xbf,0x9e,0xfb,0xef,0x7f,0xf7,
	0xfe,0xf7,0xdf,0x7f,0xff,0xfd,0xf7,0xef,
	0x7f,0xff,0xfd,0x7e,0xf1,0xfb,0xf9,0xe7,
	0x9f,0x7f,0xfe,0xf9,0xe7,0x9f,0x6f,0xfe,
	0xb9,0xe7,0x9f,0x7b,0xfe,0xe7,0x9d,0x7f,
	0xfe,0xb9,0x67,0x9e,0x7f,0xbe,0xf9,0xe7,
	0x9f,0x7f,0xf6,0xf9,0xb7,0x97,0xff,0xbf,
	0xfe,0xfe,0xfb,0xbf,0xbf,0xff,0xfe,0xfb,
	0xef,0xff,0xfe,0xfe,0xef,0xbf,0xdf,0xfb,
	0xff,0xef,0xff,0xfe,0xfe,0xfb,0xef,0xbf,
	0xff,0xfe,0xef,0xef,0xbf,0xff,0xbf,0xfc,
	0x7f,0xfb,0xfd,0xf6,0xdf,0x7e,0xff,0xfd,
	0xf7,0xdf,0x7f,0xfb,0xfd,0xb7,0xdd,0x5e,
	0xec,0xf7,0xdb,0x3f,0xfb,0xfd,0xf7,0xdf,
	0x7f,0xff,0xfd,0xb5,0xdf,0x7d,0xff,0xff,
	0xeb,0xff,0xcb,0x6f,0xbf,0xfd,0x72,0xdf,
	0x6f,0xbf,0xfd,0xf6,0x8b,0x6f,0xff,0xfc,
	0xf2,0x2b,0xbf,0x7d,0xf6,0xc9,0x7d,0xbf,
	0xfd,0xf6,0xdb,0x6f,0xbf,0xbc,0xf6,0xdb,
	0xef,0x37,0xf7,0xff,0xff,0xcf,0xff,0xff,
	0xdf,0xff,0xff,0xef,0xff,0xff,0xff,0xff,
	0xed,0xff,0xbf,0xbb,0x7f,0xbf,0xff,0x6f,
	0xfe,0xff,0xff,0xff,0xff,0xff,0xff,0xfc,
	0xff,0xff,0xab,0xf8,0xfe,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xbe,0xff,0xf3,0xff,
	0xff,0xaf,0xff,0xff,0x6f,0xff,0xde,0xf7,
	0xff,0xff,0xff,0xff,0xff,0xff,0xbf,0xff,
	0xef,0xff,0xff,0x7f,0xc1,0xf7,0xff,0xff,
	0xff,0xbf,0xff,0xf3,0xff,0xff,0xef,0xff,
	0xfe,0xff,0xff,0xdf,0xff,0xff,0xfd,0xef,
	0xff,0xff,0xbf,0xff,0xff,0xff,0xff,0xff,
	0xbf,0xff,0xff,0xfb,0xff,0x1f,0xfe,0xff,
	0xbf,0xdf,0xff,0xff,0xef,0xff,0xff,0xff,
	0xfe,0x3f,0xef,0xec,0xff,0xff,0xff,0xff,
	0xff,0xbf,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xfc,0xff,0xff,0xff,0xff,0x7f,0xf4,
	0xfd,0xef,0xbf,0xef,0xfe,0xfb,0xef,0xbf,
	0xff,0xfe,0xfb,0xeb,0xaf,0xff,0xf8,0xfb,
	0xbe,0xff,0xfe,0x7b,0xef,0xbf,0xff,0xfe,
	0xfb,0xef,0x2f,0xff,0xfe,0xfb,0xef,0xf7,
	0x83,0xff,0x6f,0xff,0xff,0x7f,0xcb,0xff,
	0xff,0xff,0xf2,0xff,0x2d,0xfe,0xff,0xf6,
	0x5b,0x9f,0xff,0xff,0xff,0x2f,0xf7,0xfb,
	0x7f,0xff,0xfd,0xff,0xff,0xfe,0xff,0xfd,
	0xff,0xff,0xfd,0x7f,0xff,0xff,0xff,0xbb,
	0xff,0x7f,0xff,0xed,0xf7,0xef,0xbe,0xff,
	0xf4,0xdf,0xff,0xfe,0xfe,0xf3,0xcf,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xfb,
	0xff,0xff,0xfb,0xe0,0xf7,0xfb,0xfb,0xff,
	0xff,0xff,0xfd,0xff,0xff,0xff,0xff,0xff,
	0xff,0xbe,0xff,0xee,0xdf,0xff,0xff,0xff,
	0xf7,0xff,0xff,0xff,0xff,0xff,0xff,0x7f,
	0xff,0xff,0xff,0xdf,0x6f,0xbf,0xff,0xef,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xbf,
	0xff,0xf7,0xff,0xff,0x7f,0xf7,0xff,0xff,
	0xff,0xef,0xef,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xf7,0xff,0xff,0xfb,0xff,
	0xff,0xff,0xff,0xeb,0xbb,0xff,0xff,0xfb,
	0xff,0xff,0xff,0xff,0xff,0xfd,0xff,0xfd,
	0xff,0xed,0xff,0xff,0xfe,0xff,0xff,0xff,
	0xff,0x7f,0xff,0xff,0xff,0xff,0xff,0xd9,
	0xef,0xff,0xff,0xfe,0xff,0xf5,0xff,0xff,
	0xff,0xff,0xfe,0xdf,0xff,0x7b,0xf1,0xdf,
	0xde,0x7f,0xff,0xff,0xc7,0xff,0x7f,0xfe,
	0xff,0xff,0xff,0x7f,0x7e,0xff,0xff,0xff,
	0x7b,0xfe,0xff,0xff,0xff,0xff,0x7f,0xdf,
	0xff,0xff,0xbf,0xff,0xff,0xff,0xff,0xdf,
	0xff,0xe9,0xfb,0xde,0xf7,0x7e,0xff,0xff,
	0xff,0xff,0xff,0xf7,0xff,0xfd,0xff,0xff,
	0xff,0x7f,0xf7,0xff,0xff,0xff,0xef,0x7d,
	0xfd,0xff,0xff,0xff,0xbd,0xff,0xfd,0xff,
	0xde,0xff,0xff,0xff,0xfd,0xfb,0xff,0xfe,
	0xfd,0xff,0xff,0xff,0xff,0xff,0xff,0xbf,
	0xfd,0xf7,0xff,0x97,0xff,0xff,0xff,0xff,
	0xdf,0xe9,0x7f,0xff,0xf7,0xff,0xdd,0xff,
	0xff,0xff,0xef,0x7d,0xfe,0xff,0xff,0xff,
	0x7f,0xdf,0xff,0xef,0xff,0xff,0xff,0xff,
	0xdf,0xfb,0xef,0xdf,0x3e,0xfc,0xff,0x7f,
	0xff,0xff,0xfb,0xee,0xff,0xff,0xff,0xfb,
	0xff,0xff,0xff,0xff,0xfb,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xfd,0xff,0xff,
	0xff,0xff,0xff,0xff,0xbf,0xff,0xe1,0xff,
	0xff,0xcf,0x37,0xef,0xfc,0xff,0xcf,0xff,
	0xef,0xff,0xf3,0xff,0xff,0xff,0xff,0xf7,
	0xff,0xff,0xfc,0xff,0x4d,0xff,0xff,0xff,
	0xff,0xcf,0x37,0xff,0xff,0xf3,0xff,0x77,
	0xff,0xff,0xff,0xfe,0xfb,0xef,0xff,0xff,
	0xfe,0xff,0xff,0xbf,0xff,0xff,0xff,0x7f,
	0xff,0xff,0xff,0xef,0xfe,0xff,0xfe,0xdf,
	0xff,0xff,0xff,0xfe,0xfb,0xff,0xbf,0xfd,
	0x3f,0xf9,0xff,0xbb,0xf7,0xde,0x7f,0xff,
	0xff,0xe3,0xff,0xff,0xee,0xb8,0xef,0xfe,
	0xff,0xaf,0xff,0xfe,0x7b,0xef,0xef,0xe7,
	0xbf,0xff,0xff,0xff,0xf7,0xdf,0xeb,0xfe,
	0xf5,0xff,0xd3,0xff,0xdf,0xfb,0xff,0xff,
	0xff,0xff,0x7f,0xff,0xff,0xf1,0xff,0x1f,
	0xff,0xef,0xbf,0xff,0xfb,0xef,0xf5,0xff,
	0x5f,0x7f,0xfd,0xff,0xff,0xf7,0xf7,0xf7,
	0xf1,0xfb,0xff,0x2f,0xfe,0xbf,0x7e,0xff,
	0xff,0xfe,0xdf,0xff,0xfd,0xff,0xff,0xff,
	0xff,0xff,0x7f,0xff,0xff,0xff,0xff,0xf7,
	0xff,0xff,0x9e,0xfb,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0x7f,0xd5,0xff,0xff,
	0xff,0xf3,0xff,0xff,0xff,0xff,0xff,0xcf,
	0xff,0xef,0xff,0xff,0xff,0xff,0xfb,0xf3,
	0xe3,0xfd,0xff,0xbf,0xf3,0xe3,0xff,0xf7,
	0xdf,0xff,0xff,0xfd,0xff,0xff,0xb4,0xd5,
	0xff,0xe7,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xf9,0xff,0x5f,0xfc,0xf3,0xf9,
	0xcf,0x3f,0xff,0xff,0xff,0xff,0x3f,0xff,
	0xff,0xff,0xff,0xfe,0xf3,0xcf,0xbf,0x5a,
	0xbc,0xfe,0xfb,0xff,0xff,0xff,0xff,0xfb,
	0xff,0xbf,0xff,0xfe,0xff,0xef,0x9f,0x7f,
	0xfe,0xef,0xbf,0xff,0xff,0xfb,0xff,0xbf,
	0xff,0xfe,0xfb,0xff,0xda,0x7f,0xfe,0xfd,
	0xf5,0x69,0xff,0xb7,0xdf,0x7f,0xff,0xfd,
	0xb7,0xdf,0x7f,0xfb,0xed,0xf7,0xdf,0x3e,
	0xfb,0xec,0xdf,0x7e,0xfb,0xfd,0xb7,0xdf,
	0x7f,0xfb,0xed,0xb7,0xdf,0x37,0xfe,0xec,
	0xf3,0xff,0x07,0xff,0xeb,0xfa,0xeb,0xef,
	0x97,0xff,0xfe,0xfb,0xef,0xbf,0xff,0xfe,
	0xfb,0x27,0x9b,0xad,0xfb,0xaf,0xbf,0xff,
	0xfe,0xfb,0xaf,0xbf,0xff,0xfe,0x7b,0x6f,
	0x93,0x5e,0xb2,0xbf,0xd9,0x5e,0xff,0xff,
	0xff,0xbf,0xfa,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xef,0x9a,0x6b,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xef,0x96,0xfe,0x5a,0xb5,0x9b,0xff,0xff,
	0xff,0xff,0xff,0xf5,0xfe,0xeb,0xff,0xff,
	0xff,0xff,0xeb,0xff,0xfd,0xf7,0xfe,0xff,
	0xff,0xff,0xff,0xff,0xaf,0xff,0xff,0xff,
	0xfb,0xff,0xbd,0xf5,0xd6,0xbf,0x0d,0xfe,
	0xff,0xdb,0xef,0xff,0xaf,0xe7,0xdf,0xff,
	0xff,0xfc,0xff,0xff,0x7f,0xbb,0xef,0x67,
	0xff,0xbf,0xff,0xff,0xcf,0xff,0xbf,0xff,
	0xff,0x9f,0xff,0xeb,0xa8,0xa6,0xfe,0x7f,
	0xe0,0xdf,0x6f,0xbd,0xed,0xb6,0x5b,0xee,
	0xbd,0xff,0xb6,0xfb,0xef,0xbf,0xf7,0xe6,
	0x1b,0xbf,0xf9,0xb6,0xfb,0xef,0xbf,0xf9,
	0xfe,0xfb,0xef,0xbd,0xff,0xe6,0x7b,0xef,
	0xef,0x2f,0xff,0x77,0xc7,0xfd,0x67,0x9f,
	0x7d,0xf6,0xf9,0xe7,0x9f,0x7d,0xfe,0xf9,
	0x77,0x9e,0xfd,0xd9,0xf7,0x9f,0x7f,0xe6,
	0xd9,0xf7,0x9f,0x7f,0xf6,0xd9,0x67,0x9f,
	0x7f,0xfe,0xff,0xf9,0xff,0xfd,0xf6,0x5b,
	0x67,0xdd,0x95,0xd6,0x5b,0x7f,0xb5,0xf5,
	0xd7,0xdf,0x6b,0xa5,0xd6,0xd9,0x6f,0xad,
	0xf5,0xd7,0xd8,0x6b,0xfd,0xd5,0xd7,0x59,
	0x67,0xbf,0xf5,0x7e,0x65,0xff,0xff,0xf3,
	0xcf,0xef,0xff,0xfd,0xf7,0xdb,0xff,0xff,
	0xfe,0xf7,0xdf,0xff,0xff,0xfd,0xdb,0xef,
	0xff,0xfd,0xf7,0xdf,0x7f,0xff,0xfd,0xff,
	0xfb,0x6f,0xbf,0xfd,0xf6,0x4e,0xff,0xff,
	0xf7,0xdf,0xff,0xff,0xfd,0xef,0xbf,0xff,
	0xff,0xff,0xdf,0x7f,0xff,0xff,0xfb,0xbf,
	0xff,0xff,0xfd,0xf7,0xdf,0x7f,0xff,0xfb,
	0xff,0xff,0xff,0xfe,0xfb,0xf7,0x7f,0xf8,
	0xff,0xe3,0x8b,0x2f,0xbe,0xd8,0xe3,0xcf,
	0x2f,0xbe,0xf8,0xe2,0x8f,0x2f,0xbe,0xd8,
	0x89,0x3f,0xfe,0xf8,0xe2,0xcf,0x2f,0xbe,
	0xf8,0x62,0xcb,0x2f,0xbe,0xd8,0xf2,0xff,
	0xdd,0xff,0x77,0x5f,0x7d,0xf5,0xd5,0x77,
	0x7f,0x7d,0xf5,0xd5,0x77,0x5f,0x7d,0xf5,
	0xd5,0x5f,0x7d,0xf7,0xd5,0x57,0x5f,0x7d,
	0xf7,0xd5,0x57,0x5f,0x7d,0xb5,0x5d,0x57,
	0xff,0xcf,0xde,0xff,0xea,0xbe,0xfb,0xbf,
	0xfe,0xda,0xfb,0xbf,0xef,0xbf,0xfe,0xeb,
	0xff,0xbe,0xfe,0xfb,0xef,0xbf,0xff,0xfb,
	0xff,0xeb,0xaf,0xff,0xfb,0x7a,0xaf,0xaf,
	0xba,0x7e,0xff,0xf3,0xff,0xf7,0x99,0x67,
	0x1e,0x7f,0xfc,0xd9,0xe7,0x1f,0x7b,0xec,
	0xd1,0xe7,0x1e,0x7f,0xb2,0xc7,0x1e,0x79,
	0xfe,0xf1,0xc7,0x1e,0x79,0xfe,0xd1,0xc7,
	0x1f,0x7f,0xf4,0xf9,0xb5,0xff,0xbf,0xff,
	0xfe,0xfb,0xfb,0xaf,0xff,0xfe,0xfb,0xef,
	0xbf,0xff,0xff,0xfb,0xff,0xbf,0xff,0xfa,
	0xff,0xbf,0xff,0xfe,0xfb,0xef,0xbf,0xbf,
	0xfe,0xfe,0xef,0xbf,0xff,0x3f,0xfc,0x7f,
	0xff,0xfd,0xf7,0x7f,0x7f,0xfd,0xfd,0xf7,
	0x7f,0xff,0xff,0xfd,0xf7,0xff,0xff,0xf7,
	0xd7,0x7f,0x7f,0xff,0xf5,0xff,0xdf,0x7f,
	0xff,0xf5,0xdf,0xff,0x7f,0xff,0xff,0xee,
	0xff,0xaf,0xff,0xfe,0xfb,0xef,0xbf,0xbf,
	0xfe,0xfa,0xef,0xaf,0xff,0xfe,0xfa,0xeb,
	0xf7,0xfe,0xfb,0xef,0xaf,0xff,0xfe,0xfa,
	0xeb,0xaf,0xff,0xfe,0xfb,0xeb,0xaf,0xff,
	0x27,0xff,0x7f,0xff,0xff,0xff,0xff,0xff,
	0x7f,0xff,0xfd,0xff,0xff,0x7f,0xff,0xfd,
	0xf7,0xff,0xff,0xff,0xdf,0xdf,0x7f,0xff,
	0xff,0xf7,0xdf,0xff,0xff,0xff,0xff,0xdf,
	0xff,0x3f,0xfa,0xff,0xff,0xbf,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xbf,0xff,0xff,0xff,0xef,
	0xff,0xff,0xff,0xff,0xff,0xff,0xef,0xff,
	0xff,0xef,0xff,0xdc,0xff,0x7f,0xfd,0xf5,
	0xd7,0x5f,0x7f,0xf9,0xc5,0x97,0x5f,0x7b,
	0xf9,0xe5,0x57,0x5f,0xee,0xf5,0x97,0x5f,
	0x7f,0xfd,0xc5,0x97,0x5f,0x7e,0xf9,0xf5,
	0x56,0x5f,0x7f,0xfd,0x9f,0xfe,0xfb,0xbc,
	0xff,0xbf,0xf7,0xff,0xfc,0xfb,0xce,0x3f,
	0xff,0xfb,0xf3,0xcf,0x3e,0xfa,0xd3,0xcf,
	0xff,0xff,0xf4,0xe3,0xcf,0x3f,0xff,0xf8,
	0xf3,0xff,0x3f,0xff,0xfc,0x7f,0xe3,0x7f,
	0xff,0xff,0xed,0xff,0x7f,0xff,0xde,0xff,
	0xdb,0xdf,0xff,0xfd,0xef,0xb6,0xfb,0xff,
	0xff,0xd7,0xff,0xff,0xf3,0xef,0xe7,0x9f,
	0xff,0xef,0xff,0xff,0x7f,0xf3,0xff,0xb5,
	0xfd,0x4f,0xe3,0xff,0xff,0xff,0xc6,0xdf,
	0x7f,0xec,0xf3,0xff,0x3f,0xf5,0xdc,0xf3,
	0x26,0x7f,0xfc,0x7d,0xc7,0xfd,0xfe,0xfc,
	0xf3,0xcf,0xfd,0xfe,0xff,0xf3,0xff,0xff,
	0xdf,0xf8,0xff,0xbf,0xe5,0xbf,0x5f,0x7f,
	0xfc,0xf5,0xc7,0x5f,0x6f,0xfd,0xf1,0xc7,
	0x5f,0xef,0xf7,0xd7,0x7f,0x7f,0xfc,0xff,
	0xf7,0x1f,0x7f,0xfc,0xf1,0xd7,0x5f,0x3f,
	0xfc,0x5f,0xe8,0xff,0xfe,0x6f,0xde,0x7f,
	0xf7,0xff,0x9f,0xfb,0xa7,0xbb,0xdf,0xdf,
	0xff,0x6f,0xb6,0xdb,0xe7,0xfd,0xff,0xff,
	0xff,0xff,0xfa,0xfe,0xff,0xef,0x7b,0xfe,
	0x77,0xfb,0xff,0x77,0xfe,0xff,0xd7,0xf7,
	0xff,0xf6,0xdb,0xff,0xff,0xfd,0xf7,0xdf,
	0xff,0xff,0xfd,0xff,0xdf,0xdf,0xed,0xd3,
	0xdd,0xf7,0xfc,0xbf,0xff,0x5f,0xfd,0x7d,
	0x7f,0xd7,0xfd,0xfb,0x9f,0xfa,0xff,0xff,
	0xff,0xff,0xff,0xf7,0xf7,0xff,0xfd,0xfd,
	0xf7,0xf7,0x7d,0xff,0xfd,0xfb,0xde,0xff,
	0xbf,0xfd,0xf7,0xda,0xff,0xfd,0xfd,0xf6,
	0x7b,0x7f,0xaf,0xfd,0xfa,0xff,0xdd,0xf7,
	0xff,0xff,0xff,0xbf,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xb7,0xdf,0xef,
	0xff,0xff,0xff,0xff,0xdf,0xfd,0xff,0xff,
	0xfb,0xff,0xff,0xff,0xff,0xff,0xfb,0x7f,
	0xbe,0xef,0xff,0xbc,0xf3,0xde,0xef,0x7f,
	0xfc,0xfd,0xcb,0xdf,0x3f,0xfc,0xff,0xdd,
	0xbb,0xfd,0xff,0xcf,0x3f,0xfe,0xff,0xf3,
	0xff,0x3f,0xff,0xfc,0xe3,0xcf,0xdf,0xff,
	0xfe,0xf4,0xfd,0xff,0xd7,0xbf,0xdf,0x7b,
	0xff,0xef,0xff,0x77,0xfe,0xff,0xef,0xef,
	0xff,0xfb,0xbd,0xff,0xff,0xbe,0xfb,0xff,
	0x9f,0xff,0xff,0xfb,0xef,0xbf,0xff,0xfe,
	0xff,0xff,0xb3,0xff,0xbf,0xfd,0xff,0xfd,
	0xcb,0xff,0x7f,0xff,0xf7,0xef,0xbf,0x7b,
	0xfe,0xdf,0xff,0xdd,0x7f,0xbf,0xe7,0xff,
	0xf7,0xff,0xfb,0xef,0xdf,0x7f,0xff,0xff,
	0xc7,0xff,0x7f,0xff,0xbd,0xff,0xfd,0xff,
	0xff,0x5f,0xff,0xff,0xff,0xff,0xfd,0xfe,
	0xfd,0xef,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfe,0x7b,0xf7,0xcf,0x0f,0xfe,0xff,0xff,
	0xff,0x5f,0xff,0xf7,0xfd,0xe8,0xfb,0xeb,
	0xff,0xff,0xff,0xef,0xff,0xff,0xff,0x6b,
	0xfe,0xeb,0xee,0xff,0x7f,0xff,0xff,0xff,
	0xff,0xfe,0xff,0xdf,0x7f,0xff,0xff,0xff,
	0xff,0xff,0xff,0xfe,0xff,0xde,0x1f,0xfd,
	0xff,0xff,0xb7,0xfe,0x7f,0xff,0xcf,0x3f,
	0xdf,0xff,0xff,0xcf,0xfd,0xff,0x7f,0xbf,
	0xf7,0xff,0xff,0xf3,0xad,0xff,0xf7,0x3f,
	0x5f,0xff,0xad,0xf6,0xfc,0x93,0xfd,0xcf,
	0x53,0xfd,0xff,0xff,0xf5,0xff,0x9b,0x3f,
	0xff,0xcc,0xff,0xff,0x7f,0xfd,0xff,0x3f,
	0xf3,0xbf,0xff,0xff,0x9f,0x77,0x7e,0xff,
	0xff,0xfb,0xfe,0xff,0xfc,0xf1,0x5f,0xfe,
	0xaf,0xc7,0xeb,0xbf,0xff,0xfd,0xff,0xff,
	0xbf,0xff,0xff,0xf5,0xef,0xbf,0xff,0xff,
	0xff,0xf7,0xff,0xff,0xfb,0xf5,0xff,0xff,
	0xfe,0xfb,0xef,0xf7,0x59,0x67,0xfd,0xf5,
	0x74,0x7f,0x2f,0xf6,0x7f,0xfb,0xff,0xf7,
	0xff,0x7f,0xfb,0xfd,0xfb,0xdf,0x7e,0xfb,
	0xfd,0xff,0xef,0xff,0xff,0xb7,0x8d,0xff,
	0xff,0xed,0xb7,0xdf,0x3e,0xff,0x78,0xea,
	0x8c,0xb7,0xff,0x6f,0xe2,0xdf,0xae,0xff,
	0xfe,0xfe,0xef,0xef,0xbf,0x7f,0xb3,0xfb,
	0xef,0xbf,0xfe,0xfb,0xcd,0xfa,0xfe,0x5e,
	0xe9,0xbf,0x97,0xfe,0xfe,0xfb,0xa5,0x97,
	0xfe,0x7b,0x69,0xbf,0xff,0xbf,0xed,0xf6,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x6e,
	0xfd,0xff,0xff,0xff,0xff,0xbb,0xf7,0xff,
	0xff,0xeb,0xff,0xbf,0xfa,0xff,0xff,0xaf,
	0xbd,0xf6,0xef,0xeb,0xff,0xfe,0x5b,0xf9,
	0xff,0xff,0x5f,0xfe,0xff,0xef,0xff,0xff,
	0x5f,0xff,0xff,0xff,0xff,0x7f,0x7b,0xdd,
	0xbf,0xff,0x5f,0xfb,0xe7,0xf5,0xff,0xff,
	0x7f,0xbd,0xf4,0x57,0x59,0x7f,0xb5,0x7e,
	0xe9,0xff,0xbf,0xfc,0xfd,0xff,0x0f,0xfe,
	0xff,0xbf,0x57,0xff,0xff,0xff,0xfe,0x17,
	0xde,0xf7,0xfc,0xff,0xea,0x7f,0xaf,0xfe,
	0xff,0xff,0xeb,0xad,0xbf,0xf6,0xea,0x6b,
	0xfd,0x47,0xfe,0xff,0xf6,0xfb,0xef,0xbf,
	0xff,0xf6,0xfb,0x6f,0xbf,0xff,0xf6,0xfb,
	0xef,0xbf,0xff,0xfb,0x6f,0xbb,0xff,0xfe,
	0xfb,0xef,0xbf,0xed,0x96,0xdb,0xef,0xbf,
	0xff,0xdf,0xff,0xfb,0xff,0x77,0x9c,0x7f,
	0xf6,0xf9,0x67,0x9d,0x7f,0xd6,0xf9,0xe7,
	0xdd,0x7f,0xfe,0xf9,0x9f,0x7f,0xfe,0xf9,
	0xe7,0xdf,0x7f,0xfe,0xd9,0x67,0x9f,0x77,
	0xfe,0xf9,0xff,0xff,0x97,0xff,0xdf,0x7f,
	0xac,0xd5,0xe7,0x5b,0x67,0xfd,0x71,0xd6,
	0x5b,0x6f,0xff,0xf1,0xc7,0x6f,0xbe,0xb5,
	0xf6,0x1f,0x7f,0xff,0xf5,0xd7,0x5d,0x67,
	0xfd,0xf1,0xf6,0xdf,0xff,0x77,0xf6,0xff,
	0xff,0xff,0xfd,0xff,0xfb,0x3f,0xff,0xff,
	0xfe,0xcf,0x3f,0xff,0xff,0xff,0xff,0xff,
	0xfe,0xf7,0xff,0x7f,0xff,0xff,0xff,0xdf,
	0x3f,0xff,0xff,0xfb,0xff,0xff,0x6f,0xf0,
	0xff,0xff,0xff,0xfd,0xff,0xf7,0x7f,0xff,
	0xff,0xff,0xdf,0x7f,0xff,0xff,0xff,0xff,
	0x7f,0xff,0xf7,0xff,0x7f,0xff,0xff,0xff,
	0x7f,0x7f,0xff,0xff,0xff,0xff,0xff,0xff,
	0x93,0xff,0x2f,0xf6,0xd1,0xe3,0x0f,0x2f,
	0xfe,0xf8,0x42,0x8f,0x2f,0xbe,0xf8,0xd7,
	0x1f,0xdd,0xf1,0xe3,0xcb,0x7f,0xfd,0xd8,
	0xe3,0x8b,0x2f,0xbe,0xfc,0xc7,0x8f,0xff,
	0xff,0x97,0xfd,0x7f,0xf5,0xdf,0x77,0x7f,
	0x7c,0xf5,0xd7,0x12,0x5f,0xfd,0xf5,0xd7,
	0x7f,0xff,0xfd,0xdf,0x77,0xdf,0xf5,0xf7,
	0xdd,0x57,0x5b,0x7d,0xf5,0xd5,0x7f,0x5f,
	0xf5,0xf3,0xff,0xc2,0xff,0xaf,0xfe,0xba,
	0xfe,0xb3,0xfb,0xbf,0xff,0xfb,0xf3,0xcf,
	0xaf,0xfb,0xff,0xb7,0x77,0xff,0xdc,0x7a,
	0xff,0xbf,0xff,0xfc,0xb3,0xbf,0xec,0xf7,
	0xef,0xfb,0xfd,0xff,0x57,0xff,0x7f,0x9f,
	0x7f,0xe4,0xd9,0x67,0x9e,0x79,0xfe,0xd1,
	0xe7,0x9f,0x79,0xfe,0xf9,0xe7,0x7f,0x7c,
	0xf1,0xe7,0x1f,0x7b,0xf4,0xd9,0xe7,0x9f,
	0x7d,0xee,0x91,0xe7,0x9f,0xff,0xfa,0xff,
	0xff,0xdf,0xbf,0xff,0xf8,0xfb,0xef,0x2f,
	0xff,0xff,0xfa,0xef,0xff,0xff,0xff,0xff,
	0xaf,0xff,0xfe,0xff,0xef,0xbf,0xbf,0xff,
	0xfb,0xeb,0x7f,0xff,0xff,0xe2,0xff,0xd7,
	0xff,0xff,0xff,0x7a,0xff,0xb7,0xf7,0xdf,
	0x7f,0xe9,0xf7,0xdf,0xff,0xff,0xef,0xbf,
	0xff,0x7b,0x7d,0xff,0xff,0x7e,0x7f,0xfd,
	0xfd,0xf7,0x5f,0xff,0xeb,0xfd,0xd7,0xfe,
	0x4f,0xfe,0xff,0xfa,0xff,0xbf,0xbf,0xff,
	0xfb,0xeb,0xff,0xff,0xfe,0xfa,0xeb,0xff,
	0xff,0x7f,0xff,0xbf,0xbe,0xfe,0xff,0xeb,
	0xaf,0xff,0xfe,0xfa,0xee,0xff,0xff,0xbe,
	0xfe,0xff,0xb1,0xff,0xf7,0xff,0xff,0xff,
	0xfd,0xff,0xf7,0xff,0xff,0xfd,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xf7,0xff,0xff,
	0xdf,0x5f,0xff,0xff,0xfd,0xdf,0xff,0x7f,
	0xff,0xf5,0xff,0xbf,0xff,0xff,0xff,0xff,
	0xef,0xff,0xff,0xff,0xff,0xef,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xef,0xff,0x3f,0xff,0xff,
	0xff,0xff,0x6f,0xef,0xbf,0xf4,0xff,0x97,
	0x5f,0x7e,0xfd,0xf5,0xd7,0x5f,0x7d,0xfd,
	0xd5,0xd7,0x5f,0x7d,0xfc,0xf1,0x1e,0x7d,
	0xed,0xc5,0x87,0x5f,0x7f,0xf5,0xd5,0x57,
	0x5f,0x7f,0xed,0xf5,0xd5,0x9f,0xef,0xff,
	0xcf,0xfb,0xff,0xdf,0x7b,0xcd,0xff,0xff,
	0xbf,0xf3,0x8f,0xff,0xff,0xf2,0xf3,0xff,
	0xfc,0xf8,0xf3,0xef,0x3f,0xff,0xfc,0xf3,
	0xbf,0x3f,0xff,0xff,0xff,0xff,0xff,0x5e,
	0xfe,0xff,0xef,0x6f,0xff,0xde,0xff,0xff,
	0xff,0xff,0xfd,0xfd,0xff,0xff,0xb7,0xd7,
	0xf3,0xef,0xff,0xd5,0x7f,0xff,0x7f,0xbf,
	0xff,0xef,0xff,0xbc,0xff,0xff,0xff,0xbf,
	0x9f,0xf9,0xff,0xfc,0xff,0xff,0xff,0xff,
	0xfc,0xff,0xff,0x3f,0xff,0xfc,0x7f,0xff,
	0xff,0x7f,0xfc,0xc7,0xff,0xff,0xff,0xf3,
	0xcf,0x3f,0xff,0xff,0xff,0xff,0xf7,0x5f,
	0xff,0xff,0xc1,0xff,0xd7,0xfb,0x6f,0xff,
	0xff,0xd7,0x5f,0xfb,0xfb,0xfd,0xfe,0x9f,
	0xff,0xff,0xff,0xfe,0x7f,0xfd,0xf5,0x7f,
	0x5f,0xdb,0xff,0xf9,0xc7,0xff,0xff,0xff,
	0xf4,0xd7,0xff,0x5f,0xfe,0xef,0xfe,0xaf,
	0xde,0xfa,0xfb,0xcd,0x79,0xff,0x7f,0xff,
	0xbf,0xff,0xff,0xff,0xfb,0xff,0x9a,0x6f,
	0xfb,0x7f,0xb7,0x7f,0xfa,0xff,0xff,0xff,
	0xff,0xbf,0xff,0xef,0xff,0xf5,0xff,0xfd,
	0xff,0xff,0xff,0xff,0xfd,0xfb,0xf5,0xff,
	0xdf,0xbf,0xdf,0xff,0xd7,0xff,0xfd,0xfd,
	0xbf,0xff,0xbf,0xf7,0x7f,0xff,0xff,0x75,
	0xdf,0x7f,0x7f,0xdf,0xff,0xff,0x97,0xff,
	0xff,0xff,0xd7,0x9f,0xff,0xff,0xff,0x7d,
	0xe7,0xdf,0xff,0xbf,0xdd,0xff,0xff,0xef,
	0xfd,0xf7,0xdf,0x7f,0xff,0xfb,0xfe,0xff,
	0xef,0xbf,0xbf,0xfe,0xff,0xfb,0xff,0x3f,
	0xfc,0xff,0xe7,0xff,0xff,0xfb,0xfd,0xff,
	0xfe,0xff,0xff,0xff,0xff,0xff,0xff,0xfa,
	0xff,0xfe,0xfb,0xf7,0xff,0xbf,0xff,0xff,
	0xef,0xbf,0xff,0xff,0xff,0xef,0xfd,0xf7,
	0xfb,0xe1,0xff,0xff,0xd7,0xdf,0x7f,0xed,
	0xf3,0x8f,0xff,0xff,0xfc,0xff,0xfd,0x7f,
	0xff,0xff,0xf9,0x3f,0xfe,0xff,0x73,0xde,
	0xdf,0xff,0xfd,0xf3,0x7f,0xff,0xfd,0xf8,
	0xf3,0xdf,0x3f,0xff,0xff,0xff,0xfe,0xff,
	0xbf,0x7f,0xff,0xfb,0xff,0xe7,0xff,0xff,
	0x7f,0xef,0xff,0xdd,0xff,0xfb,0xfe,0x9f,
	0xff,0xfb,0xff,0xbf,0x9f,0xff,0xff,0xfe,
	0xef,0x7f,0xff,0xbe,0xfb,0xff,0x3f,0xff,
	0xfd,0xf7,0xf9,0xf5,0xff,0xbf,0xfb,0xef,
	0xbf,0xff,0x2b,0xff,0xfd,0xeb,0xff,0xf7,
	0xfe,0x3b,0xdf,0xfe,0xfb,0xfd,0xf7,0x6d,
	0x7f,0xff,0xff,0xfd,0xff,0x83,0xff,0xff,
	0xf7,0xf7,0xdf,0xff,0xff,0xfb,0xff,0xfa,
	0xf7,0xfe,0xff,0xff,0xff,0xd5,0xff,0xff,
	0xff,0xb1,0x87,0xff,0xff,0xff,0xff,0xfb,
	0x5e,0xff,0xff,0xfe,0xff,0xff,0x77,0xfe,
	0xff,0xdf,0xef,0xff,0xff,0xfe,0xfb,0xef,
	0xff,0xbf,0xff,0xff,0xff,0xff,0xfd,0x78,
	0xef,0x7f,0xff,0xf7,0x7f,0x7f,0xff,0xff,
	0xf7,0xff,0xfe,0xff,0xff,0xfe,0xff,0x7f,
	0xc2,0x9f,0x9f,0xff,0xeb,0xa9,0xf7,0xff,
	0xff,0xeb,0xff,0xff,0xff,0xdf,0xff,0x2b,
	0xfd,0xfb,0xff,0xfd,0xff,0xff,0x5f,0x7f,
	0xff,0xfd,0xff,0xd7,0xfe,0xcf,0x3f,0xd7,
	0xff,0xbc,0xd5,0xf9,0xff,0x9f,0x7f,0xde,
	0xff,0xff,0x5f,0xff,0xff,0xff,0xff,0xff,
	0x7f,0xf9,0xff,0xff,0xf7,0xff,0xff,0xff,
	0x9b,0xfc,0xff,0xcf,0xbf,0xff,0x7f,0xfd,
	0xf5,0xbf,0xfa,0xbd,0xfe,0xf7,0xef,0x7f,
	0xff,0xff,0xfb,0xef,0xbf,0xff,0xfe,0xfb,
	0xe5,0xff,0xdf,0xfe,0xef,0xff,0xff,0x4e,
	0xf9,0xff,0x9f,0xff,0xff,0xf9,0xed,0x92,
	0xff,0xff,0xf7,0xf5,0x6b,0xff,0xff,0xdf,
	0xfe,0xff,0xff,0xb7,0xdf,0x7e,0xfb,0xed,
	0xb7,0x89,0xfe,0xff,0xec,0xdf,0xfe,0xff,
	0xed,0xa3,0xff,0x3f,0xfb,0xff,0xb3,0xcf,
	0x3e,0xfa,0xff,0xff,0xff,0x07,0xff,0xff,
	0xfb,0xeb,0xbf,0xff,0xfe,0xfa,0xeb,0xaf,
	0xbf,0xff,0x7e,0xe9,0xbf,0xb7,0xbe,0xfb,
	0xbf,0xbf,0xff,0xfe,0xef,0xad,0xff,0x5e,
	0x2a,0xe9,0xa5,0xff,0xfe,0xfb,0x3f,0xdb,
	0xfe,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xeb,0xff,0xbf,0xfb,
	0xff,0xff,0xff,0xff,0xef,0xff,0xef,0xd5,
	0xff,0xba,0xe9,0xaf,0xff,0xff,0xff,0xb7,
	0x85,0xe7,0xff,0xeb,0xff,0xb7,0xdf,0xff,
	0xff,0xff,0xbf,0xfe,0xff,0x5b,0xef,0xb7,
	0xfc,0xff,0xff,0xbf,0xfe,0xd2,0xff,0x6b,
	0xff,0xff,0xd7,0x5f,0x6f,0xbd,0xff,0xfe,
	0xff,0x9d,0x7e,0xff,0x53,0xef,0x7f,0xf1,
	0xfd,0xfb,0xef,0xbf,0xfd,0xff,0xbf,0x2a,
	0x7e,0xec,0x7e,0x7f,0x3f,0xfd,0xb7,0xfe,
	0x4f,0xfa,0xff,0xbc,0xfa,0x6a,0xa8,0xf8,
	0xc4,0xc3,0xff,0xf1,0xff,0xef,0xbf,0xed,
	0xfe,0xfb,0x6f,0xbe,0xff,0xd6,0xfb,0xef,
	0xbf,0xff,0xfe,0xfb,0xbf,0xe5,0xde,0x7b,
	0x6f,0xbb,0xf7,0xfe,0xfb,0x6f,0xbb,0xff,
	0xf6,0xfb,0xef,0xff,0x93,0xff,0x7f,0xfe,
	0xdd,0xe7,0x9f,0x7f,0xe7,0x9d,0x77,0x9c,
	0x7f,0xfe,0xf9,0xe7,0xdf,0xff,0x99,0xe7,
	0x9f,0x7f,0xf6,0xf9,0x67,0x1e,0x7f,0xfe,
	0xf9,0xe7,0x9d,0x7f,0xfe,0xbf,0xf9,0xff,
	0xf9,0xf6,0x1d,0x7f,0xfc,0xbd,0xf6,0xdb,
	0x6f,0xfd,0xf5,0xf6,0x1f,0x7f,0xbf,0xd6,
	0x9f,0x6f,0xfd,0x75,0xe7,0x5b,0x7f,0xfe,
	0xfd,0xf6,0xdb,0x7f,0xfe,0xf9,0x7e,0x62,
	0xff,0xff,0xf3,0xff,0xff,0xff,0xfd,0xf7,
	0xef,0xef,0xff,0xfd,0xf6,0xff,0x7f,0xff,
	0xfc,0xff,0x6f,0xff,0xff,0xff,0xef,0xff,
	0xff,0xff,0xfb,0xcf,0xff,0xff,0xff,0xfb,
	0x5e,0xff,0xff,0xf5,0xff,0xff,0xff,0xfd,
	0xef,0xff,0xff,0xff,0xfb,0xef,0xff,0x7f,
	0xff,0xfd,0xff,0xdf,0xfd,0xff,0xff,0xf7,
	0xff,0x7f,0xff,0xff,0x7f,0xff,0x7f,0xff,
	0xfd,0xff,0xe9,0xff,0xc7,0x8b,0x7f,0xf4,
	0xf1,0xe2,0x8f,0x3f,0xfe,0xd8,0xe2,0x8b,
	0x37,0xfc,0xf8,0x89,0x7f,0xbc,0xf8,0x62,
	0x1f,0x2f,0xf6,0xf1,0xe2,0x8b,0x3f,0xfe,
	0xf1,0xc7,0x7f,0xcf,0xff,0x7e,0x5f,0xed,
	0xf7,0xdf,0x57,0xdf,0xfd,0xf5,0xdd,0x57,
	0x5b,0xf5,0xf1,0xd5,0x7f,0xfd,0xb7,0xdd,
	0x57,0xff,0x7d,0xf5,0xdf,0x57,0xdd,0x65,
	0xd7,0xdf,0x7f,0xff,0x3f,0xfe,0xff,0xfd,
	0xb3,0xbf,0xff,0xff,0xfc,0xb2,0xaf,0x3d,
	0xff,0xfb,0x2a,0xcf,0xbf,0x7f,0xeb,0xbf,
	0xae,0xb6,0xfd,0xf7,0xfb,0x7f,0xf7,0xfd,
	0xea,0xaf,0xfd,0xff,0xfb,0x7f,0xf7,0xff,
	0xff,0xf1,0xe7,0x9f,0x7b,0xfc,0xd1,0xc7,
	0x1f,0x7f,0xfe,0xd1,0x67,0x1f,0x79,0xfe,
	0xe7,0x1f,0x7f,0xfc,0xf9,0x67,0x9e,0x7f,
	0xfe,0xf1,0xc7,0x9f,0x7b,0xfe,0xf9,0x8b,
	0xff,0x7f,0xbe,0xfd,0xf7,0xdf,0xef,0xff,
	0xfe,0xfb,0xef,0xbf,0xff,0xfe,0xfb,0xff,
	0xbf,0xfe,0xf7,0xef,0xbf,0xff,0xf9,0xfb,
	0xdf,0xaf,0xff,0xfe,0xfa,0xdf,0x7f,0xff,
	0x3f,0xfc,0xff,0xef,0xf7,0xbf,0xfe,0xfb,
	0xfd,0xfd,0xdf,0x7f,0x7f,0xff,0xfd,0xdf,
	0xde,0xff,0xf7,0xff,0xfe,0xff,0xfd,0xbf,
	0xf7,0xff,0x7b,0xfd,0xff,0xd7,0xff,0xfa,
	0xeb,0xff,0xe4,0xff,0xff,0x7f,0xfe,0xff,
	0xf7,0xbf,0xff,0xfe,0xfa,0xeb,0xaf,0xbf,
	0xfe,0xfe,0xef,0xb7,0xfe,0xfd,0xeb,0xbf,
	0xff,0xff,0xfb,0xf7,0xab,0xbf,0xfe,0xfb,
	0xff,0xff,0xff,0x1b,0xff,0xff,0xff,0xfb,
	0xff,0xff,0xff,0xff,0xff,0xed,0xdf,0xdf,
	0x7f,0xff,0xf7,0xf7,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0x5f,0xff,0xff,
	0xff,0xff,0xff,0xff,0x7b,0xba,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xf7,
	0xff,0xff,0xff,0xff,0x3f,0xfd,0xbf,0xff,
	0xff,0xff,0xff,0xff,0x4f,0xfd,0xff,0xef,
	0xbb,0xff,0xff,0xfb,0xff,0xff,0x4b,0xff,
	0x7d,0xfd,0xd5,0xd7,0x5e,0x7f,0xec,0xe5,
	0x97,0x5f,0x7b,0xfd,0xb5,0xd7,0x5f,0xed,
	0xf5,0xd7,0x5e,0x7f,0xf9,0x65,0x97,0x5f,
	0x7e,0xcd,0xe5,0x17,0x5f,0x7e,0xf5,0x1f,
	0xfe,0xff,0x9f,0xf3,0xff,0x3f,0xfe,0xbc,
	0xf3,0xcf,0x3f,0xff,0xfc,0xf3,0xff,0x14,
	0xff,0xf3,0xcf,0x3f,0xff,0xfc,0xfb,0x8f,
	0x3f,0xdf,0xdc,0xfb,0xcf,0xff,0xff,0xdf,
	0xff,0xe0,0xff,0xfe,0xff,0xee,0x7f,0xdd,
	0xff,0xff,0xfb,0xbb,0x7f,0xff,0xfb,0xff,
	0xff,0xff,0xff,0x7f,0xff,0x7f,0xff,0xff,
	0xfb,0xff,0xbd,0xff,0xfd,0xf7,0xff,0xff,
	0xff,0xff,0xa1,0xff,0xff,0x3f,0xdf,0xff,
	0xff,0xcf,0x3f,0xff,0xfc,0xf3,0xcf,0x3f,
	0xff,0xff,0xf3,0x2f,0xff,0xde,0xf3,0xcf,
	0xff,0x7e,0xfc,0xf9,0xcf,0x7f,0xff,0xfc,
	0xff,0xf7,0xcf,0x1e,0xfd,0xff,0xff,0xfd,
	0xff,0xfb,0x7f,0xff,0xed,0xff,0x7f,0x7f,
	0xfc,0xf5,0xfe,0x1f,0xef,0xff,0xff,0x5f,
	0x7f,0xfd,0xff,0xc6,0xff,0x3f,0xfd,0xf5,
	0xc7,0xff,0xff,0xff,0x7b,0xed,0xff,0xff,
	0x7f,0xff,0xff,0xef,0xdf,0xbf,0xff,0xff,
	0x9e,0xf7,0x5f,0xef,0xff,0xfe,0xff,0xef,
	0xbf,0xfb,0xef,0xfe,0xbf,0xff,0xff,0xe7,
	0x6f,0xff,0x7f,0xff,0xfe,0xdf,0x57,0xff,
	0xff,0xff,0xff,0x7f,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xfe,0xfd,0x7f,0xdf,0xdd,
	0xdf,0x7f,0xb3,0xdf,0xff,0xd5,0xe7,0x9f,
	0xff,0xbf,0xd9,0xfd,0xbf,0x7f,0xff,0x1f,
	0xf9,0x77,0xff,0x7f,0xff,0xfd,0xff,0xf7,
	0xff,0x7f,0xff,0xff,0xff,0xfb,0x7f,0xbf,
	0xff,0xde,0x77,0xff,0xff,0xdf,0xfa,0xef,
	0xbf,0xff,0xdf,0xfb,0xff,0xff,0xff,0xee,
	0xf7,0xc1,0xf7,0xff,0xff,0xbf,0xf7,0xff,
	0xff,0xff,0xff,0xff,0xdf,0xfb,0xfd,0xff,
	0xb7,0xff,0xfc,0xff,0xff,0xff,0x7f,0xfa,
	0xff,0xff,0xbf,0xfd,0xff,0xef,0xff,0xfe,
	0xff,0xfd,0x5f,0xbe,0x2f,0xef,0xfd,0xd7,
	0x8f,0xdf,0x7f,0xfc,0xfd,0x87,0xff,0xf7,
	0xff,0xfd,0xd7,0xbb,0xfc,0xf2,0xf7,0xbf,
	0xef,0xbf,0xff,0xf7,0x5f,0x7f,0xff,0xfb,
	0x8f,0xff,0xff,0xff,0xf5,0xfd,0xf5,0xbf,
	0xbd,0xff,0xfd,0xff,0xd7,0xff,0xff,0xfd,
	0xff,0xfd,0xff,0xff,0xfb,0xdd,0x7f,0xff,
	0xff,0xfd,0xfd,0xef,0xff,0xff,0xfb,0xff,
	0xdf,0xff,0xfb,0xff,0xff,0x8b,0xff,0xff,
	0xff,0xdd,0xff,0xdf,0xff,0xff,0xff,0x9f,
	0xef,0xff,0xff,0xaf,0xff,0xdf,0x5d,0xff,
	0xf5,0xef,0xff,0xff,0xff,0xf7,0xff,0xf7,
	0xdd,0x7f,0xf7,0xff,0xbf,0xff,0xbf,0xf8,
	0xff,0xff,0xff,0xd7,0xf7,0xff,0xff,0xff,
	0xff,0x3f,0xfb,0xfd,0xff,0x9f,0xff,0xff,
	0xff,0xbf,0xfd,0xbf,0xff,0xbf,0xde,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xfd,0xff,
	0xe2,0xbf,0xbf,0xff,0xfe,0xdb,0xff,0xff,
	0xff,0xff,0xfb,0xfb,0xef,0xff,0xff,0xfe,
	0xff,0xf7,0xfe,0xfb,0xfe,0xff,0xff,0xff,
	0xf6,0xff,0xff,0xff,0xfe,0xff,0xff,0x9f,
	0xef,0x3f,0xfd,0xff,0xad,0xe7,0xff,0x7f,
	0x72,0xff,0x3f,0xbf,0xff,0xff,0xaf,0xfd,
	0xff,0xff,0xbf,0xb7,0xde,0xfc,0xff,0xf9,
	0xff,0xfd,0xdf,0xf3,0x7d,0x9d,0xd7,0x5a,
	0x5f,0xff,0x2f,0x51,0xff,0x6f,0xfe,0xff,
	0xff,0xcd,0xff,0xff,0xfc,0xff,0xff,0x7f,
	0xfd,0xff,0xff,0xcf,0xff,0xff,0xfe,0xff,
	0xff,0xff,0xef,0xff,0x5f,0xef,0xfb,0xf5,
	0xf3,0xff,0xfe,0xab,0xc6,0xfb,0xff,0xff,
	0xfd,0xff,0xff,0xbf,0xff,0xfe,0xfb,0xef,
	0xbf,0x5f,0xfe,0xfb,0xe5,0xff,0xfd,0xff,
	0xef,0x7f,0x59,0x7e,0x9b,0xe5,0xff,0xdf,
	0xfe,0xf7,0xdf,0xf7,0x4b,0x9f,0xf6,0xff,
	0xff,0xff,0xf7,0xff,0x7f,0xfb,0xed,0xb7,
	0xdf,0x7e,0xcb,0xe8,0xb7,0x8d,0xfe,0xff,
	0xff,0xdf,0xfe,0xcf,0xe8,0xb3,0x8c,0xfe,
	0xff,0xec,0xff,0xff,0xbf,0xcf,0xef,0xe1,
	0xdf,0xbe,0xff,0xfe,0xfa,0xef,0xaf,0xbf,
	0xfe,0xfa,0xeb,0xef,0x97,0xfe,0x5a,0xe9,
	0xfb,0xfe,0xfb,0xfb,0xbf,0x97,0x5e,0x7a,
	0xe9,0xbf,0x97,0xfe,0xfb,0xef,0xbf,0xff,
	0x95,0xfd,0xf6,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xbf,0xf6,0xff,
	0x6b,0xff,0xff,0xff,0xff,0xff,0xbf,0xfe,
	0xda,0xeb,0xff,0xbf,0xf6,0xff,0xff,0xff,
	0x7d,0x5b,0xf9,0xff,0x9f,0x7f,0xfe,0xff,
	0xad,0xff,0xff,0xff,0xff,0xff,0xff,0xf5,
	0xff,0x5b,0xff,0xff,0xfa,0xfe,0xff,0xbf,
	0xf4,0xd7,0x5f,0xef,0xf7,0xf5,0xff,0xfb,
	0xff,0xff,0x7f,0xe6,0xff,0x3f,0xff,0xfc,
	0xfb,0xdf,0xbf,0xff,0xfe,0xfb,0xef,0xff,
	0xa5,0xff,0x8b,0xfa,0xf7,0xfc,0xf3,0xff,
	0x3f,0xac,0xbe,0xfe,0xea,0x7e,0xaf,0xf3,
	0x13,0xcf,0xff,0xff,0x4f,0xfe,0xff,0xfe,
	0xfb,0xef,0xbf,0xff,0xb6,0xdb,0xef,0xbf,
	0xff,0xfe,0xdb,0xef,0xbf,0xed,0xfb,0xef,
	0xbf,0xed,0xfe,0xfb,0xef,0xbf,0xed,0xfe,
	0xdb,0xef,0xbf,0xff,0xff,0x9f,0xfa,0xff,
	0xe7,0x9f,0x7f,0xf7,0xf9,0xf7,0xdf,0x75,
	0xf7,0xfd,0xe7,0x9f,0x77,0xff,0xd9,0x9f,
	0x7f,0xfe,0x99,0xe7,0x9f,0x7d,0xfe,0x99,
	0xe7,0x9f,0x77,0xfe,0xf9,0xff,0xff,0x99,
	0xff,0x9f,0x7f,0xbe,0x7d,0xc7,0xdf,0x6f,
	0xb7,0x7d,0xf6,0x5b,0x7f,0xfd,0xfd,0xd7,
	0x6d,0xfe,0xf9,0xd6,0x9f,0x7f,0x8f,0xf5,
	0xf7,0x1f,0x6f,0xfd,0xf9,0xe7,0xda,0xff,
	0xd7,0xf6,0xff,0xff,0xff,0xfc,0xff,0xdf,
	0x3f,0xff,0xfe,0xf3,0xfb,0x7f,0xff,0xfd,
	0xf7,0xff,0xff,0xff,0xf6,0xff,0x7f,0xff,
	0xff,0xf7,0xff,0xbf,0xff,0xff,0xff,0xff,
	0xff,0xef,0xf6,0xff,0xdf,0x7f,0xfd,0xff,
	0x7f,0xff,0xfe,0xff,0xf7,0xff,0x7f,0xff,
	0xfb,0xef,0xff,0x7f,0xff,0xed,0xff,0xdf,
	0xfe,0xff,0xdf,0xff,0xff,0xff,0xff,0xfd,
	0xf7,0xff,0xff,0xbf,0xff,0x7f,0xf4,0xf1,
	0xf3,0x5f,0x2f,0xfe,0xf8,0xe3,0x8b,0x2f,
	0xfe,0xf8,0xe3,0x8f,0x9d,0xf0,0xc7,0x8f,
	0x7d,0xfc,0xd8,0xe3,0x8b,0x7f,0xfc,0xf8,
	0xc7,0x5f,0xff,0xff,0x3f,0xfd,0xff,0xf7,
	0xdf,0x57,0xff,0x7d,0xf5,0xdd,0x77,0x7f,
	0xfd,0xf5,0xdd,0x77,0xdf,0xfd,0xc5,0x7f,
	0x5f,0xfd,0xf7,0x5d,0x77,0x5b,0xf5,0xf7,
	0xdd,0x7f,0xff,0x7d,0xf7,0xff,0xef,0xff,
	0xbf,0xfe,0xfe,0xff,0xff,0xef,0x3e,0xfb,
	0xfe,0xf2,0xab,0xbf,0xfb,0xfa,0xf3,0xa5,
	0xf2,0xdf,0xff,0xdf,0xbd,0xfb,0xfe,0xaf,
	0xdf,0xbf,0xfe,0xfd,0xff,0xff,0xfe,0x7b,
	0xfe,0xff,0x97,0x7f,0xf6,0xf9,0xc7,0x1e,
	0x7f,0xec,0xd9,0x67,0x1f,0x7b,0xfc,0xf1,
	0x67,0x3d,0xee,0xb1,0xe7,0x1f,0x3b,0xec,
	0xf9,0xe7,0x1f,0x7f,0xfe,0xf9,0xc7,0x9f,
	0x7f,0xfa,0xff,0xf7,0xdf,0xbf,0xff,0xf9,
	0xfe,0xeb,0xaf,0xff,0xfe,0xfb,0xeb,0xbf,
	0xbf,0xfe,0xdf,0x77,0xff,0xfe,0xe7,0xeb,
	0xb7,0xff,0xfe,0xff,0xeb,0x7f,0xfe,0xfd,
	0xff,0xff,0xc9,0xff,0xbf,0xf4,0x7a,0xfd,
	0xbf,0xdf,0x5f,0x7f,0xfd,0xff,0xff,0x5f,
	0xff,0xfd,0xf5,0xff,0xea,0xeb,0xf5,0xff,
	0x5e,0xff,0xfd,0xfd,0xff,0x5e,0xff,0xef,
	0xaf,0xf7,0xff,0xdf,0xfe,0xff,0xff,0xff,
	0xbf,0xff,0xff,0xfb,0xef,0xbf,0xbf,0xfe,
	0xfa,0xef,0xaf,0xff,0x7e,0xff,0xff,0xff,
	0xfe,0xff,0x6f,0xaf,0xbf,0xfe,0xff,0xef,
	0xff,0xff,0xff,0xfb,0x7f,0xf0,0xff,0x7f,
	0xff,0x7f,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0x7f,0xff,0xff,0xff,0xfb,
	0x7f,0xff,0xff,0xff,0x7b,0x7f,0xff,0xff,
	0xff,0xff,0xff,0xff,0xfd,0xff,0xa3,0xff,
	0xff,0xff,0xfb,0xff,0x7f,0xfd,0xff,0xff,
	0xff,0xff,0xff,0xfe,0xff,0xef,0x3f,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xe7,0xff,
	0xff,0xbf,0xfc,0xef,0xbf,0x6f,0xff,0x7f,
	0xf5,0xff,0xd7,0x5f,0x7f,0xfd,0xf5,0xd7,
	0x5f,0x7f,0xf5,0xd5,0xd7,0x5f,0x7b,0xfd,
	0xf5,0x5e,0x7f,0xfd,0xe5,0x17,0x1f,0x7f,
	0xfd,0xf5,0xd7,0x5f,0x7f,0xed,0xf5,0xd7,
	0xbe,0xea,0xff,0xef,0xff,0xef,0xdc,0xef,
	0xcf,0x37,0xef,0xfc,0xf3,0xcf,0x3f,0xff,
	0xe4,0x73,0xff,0xff,0xfc,0xf3,0xbe,0x3f,
	0xaf,0xec,0xf3,0xac,0x1d,0xff,0xfb,0xa3,
	0xb3,0xff,0x2f,0xfe,0x7f,0xbf,0xff,0xf2,
	0xfb,0xe6,0xdf,0x7f,0xff,0xfb,0xff,0xff,
	0x7f,0xff,0xfb,0xff,0xff,0xfa,0xfb,0xef,
	0xf7,0xff,0xff,0xfb,0xfd,0xff,0xbf,0xff,
	0xd7,0xff,0xff,0x5f,0xfa,0xff,0xff,0xff,
	0xff,0xff,0xff,0xfc,0xf3,0xcd,0x3f,0xff,
	0x7c,0xf3,0xcf,0x3f,0xff,0xfe,0xdf,0x3f,
	0xb7,0xdf,0x71,0xc7,0x3f,0xff,0xfb,0xf3,
	0xff,0xff,0xff,0xff,0xff,0xdf,0xff,0xff,
	0xff,0xff,0xdb,0xff,0xd7,0x5f,0x7f,0xf8,
	0xe7,0xdb,0x5f,0x6f,0xfd,0xf5,0x5e,0xff,
	0xfb,0xef,0xff,0x5f,0x7e,0x7c,0xb1,0xff,
	0x1f,0xff,0xff,0xff,0xfd,0xff,0xff,0xfe,
	0xff,0xff,0xff,0xbf,0xff,0xbf,0xfd,0xf9,
	0xff,0xde,0x7f,0xff,0xf6,0xda,0x9f,0xff,
	0xf9,0x7e,0xfb,0xfd,0xff,0xb7,0xff,0xff,
	0xfe,0xef,0xbf,0xff,0xff,0xfb,0xff,0x7f,
	0xf5,0xff,0xff,0xff,0xff,0xff,0xbf,0xf7,
	0xfa,0xdf,0xff,0xdf,0xff,0xdf,0xfd,0xff,
	0xff,0xfd,0xfd,0xb7,0xff,0x7f,0xf7,0xdf,
	0x7f,0xfd,0xff,0xd7,0x7f,0xff,0xfd,0xf7,
	0xff,0xab,0xff,0xff,0xff,0xff,0xff,0x7f,
	0x6f,0xff,0xfd,0xff,0xde,0xff,0xbe,0xff,
	0xdf,0xff,0xcf,0xfd,0xf7,0xdf,0x7f,0xf7,
	0xf7,0xf7,0x7b,0xff,0xbd,0xf7,0xfe,0xfa,
	0xef,0xff,0xff,0x7d,0xff,0xdf,0xff,0x7f,
	0xfd,0xbf,0xf7,0xff,0x7b,0xff,0xff,0xdf,
	0xdf,0xfa,0xff,0xff,0xfe,0xfd,0xfb,0xbf,
	0xff,0xff,0xff,0xff,0xff,0xe7,0xff,0xfa,
	0xff,0xff,0xff,0xff,0xec,0xfb,0xe3,0x7b,
	0xff,0xfd,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0x7f,0xbf,0x3c,0xfb,0xdf,0xff,
	0xff,0xff,0xaf,0xfb,0xef,0xff,0xeb,0xff,
	0x3f,0xfe,0xff,0xff,0xff,0x27,0xdf,0xff,
	0xfe,0xff,0xff,0xfb,0xfe,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xef,0xd7,0xd7,0xff,
	0xff,0xff,0xff,0xff,0x7f,0xff,0xfb,0xfe,
	0xff,0xff,0xef,0xff,0xff,0xff,0xbf,0xf8,
	0xff,0xff,0xff,0xfe,0xff,0xfd,0xfa,0xcf,
	0xff,0xfd,0xfe,0xff,0xff,0xff,0xcf,0xef,
	0xfd,0x7d,0xf3,0xcf,0xbb,0xfd,0xff,0xff,
	0xfe,0xff,0xcd,0xbb,0xff,0xfe,0xfb,0xff,
	0x89,0xff,0xff,0x5f,0xff,0xff,0xff,0xd7,
	0x7f,0xff,0xff,0xb1,0xff,0x6f,0xfa,0xef,
	0xef,0xff,0xdf,0xfb,0xfe,0xdf,0x5f,0xff,
	0xff,0xf7,0xff,0x5e,0xff,0xfd,0xf7,0xdf,
	0xff,0x07,0xfe,0xff,0xdf,0xff,0x7f,0xff,
	0xfe,0xff,0xfa,0xfd,0xff,0xff,0xf3,0xfb,
	0xff,0xff,0xfd,0xef,0xff,0xff,0xbf,0x7e,
	0xdf,0xff,0xcf,0xff,0xff,0xce,0xf3,0xcf,
	0xff,0xfd,0xff,0xc7,0xff,0xff,0xfa,0xf9,
	0xfd,0x3f,0xfd,0xff,0x79,0xe5,0x33,0xdf,
	0xf7,0xeb,0xe5,0x3f,0xdb,0xff,0xcb,0xfd,
	0x77,0xf6,0xff,0xfb,0xff,0x9f,0xff,0xff,
	0x7f,0xfd,0xf5,0xff,0xac,0xd5,0xff,0xd7,
	0x5f,0xff,0xff,0xe9,0xe7,0x5f,0x7f,0xfd,
	0xf5,0xbf,0x5f,0xef,0xff,0xff,0x5f,0x7f,
	0xfd,0xf5,0x33,0xce,0xff,0xff,0x7f,0xff,
	0xff,0xff,0xfb,0xef,0xff,0xda,0xbc,0xfe,
	0xfb,0xef,0x7f,0xff,0x7e,0xfb,0xef,0x7f,
	0xff,0xfc,0xf7,0xed,0xff,0xff,0xff,0xef,
	0xff,0xff,0xfe,0x3b,0xfd,0xff,0x5f,0xfe,
	0xff,0xe5,0x94,0xdf,0x7e,0xff,0xf7,0x67,
	0xff,0xb7,0xdf,0xfe,0xff,0xed,0xb3,0xdf,
	0xfe,0xff,0xed,0xff,0xcf,0xfe,0xff,0xff,
	0xdf,0xfe,0xff,0xed,0xb7,0xff,0xff,0xff,
	0xe8,0xff,0x8d,0x3e,0xfa,0xec,0xfb,0xff,
	0x16,0xff,0xff,0xfe,0xeb,0xbf,0xbf,0x5e,
	0xfa,0xeb,0xbf,0xbf,0xff,0x7b,0xe9,0xbf,
	0xff,0xbe,0xeb,0xbf,0xbf,0xff,0x7a,0xe9,
	0xbf,0x97,0xfe,0x5b,0xe9,0xa5,0xbf,0x5e,
	0xfb,0x7f,0xdb,0xfe,0xff,0xff,0xff,0xff,
	0xff,0xda,0xff,0xff,0xff,0xff,0xff,0xeb,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xeb,0xf7,0xbf,0xfe,0xff,0xeb,0xaf,0xff,
	0xff,0xda,0xbf,0x91,0xe7,0xff,0xff,0xff,
	0xbf,0xfe,0xd7,0xff,0xff,0xf7,0xbf,0xff,
	0x59,0xff,0xff,0x5f,0xfe,0xeb,0x97,0xff,
	0xff,0xde,0xdb,0x9f,0xf5,0xff,0x5b,0x7f,
	0xfd,0xff,0xd6,0xff,0xf7,0x7e,0xff,0xff,
	0xef,0x3f,0xfc,0x96,0xfe,0xef,0x7f,0xff,
	0xf3,0x53,0x6a,0x3d,0xff,0x7d,0x6f,0x7c,
	0xfd,0xff,0xfb,0xed,0x3f,0xa5,0xf2,0x93,
	0x6a,0xa9,0xff,0xa2,0xfe,0xff,0xf2,0xff,
	0x6f,0xbb,0xff,0xfe,0xdb,0xee,0xbf,0xf9,
	0xfe,0xfb,0xef,0xbf,0xf9,0xfe,0xfb,0xbf,
	0xff,0xde,0xdb,0xee,0xff,0xff,0xfe,0xfb,
	0xef,0xbd,0xff,0xfe,0xfb,0xff,0xff,0x83,
	0xff,0x7f,0xf6,0xfd,0xe7,0xdf,0x7d,0xfe,
	0xdd,0xe7,0x9f,0x7f,0xfe,0xf9,0xe7,0x9f,
	0xff,0xdd,0xe7,0x9f,0x7d,0xf7,0xfd,0xe7,
	0x9f,0x7f,0xfe,0xf9,0xe7,0x9f,0xff,0xf7,
	0x9f,0xf9,0xbf,0x75,0xf6,0x9f,0x7f,0xdf,
	0xf5,0x76,0x1f,0x6f,0xfd,0xf9,0xf7,0x9f,
	0x7f,0xbc,0xf6,0x1d,0x6f,0xf5,0xdd,0xe6,
	0x9f,0x7f,0xff,0xf9,0xf7,0xdf,0x6b,0xff,
	0x7d,0x7f,0x6b,0xbf,0xfd,0xf3,0xff,0x7f,
	0xff,0xff,0xfe,0xff,0xef,0xff,0xff,0xff,
	0xff,0xff,0xbf,0xfd,0xff,0x3f,0xff,0xff,
	0xf3,0xfd,0xff,0xff,0xff,0xf7,0xdf,0x7f,
	0xff,0xff,0xff,0x56,0xff,0xfb,0xf7,0xff,
	0xdf,0xfd,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfd,0xff,0xdf,0xff,0xfd,0xff,0x7f,0xff,
	0xff,0xdf,0xff,0xdf,0xff,0xff,0xf5,0x7f,
	0xff,0xfd,0xff,0xff,0x7f,0xea,0xff,0xe2,
	0x8b,0x7d,0xfc,0xf8,0xe2,0x8b,0x7f,0xbc,
	0xd8,0xc7,0x8f,0x7f,0xfc,0xf1,0x89,0x7f,
	0xb5,0xf8,0xe2,0x8f,0x7d,0xfd,0xd8,0xd7,
	0x8b,0x3f,0xfe,0xf8,0xfe,0x7f,0xcd,0xff,
	0x57,0x7f,0xfd,0xf7,0xdd,0x57,0xdf,0xfd,
	0xf7,0xdd,0x7f,0x7f,0xf5,0xb7,0xdf,0x5f,
	0xfd,0xf7,0xd5,0x57,0x5f,0xfd,0xf7,0x57,
	0x7f,0xdb,0xe5,0xd5,0x5d,0x77,0xff,0xcf,
	0xfe,0xff,0xfd,0xfa,0xdf,0xbf,0xbe,0xfe,
	0xfb,0xbf,0xaf,0xfe,0xfb,0xeb,0xdf,0xfe,
	0x73,0x7b,0xff,0xef,0xff,0xdf,0x7b,0xff,
	0xbf,0xfa,0xff,0xea,0xcf,0xbf,0xfe,0xef,
	0x7f,0xf3,0xff,0xff,0x99,0xe7,0x1f,0x7f,
	0xe4,0xb9,0xe7,0x1f,0x7f,0xfe,0xd9,0xe7,
	0x9f,0x7b,0xb6,0xe7,0x9f,0x79,0xee,0x91,
	0xe6,0x9f,0x7d,0xfe,0xf1,0x67,0x1f,0x7f,
	0xfe,0xf9,0x97,0xff,0xbf,0xff,0xfe,0xe7,
	0xeb,0xbf,0xbf,0xfe,0xf7,0xef,0x7f,0xbf,
	0xfe,0xff,0xdf,0xbf,0xf9,0xff,0xef,0x3f,
	0xff,0xfc,0xff,0xeb,0x7f,0xfe,0xfe,0xfb,
	0xef,0xef,0xff,0x3f,0xfd,0x7f,0xff,0xff,
	0xff,0x5e,0xff,0xfd,0xfd,0xff,0xfe,0xff,
	0xef,0xf5,0xff,0xfe,0xfa,0xf7,0xff,0xde,
	0xff,0xfd,0xf5,0xff,0x5e,0xff,0xef,0xff,
	0xf7,0x7f,0x7f,0xff,0xff,0xe6,0xff,0xaf,
	0xbf,0xfe,0xff,0xef,0xaf,0xbf,0xfe,0xfd,
	0xeb,0xdf,0xff,0xfe,0xff,0xff,0xf7,0xfe,
	0xff,0xef,0x8f,0x3f,0xfe,0xff,0xea,0xff,
	0xbf,0xfe,0xfa,0xeb,0xbf,0xff,0x7f,0xff,
	0xdf,0xff,0xff,0xff,0xff,0x7f,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfb,0xff,0xff,0x7f,0x7e,0xfd,0xff,0xd7,
	0xff,0xff,0xff,0xfd,0xdf,0xff,0xff,0x3f,
	0xf8,0xff,0xff,0xff,0xff,0xff,0xff,0xf6,
	0xff,0xff,0xff,0xff,0xff,0xbb,0xff,0xfe,
	0xff,0xff,0xff,0xff,0xff,0xff,0xdb,0xff,
	0xff,0xff,0xff,0xff,0xff,0xbe,0xfc,0xe6,
	0xff,0x57,0xff,0x7f,0xfd,0xf5,0x57,0x5f,
	0x7f,0xfd,0xf5,0x97,0x5e,0x7e,0xd9,0xe5,
	0x57,0x5f,0xeb,0xd5,0xd7,0x5f,0x7c,0xf5,
	0xd5,0xd7,0x5f,0x7d,0xfd,0xf5,0xd7,0x1b,
	0x7f,0xfd,0x7b,0xfe,0xdf,0xd8,0xf3,0xfe,
	0x3f,0x23,0xfc,0xf3,0xce,0x3f,0xff,0xfc,
	0xf3,0x0f,0xff,0xfa,0xd3,0xdf,0x3f,0xff,
	0xfc,0xe8,0xef,0x3f,0xfd,0xfd,0xf3,0xcf,
	0x3f,0x3e,0xfb,0x7f,0xe6,0xff,0x7f,0xfc,
	0xfd,0xb3,0xdf,0xbe,0xfd,0xaf,0xf3,0xdb,
	0x7b,0xfb,0x59,0xbf,0x7f,0xbf,0xbf,0xd7,
	0xff,0xff,0xff,0x7f,0xef,0xff,0xfd,0xff,
	0xfb,0xff,0xff,0xff,0xff,0xa5,0xff,0xcf,
	0x3f,0xff,0xff,0xf3,0xcf,0x3f,0xff,0xfb,
	0xf3,0xbf,0x3f,0xbf,0xff,0xff,0x0d,0xff,
	0xff,0xf3,0xce,0x5f,0x7f,0xff,0xf3,0xff,
	0x3f,0xff,0xfc,0xf3,0xff,0xff,0x7f,0xfc,
	0x7f,0xff,0xf7,0xff,0xdf,0x7f,0xbd,0xff,
	0xff,0x1f,0xff,0xfb,0xe4,0xbf,0xf7,0xeb,
	0xfd,0xbf,0x6f,0x7f,0xfd,0xff,0xff,0x4f,
	0xff,0x7f,0xfd,0xc7,0x5f,0xff,0xff,0xfb,
	0xe0,0xff,0x6f,0xef,0xff,0xff,0xf7,0xd7,
	0xff,0xeb,0xfb,0xfd,0xff,0x6e,0xeb,0xff,
	0xff,0x7f,0xf7,0xff,0xff,0xdf,0xfa,0xff,
	0xff,0xf9,0xfd,0x7f,0xfb,0xff,0xf6,0xff,
	0xdf,0x2f,0xff,0xff,0xe6,0xdf,0x7f,0xff,
	0xff,0xff,0xff,0xf7,0xbf,0xff,0xff,0xff,
	0x7e,0xff,0xdd,0x6d,0x7f,0xd3,0x56,0xff,
	0xff,0xdf,0xff,0xff,0xff,0xff,0xfd,0xdf,
	0x7f,0xff,0x3f,0xfb,0xff,0xff,0xff,0xff,
	0xff,0xff,0xdf,0x7f,0x6f,0xff,0xbf,0xdf,
	0x7f,0x77,0xff,0xfd,0xba,0x7f,0xbf,0xff,
	0xfe,0xff,0xff,0xfd,0xff,0xff,0xff,0xff,
	0xbf,0xff,0xfe,0xf7,0xdd,0xf7,0x7f,0xff,
	0xfd,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xef,0xfb,0xff,0xff,0xef,0xff,0xbf,
	0xfe,0xff,0xfe,0xdf,0xff,0xff,0xff,0xff,
	0xef,0xef,0xff,0xff,0xff,0x7f,0xbe,0xff,
	0xff,0xff,0xff,0xcf,0xbf,0xff,0xff,0xff,
	0xdf,0xdf,0xff,0xff,0xeb,0xaf,0x8b,0x7f,
	0xfd,0xff,0xff,0x7f,0xff,0xdf,0xdf,0x7f,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xf0,
	0xfd,0xff,0xff,0xff,0x5f,0xfd,0xfd,0xff,
	0xff,0xf7,0xfd,0x7f,0xff,0xff,0xef,0xff,
	0xfd,0xff,0xfb,0xff,0xff,0xff,0xff,0xdf,
	0xbb,0xfb,0xfd,0xff,0xff,0xff,0xff,0xff,
	0xb7,0xff,0xff,0xdf,0xff,0xff,0xcb,0xf7,
	0xfb,0xef,0xfb,0xdf,0xff,0xff,0xff,0xbf,
	0xdc,0x9e,0xee,0x7b,0xff,0xbd,0xdb,0xed,
	0x3f,0xdf,0xbb,0xff,0xfd,0xff,0xff,0xbf,
	0xff,0x7e,0xf9,0xff,0xfd,0xff,0xff,0x7f,
	0xff,0xfb,0xbf,0xdf,0xfe,0xfe,0xff,0xee,
	0xbf,0xff,0xff,0xff,0xcf,0x5f,0xbf,0xff,
	0xff,0xd7,0xff,0xfe,0xfd,0xff,0xff,0xf7,
	0x7b,0xed,0x7f,0xe6,0xff,0xef,0xff,0xff,
	0xfb,0xfb,0xff,0xf7,0x3d,0xfb,0xff,0xff,
	0xff,0xfd,0xff,0xff,0xff,0xdf,0xfa,0xff,
	0xbb,0xfe,0xff,0xff,0x7f,0xf3,0xfd,0xff,
	0xff,0xff,0xfb,0xff,0x67,0xfc,0xf3,0xaf,
	0xf7,0xff,0xff,0x7f,0xff,0xff,0x7f,0x7e,
	0xeb,0xe3,0xe5,0xdf,0x7f,0xbf,0xff,0xff,
	0xd2,0x7e,0x2f,0xf7,0xff,0xff,0xff,0xcf,
	0xff,0xff,0xda,0x5f,0xff,0x4f,0x5b,0x9d,
	0x7f,0xbe,0xff,0xff,0xff,0xff,0xff,0xff,
	0xd7,0x5d,0x7f,0x3d,0xfc,0xf0,0xfb,0xff,
	0xff,0xb1,0xff,0x7f,0xb9,0xff,0xf3,0xff,
	0x7f,0xfd,0xff,0xd7,0x5f,0xfe,0xaf,0xd3,
	0xeb,0xbf,0xff,0xff,0xfb,0xef,0xbf,0xff,
	0x2e,0xf9,0xff,0xbf,0x53,0x67,0xfd,0xff,
	0x5c,0x7e,0xe9,0x6d,0x96,0xdf,0xff,0xfb,
	0xff,0xbf,0xff,0xff,0x3f,0xe5,0xb7,0x7f,
	0x6f,0xf6,0x7f,0xfb,0xff,0xb7,0xdf,0x7e,
	0xfb,0x6d,0xa2,0xff,0x7f,0x9b,0x3a,0xeb,
	0xff,0xf3,0x68,0xa3,0xcf,0x3e,0xfa,0xfe,
	0xb7,0xff,0x7f,0xfb,0xff,0xff,0x89,0x3e,
	0xfb,0xef,0xf0,0xff,0xaf,0xff,0xfe,0xfe,
	0xeb,0xaf,0xbf,0x5f,0xfa,0xef,0xaf,0xff,
	0xfe,0xfb,0xef,0x93,0xf6,0x7e,0xe9,0xa5,
	0xf7,0xfe,0xfa,0xef,0xef,0xff,0xfe,0xfb,
	0xeb,0xaf,0xff,0xa3,0xed,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xda,0xff,0xff,
	0xff,0xfb,0xef,0xff,0xbf,0xfe,0xef,0xeb,
	0xaf,0xbf,0xfb,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xfd,0xbb,0xf8,0xff,0xff,
	0x7f,0xfe,0xfb,0xff,0xff,0xff,0xd6,0xfe,
	0xaf,0xff,0xf5,0x52,0xff,0xf9,0xf5,0xd6,
	0x4b,0x7f,0xfd,0xfd,0xfe,0xff,0xa9,0xff,
	0x7f,0x7e,0xf9,0xff,0xff,0x7f,0xef,0xff,
	0xbf,0xff,0xfc,0xbf,0xef,0xbf,0xff,0xaf,
	0xfe,0x0f,0xbf,0xaf,0xbc,0xf6,0xdf,0xa7,
	0x86,0xde,0xea,0xab,0xef,0xc4,0xfb,0x5f,
	0xff,0xff,0xf4,0xd7,0xef,0xff,0xff,0x47,
	0xff,0xff,0xf6,0xfb,0x6f,0xbb,0xff,0xf6,
	0xdb,0x6e,0xbf,0xff,0xfe,0xfb,0xef,0xbf,
	0xff,0xfb,0x6f,0xb9,0xfd,0xb6,0xdb,0x6f,
	0xbf,0xf7,0xfe,0x7b,0xef,0xbd,0xff,0x7e,
	0xbe,0xfa,0xff,0x77,0x9d,0x7f,0xe6,0xfd,
	0x77,0x9d,0x79,0xde,0xf9,0x77,0x9f,0x7f,
	0xfe,0xf9,0x9f,0x7f,0xe6,0x79,0xe7,0x9f,
	0x75,0xdf,0xf9,0xe7,0x9f,0x7f,0xfe,0xd9,
	0xe7,0xf7,0x8d,0xff,0xdf,0x7d,0xbe,0xf5,
	0xf6,0xdf,0x7d,0xfd,0xf5,0xe7,0xdf,0x77,
	0xfe,0xf1,0xc7,0x6a,0xff,0xf5,0xf6,0xdf,
	0x7f,0xda,0xfd,0xc7,0x5f,0x7f,0xfe,0xf1,
	0xf7,0x59,0xff,0x67,0xf6,0xff,0xff,0xbf,
	0xfd,0xf3,0xdf,0x7f,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xdf,0xff,0xff,0xf3,
	0xff,0xff,0xff,0xff,0xff,0xdf,0xff,0xff,
	0xff,0xf7,0xcf,0xff,0x6f,0xf3,0xff,0xff,
	0x7f,0xfb,0xef,0x7f,0xff,0xfd,0xff,0xff,
	0xf7,0xff,0x7f,0xff,0xff,0xbf,0xff,0xff,
	0xf7,0xff,0xff,0x7f,0xff,0xff,0x7f,0xff,
	0x7f,0xff,0xf7,0x7f,0xff,0xff,0xa7,0xfe,
	0x2f,0xfe,0xf1,0xe2,0x8f,0x2f,0xfe,0xd8,
	0xe3,0x1f,0x2f,0xfe,0xf1,0xc7,0x1f,0x9f,
	0xf8,0x62,0x8b,0x2f,0x9e,0xf0,0xe3,0x1f,
	0x3f,0xf6,0xd5,0xd7,0x8b,0x3d,0xfe,0xb7,
	0xfc,0x7f,0xf5,0xdf,0x57,0xdf,0x7d,0xf5,
	0xdd,0x5f,0xfb,0x7d,0xf5,0xdf,0x7e,0xfb,
	0xfd,0x57,0x57,0x7f,0x75,0xd5,0xc5,0x5f,
	0xfb,0x7d,0xf7,0xdf,0x7f,0x5f,0x75,0xf5,
	0xff,0xe2,0xff,0xff,0xff,0xbf,0xfb,0xf3,
	0xcf,0xbe,0xfe,0xfc,0xff,0xff,0xfd,0xff,
	0xed,0xf7,0xa7,0xfa,0xfb,0xfb,0xcf,0xfc,
	0xfa,0xee,0xf7,0xcf,0xfd,0xfb,0xef,0xbf,
	0xef,0xff,0x4f,0xff,0xff,0x9e,0x7b,0xf6,
	0xf1,0xc7,0x0f,0x7f,0xf6,0xb9,0xe7,0x9e,
	0x7b,0xfe,0xf9,0xe7,0x5d,0xfe,0xb9,0x67,
	0x9f,0x7f,0xe6,0xf9,0xc7,0x9f,0x7f,0xfe,
	0xb9,0x47,0x9e,0x7f,0xf9,0xff,0xfb,0x9f,
	0xbf,0xff,0xfe,0xfe,0xe9,0xaf,0xff,0xf9,
	0xfb,0x9f,0xff,0xff,0xff,0xef,0xbf,0xbf,
	0xfe,0xfe,0xeb,0xaf,0xff,0xff,0xfb,0xff,
	0x7f,0xff,0xfe,0xff,0xff,0xcd,0xff,0xdf,
	0xff,0x7b,0xff,0xf7,0xdf,0x5f,0x7f,0xfd,
	0xbf,0xdf,0xff,0xfb,0xef,0xbf,0xff,0xd7,
	0xfd,0xf7,0xf7,0x5f,0x7b,0xfd,0xbf,0xdf,
	0xff,0xfb,0xeb,0xfd,0xf7,0xff,0xff,0xfe,
	0xff,0xfa,0xff,0xbf,0xbf,0xfe,0xdb,0xef,
	0xbf,0xff,0xff,0xfa,0xff,0xff,0xff,0x7f,
	0xeb,0xaf,0xbf,0xfe,0xfb,0xff,0xbf,0xff,
	0xff,0xfa,0xff,0xff,0xbf,0xfe,0xfb,0xff,
	0xf2,0xff,0xf7,0xff,0xff,0xff,0xfd,0xff,
	0xfe,0xff,0xff,0xff,0xf7,0xff,0xff,0xff,
	0xff,0xff,0x7d,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xf7,0xff,0xff,0x7f,0xff,0xfd,
	0xff,0xab,0xff,0xfd,0xff,0xff,0xff,0xff,
	0xff,0xfe,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xbf,0xff,0xff,0xff,0xfb,0xff,0xbf,
	0xfc,0xff,0xff,0xff,0xff,0xff,0xef,0xff,
	0x4f,0xfe,0x9f,0xf5,0xff,0xd7,0x5f,0x7e,
	0xfd,0xd5,0xd7,0x5f,0x7f,0xfd,0xf5,0xd7,
	0x5f,0x7f,0xf1,0xa5,0x1e,0x7b,0xd5,0xe5,
	0x17,0x5b,0x7e,0xfd,0xf1,0x57,0x5f,0x7f,
	0xfd,0x75,0xd7,0xbf,0xea,0xff,0xcf,0xfd,
	0xcf,0xf8,0xf3,0xcf,0x3f,0xff,0xfc,0xef,
	0xcf,0x3f,0xff,0xef,0xa3,0x3f,0xfd,0xec,
	0xf3,0x0f,0x7f,0xfe,0x7c,0xcf,0xcf,0xff,
	0xff,0xfe,0xf3,0xcf,0xff,0x7f,0xfe,0xef,
	0xbf,0xff,0xff,0xfb,0xef,0xe7,0xbf,0xff,
	0xfb,0xff,0x5f,0xff,0xfe,0xeb,0xf5,0xff,
	0x7f,0xff,0xed,0xbf,0xfe,0xfe,0xff,0xee,
	0xff,0xfb,0xf7,0xff,0xff,0xff,0x1f,0xfa,
	0xff,0xdc,0xff,0xcf,0x3f,0xfd,0xfc,0xf3,
	0xcf,0xff,0xff,0xfc,0xef,0xff,0x5f,0x7f,
	0xf0,0xc5,0x3e,0xff,0xfc,0x3f,0xcf,0xff,
	0xff,0xfc,0xff,0xff,0x31,0xff,0xfc,0xff,
	0xd7,0xff,0xc7,0xff,0x7f,0xf9,0xf1,0xd7,
	0x17,0x6f,0xfc,0xff,0x46,0xff,0xff,0xef,
	0xff,0x5e,0x7f,0xfd,0xff,0xc7,0xff,0x7e,
	0xf9,0xff,0xc7,0xff,0xff,0xff,0xf1,0xd7,
	0xff,0x0f,0xfe,0xbf,0xff,0xff,0xdf,0xef,
	0x7f,0xfe,0xfe,0xb6,0xfb,0xbf,0x7f,0xff,
	0xff,0xf7,0xff,0xd7,0xdb,0xfb,0x7b,0xbf,
	0xff,0xde,0xff,0xff,0xbf,0xff,0xbd,0xfb,
	0xfb,0xfe,0xff,0xf7,0xff,0xff,0x7f,0xdd,
	0xfb,0xff,0xd7,0xdf,0xff,0xff,0xdf,0x7f,
	0xdf,0xfe,0xff,0xdf,0xf1,0xfe,0xba,0xfd,
	0xf7,0xff,0xdf,0xff,0xff,0x7d,0xff,0xff,
	0x7f,0xfd,0x75,0xff,0xa7,0xff,0xff,0xdf,
	0xdf,0xff,0xff,0xef,0xbd,0xfd,0xff,0xdf,
	0xff,0xaf,0xff,0xf7,0xdf,0xef,0xff,0xef,
	0xfb,0xef,0xff,0xff,0xff,0x7f,0xff,0xff,
	0xbf,0xff,0xf9,0xef,0xff,0xbf,0x7d,0xff,
	0xff,0xdf,0xff,0xff,0xff,0xff,0xe7,0xff,
	0xff,0xff,0xff,0xfe,0x7f,0xe7,0xbf,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xfb,0xff,
	0xbf,0xff,0xfe,0xfe,0xed,0xff,0xff,0xe4,
	0xfb,0xff,0xdd,0x3b,0xff,0xff,0xfb,0xff,
	0xff,0xdf,0xff,0xfd,0xfb,0x3f,0xf6,0xbf,
	0xfb,0x3f,0xff,0x77,0xdf,0xfb,0xdf,0x6f,
	0xfd,0xff,0xff,0xff,0xaf,0xf8,0xff,0xff,
	0x67,0xdf,0xff,0xdf,0xfb,0xf9,0xfd,0xdf,
	0xf7,0x7f,0xff,0xff,0xff,0xfb,0xff,0xfb,
	0xff,0xdf,0xff,0xf5,0xff,0xff,0xdf,0xff,
	0xff,0xbf,0xff,0xf7,0xff,0xff,0xbf,0xff,
	0xff,0xbf,0xfb,0xef,0xbb,0xdf,0xff,0xbf,
	0x7c,0xfb,0xef,0x3f,0xff,0xdc,0xff,0xfe,
	0x7f,0xfb,0xfe,0xdd,0xff,0xff,0x73,0xff,
	0xeb,0xbf,0xf7,0xfd,0xfb,0xff,0xfb,0xff,
	0xfe,0xff,0xff,0x97,0xff,0xff,0xff,0xff,
	0xff,0xff,0xef,0x7d,0xfe,0xdc,0xf7,0xff,
	0xfd,0xfb,0xff,0xf2,0xf9,0xfd,0xff,0xfe,
	0xff,0xff,0xff,0xff,0xff,0xdf,0xff,0xff,
	0xff,0xf5,0xfb,0xff,0xa7,0xfe,0xbf,0xff,
	0xef,0xff,0xff,0xff,0xff,0xfc,0xf7,0xcf,
	0xff,0xff,0xef,0xbf,0xdf,0xff,0xbf,0x3f,
	0xff,0xff,0xff,0xff,0xf9,0xfd,0x3f,0xfb,
	0xff,0xbf,0xef,0xfe,0xff,0x7f,0xc0,0xff,
	0xff,0xff,0xff,0xff,0xff,0xfd,0xff,0xff,
	0xcf,0xff,0xfd,0xff,0xff,0xe7,0xbf,0xd2,
	0x78,0xad,0xed,0xf7,0xd7,0xf3,0xff,0x27,
	0x3f,0xfe,0xff,0xfb,0x3f,0xff,0xff,0x98,
	0xd5,0xff,0xff,0xff,0xff,0xff,0xef,0xff,
	0xff,0x3f,0xff,0xef,0xe7,0xff,0x3f,0xff,
	0xb5,0x5f,0x7c,0xfd,0xff,0xa7,0xcf,0x7f,
	0xbd,0xfc,0xf3,0xcf,0xff,0xf3,0xc9,0xff,
	0x5a,0xbd,0xfe,0xfb,0xff,0xbf,0xff,0x7e,
	0xfb,0xaf,0xbf,0xff,0x7e,0xfb,0xe5,0xff,
	0xff,0xfe,0xfd,0x5f,0x59,0x66,0xf9,0xed,
	0xbf,0xff,0xff,0xfb,0xef,0xff,0x7e,0xfe,
	0xf9,0xf7,0x66,0xff,0xb7,0xff,0x7f,0xfb,
	0xed,0xb3,0xdf,0x7e,0xfb,0xed,0xb3,0x8c,
	0xfe,0xff,0xed,0xef,0xbf,0xff,0x68,0xa2,
	0xcf,0x7e,0xfb,0xff,0xb7,0xdf,0xfe,0xff,
	0xec,0xb3,0xff,0x5e,0xff,0xff,0xfe,0xef,
	0xaf,0xbf,0x5e,0x7a,0xf9,0xaf,0xbf,0xde,
	0x7e,0xe9,0xbf,0xbf,0x3e,0xef,0x77,0x93,
	0x5e,0x7a,0xeb,0xef,0xff,0xfe,0xfa,0xfb,
	0xbf,0xbf,0xfe,0xfa,0xbf,0xda,0xfe,0xff,
	0xff,0xff,0xff,0xff,0xda,0xab,0xff,0xff,
	0xff,0xee,0xeb,0xff,0xff,0xff,0xbb,0xef,
	0x9a,0xfe,0xda,0xbb,0xff,0xff,0xff,0xff,
	0xff,0xff,0xfe,0x5f,0xff,0xbd,0x97,0xe7,
	0xff,0xf9,0xe7,0xef,0xff,0xd7,0x5f,0xff,
	0xff,0xff,0xf7,0x5f,0xff,0x97,0xff,0xdf,
	0xe5,0xfd,0xf5,0xd7,0xdb,0xfb,0xff,0xff,
	0xff,0xff,0xff,0x97,0xff,0xff,0xb7,0xb7,
	0x7e,0xff,0xdf,0x4f,0xbd,0xff,0xb2,0xfe,
	0xfa,0xbf,0xff,0xbe,0xff,0xea,0x7f,0xfd,
	0x7e,0x5e,0xfb,0xaf,0xbe,0x1e,0xee,0xff,
	0xff,0xfc,0xfb,0xff,0x7f,0xfd,0xfe,0xfb,
	0xff,0xf0,0xff,0xef,0xbd,0xff,0xf6,0xfb,
	0x6f,0xbf,0xff,0xfe,0xfb,0xef,0xbf,0xff,
	0xde,0xfb,0xbf,0xf7,0xb7,0xdb,0xee,0xbf,
	0xff,0xf6,0xfb,0xef,0xbf,0xed,0xde,0xfb,
	0xef,0xff,0xb3,0xff,0x7f,0xfe,0xf9,0x77,
	0xdc,0x7f,0xde,0xf9,0x77,0xdf,0x7f,0xf6,
	0xf9,0xe7,0xdf,0xff,0xf9,0x6f,0x9f,0x7d,
	0xfe,0xfd,0xe7,0x9d,0x7f,0xff,0xf9,0xe7,
	0x9f,0x7f,0xf6,0x9f,0xf8,0xbf,0xf5,0xe7,
	0xdb,0x7f,0xbf,0xf5,0xd6,0xdf,0x6d,0xaf,
	0xd5,0xf7,0x1f,0x6b,0xaf,0xc6,0xdf,0x6d,
	0x9f,0xf5,0xf7,0x5f,0x7f,0xfe,0xfd,0xd7,
	0x1f,0x6b,0xaf,0x7d,0x7e,0x6b,0xff,0xfc,
	0xff,0xdb,0x7f,0xbf,0xff,0xf6,0xdf,0x3f,
	0xff,0xff,0xf7,0xff,0xff,0xff,0xff,0xff,
	0x3f,0xbf,0xff,0xf7,0xff,0xff,0xff,0xff,
	0xf7,0xff,0x7f,0xff,0xff,0xf3,0x0e,0xff,
	0xfd,0xff,0xb7,0x7f,0xff,0xff,0xf7,0xdf,
	0x7f,0xff,0xff,0xdf,0xff,0xff,0xff,0xff,
	0xff,0x5f,0xff,0xff,0xdf,0xff,0xff,0x7f,
	0xff,0xf7,0xff,0x7f,0xff,0xff,0xef,0xbf,
	0xeb,0xff,0xe3,0x1f,0x3f,0xbe,0xd8,0xe2,
	0x8b,0x3f,0xfe,0xf8,0xe3,0x8b,0x7f,0xfd,
	0xf8,0x4d,0xa7,0xbd,0xf8,0xe3,0x8f,0x2f,
	0xf6,0xf1,0xe3,0x8f,0x7f,0xbd,0xf8,0xe2,
	0x7f,0xd9,0xff,0x57,0xfb,0x7d,0xf7,0xdd,
	0x5f,0x7f,0x7d,0xf5,0xd5,0x77,0x5f,0xf4,
	0xf7,0xd5,0xde,0x7c,0xf1,0x55,0x77,0xdf,
	0x7d,0xf5,0xdf,0x77,0xdf,0xfd,0xb7,0x5d,
	0x56,0xfd,0x2f,0xfe,0xff,0xfa,0xf7,0xcb,
	0xaf,0xfe,0xfc,0xfa,0xdf,0xbf,0xf7,0xdc,
	0x3a,0xff,0xbf,0x7f,0xfb,0xff,0xed,0xff,
	0xea,0x7b,0xfb,0x7f,0xff,0xfe,0x7b,0xff,
	0x2f,0xff,0xff,0xff,0xf6,0xff,0xf5,0xf9,
	0x47,0x1f,0x7f,0xfe,0x99,0xc7,0x1f,0x79,
	0xfc,0x91,0xe7,0x1f,0x79,0xf6,0xe7,0x9f,
	0x79,0xfc,0xb1,0xc7,0x9e,0x7f,0xec,0xb1,
	0xe7,0x1f,0x7f,0xee,0xf9,0x8b,0xff,0xff,
	0xff,0xfd,0xfb,0xef,0xaf,0xff,0xfe,0xfb,
	0xff,0xaf,0xff,0xfe,0xf7,0xff,0xff,0xfe,
	0xf2,0xef,0xaf,0xff,0xfe,0xfb,0xff,0xbf,
	0xbf,0xfe,0xff,0xef,0xbf,0xff,0xff,0xfd,
	0xff,0xfd,0xbf,0xf7,0xff,0xff,0xfd,0xff,
	0xd7,0x7f,0x7f,0xfd,0xfd,0xbf,0xde,0xff,
	0xbd,0x97,0xde,0x7f,0xfd,0xf7,0xdf,0xff,
	0xfb,0xfd,0xf5,0xff,0xfe,0x7f,0xff,0xff,
	0xe6,0xff,0xbf,0x7f,0xff,0xfb,0xeb,0xaf,
	0xbf,0xfe,0xfb,0xef,0xbf,0xbf,0xfe,0xff,
	0xef,0xb7,0xff,0xff,0xef,0xab,0xbf,0xfe,
	0xfb,0xff,0xaf,0xff,0xfe,0xff,0xeb,0xaf,
	0xff,0x4f,0xff,0x7f,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xfd,0xdf,0xff,0x7f,0xff,
	0xff,0xf7,0x7f,0xff,0xff,0xff,0x5f,0xff,
	0xfd,0xff,0xff,0x7f,0xff,0xff,0xff,0xff,
	0xdf,0xff,0x3f,0xf8,0xff,0xff,0xff,0xff,
	0xff,0xfb,0xfe,0xff,0xff,0xff,0xfb,0xf2,
	0xff,0xff,0xfe,0xff,0xcb,0xff,0xff,0xfb,
	0xef,0xcb,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xef,0xff,0x48,0xff,0x7f,0xfd,
	0xd5,0x17,0x5f,0x76,0xf1,0xf5,0x57,0x5f,
	0x75,0xfd,0xd5,0xd7,0x5e,0xeb,0xf1,0xd7,
	0x5f,0x7e,0xd9,0xe5,0x17,0x1f,0x7e,0xfd,
	0xf5,0xd7,0x5f,0x7f,0xfd,0x9f,0xfe,0xff,
	0xfc,0xff,0xce,0x3f,0xff,0xfc,0x33,0xcf,
	0x3f,0xfb,0xfc,0xf3,0xef,0x3e,0xef,0xf9,
	0xbf,0x3c,0xff,0x68,0xf3,0xcf,0xbf,0xfe,
	0xfc,0xf3,0xef,0x36,0xfb,0xfc,0x7f,0xe4,
	0xff,0xff,0xf9,0xff,0xbf,0x6f,0xff,0xfb,
	0xef,0xff,0xff,0xff,0xff,0x6f,0xff,0x5f,
	0xff,0xf7,0xdf,0xff,0xff,0xff,0xf7,0xff,
	0xfd,0xff,0xfe,0x6f,0xff,0xff,0xfe,0xff,
	0x8d,0xff,0xcf,0xff,0xff,0xdc,0xf3,0xcf,
	0x3d,0xff,0xdc,0xf3,0xcf,0x3f,0xff,0xdf,
	0xf3,0xcf,0xfe,0xdf,0x73,0xcf,0x3f,0xff,
	0xfc,0xff,0xcf,0x3f,0xfb,0xdf,0xf3,0xcf,
	0xff,0x9f,0xfd,0x7f,0xfc,0xff,0x97,0x5f,
	0xff,0xff,0xf5,0xff,0x1f,0x6f,0xff,0xf5,
	0xff,0x1f,0xff,0xee,0xbe,0x1f,0xfe,0xfd,
	0xf1,0xf7,0xff,0x7f,0xfd,0xf5,0xff,0x5f,
	0x7f,0xfc,0x7b,0xe9,0xff,0xf7,0xfb,0xff,
	0xfe,0xba,0xff,0xaf,0xef,0xbf,0xfd,0xef,
	0x5f,0xfb,0xa7,0xbe,0xfe,0xff,0xef,0xfd,
	0xf7,0xff,0xff,0xfd,0xff,0x5e,0x7f,0xfb,
	0xf7,0xf7,0xff,0xfe,0x07,0xff,0xff,0xf7,
	0xff,0x7d,0xff,0xff,0xff,0xdf,0xff,0xff,
	0xfe,0xff,0xff,0xff,0xff,0xdf,0xfd,0xf7,
	0xdb,0x7d,0x7f,0xdd,0xdf,0xff,0xdf,0x4b,
	0xff,0xcf,0xf7,0xdf,0xff,0x1f,0xfa,0xff,
	0xf7,0xdf,0xff,0xff,0xff,0xdf,0xfd,0xff,
	0xf7,0xff,0xff,0xff,0xff,0xff,0xff,0xfa,
	0xef,0xbf,0xfd,0xfe,0xdb,0xff,0xfd,0xfd,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xd1,
	0xf7,0xff,0xff,0xef,0xf7,0xff,0xff,0xff,
	0xff,0xff,0xfe,0xfe,0xff,0xdf,0xff,0xff,
	0xff,0xfb,0xf7,0xff,0xff,0xff,0xef,0xbf,
	0xff,0xfb,0xf7,0xf7,0xff,0xdf,0xfa,0xff,
	0xcb,0xbe,0x3f,0xfe,0xbf,0xfd,0xff,0xdf,
	0xff,0xff,0xfd,0xff,0xff,0x5d,0xff,0xff,
	0xcf,0xbb,0xef,0xff,0xfb,0x3f,0xff,0xff,
	0xff,0x7f,0xff,0xef,0xf8,0xfd,0xd7,0xbf,
	0xff,0x7f,0xf2,0xfd,0xef,0xfe,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x7f,
	0xff,0x7f,0xfe,0xfd,0xfb,0xff,0xff,0xf9,
	0xff,0xf7,0xff,0xff,0xff,0xef,0xff,0xef,
	0xfb,0xfd,0xff,0x8b,0xff,0xff,0xff,0xdd,
	0x7f,0xff,0xf7,0xff,0xdd,0xff,0xff,0xff,
	0xef,0xf5,0xaa,0x69,0xd6,0xaf,0xbb,0xed,
	0xaf,0xff,0xfd,0xb7,0xdf,0xfc,0xfb,0xfd,
	0x7f,0xff,0xff,0xff,0x2f,0xf8,0xff,0xff,
	0x7f,0xff,0xef,0xff,0xdd,0xff,0xff,0xff,
	0xbb,0xfd,0xff,0x9f,0x7f,0xff,0xf7,0xdf,
	0x3f,0x7f,0xfd,0xfe,0xd7,0xfa,0xff,0xff,
	0xef,0xff,0xf7,0xff,0xff,0xff,0xea,0xff,
	0xbf,0xff,0xfe,0xff,0xff,0xef,0xff,0xfe,
	0xff,0xdf,0xef,0xff,0xfb,0xde,0xfb,0xbf,
	0x3e,0xff,0xfb,0xff,0xff,0xff,0xff,0xff,
	0xff,0xfd,0xdf,0xff,0x7f,0xff,0xff,0x77,
	0xfc,0xff,0xff,0x37,0xff,0xff,0xff,0xff,
	0x9f,0x7f,0x7e,0xf3,0xe7,0xef,0xff,0xff,
	0xbf,0xfd,0xff,0xd7,0xff,0xff,0xff,0xff,
	0xff,0xeb,0xfd,0x9f,0xdf,0x5f,0x73,0xfd,
	0xef,0x5b,0xfd,0xff,0xbf,0xfc,0xff,0xff,
	0xff,0xff,0xfc,0xd7,0xfb,0x7f,0xfd,0xff,
	0xff,0xff,0xff,0xff,0xbf,0xff,0x7c,0xf2,
	0xff,0xff,0xcf,0xef,0xf3,0xf5,0xf3,0x5b,
	0xff,0xaf,0xd6,0xeb,0xbf,0xff,0xff,0xfb,
	0xef,0xbf,0xff,0xfe,0xfb,0xff,0xbf,0xff,
	0xff,0xfb,0xef,0x5f,0x2e,0xf9,0xed,0x9f,
	0x7f,0xfe,0xfb,0xef,0xff,0x7f,0xff,0xf7,
	0xff,0x7f,0x6b,0xaf,0xf6,0x7f,0xfb,0xff,
	0xb7,0xdf,0x7e,0xfb,0xed,0xb7,0xff,0x7f,
	0xfb,0xff,0xb7,0xdf,0xe6,0xe8,0xa3,0xcf,
	0x3e,0xfb,0xec,0xb7,0xdf,0xfe,0xff,0xfe,
	0xff,0xff,0xff,0xff,0x6f,0xf0,0xfd,0xef,
	0xff,0xfe,0xfa,0xeb,0xaf,0xbf,0xfe,0xfe,
	0xef,0xef,0xff,0xfe,0xfa,0xfb,0x93,0x5e,
	0x7a,0xe9,0xa5,0x97,0xee,0xfa,0xeb,0xbf,
	0xdf,0xec,0xfb,0xef,0xbf,0xff,0xa9,0xed,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xbf,
	0xf6,0xfa,0xeb,0xaf,0xbf,0x2e,0xff,0xff,
	0xff,0xbf,0x5b,0xff,0xff,0xff,0xff,0x3b,
	0xf9,0xff,0xff,0xff,0xfa,0xff,0xff,0xef,
	0xff,0xfa,0xff,0xaf,0xff,0x5f,0xfe,0xeb,
	0xff,0xf5,0xd2,0x5b,0x2f,0xfd,0xf5,0xff,
	0xff,0xff,0x9f,0x77,0x7e,0xeb,0xff,0x97,
	0x7f,0xeb,0xff,0xff,0xff,0xf4,0xfb,0xef,
	0xbf,0xff,0xf6,0xff,0x0f,0xff,0xff,0xfd,
	0xdb,0xff,0xa7,0xb7,0x8a,0x2a,0xab,0xaf,
	0xfe,0xfb,0x2f,0x3f,0xec,0xd4,0xd7,0xcf,
	0x7f,0xff,0x37,0xff,0xff,0xf6,0x7b,0xef,
	0xbf,0xfd,0xfe,0x5b,0x6e,0xbb,0xff,0xfe,
	0xfb,0x6f,0xbd,0xff,0x7b,0xef,0xbf,0xff,
	0xe6,0xfb,0xef,0xbf,0xfd,0xfe,0xfb,0xef,
	0xbd,0xff,0xfe,0x1f,0xf8,0xff,0x67,0x9c,
	0x7f,0xf7,0x5d,0xf7,0xdf,0x7d,0xe6,0xf9,
	0x67,0x9f,0x7f,0xc7,0xf9,0x9f,0x7f,0xfe,
	0xd9,0x67,0x9e,0x7d,0xf7,0x7d,0xe7,0x9f,
	0x7f,0xfe,0xf9,0xe7,0xff,0x97,0xff,0x5b,
	0x7f,0xfe,0x3d,0x76,0xdf,0x6f,0xa7,0xb5,
	0xe6,0x5f,0x7d,0xfc,0xfd,0xd6,0x6b,0xbd,
	0xfd,0x77,0xdf,0x6b,0xf7,0xdd,0xf6,0x9f,
	0x7f,0xfe,0xf1,0xe7,0x1f,0xff,0x17,0xf6,
	0xcf,0xff,0xff,0xfd,0xff,0xdf,0x6f,0xff,
	0xfd,0xf7,0xff,0xff,0xff,0xff,0xf3,0xdb,
	0xbf,0xff,0xff,0xff,0x7f,0xff,0xff,0xfb,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x6f,
	0xf1,0xbf,0xff,0x7f,0xfd,0xff,0x7f,0x7f,
	0xff,0xfd,0xef,0xf7,0xff,0xff,0xff,0xf7,
	0xdf,0xff,0xff,0xff,0xff,0xff,0xfe,0xff,
	0xff,0xff,0xdf,0x7f,0xff,0xff,0xf7,0xff,
	0xff,0xa7,0xfe,0x2d,0xfe,0xf1,0xe2,0x8f,
	0x2f,0xbe,0xf8,0xe3,0x0f,0x2f,0xf6,0xf1,
	0xe3,0x8f,0x9d,0xf8,0xe2,0x8f,0x2f,0xfe,
	0xf8,0xe2,0x8b,0x7f,0xfc,0xf5,0xc7,0x1f,
	0x7f,0xfc,0x1f,0xfd,0x7f,0xf5,0xdf,0x57,
	0xdf,0xfd,0xf5,0xd5,0x77,0xdf,0x7c,0xb5,
	0xdf,0x5f,0x5f,0xfd,0xd7,0x5f,0x79,0x75,
	0xd5,0x5d,0x57,0x5f,0xed,0xf7,0xdf,0x7f,
	0xff,0xfd,0xf7,0xff,0xe5,0xff,0xcf,0xfd,
	0xfa,0xff,0xfb,0xcb,0xef,0xf6,0xde,0xfb,
	0xff,0xff,0xfb,0xda,0xf3,0xa7,0xba,0xfa,
	0xeb,0xdf,0xbf,0xff,0xfd,0xeb,0xbf,0xfe,
	0xff,0xfb,0x77,0xbf,0xff,0x5b,0xff,0xdf,
	0x9f,0x7f,0xee,0xb1,0x67,0x9f,0x7d,0xec,
	0xb1,0xe7,0x9e,0x7b,0xfc,0xd1,0xe7,0x7d,
	0xf6,0xd9,0xe7,0x1f,0x7b,0xfe,0xf1,0xe7,
	0x9f,0x7f,0xfe,0xf8,0xe7,0x9f,0xbf,0xfa,
	0xff,0xff,0xdf,0xbf,0xbf,0xfe,0xfb,0xef,
	0xaf,0xff,0xf8,0xfb,0xdf,0xbf,0xff,0xff,
	0xef,0xaf,0xbf,0xfe,0xfb,0xeb,0xbf,0xbf,
	0xff,0xff,0xdf,0x7f,0xdf,0xff,0xf7,0xff,
	0xd3,0xff,0xff,0xff,0x7a,0xff,0xf5,0xff,
	0xdf,0x7f,0xfd,0xbd,0xdf,0xff,0xfb,0xfd,
	0xf7,0xff,0xff,0xfd,0xf5,0xdf,0x5f,0xff,
	0xfd,0xf7,0xff,0xfe,0xfa,0xeb,0xbf,0xbf,
	0xfe,0x6f,0xfe,0xff,0xfa,0xff,0xaf,0xff,
	0xfe,0xfa,0xef,0xbf,0xbf,0xff,0xfa,0xf7,
	0xaf,0xff,0x7e,0xeb,0xab,0xff,0xfe,0xfa,
	0xef,0xaf,0xff,0xfe,0xff,0xff,0xff,0xfd,
	0xff,0xff,0x7f,0xf0,0xff,0xfd,0xff,0xdf,
	0xff,0xff,0xff,0xff,0xff,0x7f,0xff,0xf7,
	0xff,0xdf,0xff,0xfd,0xff,0x7f,0xff,0xff,
	0xf7,0xff,0x7f,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0x9b,0xff,0xff,0xff,
	0xfb,0xef,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xef,0xff,0xff,0xff,0xff,
	0xcb,0xff,0xff,0xfb,0xef,0xff,0xfb,0xff,
	0xfb,0xff,0xff,0xff,0xfe,0x5f,0xf5,0xff,
	0xd7,0x5e,0x7f,0xfd,0xf5,0x97,0x5e,0x7d,
	0xf5,0xd1,0xd7,0x5f,0x7a,0xe5,0xf5,0x5e,
	0x7f,0xbd,0xa5,0x97,0x5f,0x7a,0xf5,0xf5,
	0xd7,0x5e,0x7f,0xfd,0xf1,0xd7,0xbe,0xef,
	0xfd,0xcf,0xbf,0xff,0xec,0xf3,0xcf,0x3f,
	0xfb,0xfc,0x67,0xcf,0x3f,0xff,0xfc,0xf3,
	0x3e,0xfd,0xe8,0xf3,0x8f,0x37,0xff,0xfc,
	0xd3,0xad,0xff,0xff,0xef,0xe7,0xff,0xff,
	0x4f,0xfe,0xff,0xff,0x7d,0xff,0xfd,0xef,
	0xdf,0xff,0xfe,0x6b,0xff,0xff,0xfe,0xfe,
	0xfe,0xff,0xaf,0xbf,0xfd,0xfc,0xff,0xff,
	0xff,0xff,0x7f,0xfb,0xff,0xff,0xef,0xff,
	0xff,0x9f,0xf8,0xff,0xfc,0xff,0xcf,0x3f,
	0xff,0xfc,0xf3,0xcf,0xed,0xfb,0xfc,0xef,
	0x4f,0x3f,0xff,0xf0,0xcb,0x3f,0xff,0xfc,
	0xf3,0xcf,0x3f,0xff,0xff,0xfe,0xff,0xff,
	0xff,0xff,0xff,0xdd,0xff,0xf7,0xff,0x7f,
	0xff,0xf1,0xd6,0x5f,0x7f,0x7c,0xef,0x87,
	0xff,0xff,0x7f,0xf5,0xff,0xeb,0x6f,0xf1,
	0xc7,0x1f,0x7f,0xfd,0xf7,0xff,0xfb,0xdf,
	0xff,0xff,0xfd,0xff,0xcf,0xfe,0xff,0xbe,
	0xff,0xdf,0xdf,0xff,0xfd,0x7b,0xf7,0xff,
	0x7e,0xff,0xff,0xfe,0x9f,0xfb,0xff,0xfe,
	0xff,0xff,0xff,0xff,0xdf,0xef,0xff,0xff,
	0xff,0xff,0xfe,0xff,0xef,0x7f,0xf0,0xff,
	0xff,0xff,0xfd,0xf7,0xff,0x7d,0xf7,0xfe,
	0xf7,0xdf,0x7f,0xff,0xeb,0xff,0x77,0xdd,
	0xfd,0x3e,0xfd,0xfd,0xf7,0xde,0xf7,0xff,
	0x7f,0xff,0xff,0xff,0xfd,0xf7,0xff,0xb3,
	0xff,0xff,0xff,0xf7,0xf7,0xdf,0xff,0xfd,
	0xff,0xf7,0xdf,0xff,0xff,0xf9,0xf7,0xff,
	0xaf,0xfd,0xff,0xdb,0x7f,0xff,0xff,0xf7,
	0xff,0xff,0xf7,0xff,0xff,0xda,0xef,0xff,
	0xbf,0x7c,0xff,0xff,0xff,0xfb,0xff,0xff,
	0xff,0xff,0xff,0xff,0xed,0xff,0x7f,0xff,
	0xff,0xfb,0xbe,0xff,0xff,0xff,0x3f,0xff,
	0xff,0xfb,0xef,0xdf,0xff,0xff,0xff,0xff,
	0xff,0xff,0xe5,0xfb,0xe3,0xf7,0x73,0xff,
	0xff,0xff,0x7f,0xff,0xff,0xff,0xff,0xaf,
	0x3f,0xfe,0xff,0xfb,0x7f,0xff,0xdc,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xef,
	0xfe,0xff,0xfe,0x17,0xdf,0xff,0xfe,0xff,
	0xef,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xfb,0xff,0xd7,0xff,0xf7,0xe5,
	0xff,0xff,0xff,0xff,0xff,0xff,0xf7,0xff,
	0xff,0xf7,0xff,0xff,0xbf,0xf9,0xff,0x77,
	0xff,0xff,0x9f,0x7e,0xdf,0xeb,0xbe,0xbb,
	0x6e,0xfb,0xff,0xaf,0xdf,0xfe,0xbd,0xff,
	0xef,0xfc,0xfa,0xff,0xfe,0xfb,0xef,0xdb,
	0xff,0xbf,0xff,0xff,0xfa,0xff,0x9f,0xff,
	0xff,0x7f,0xff,0xff,0xf7,0xdf,0x6f,0xff,
	0xfe,0xf7,0x9f,0xff,0xff,0x7d,0xf7,0xef,
	0xfb,0xef,0xf7,0xdf,0xff,0xfe,0xff,0xff,
	0xd7,0xff,0xff,0xfd,0xff,0xdf,0xff,0x77,
	0xfe,0xff,0xdf,0xeb,0xff,0xdf,0xdf,0xfe,
	0xfb,0xff,0xbf,0xbf,0xfe,0xef,0xf7,0xcf,
	0xff,0xff,0xff,0xaf,0x7f,0xff,0xff,0xff,
	0xff,0xbf,0xdf,0xcf,0x9f,0xff,0x7f,0xff,
	0x7f,0xc5,0xff,0xff,0xfa,0xff,0xa9,0xff,
	0xbf,0xff,0xf8,0x63,0xbf,0x7e,0x7e,0xff,
	0xab,0xff,0xf7,0xeb,0xfd,0x9f,0xff,0xfa,
	0xeb,0xaf,0xf7,0xff,0x7a,0xff,0xfb,0x8f,
	0xfd,0xff,0xbc,0xf5,0xff,0xd7,0xff,0x7f,
	0xfd,0xff,0xff,0xcf,0x3f,0xfb,0xf9,0xe7,
	0xfb,0x7f,0xfd,0x9f,0x5f,0xf7,0xff,0xff,
	0xff,0x9f,0x3f,0xff,0xff,0xe7,0xff,0xff,
	0xff,0xec,0xff,0xda,0xbd,0xff,0xfb,0xef,
	0x7f,0xff,0xfe,0xfb,0xef,0xbf,0xdf,0xfe,
	0xfb,0xff,0xff,0xff,0xfe,0xed,0xff,0x5f,
	0xfe,0xfb,0xef,0xff,0xff,0xfd,0xfb,0xdf,
	0xff,0x5f,0x7e,0xfb,0xf7,0x68,0xff,0xb7,
	0xdf,0xfe,0xff,0xed,0xb7,0xdf,0x7e,0xfb,
	0xec,0xb7,0xff,0xff,0xff,0xed,0xc7,0xfe,
	0xcf,0xe8,0xb7,0xff,0xfe,0xff,0xff,0xb7,
	0xff,0xff,0xcf,0xe8,0xb3,0xff,0x27,0xff,
	0xff,0xfa,0xeb,0xbf,0xbf,0xff,0xfa,0xeb,
	0xaf,0xb7,0xfe,0xfe,0xef,0xbf,0xbf,0x7f,
	0xe9,0xbf,0x97,0xfe,0xfe,0xef,0xbf,0xff,
	0xfe,0xfe,0xef,0xff,0xbf,0xfe,0xfa,0x7f,
	0xd9,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xbf,0xfb,0xff,0xff,0xff,0xff,
	0xff,0xeb,0xff,0xbf,0xfe,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xdf,
	0xbf,0x81,0xff,0xff,0xff,0xff,0xf7,0xff,
	0xff,0xff,0xff,0xff,0xfd,0xff,0xff,0xff,
	0x9f,0xff,0x5f,0xfb,0xef,0xf5,0xff,0xff,
	0xfd,0xb7,0x7f,0xfe,0xff,0xff,0xff,0xff,
	0xff,0xbf,0x4d,0xfe,0xff,0xfb,0xef,0x7f,
	0xff,0xff,0xcb,0xef,0xbf,0xef,0xfe,0xff,
	0xcf,0x3f,0xff,0x7f,0x6a,0x3f,0xad,0xfe,
	0xff,0xff,0x7f,0xf9,0xfc,0xff,0x0f,0xff,
	0xfd,0xfe,0xff,0x7f,0xf2,0xff,0xef,0xbf,
	0xff,0xfe,0xfb,0xef,0xbf,0xed,0xb6,0xdb,
	0xee,0xbf,0xff,0xfe,0xfb,0xbf,0xf7,0xde,
	0xfb,0xef,0xff,0xfe,0xde,0xfb,0xef,0xbf,
	0xff,0xdf,0xfb,0xef,0xff,0xa3,0xff,0x7f,
	0xf7,0xfd,0xe7,0x9f,0x7d,0xff,0xdd,0x77,
	0xdf,0x7d,0xfe,0xf9,0xe7,0x9f,0xff,0xd9,
	0xe7,0x9f,0x7f,0xf6,0xf3,0xe7,0x9f,0x7f,
	0xfe,0xf9,0xef,0x9f,0x7f,0xfe,0xbf,0xf9,
	0xff,0x7d,0xf6,0x1f,0x6f,0xa5,0xfd,0xf7,
	0xd8,0x77,0xdf,0xf5,0xe7,0x9f,0x6f,0xbd,
	0x76,0x9f,0x6f,0xff,0xd5,0xf7,0x1f,0x7f,
	0xae,0xf5,0xe7,0xdf,0x6f,0xff,0xf5,0x7e,
	0x6e,0xff,0xfd,0xf3,0xff,0x3f,0xff,0xff,
	0xff,0xdf,0xff,0xff,0xfd,0xff,0xff,0xef,
	0xff,0xfe,0xff,0x6f,0xff,0xff,0xff,0xfc,
	0xff,0xff,0xfd,0xff,0xff,0x3f,0xff,0xfd,
	0xfb,0x76,0xff,0xfd,0xf7,0xff,0x7f,0xff,
	0xff,0xff,0xdf,0xff,0xff,0xfd,0xff,0xf7,
	0xdf,0xff,0xff,0xff,0x5f,0xff,0xff,0xff,
	0xf7,0xff,0x7f,0xfd,0xff,0xf7,0x5f,0xff,
	0xfd,0xff,0xbf,0xeb,0xff,0xe2,0x8b,0x7d,
	0xbc,0xf8,0xe3,0x8b,0x2f,0xbe,0xd8,0xe3,
	0x1f,0x3f,0xf5,0xf8,0x8d,0x7f,0xfc,0xf8,
	0xe3,0xff,0x7d,0xfc,0xf5,0x63,0x5f,0xef,
	0xbf,0xd8,0xe2,0x7f,0xc9,0xff,0x57,0x7f,
	0xfd,0xf7,0xd7,0x5f,0x5f,0xfd,0xf5,0xd5,
	0x57,0xff,0xed,0xf1,0xd5,0x5f,0xe5,0xb7,
	0x55,0x5f,0xff,0xff,0xf7,0xdf,0x57,0xff,
	0x7d,0xf7,0x55,0x5f,0xff,0x3f,0xfe,0xff,
	0xfd,0xfa,0xdf,0xbf,0xff,0xfa,0xf3,0xef,
	0xff,0xff,0xde,0x37,0xef,0xbf,0xff,0xef,
	0xff,0x3f,0xff,0xde,0xff,0xff,0xfd,0xfb,
	0xfc,0xbf,0xfb,0x2f,0xbf,0xfe,0x7f,0xf3,
	0xff,0xff,0x99,0xe7,0x9f,0x7b,0xf6,0xd9,
	0xe7,0x9e,0x7b,0xe4,0xf9,0x67,0x1f,0x79,
	0xda,0xe7,0x1e,0x7d,0xe6,0xf9,0xe7,0x9e,
	0x7f,0xf4,0xf9,0xc7,0x1f,0x7d,0xe6,0xf9,
	0xb3,0xff,0xbf,0xff,0xfe,0xf7,0xeb,0xbf,
	0xbf,0xff,0xfa,0xef,0xff,0xff,0xff,0xf3,
	0xff,0xff,0xfe,0xe7,0xff,0xaf,0xff,0xff,
	0xf7,0x9f,0xff,0xff,0xf9,0xe7,0xef,0xbf,
	0xff,0xdf,0xfd,0x7f,0xff,0xff,0xff,0x7f,
	0x5f,0xff,0xfd,0xdf,0x7f,0xff,0xfd,0xff,
	0x9f,0x7f,0xff,0xf4,0xff,0x7f,0x7f,0xfd,
	0xff,0xff,0xff,0x7f,0xff,0xff,0xff,0xdf,
	0xff,0xff,0xff,0xe4,0xff,0xef,0xbf,0xff,
	0xf9,0xfb,0xef,0xff,0xff,0xfe,0xfb,0xff,
	0xff,0xfe,0xfa,0xff,0xff,0xff,0xfb,0xff,
	0xff,0xff,0xff,0xf9,0xef,0xff,0xff,0xfe,
	0xff,0xfb,0xef,0xff,0x3f,0xff,0xdf,0xfb,
	0xef,0xff,0xff,0xdc,0xfb,0xef,0xbf,0xdf,
	0x7e,0xfb,0xff,0xf7,0xdf,0xfe,0xed,0xff,
	0xdf,0xfe,0xfb,0xff,0xff,0xff,0xdf,0xfb,
	0xff,0xbf,0xf7,0xfe,0xfb,0x3f,0xfb,0xff,
	0xff,0xff,0xff,0xff,0xff,0xef,0xbe,0xff,
	0xbf,0xf8,0xff,0xff,0xff,0xff,0xff,0xd7,
	0xff,0xfe,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0x7f,0xff,0xff,0xff,0xe4,0xff,0x58,
	0xff,0x7f,0xfd,0xf5,0xc7,0x5f,0x7e,0xe9,
	0xe5,0xd7,0x5d,0x7f,0xfd,0xf5,0x57,0x5f,
	0xea,0xf5,0x57,0x0f,0x7f,0xfd,0xe5,0x57,
	0x5f,0x7f,0xfd,0xf5,0xd7,0x5f,0x77,0xfd,
	0xcf,0xfe,0xff,0xdc,0xf3,0x3f,0x3d,0xda,
	0xfc,0xb3,0xcd,0x3f,0xff,0xfc,0x7b,0xac,
	0x3d,0xff,0xe3,0xff,0x3e,0xff,0xf8,0xf3,
	0xae,0xbf,0xfb,0xfc,0xe3,0xbd,0x3f,0xff,
	0xdc,0xff,0xe1,0xff,0xff,0xfa,0xff,0xaf,
	0xff,0xff,0xff,0xff,0xbf,0xff,0x7f,0xff,
	0x7d,0xbb,0xff,0xff,0xfd,0xf7,0xff,0xff,
	0xed,0x7f,0xbf,0xfd,0xff,0xde,0xff,0xff,
	0xff,0xff,0xfb,0xa5,0xff,0xcf,0x3f,0xf7,
	0xff,0xf3,0xcd,0x2f,0xdf,0xdc,0x73,0xcf,
	0x7f,0xff,0xff,0xf3,0x2b,0xdf,0xff,0xf3,
	0xcf,0x7f,0xff,0xff,0xff,0xcf,0xbf,0xff,
	0xff,0xf3,0xcf,0xff,0xff,0x7c,0x7f,0xff,
	0xfd,0xff,0xff,0xff,0xed,0xf1,0xfb,0x5f,
	0xff,0xfd,0xdf,0xff,0x5d,0xef,0xff,0xfd,
	0x5f,0x7f,0xfd,0xef,0xff,0xff,0x7f,0x9c,
	0xff,0xdf,0x5f,0x7f,0xff,0x7f,0xe6,0xfb,
	0xef,0x7f,0xd9,0xff,0xff,0xfa,0xff,0xff,
	0xff,0xf7,0xff,0xff,0xff,0xf7,0xf7,0xff,
	0xfb,0xff,0xf6,0xdf,0xff,0xff,0xdf,0xff,
	0xf7,0xff,0xff,0xff,0xfa,0xef,0xff,0x57,
	0xff,0xff,0xdf,0xff,0x7f,0xff,0xff,0x7d,
	0xff,0xff,0xf6,0xdd,0xf7,0xab,0xff,0xf7,
	0x9d,0xff,0xdf,0xff,0xdb,0xfb,0xf7,0xff,
	0xff,0xff,0xf5,0xfb,0xbd,0xff,0xee,0xfd,
	0x5f,0xfb,0xf6,0xff,0xff,0xfb,0xff,0xf7,
	0xff,0xff,0xff,0xff,0xff,0xf7,0xff,0xff,
	0xfd,0xfd,0xfe,0x7f,0xff,0xff,0xff,0xff,
	0xff,0xff,0x5f,0xf7,0xff,0xff,0xff,0xdf,
	0xff,0xff,0xcf,0xf7,0x7f,0xff,0xf7,0xf7,
	0xff,0xff,0xff,0xbf,0xf7,0xff,0xff,0xfb,
	0xef,0xff,0xde,0xff,0xff,0xef,0xff,0xff,
	0xff,0xef,0xff,0xff,0xff,0xff,0xff,0x3f,
	0xff,0xff,0xff,0xbe,0xbe,0x3f,0xff,0xff,
	0xdf,0xff,0xff,0xef,0xff,0xff,0xff,0x3f,
	0xff,0xfc,0xf3,0x8e,0xbf,0xdf,0xf3,0xce,
	0x3f,0xff,0xbc,0xf3,0xff,0xdb,0x7f,0xfd,
	0xf1,0xce,0x7f,0xff,0xff,0xf5,0xfd,0xfb,
	0xff,0xff,0xdf,0x1f,0xff,0xff,0xff,0xf7,
	0xff,0x7b,0xd7,0xbf,0xfd,0xeb,0xfd,0x5e,
	0xff,0xfe,0xf9,0xef,0xbf,0xff,0xff,0xff,
	0xef,0xbf,0xff,0xfe,0xef,0xff,0xbb,0xff,
	0xdf,0xff,0xbe,0xff,0xdf,0x7f,0xd3,0xfd,
	0xff,0xfd,0x8f,0xff,0xff,0xf1,0xff,0xdf,
	0xfe,0xff,0xe3,0xbf,0x7f,0xfe,0xf9,0xfe,
	0xbf,0xff,0xff,0xf5,0xe7,0xff,0xf7,0xdf,
	0xf9,0xff,0xff,0xfb,0xd7,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xfd,0xef,0xdf,0xff,
	0xff,0xff,0xff,0x7e,0xff,0xfd,0xbf,0x5f,
	0xee,0x7f,0xfd,0xff,0xff,0x7f,0xff,0xff,
	0xf7,0xe5,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xfe,0xfb,0x7f,0xff,0xef,0xff,0x3f,
	0xdf,0xff,0xef,0xff,0xff,0xfd,0xeb,0xef,
	0x3f,0xff,0xdf,0xbb,0xff,0xff,0xff,0xea,
	0xff,0xff,0x1f,0xfc,0xff,0xaf,0xff,0xff,
	0xfc,0xf9,0xe5,0xf7,0xdf,0x7f,0xf3,0xe7,
	0xef,0xff,0x7f,0xeb,0xff,0xdf,0xff,0xff,
	0xe3,0xff,0xfd,0xff,0xff,0xff,0xff,0xff,
	0xff,0xf3,0xfd,0x0f,0x5a,0xfd,0x7f,0xfd,
	0xff,0xf3,0x5f,0x77,0xde,0xfc,0xfe,0xfd,
	0x7f,0xfd,0xff,0xff,0xff,0xfd,0xff,0xfe,
	0xff,0x3f,0xff,0xef,0xd7,0x5f,0xff,0xff,
	0xff,0xe7,0x5f,0xff,0xaf,0xdb,0xeb,0xff,
	0xff,0xfe,0xfb,0xef,0xff,0xff,0xff,0xff,
	0xff,0xff,0x59,0xff,0xff,0xdf,0xff,0xfe,
	0xff,0xff,0xbf,0xff,0x7f,0xff,0xef,0xbf,
	0xff,0xfe,0xff,0xff,0x7f,0x7f,0x7f,0xf6,
	0x7f,0xff,0xed,0xb7,0xdf,0xfe,0xff,0xff,
	0xff,0xff,0x7f,0xdf,0xfa,0xf7,0xff,0x7f,
	0xed,0xff,0xdf,0x7f,0xfb,0xfd,0xf3,0xdf,
	0x7e,0xfb,0xed,0xf7,0xdf,0xff,0xff,0x6f,
	0xf2,0xff,0xaf,0xbf,0xff,0xfa,0xeb,0xbf,
	0xff,0xfe,0xfb,0xef,0xef,0xfd,0xfe,0xfa,
	0xef,0xbf,0xff,0xfb,0xfb,0xaf,0xbf,0xdf,
	0xfa,0xeb,0xef,0xbf,0xff,0xfa,0xfb,0xbf,
	0xff,0xbf,0xed,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xfb,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xee,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xdb,0xf8,0xff,0xbf,0xfe,0xff,
	0xfb,0xff,0xe7,0xff,0xff,0xfa,0xaf,0xff,
	0xf5,0xff,0xff,0xff,0xff,0x7f,0xff,0xff,
	0xff,0xff,0xf7,0xfb,0xff,0xff,0xff,0xff,
	0xfb,0xff,0xbf,0x7a,0xe8,0xff,0xbf,0xfd,
	0xff,0x9b,0xef,0x7f,0xff,0xf4,0x13,0x4f,
	0xff,0xaf,0xd4,0xfb,0xcf,0xf7,0xff,0xd7,
	0xff,0xbf,0xff,0xbf,0xbb,0xef,0xff,0xff,
	0xff,0x1b,0xff,0x3f,0xff,0x2f,0xff,0xff,
	0xd6,0xfb,0xef,0xbd,0xf9,0xfe,0x7b,0xef,
	0xbf,0xf7,0xfe,0xfb,0x6f,0xbf,0xff,0xdb,
	0xef,0xbd,0xed,0xb6,0xfb,0xef,0xbf,0xf9,
	0xfe,0x9b,0x6f,0xbb,0xff,0xfe,0x7f,0xf8,
	0xff,0x77,0x9c,0x7f,0xff,0xfd,0xe7,0x9f,
	0x7f,0xfe,0xf9,0xe7,0x9f,0x7f,0xdf,0xf9,
	0x9f,0x71,0xfe,0x99,0x77,0x9e,0x7f,0xff,
	0x9d,0xe7,0x9f,0x79,0xe7,0xd9,0xe7,0xff,
	0x8b,0xff,0xdb,0x6f,0xbd,0xfd,0xf6,0x1f,
	0x7f,0xfe,0xf9,0xe7,0x5a,0x7f,0xfe,0xfd,
	0xe7,0x6f,0xfd,0xf1,0xd7,0xdf,0x6b,0xbd,
	0xed,0xf7,0x5b,0x7f,0xbd,0xfd,0xd7,0x9d,
	0xff,0x87,0xf6,0xcf,0x3f,0xff,0xfe,0xf3,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x7f,
	0xff,0xfe,0xf7,0xdb,0x7f,0xbf,0xfd,0xff,
	0xff,0x7f,0x6f,0xf7,0xbf,0x7f,0xff,0xff,
	0xf7,0xff,0xff,0x7f,0xff,0xfd,0xf7,0xff,
	0x7f,0xff,0xff,0xf7,0xff,0xff,0xff,0xff,
	0x7f,0xff,0xff,0xdf,0xbf,0x7f,0xff,0xfb,
	0xff,0xff,0xdf,0xfd,0x97,0xfe,0x3f,0xfe,
	0xf8,0xe3,0x8b,0x7f,0xf4,0xd1,0x42,0x1f,
	0x3f,0xfe,0xf1,0xf2,0x1f,0xdd,0xf8,0x47,
	0xcf,0x3f,0xf6,0xf8,0xe2,0x8f,0x2f,0xfe,
	0xf8,0xe2,0x8f,0x7f,0xfc,0xd7,0xfd,0x7f,
	0xf7,0xd5,0x77,0x5f,0xfd,0xf7,0xdf,0x16,
	0xff,0x7d,0xb7,0xdf,0x57,0xff,0xfd,0xdd,
	0x7e,0x7f,0xfd,0xf5,0xd7,0x5f,0xdf,0x7d,
	0xf7,0xdd,0x57,0x7b,0xfd,0xf7,0xff,0xe9,
	0xff,0xab,0xbd,0xff,0xfa,0x7b,0xff,0xfd,
	0xba,0xfe,0xaf,0xef,0xff,0xff,0xdf,0xf7,
	0xbf,0xff,0xed,0xfb,0xcf,0xbf,0xbf,0xea,
	0xbb,0xeb,0x3f,0xff,0xfa,0x7b,0xdf,0xff,
	0x0f,0xff,0x5f,0x1f,0x79,0xfc,0xb1,0xe7,
	0x9e,0x7f,0xe4,0xf8,0xc7,0x9e,0x7b,0xfe,
	0xf9,0xa7,0x79,0xfe,0xd9,0x67,0x9f,0x79,
	0xf6,0xb1,0xc7,0x1e,0x7f,0xfc,0x99,0xe7,
	0x9f,0xff,0xf9,0xff,0xfb,0xff,0xbf,0xbf,
	0xff,0xf7,0xff,0x7f,0xde,0xfd,0xfb,0xdf,
	0xbf,0xff,0xfd,0xef,0x7f,0xbe,0xfe,0xfa,
	0xef,0xbf,0xbf,0xfe,0xfb,0xeb,0xff,0xbf,
	0xfe,0xe7,0xff,0xc3,0xff,0xf7,0x7f,0xff,
	0xfd,0xf7,0xff,0xff,0xff,0xbf,0xff,0xdf,
	0xff,0xfe,0xfd,0xff,0xcf,0xff,0xff,0xf5,
	0xd7,0xdf,0xff,0xff,0xf5,0xff,0x5f,0xff,
	0xff,0xf5,0xff,0xff,0x5f,0xfe,0xff,0xff,
	0xff,0xef,0xff,0xff,0xf9,0xef,0xbf,0x7f,
	0xfe,0xfe,0xef,0xef,0x7f,0xfe,0xfb,0xbf,
	0xff,0xff,0xff,0xfb,0xef,0xff,0xff,0xfe,
	0xff,0xef,0xff,0xff,0xfb,0x7f,0xb6,0xff,
	0xbf,0xdf,0x7e,0xfb,0xef,0xff,0xff,0xff,
	0xef,0xff,0xb7,0xff,0x7f,0xfb,0xff,0xdf,
	0xfe,0xff,0xef,0xbf,0xf7,0xfe,0xfb,0xef,
	0xbf,0xff,0xde,0xfb,0xef,0xff,0xff,0xaf,
	0xff,0xff,0xff,0xff,0xef,0xff,0xfb,0xfe,
	0xbf,0xef,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xcf,0xff,0xff,0xfe,0xff,0xf2,
	0xff,0xff,0xff,0xbf,0xff,0xbf,0xff,0xf7,
	0xaf,0xf5,0xff,0xd7,0x5f,0x7e,0xf8,0xf5,
	0xd7,0x5f,0x7d,0xfd,0xb1,0xd7,0x5f,0x7f,
	0xfd,0xd5,0x5e,0x7f,0xfd,0xf5,0x57,0x5f,
	0x7f,0xed,0xf5,0xd7,0x5f,0x7d,0xfd,0xf5,
	0xd7,0xff,0xe6,0xff,0xff,0x3d,0xff,0xf4,
	0xb3,0xdf,0xff,0xff,0xf7,0xf3,0xbf,0xff,
	0xf7,0xfc,0x63,0x3f,0xff,0xfe,0xf3,0xcf,
	0xff,0x7f,0xdf,0xf3,0x8f,0x3f,0xff,0xff,
	0xff,0xfd,0xff,0x16,0xfe,0xe7,0xdf,0x7f,
	0xfe,0xfd,0xa5,0xdf,0xff,0xff,0xdd,0xff,
	0xff,0xff,0xf2,0xeb,0xed,0xdf,0xf6,0xcf,
	0xff,0xbf,0xff,0xff,0xff,0xf7,0xff,0xff,
	0xff,0xff,0xff,0xbf,0x1f,0xfb,0xff,0xff,
	0xf3,0xcf,0x3f,0xff,0xef,0xff,0xff,0xbf,
	0xf7,0xff,0xff,0xff,0x6f,0xeb,0xf0,0xff,
	0xff,0xff,0xfc,0xff,0xff,0x3f,0xbf,0xec,
	0xf3,0xff,0xce,0xff,0xff,0xff,0xd1,0xf7,
	0xf7,0x5f,0x6f,0xff,0xbd,0xff,0xff,0x7b,
	0xbd,0xff,0xf7,0xff,0x7f,0xfd,0xff,0xff,
	0xff,0xff,0xf7,0xc6,0x5f,0xef,0xfd,0xf5,
	0xc7,0x5b,0x7b,0xfd,0xf5,0xfd,0xff,0x27,
	0xbe,0xff,0xfe,0xf7,0xdf,0x7e,0xef,0xf7,
	0xff,0xe7,0xfe,0xfb,0xae,0xff,0xdf,0xff,
	0xef,0xff,0xfe,0xef,0xff,0xff,0xe7,0xfb,
	0x7b,0x7f,0xff,0x77,0xe7,0x7e,0xff,0xff,
	0x5f,0xf3,0xff,0xff,0xf7,0xfd,0xff,0xdf,
	0x7f,0xff,0xff,0xff,0xf7,0xff,0xff,0xd9,
	0xeb,0xff,0xf9,0xff,0xff,0xff,0x7d,0xb6,
	0xff,0x6f,0xf7,0x7f,0xc7,0x7f,0x6d,0xdf,
	0xff,0xff,0xb9,0xef,0xff,0xff,0xff,0x9f,
	0xff,0x7f,0xff,0xfd,0xff,0xff,0xf7,0xff,
	0xff,0xff,0xff,0xef,0xff,0x5f,0xff,0xfe,
	0xff,0x5f,0xdf,0xff,0xf7,0xaf,0xff,0xfe,
	0xff,0xff,0xff,0xdf,0xfd,0xff,0xf7,0xff,
	0xff,0xef,0xf9,0xff,0xfe,0xff,0xff,0xff,
	0xff,0xdf,0xfe,0xff,0xff,0xbe,0xff,0xfe,
	0xff,0xff,0x7f,0x7f,0xff,0xff,0xff,0xfd,
	0xb7,0xfb,0xff,0x3f,0xff,0xe4,0xff,0xff,
	0xcf,0x1f,0xff,0xbc,0xf1,0xfa,0x7b,0xeb,
	0xbf,0xf2,0xfd,0x3f,0x67,0xff,0xfb,0x7f,
	0xef,0xfc,0xf7,0xff,0xf7,0xff,0xfc,0xf3,
	0xff,0xff,0xef,0xff,0xff,0xff,0x7f,0xff,
	0xff,0xff,0xfe,0xf9,0xe7,0x9b,0xff,0xff,
	0xf7,0xfd,0x9f,0xff,0xef,0xfb,0xff,0xdf,
	0xff,0xf7,0xef,0xff,0xfe,0xff,0xff,0xef,
	0x9b,0xff,0xbf,0xff,0xff,0xff,0xff,0xff,
	0xf9,0xff,0xff,0xc7,0xaf,0xff,0xfd,0xff,
	0xff,0xfb,0xff,0xfd,0xfb,0xff,0x9f,0xff,
	0xfc,0xff,0xff,0x7b,0xbc,0x7f,0xff,0xbf,
	0x7f,0xfe,0xfb,0xef,0xff,0xff,0xff,0xff,
	0xff,0x94,0xff,0xff,0x7f,0xff,0xfd,0xff,
	0xff,0xff,0xff,0x3f,0xff,0xcf,0x7d,0xff,
	0xfc,0xf7,0xff,0xff,0xeb,0xb3,0xfd,0xff,
	0xef,0xbc,0xf5,0xcf,0xbf,0xbd,0xff,0xff,
	0xff,0xf7,0xef,0xfe,0xff,0xff,0xfa,0xef,
	0xff,0xff,0xfb,0xef,0xff,0xff,0xfa,0xfe,
	0xed,0xf7,0xaf,0xff,0xff,0xff,0xdf,0xff,
	0xf7,0xff,0xef,0xee,0x7f,0xdf,0xaf,0xff,
	0xff,0xff,0xff,0x7f,0xd3,0xff,0xff,0xff,
	0xff,0xff,0xa7,0xfc,0x7f,0xf9,0xe5,0xf7,
	0xff,0xf4,0xff,0x7f,0x3f,0xff,0xf9,0xfd,
	0x9f,0x9f,0xff,0xff,0xff,0xff,0xfd,0xf7,
	0xff,0xff,0xff,0xff,0xff,0xa0,0xf5,0xff,
	0xff,0xff,0xff,0xbf,0xec,0xf3,0x9b,0x77,
	0xbd,0xff,0xb3,0xff,0xff,0xfb,0xd9,0xcf,
	0xef,0xff,0x79,0xf3,0xcf,0xff,0xff,0xef,
	0x17,0xfe,0x3f,0xff,0xff,0xff,0x5a,0xbc,
	0xff,0xfb,0x7f,0x96,0xfe,0x7f,0xff,0xff,
	0xff,0xff,0xff,0xfb,0xed,0xff,0xdf,0xfe,
	0xef,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0x7f,0xfb,0xe5,0xfa,0xff,0xff,0xff,0xf7,
	0x6e,0xff,0xb7,0xdf,0x37,0xfa,0xff,0xfb,
	0xff,0xff,0xff,0xff,0xb7,0xcf,0x7e,0xff,
	0xec,0xd7,0xfe,0xff,0xfd,0xff,0xff,0xff,
	0xff,0xfd,0xb3,0x8f,0xfe,0xff,0xff,0xff,
	0xff,0x1f,0xfe,0xff,0xfa,0x6b,0xaf,0xff,
	0xde,0xfb,0xef,0xbf,0xff,0xfe,0x7e,0xeb,
	0xef,0xb7,0xfe,0xeb,0xbf,0xbf,0xfe,0xfb,
	0xef,0xbf,0xbf,0xdf,0xfe,0xeb,0xfd,0xfa,
	0xfe,0xff,0x3f,0xd9,0x5f,0xff,0xff,0xff,
	0xfe,0xff,0xee,0xff,0xff,0xff,0xff,0xff,
	0xbb,0xff,0xbf,0xfb,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xee,0xbf,0xff,
	0xde,0xff,0xff,0xbf,0x89,0xff,0xff,0xfb,
	0x7f,0xfd,0xbf,0xf2,0xf9,0xf9,0xff,0xbf,
	0xfa,0xdf,0xff,0xff,0xfd,0xff,0xff,0xef,
	0xff,0x7e,0xeb,0xab,0xb7,0xff,0xf6,0xff,
	0xfb,0x7f,0xff,0xff,0xff,0xbd,0xfe,0xff,
	0x8b,0xef,0xab,0xff,0xb0,0xd3,0x5f,0x3d,
	0xff,0xf0,0xff,0xee,0xff,0xef,0x56,0xef,
	0x3f,0xff,0xe2,0xc7,0x4f,0x7f,0xf9,0x87,
	0xff,0xef,0xff,0xff,0xfc,0xff,0x7f,0xe7,
	0xff,0x6f,0xbb,0xed,0xe6,0xfb,0xef,0xbd,
	0xff,0xfe,0xfb,0x6f,0xbb,0xff,0xfe,0xfb,
	0xbf,0xed,0xfe,0xfb,0xef,0xbf,0xf7,0xde,
	0xdb,0xee,0xbf,0xff,0xff,0xfb,0xff,0xff,
	0xbb,0xff,0x7f,0xf7,0xdd,0xf7,0x9f,0x7f,
	0xfe,0xf9,0xe7,0x9f,0x7f,0xe6,0xfd,0xe7,
	0xdf,0xff,0xfd,0xe7,0xdf,0x7f,0xfe,0xf9,
	0xe7,0x9f,0x79,0xf6,0xfd,0x7f,0x9e,0xff,
	0xf7,0xdf,0xf9,0xbf,0x7d,0xf7,0xdd,0x7f,
	0xfe,0xf9,0xc7,0x9f,0x7f,0xfe,0xf5,0xb7,
	0x5b,0x6b,0xaf,0xf6,0x9f,0x6f,0xff,0xf1,
	0xe7,0x1f,0x7f,0xfd,0x75,0xf6,0xdf,0x7f,
	0xbe,0xdd,0x7e,0x65,0xbf,0xfd,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xfd,
	0xff,0xef,0x7f,0xff,0xfd,0xff,0xef,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xfe,0xdf,
	0xff,0xff,0xfe,0xf6,0x66,0xff,0xfb,0xff,
	0xff,0xff,0x7f,0xff,0xfd,0xff,0xdf,0x7f,
	0xfb,0xff,0xff,0xff,0xfd,0xfb,0xff,0xdf,
	0xff,0xff,0xff,0xf7,0xff,0xff,0xff,0xff,
	0xdf,0xff,0x7f,0xff,0xdf,0xff,0xf9,0xff,
	0xe3,0xcf,0x2f,0xf6,0xf1,0xc7,0x1f,0x7f,
	0xf5,0xf1,0xe3,0x8f,0x3f,0xff,0xf8,0x89,
	0x7f,0xfc,0xf8,0x47,0x1f,0x3d,0xbc,0xf8,
	0xe3,0x8b,0xff,0xff,0xf5,0xfa,0x7f,0xc3,
	0xff,0x77,0x7f,0x7d,0xf5,0xdf,0x7f,0xff,
	0xed,0xf7,0xdf,0x77,0x7f,0x7d,0xf5,0xdd,
	0x5f,0xfd,0xf7,0xdd,0x7e,0xff,0xbd,0xb1,
	0xd5,0x5e,0x7f,0xfd,0xfb,0xdf,0x97,0xff,
	0x0f,0xfe,0xff,0xfa,0xfb,0xdf,0x7f,0xff,
	0xfd,0xff,0xff,0xfe,0xff,0xfc,0xfb,0xef,
	0xbf,0xfe,0x3e,0xff,0xbf,0xfe,0xff,0xbf,
	0xcf,0xee,0xfe,0xfc,0xfa,0xff,0xfd,0xfb,
	0xfe,0x7f,0xf4,0xff,0xfd,0xd9,0xe7,0x9f,
	0x7f,0xfe,0xb9,0xe5,0x9f,0x7b,0xf6,0x99,
	0x47,0x1f,0x7f,0x9a,0xe7,0x1e,0x7f,0xee,
	0xb9,0x67,0x1f,0x7f,0xf6,0x99,0x67,0x9f,
	0x7f,0xfc,0xf9,0xa3,0xff,0xaf,0xbf,0xfe,
	0xfb,0x9f,0xff,0xff,0xf9,0xff,0x9f,0xbf,
	0xff,0xfe,0xff,0xef,0xff,0xfe,0xf7,0xef,
	0x7f,0xfe,0xfd,0xfb,0xef,0xaf,0xbf,0xfe,
	0xff,0xff,0xaf,0xff,0x1f,0xfd,0x7f,0xfd,
	0xf5,0xdf,0xff,0xff,0xff,0x7f,0xfd,0xff,
	0x7f,0xff,0xfd,0xdf,0x7f,0xff,0xfc,0xbf,
	0x7f,0xdf,0xff,0xff,0xdf,0x7f,0x7f,0xfd,
	0xf7,0xff,0xff,0x7f,0xfd,0xff,0xe9,0xff,
	0xff,0xff,0xff,0xfe,0xef,0xbf,0xff,0xfe,
	0xfb,0xef,0xef,0xbf,0xff,0xff,0xfb,0xff,
	0xff,0xfb,0xfb,0xbf,0x7f,0xfe,0xfa,0xff,
	0xff,0xaf,0xff,0xff,0xef,0xef,0xff,0x07,
	0xff,0xff,0xfb,0xef,0xb7,0xff,0xff,0xff,
	0xdf,0xff,0xff,0x7f,0x7b,0xef,0xb7,0xdf,
	0xfe,0xef,0xff,0xdf,0xfc,0xff,0xff,0xf7,
	0xff,0xfe,0xfb,0xed,0xff,0xff,0xdf,0xff,
	0x7d,0xd8,0xff,0xef,0xff,0xff,0xff,0xff,
	0xf2,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0x3f,0xf9,0xff,0xff,0xff,0xfb,0xff,0xff,
	0xff,0xfe,0xff,0xf2,0xff,0x2f,0xff,0xff,
	0xff,0xcf,0xd6,0xff,0x7f,0xfd,0xf4,0xd7,
	0x1f,0x77,0xfd,0xf4,0xc7,0x5f,0x7f,0xfd,
	0xf4,0x97,0x5e,0xef,0xf5,0x57,0x5f,0x7b,
	0xf5,0xf5,0xd3,0x5f,0x77,0xfd,0xd5,0x56,
	0x5f,0x7d,0xf5,0x9e,0xfe,0xff,0xfc,0xf3,
	0xcf,0x3e,0xfc,0xfb,0xf3,0xbf,0x7f,0xff,
	0xf8,0xf3,0xcf,0x1f,0xfd,0xf3,0xed,0xfe,
	0xf7,0xff,0xf7,0x7f,0xff,0xff,0xfc,0xf3,
	0xff,0xbf,0xff,0xff,0x7f,0x65,0x7f,0xff,
	0xe9,0xff,0xff,0xfd,0xff,0xdf,0xff,0xff,
	0xde,0xfe,0xff,0x3f,0xff,0xff,0xfa,0x7f,
	0xff,0xff,0xff,0xef,0xef,0xbf,0x7f,0xff,
	0xff,0xf7,0xbf,0x7d,0xff,0xfb,0x9d,0xff,
	0xcf,0xff,0xfe,0xfc,0xfb,0xff,0xbf,0xff,
	0xff,0xfd,0xcf,0x3f,0xff,0xff,0x73,0x2f,
	0xff,0xff,0xff,0xff,0xff,0x7f,0xff,0xbe,
	0xcf,0x3c,0xff,0xdf,0xff,0xff,0xff,0xdb,
	0x9d,0x7f,0xff,0xfd,0xf6,0xfb,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xf5,0xf6,0x5b,
	0xdf,0xf6,0xff,0x5f,0xff,0xbf,0xff,0xff,
	0x7f,0x7f,0xbd,0xff,0xf6,0xfb,0xe7,0xff,
	0x7f,0xc6,0xff,0xde,0x7f,0xff,0xfe,0xbf,
	0x7e,0xff,0xff,0xff,0xff,0xff,0x7e,0xfd,
	0x6d,0x97,0xff,0xef,0xff,0xf6,0xff,0xff,
	0xff,0xff,0xff,0xdf,0xff,0xff,0xfd,0xff,
	0xbf,0xff,0x57,0xfe,0xff,0xff,0xff,0x7f,
	0x7f,0xff,0xff,0xdf,0xdf,0xff,0xff,0x6f,
	0xbf,0xdf,0xf7,0x1f,0xff,0x7f,0xde,0xfe,
	0xff,0xff,0xff,0xf7,0xdb,0xff,0xff,0xaf,
	0xbf,0xff,0xff,0xcf,0xf2,0xff,0xff,0x7f,
	0xfd,0xfb,0xfd,0xff,0xdf,0xfb,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xfe,0xff,0x1b,
	0xfe,0xff,0xff,0xf3,0xff,0xff,0xf7,0xbb,
	0xaf,0xbf,0xdf,0xfe,0xff,0xcf,0xff,0xff,
	0xff,0xfd,0xef,0xff,0xfd,0xff,0xf7,0xff,
	0xff,0xff,0xff,0xeb,0xef,0xff,0xff,0xef,
	0xef,0x7f,0xff,0xfe,0xff,0xef,0xff,0xfd,
	0x7f,0xff,0xff,0xf7,0xdf,0xfb,0x4d,0xfc,
	0x7f,0xff,0xf7,0xf7,0xf7,0xdf,0xef,0xf7,
	0xff,0xc7,0x5b,0x7f,0xfe,0xe3,0xff,0xbf,
	0xff,0xff,0xfa,0x3d,0xef,0xbc,0xff,0xff,
	0x3f,0xfe,0x3f,0xbf,0xff,0xff,0xfb,0xfe,
	0xf0,0xff,0xef,0xfd,0x66,0xff,0xff,0xff,
	0xff,0xff,0x7f,0xfe,0xf7,0xf6,0x7f,0xf7,
	0xff,0xfd,0xff,0xff,0xff,0xfb,0xef,0xfd,
	0xff,0xff,0xfb,0xff,0xff,0xff,0xcf,0xff,
	0xff,0x87,0xff,0xff,0xb7,0xfa,0xff,0xff,
	0x7f,0xff,0xfb,0xf7,0xeb,0xf7,0xdf,0xdf,
	0x73,0xff,0xfd,0xdf,0xfb,0xff,0x9f,0x7b,
	0x7e,0xff,0xeb,0xfb,0xff,0x7b,0xfb,0xff,
	0xff,0xfb,0x9f,0xfc,0xff,0xff,0xf5,0xff,
	0xff,0xff,0xff,0xff,0xff,0x76,0xfb,0xbf,
	0xff,0xd6,0xff,0xff,0xfe,0xdf,0xff,0xff,
	0xfd,0xf3,0xff,0x7f,0xef,0xbf,0xb7,0x5f,
	0x7b,0xff,0xfd,0xff,0xea,0xff,0xff,0xff,
	0xf7,0xff,0xff,0xff,0xff,0xff,0xf7,0xeb,
	0xfd,0xff,0xbf,0xfb,0xff,0xfd,0xbf,0xfa,
	0xff,0xeb,0xbf,0xff,0xff,0xfd,0xff,0xbf,
	0xbf,0xfe,0xab,0xef,0xff,0x4f,0xfd,0xff,
	0xcd,0x3f,0x9f,0xff,0xf9,0x29,0xaf,0xde,
	0xff,0xeb,0xff,0xff,0xff,0x7f,0xff,0xff,
	0xff,0xff,0xff,0xcf,0x3f,0xdf,0xf7,0xff,
	0x7c,0xff,0xff,0xff,0xff,0xfe,0x8f,0x5a,
	0xff,0x2f,0xff,0x75,0xff,0x5f,0x6f,0xf9,
	0xfc,0xff,0xcf,0xff,0xff,0xff,0xd7,0xfd,
	0xfd,0xff,0xf3,0xff,0x7f,0xfe,0xfc,0xbe,
	0xff,0xf3,0xfb,0xff,0xff,0xff,0xfd,0xaf,
	0xc5,0xfb,0xff,0xff,0xfe,0xff,0xef,0xff,
	0xdf,0xfe,0xf7,0xef,0xff,0xff,0xfe,0xfb,
	0xff,0xff,0xfe,0xfb,0xff,0xff,0xff,0xff,
	0xff,0xfd,0xff,0xf9,0xff,0xff,0xbf,0xfc,
	0x7f,0xef,0xf6,0xff,0xff,0xed,0xff,0xdf,
	0xfe,0xff,0xec,0xff,0xdf,0xfe,0xff,0xed,
	0xb7,0xff,0x7f,0xed,0xb7,0xff,0x7f,0xff,
	0xfd,0xff,0xef,0xf3,0x9f,0xff,0xff,0xff,
	0xff,0xff,0x7f,0xe2,0xfb,0xbe,0xbf,0xfe,
	0xfb,0xfb,0xbf,0xb7,0xfe,0xfb,0xeb,0xbf,
	0xbf,0xfe,0xfa,0xef,0xbf,0xff,0x7e,0xed,
	0xef,0xbf,0xfe,0x7b,0xff,0xff,0xff,0xff,
	0x2f,0xff,0x7f,0xfb,0x97,0xfd,0xf2,0xff,
	0xff,0xff,0xff,0xff,0xbf,0xfb,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xab,0xff,0xff,0xff,0xff,0xbb,0xbf,0xd6,
	0xfb,0xab,0xbd,0xff,0xd6,0xff,0xf9,0xff,
	0xb7,0xfe,0xff,0xff,0xef,0x9f,0xfd,0x7f,
	0xf8,0xff,0xbf,0xfe,0xff,0xff,0xe7,0xff,
	0xff,0x5b,0xff,0xff,0xff,0xff,0xdf,0xff,
	0xff,0xff,0xff,0xdf,0xff,0xaf,0xfe,0xe2,
	0xff,0x7f,0xfc,0xfe,0xd3,0x7f,0x3c,0xed,
	0xfe,0xf7,0xef,0x3f,0xfd,0xfe,0xfb,0x4f,
	0xf7,0xf3,0x8f,0x0a,0xff,0xff,0xfe,0xf3,
	0xfe,0xff,0xff,0xff,0xfe,0xff,0xff,0xfc,
	0x4f,0xfe,0xff,0xfe,0x9b,0xef,0xbd,0xf9,
	0xde,0xfb,0xef,0xbf,0xff,0xde,0xfb,0x6f,
	0xbe,0xf7,0xdb,0x6f,0xbe,0xff,0xfe,0xfb,
	0xef,0xff,0xff,0xb7,0xff,0xff,0xff,0xed,
	0xe7,0x7f,0xf9,0xff,0xe7,0xdf,0x7f,0xfe,
	0x99,0xe7,0xdf,0x7d,0xfe,0xfd,0xe7,0xdf,
	0x7f,0xff,0xf9,0x9f,0x77,0xe6,0xf9,0xe7,
	0xdf,0x7d,0xfe,0xff,0x7f,0xff,0xff,0xf7,
	0xdf,0xff,0xff,0x8f,0xff,0x1f,0x6f,0xff,
	0xf9,0xd6,0x9f,0x6b,0xf3,0xf1,0xf6,0x9f,
	0x7f,0xbf,0xfd,0xe7,0x6b,0xad,0xf5,0xa7,
	0x5a,0x7f,0xdf,0xf9,0xf7,0xda,0x6d,0xfb,
	0x7d,0xf7,0xd9,0xee,0xc7,0xf6,0xff,0x3f,
	0xff,0xff,0xf6,0xff,0x7f,0xff,0xff,0xf6,
	0xff,0x7f,0xbf,0xfd,0xff,0xef,0xff,0xfd,
	0xff,0xdf,0xff,0xff,0xff,0xff,0xdf,0x6f,
	0xff,0xff,0xff,0xcf,0xbf,0x6f,0xf6,0xff,
	0x7f,0xff,0xff,0xed,0xff,0x5f,0xff,0xff,
	0xef,0xff,0x5f,0xff,0xfb,0xff,0xf7,0xff,
	0xf7,0xff,0x77,0xff,0xff,0xff,0xfd,0x7f,
	0x7f,0xff,0xff,0xff,0xdf,0xff,0xff,0x8b,
	0xff,0x7d,0xbc,0xf8,0xc7,0x8b,0x7f,0xbc,
	0xd8,0xc7,0x8f,0x7f,0xfc,0xf8,0xe2,0x1f,
	0x9f,0xf8,0xe2,0x1f,0x3d,0xff,0xd8,0xc2,
	0xeb,0xff,0xbf,0xff,0xff,0xff,0xf7,0xff,
	0x9f,0xfc,0xff,0xf7,0xd7,0x7f,0x5b,0xfd,
	0xf7,0xd5,0x7f,0x5f,0xfd,0xf7,0xdd,0x77,
	0xfb,0xfd,0xd5,0x5f,0xff,0x7d,0xf5,0xd7,
	0x17,0x5f,0xfe,0xfb,0xef,0xbb,0xef,0xfe,
	0xfb,0xff,0xef,0xff,0xdf,0xbe,0xff,0xfb,
	0x6f,0xff,0xef,0xff,0xfb,0x33,0xdf,0x3e,
	0xfb,0xfc,0xbf,0xbf,0xfe,0xfa,0xff,0xef,
	0xbf,0xbb,0x7e,0xfa,0xef,0xee,0xff,0xff,
	0xfb,0xff,0xff,0x1f,0xef,0xff,0x9f,0x7b,
	0xfe,0xf9,0xe7,0x9e,0x79,0xfe,0xd1,0xe7,
	0x1f,0x7f,0xfe,0xb9,0x27,0x7f,0xfe,0xb9,
	0x47,0x9f,0x79,0xe4,0xd1,0xc7,0x9f,0x7f,
	0xfe,0xd9,0xe7,0x9f,0x5f,0xf8,0xff,0xe7,
	0xeb,0xff,0xff,0xfe,0xe7,0xef,0x7f,0xff,
	0xff,0xff,0xef,0xaf,0xff,0xf9,0xfb,0xaf,
	0xff,0xfd,0xff,0xeb,0x7f,0xff,0xfe,0xff,
	0xff,0xff,0xff,0xff,0xfe,0xff,0xd9,0xff,
	0xff,0x7f,0xff,0xff,0xfd,0xff,0xdf,0xff,
	0xff,0xfd,0xff,0x7f,0x7f,0xff,0xff,0x6f,
	0xff,0xfd,0xff,0xdf,0x5f,0xff,0xfb,0xf7,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x6f,
	0xfe,0xff,0xfb,0xfb,0xbf,0xbf,0xff,0xfb,
	0xff,0x9f,0xff,0xff,0xfb,0xfb,0xef,0xff,
	0xfe,0xff,0xef,0x7f,0xfe,0xff,0xff,0xbf,
	0xbf,0xff,0xff,0xff,0xfb,0xff,0xff,0xff,
	0xbf,0xf5,0xff,0xff,0xff,0xfe,0x7f,0xef,
	0xff,0xff,0xfe,0x7f,0xef,0xff,0xdf,0xfe,
	0xfb,0xff,0xff,0xfe,0xfb,0xff,0xb7,0xff,
	0xfe,0xff,0xfd,0xff,0xff,0xdf,0xff,0xff,
	0xff,0xff,0x9f,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xbd,0xfc,0xff,0xff,0xff,0xff,
	0xff,0xff,0xfe,0xff,0xff,0x7f,0xff,0xfb,
	0xfe,0xbf,0xff,0xfb,0xef,0xbf,0xdf,0xfe,
	0xdf,0x2f,0xff,0x1d,0xfc,0xff,0xd7,0x5f,
	0x7f,0xf5,0xf5,0x57,0x5f,0x7f,0xf5,0xf5,
	0x57,0x5e,0x7b,0xf1,0xf1,0x5e,0x3c,0xfd,
	0xf5,0xd7,0x5f,0x7b,0xbc,0x75,0xd7,0x5d,
	0x75,0xfd,0xd5,0xd6,0xdf,0xef,0xff,0xef,
	0x3f,0xf7,0xff,0xb3,0xaf,0x1d,0xff,0xf9,
	0xf3,0xff,0x3f,0xff,0xfc,0x6f,0x3f,0xfd,
	0xdc,0xf3,0xcf,0xff,0xff,0xf8,0xff,0xff,
	0xff,0xdf,0xff,0xbe,0xfe,0xfb,0x07,0xde,
	0x7f,0xdf,0xff,0xfe,0xfb,0x6f,0xbf,0xff,
	0xea,0xfb,0xf7,0xff,0xff,0xff,0xfb,0xef,
	0xff,0x7e,0xdf,0xd7,0xff,0x7f,0xd5,0x7f,
	0xff,0xef,0x77,0xef,0xff,0xff,0xbf,0xd7,
	0xe8,0xff,0xff,0xf3,0xff,0x1f,0x7f,0xff,
	0xf1,0xf7,0x3f,0xff,0xff,0x32,0xcf,0xff,
	0xff,0xf0,0xcf,0xbf,0xff,0xbb,0xff,0xff,
	0xff,0x7f,0xcf,0xff,0xfe,0xff,0xfb,0xff,
	0x77,0xc1,0xff,0xff,0xff,0xdf,0xbf,0xf1,
	0xfb,0x9f,0xf7,0xff,0xf7,0xff,0xdf,0xff,
	0xaf,0xff,0x0f,0xdf,0xff,0xff,0xc7,0x5b,
	0x7f,0xed,0xdf,0xd7,0x5f,0x6f,0xed,0xe7,
	0xf3,0xfb,0xf7,0xfc,0xef,0xff,0xff,0xfd,
	0xdb,0xff,0xff,0xff,0xff,0xdf,0xef,0xff,
	0xf7,0xff,0xff,0xef,0xfd,0xff,0xff,0x7f,
	0xe7,0x77,0xaf,0xff,0xef,0xbe,0xfb,0xef,
	0xef,0xff,0xfd,0xff,0xe1,0xfe,0xff,0xff,
	0xfd,0xff,0xdf,0x7f,0xff,0xff,0xff,0xff,
	0xff,0xff,0xf9,0xfe,0x9f,0xf9,0xff,0xf7,
	0xff,0xff,0x77,0xdf,0xff,0xfd,0xbd,0xf7,
	0xdd,0xfe,0xff,0xff,0xff,0x39,0xfb,0xff,
	0xef,0xb7,0xbf,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xf7,0xff,0x7f,0xdf,0xef,0xb7,
	0xbf,0xff,0xfd,0xff,0xff,0xff,0xff,0xed,
	0xb7,0xfe,0xfe,0xab,0x7f,0xff,0xff,0xfc,
	0xef,0xff,0xde,0xfe,0xff,0xdf,0xf7,0xff,
	0xff,0xff,0xaf,0xff,0xbf,0xff,0xff,0xff,
	0xbf,0xff,0xff,0xfb,0xff,0xff,0xff,0xfe,
	0xfa,0xfd,0xff,0xdf,0xff,0xbf,0xef,0xff,
	0x85,0xff,0xf3,0x8f,0x3f,0xfd,0xfc,0xdf,
	0xc6,0x3f,0xef,0xff,0xf3,0xcf,0xaf,0x6f,
	0xfd,0xfb,0x3f,0xfe,0xfc,0xd3,0xfe,0xff,
	0xfd,0xff,0xfd,0xf7,0xd9,0x7f,0x7e,0xff,
	0xfd,0x47,0xff,0xbf,0xef,0xee,0xfb,0xef,
	0xfb,0xff,0xfd,0xfb,0xfe,0xbf,0xdf,0xfe,
	0xfd,0xbf,0xdf,0xdf,0xfb,0xef,0xbf,0xff,
	0xff,0xff,0xff,0xff,0xf5,0xff,0xff,0xf7,
	0xf7,0xbf,0xbf,0xfa,0xff,0xf0,0xff,0x95,
	0x73,0xef,0x9f,0xd7,0x9f,0xfd,0xf9,0xf9,
	0xd7,0xbf,0xed,0xd9,0xef,0xff,0x7f,0xfe,
	0xfd,0xff,0xff,0xfe,0xf7,0xef,0xef,0xff,
	0xdd,0xfe,0x6f,0xbe,0xd1,0xff,0xd5,0xff,
	0xff,0xf8,0xf7,0xd3,0xff,0xff,0xfd,0x7e,
	0xcf,0x7f,0xaf,0xbd,0xff,0xbe,0xff,0xfd,
	0xf7,0xd7,0xff,0x3f,0xfd,0xff,0xdf,0x7f,
	0xff,0xfd,0xb7,0xff,0xfb,0x5f,0xfe,0xff,
	0xdf,0xff,0xef,0xdf,0xff,0xff,0xdf,0xf3,
	0xff,0x77,0xdf,0xfd,0xeb,0xfd,0xff,0x6f,
	0xaf,0xcf,0xfa,0xff,0xff,0xff,0xf7,0xfb,
	0x6e,0xda,0xef,0xaf,0xff,0xff,0xff,0x77,
	0xfe,0xec,0xbf,0xff,0xfe,0xff,0xff,0xff,
	0xff,0xfe,0xff,0xff,0x8f,0xff,0xff,0xf9,
	0xbf,0xbf,0xfe,0xff,0xef,0xff,0xff,0xff,
	0xff,0xfb,0xef,0xff,0xfe,0xfb,0xdb,0xef,
	0x9e,0xff,0xff,0xff,0xfb,0xf6,0xff,0xef,
	0xff,0xff,0xef,0xdf,0xff,0xbe,0xfd,0xe7,
	0xdf,0xff,0xff,0xfe,0xbf,0x7f,0xbf,0xfb,
	0xf7,0xdb,0xff,0xfe,0xff,0xee,0x9f,0x7f,
	0xbb,0x7f,0xf4,0xff,0xff,0x9f,0xfc,0xff,
	0xe7,0xff,0xdf,0xff,0xea,0xff,0xff,0xfd,
	0xf7,0xff,0xf9,0xfe,0xff,0xf8,0xff,0xbf,
	0xde,0xff,0xff,0xff,0x9f,0x7f,0x3e,0xde,
	0xf8,0xff,0x3f,0x6a,0xfd,0xff,0xfc,0xe7,
	0xff,0x3f,0xff,0xfc,0xfe,0x9f,0x7f,0xfd,
	0xef,0xbf,0xff,0xef,0xff,0xff,0xcf,0xff,
	0xff,0xff,0xfe,0xff,0xff,0xf7,0xff,0xe7,
	0xff,0xf7,0xff,0xbf,0x4e,0xef,0xff,0xff,
	0xf7,0xff,0xff,0xff,0xfe,0xff,0xff,0xff,
	0xff,0x7f,0xff,0xed,0xff,0xff,0xfb,0xef,
	0xff,0xff,0xff,0xff,0xff,0xfc,0xf3,0xff,
	0xef,0x5f,0xfe,0xeb,0xff,0x3d,0xda,0xff,
	0xfd,0xff,0xdf,0x7f,0xff,0xed,0xff,0xdf,
	0xff,0xff,0xfd,0xf3,0xcf,0xfe,0xff,0xb7,
	0xdf,0x7e,0xff,0xfd,0xff,0xff,0xe7,0xdf,
	0x7f,0xff,0xb9,0xff,0xff,0xfb,0xbf,0x89,
	0xff,0xbf,0xfe,0xfb,0xfb,0xef,0xbf,0xfe,
	0xfb,0xeb,0xbf,0xbf,0xdf,0x7a,0xeb,0xbf,
	0xff,0xfe,0xeb,0xef,0xbf,0xfe,0xfb,0xff,
	0xff,0xfd,0xf7,0xff,0xbf,0xfe,0xff,0xdf,
	0xbf,0xf6,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xee,0xbb,
	0xff,0x7f,0xff,0xff,0xff,0xff,0xff,0xff,
	0xdf,0xff,0xfe,0x7b,0xaf,0xbd,0xf7,0xda,
	0xfb,0xef,0xe5,0xff,0xff,0xff,0xeb,0xff,
	0xff,0xff,0xff,0xff,0xff,0xb7,0xff,0xf7,
	0xdf,0xff,0xb7,0xff,0xff,0xff,0xff,0xff,
	0x7f,0xf8,0xfe,0xfd,0xff,0xfd,0xff,0xdf,
	0xff,0xff,0xff,0xad,0xff,0xff,0xfe,0xd3,
	0xff,0xff,0xff,0xfe,0xf3,0xef,0x7f,0xf1,
	0xbf,0xfb,0xee,0x7f,0xdb,0x5f,0xef,0xff,
	0xff,0xfe,0xc7,0xff,0xef,0xff,0xff,0xfe,
	0xff,0xff,0xbf,0xff,0x7f,0xf8,0xf7,0xfb,
	0xef,0xbd,0xff,0xfe,0xdb,0xee,0xbf,0xff,
	0xfe,0xfb,0xef,0xbf,0xed,0xfe,0xef,0xbf,
	0xff,0xfe,0xfb,0xef,0xff,0xff,0xff,0x9f,
	0x7f,0xfb,0xff,0xb7,0xdf,0xf8,0xe3,0x7f,
	0xdd,0x7d,0xfe,0xf9,0xe7,0xdf,0x7f,0xfe,
	0xdd,0xe7,0x9f,0x7d,0xff,0x9d,0xe7,0x7f,
	0xf6,0xfd,0xe7,0xdf,0x7f,0xfe,0xfd,0xff,
	0xff,0xff,0xff,0x9f,0xff,0xff,0xdd,0x37,
	0xfe,0x7d,0xdf,0xb9,0xd6,0x5f,0x7f,0xff,
	0xf9,0x77,0x1f,0x7f,0xdd,0xfd,0xb7,0x1f,
	0xbf,0x75,0xf7,0x5b,0x6f,0xff,0xf1,0xa6,
	0xdf,0x7f,0xaf,0xfd,0xb6,0xdb,0x7f,0xdb,
	0x17,0xd8,0xff,0xff,0xff,0xf7,0xdf,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xfd,0xf7,
	0xff,0xff,0xff,0xf7,0xdb,0x6f,0xff,0xff,
	0xf3,0xff,0x7f,0xff,0xfd,0xfe,0xfb,0x7f,
	0xff,0xbf,0xc7,0xff,0xff,0xff,0xdd,0xdf,
	0xff,0xff,0xff,0xfd,0xff,0xff,0xff,0xf7,
	0xf7,0xff,0xff,0xef,0xf7,0xbf,0xff,0xfe,
	0xff,0xdf,0xff,0xff,0xfd,0xfd,0xff,0xff,
	0x7f,0xff,0xff,0x0f,0xfe,0xbf,0xfc,0xc7,
	0x8f,0x3d,0xff,0xf8,0xc7,0xcb,0x7d,0xf4,
	0xf8,0xf3,0x8f,0x7f,0xf4,0xe2,0x8f,0x3d,
	0xfe,0xfc,0xc7,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0x7f,0xf1,0xff,0xd5,
	0x7f,0x7f,0xfd,0xf5,0xd7,0x7f,0x5f,0xfd,
	0xf7,0xd7,0x5f,0x5f,0xed,0xf7,0x57,0x5f,
	0xfd,0xf5,0xdd,0x7f,0x6f,0xfe,0xfb,0xef,
	0xbf,0xff,0xfe,0xfb,0xef,0xfb,0x9f,0xff,
	0xff,0xf7,0xfb,0xfb,0xef,0xbf,0xff,0xef,
	0xfb,0xff,0xbf,0xff,0xfe,0xb7,0xbf,0xff,
	0xfb,0xf3,0xef,0xbf,0xf7,0xff,0xff,0xff,
	0xff,0xf7,0xff,0xff,0xfd,0xff,0xff,0x7f,
	0xfd,0xff,0x7f,0xfe,0x99,0x67,0x9f,0x79,
	0xee,0xd9,0xe7,0x9e,0x79,0xf6,0xf1,0xe7,
	0x9f,0xfe,0xd1,0x67,0x1e,0x7f,0xee,0xd9,
	0xe7,0x9f,0x7f,0xfe,0xf9,0x67,0x9f,0x7f,
	0x76,0xea,0xff,0xef,0x7f,0xff,0xfe,0xfb,
	0xeb,0x7f,0xff,0xff,0xf7,0xeb,0xbf,0xff,
	0xfc,0xf7,0xbf,0xff,0xff,0xfb,0xef,0x7f,
	0xfe,0xff,0xfb,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0x57,0xff,0x7f,0xfb,0xfb,0xbd,
	0xf7,0x5e,0xfb,0xfb,0xbf,0xbf,0x5f,0x7b,
	0xef,0xb5,0xbf,0xff,0xed,0xbd,0xf7,0x7e,
	0xfb,0xff,0xf7,0xff,0x7f,0xff,0xff,0xff,
	0xff,0xff,0xff,0x3f,0xf9,0xff,0xfb,0xff,
	0xbf,0xff,0xfe,0xff,0xff,0xbf,0xff,0xff,
	0xff,0xef,0x7f,0xff,0xff,0xed,0xff,0xff,
	0xfe,0xfb,0xbf,0xff,0xff,0xfe,0xff,0xef,
	0xbf,0xff,0xff,0xfe,0xff,0xdd,0xff,0xdf,
	0xff,0x7b,0xff,0xfd,0xff,0xff,0xfb,0xfd,
	0xbf,0xff,0xdf,0x7f,0xfb,0xbf,0x7f,0x7f,
	0xff,0xfd,0xdf,0xff,0xff,0xfd,0xff,0xff,
	0xff,0xfe,0xff,0xff,0xdf,0xff,0x3f,0xfe,
	0xef,0xbb,0xef,0x7f,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xf2,0xbf,0xff,0x6f,
	0xff,0xff,0xfd,0xff,0xfb,0xff,0xbf,0x4b,
	0x3f,0xf1,0xf4,0x8b,0x6b,0xbf,0xfc,0x37,
	0xf3,0xff,0x5f,0x7f,0xf1,0xd5,0x57,0x5f,
	0x7f,0xf5,0xb5,0x57,0x5f,0x77,0xfd,0xf5,
	0x97,0x7f,0xfd,0xf5,0x57,0x5f,0x7f,0xed,
	0xf5,0x53,0x5b,0x6f,0xbc,0xb5,0xc7,0x5a,
	0xfb,0x8b,0xff,0xff,0xff,0xfb,0x7f,0xcf,
	0x3e,0xdb,0xfe,0xf3,0xbf,0xff,0xdf,0xfc,
	0xf3,0xdf,0xff,0xec,0x73,0xff,0x3f,0xff,
	0xfc,0xdf,0xf7,0xff,0xfb,0xff,0xff,0xf3,
	0xff,0xff,0xbf,0xf8,0xbf,0x7f,0x7e,0xfb,
	0x2f,0xff,0xff,0xf7,0xee,0xfd,0xff,0xff,
	0xeb,0xff,0xbf,0xff,0xff,0xff,0xff,0x7f,
	0xff,0xef,0xfd,0xf7,0xf7,0xff,0xff,0xfd,
	0xf7,0xdf,0xff,0x7f,0xe9,0xff,0xbf,0xff,
	0xfb,0xff,0xff,0xf3,0xff,0xff,0xfe,0xff,
	0xff,0xbf,0x3f,0xf7,0xff,0xcb,0x3f,0xff,
	0xff,0xef,0xcf,0xff,0xff,0xef,0x7d,0xff,
	0xfb,0xfb,0xff,0xff,0xff,0x1f,0xfe,0x5b,
	0xbe,0xff,0xfd,0xc7,0xdf,0xff,0xff,0xf1,
	0xff,0x5f,0x7f,0xff,0xff,0xff,0x74,0xfd,
	0xfd,0xd7,0x5f,0xef,0xbf,0xef,0xc7,0x77,
	0xbe,0xbd,0xbd,0xd7,0xdd,0xff,0xef,0xfa,
	0xbf,0xfa,0xbf,0xbf,0x6f,0xff,0xbb,0xff,
	0xef,0xff,0xef,0xf7,0xef,0xfe,0xfd,0xf7,
	0xdd,0x7f,0xff,0xfd,0xf6,0xff,0xfe,0xff,
	0xff,0xdd,0xf7,0x7f,0xfd,0xbe,0xf7,0xff,
	0x99,0xff,0xff,0xff,0xd7,0xd7,0x9f,0xfd,
	0xff,0xf7,0xe7,0xff,0xb6,0xff,0xff,0xbf,
	0x7f,0xc7,0xb7,0xbf,0xfd,0x7f,0xff,0xff,
	0xdf,0xf1,0xcf,0x3f,0xff,0xff,0xfd,0xff,
	0xf6,0x5f,0xfe,0xbf,0xd6,0xff,0x7f,0xff,
	0xfb,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xfd,0xbf,0xff,0xff,0xfe,0xdf,0xff,
	0xdf,0xff,0xff,0xdf,0xfb,0xfa,0x7b,0xff,
	0xbf,0xff,0xff,0xf5,0xbf,0xff,0x7f,0xff,
	0xff,0xef,0xbf,0xff,0xff,0xbd,0x3f,0xff,
	0x7f,0xee,0xfb,0xff,0xff,0xff,0xff,0xef,
	0xff,0xff,0xfb,0xfb,0xff,0xff,0xff,0x7f,
	0xff,0xff,0xf7,0xff,0x97,0xff,0xce,0xbb,
	0xff,0xb8,0xff,0x8c,0xff,0xff,0xfe,0xff,
	0xef,0xd7,0xff,0x3a,0xf5,0x67,0xed,0xff,
	0xff,0xef,0x35,0xfe,0xff,0xb7,0x7f,0x3f,
	0xec,0xe8,0xff,0x7f,0xfe,0x1d,0xf5,0xff,
	0xfe,0xfd,0xec,0xff,0xff,0xfe,0xfe,0xf7,
	0xff,0x7f,0xff,0x7f,0xff,0xbf,0xff,0xef,
	0xff,0xff,0x7f,0xff,0xef,0xff,0xbf,0xef,
	0xf7,0xd7,0xee,0xf3,0xff,0xff,0xf7,0xe4,
	0xff,0xe5,0xfe,0xf9,0xd9,0x33,0x9c,0xfc,
	0xff,0xfe,0xfb,0x7f,0xaf,0xf6,0xef,0x5a,
	0x2d,0xff,0xfb,0xff,0xff,0xff,0xf7,0xff,
	0xf7,0x7f,0xfd,0xfd,0xff,0x7f,0xfe,0x77,
	0x17,0xff,0x7f,0xff,0xff,0xaf,0xcf,0xfe,
	0xff,0xfb,0xf6,0xc7,0xfb,0xff,0xfd,0xfe,
	0x97,0x7f,0xed,0xff,0xfd,0x7f,0xff,0xbf,
	0xff,0x7b,0xfe,0xff,0xf7,0xf7,0xff,0x7f,
	0xfd,0x5b,0xfa,0xdf,0xfa,0xff,0xff,0xff,
	0xfe,0x7f,0xbf,0xcf,0xff,0xff,0xff,0xf7,
	0xff,0xfe,0xff,0xff,0xff,0xff,0xff,0xfb,
	0xff,0xf7,0xfb,0xef,0xff,0xff,0xbf,0xf7,
	0xff,0xfb,0xff,0x4d,0xff,0xff,0xff,0xab,
	0x37,0xff,0xff,0xeb,0xff,0x3f,0xff,0xff,
	0xff,0xe7,0xf7,0xff,0xff,0x2d,0x3f,0xff,
	0xff,0xf3,0xff,0xff,0xfd,0xbf,0xff,0xff,
	0xf3,0xff,0xff,0xff,0x43,0xd6,0xff,0xff,
	0x7f,0xfe,0xf5,0xf3,0x5f,0xff,0xff,0xff,
	0xff,0xff,0x3f,0xff,0xff,0x7f,0x6f,0xfa,
	0xff,0xff,0xcf,0x3f,0xff,0xef,0x3f,0xff,
	0xdc,0xb7,0xff,0xff,0xff,0x6b,0xf4,0xfe,
	0xbd,0xff,0xff,0xfd,0xfb,0xff,0xbf,0xff,
	0xfe,0xfb,0xef,0xbf,0xff,0x7d,0xfd,0xff,
	0xdf,0xff,0xff,0xef,0xff,0xff,0x7f,0xff,
	0xb7,0xdf,0xf3,0xff,0xff,0xff,0xdf,0x8f,
	0xfd,0xf9,0xff,0xff,0xff,0xf7,0xff,0x7f,
	0xfb,0xed,0xd7,0xdf,0x7e,0xff,0x7f,0xe3,
	0xdf,0xff,0xfe,0xff,0xdf,0xfe,0xff,0xff,
	0xfb,0xef,0xbf,0xdf,0xff,0xff,0xff,0xff,
	0xff,0xf9,0x37,0xa9,0xbe,0xff,0xfe,0xf8,
	0xef,0xef,0xbf,0xfe,0xfe,0xfb,0x8f,0xff,
	0xf6,0xfe,0xbf,0xf7,0xfe,0xff,0xeb,0xbf,
	0xff,0x5e,0x7b,0xfd,0x35,0xf9,0xef,0xff,
	0xff,0xff,0x6c,0xff,0x6b,0xf6,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xef,0xff,0xbf,0xfb,0xff,0xff,
	0xff,0xff,0xff,0xea,0xab,0xaf,0xfe,0x2b,
	0xff,0xaf,0xff,0x57,0xfe,0x7f,0xff,0xf7,
	0x3f,0xfe,0xfb,0xff,0xff,0xff,0xff,0xfb,
	0xff,0xbf,0xff,0xd2,0xbb,0x9f,0xfd,0xff,
	0xff,0xff,0xff,0xdf,0xd7,0x5f,0x7f,0xfd,
	0xff,0xff,0xff,0xff,0x7f,0xf8,0xff,0xf2,
	0xff,0xff,0xfc,0x8b,0xcf,0xff,0xff,0xfe,
	0x8f,0xff,0xbf,0xf8,0xb4,0x9e,0x3f,0xef,
	0xfe,0xff,0xef,0xbf,0xff,0xbf,0xfa,0xfa,
	0xab,0xff,0xff,0xff,0xfb,0xff,0x91,0x67,
	0xbb,0xfe,0xfb,0xfb,0xef,0xbf,0xff,0xb6,
	0xfb,0xef,0xbf,0xff,0xb6,0x7b,0x6f,0xed,
	0xfe,0xfb,0xff,0xbf,0xff,0xfe,0xff,0x6f,
	0xff,0xfd,0xf6,0x9f,0xff,0xff,0xed,0xaf,
	0xb6,0xff,0xfb,0xef,0x9f,0x7f,0xff,0xf9,
	0x67,0xdf,0x7f,0xfe,0xf9,0x77,0x9f,0x7f,
	0xde,0xe7,0x9f,0xff,0xfe,0xfd,0xe7,0xff,
	0x7f,0xde,0x7f,0xe7,0xfd,0xff,0xff,0xdf,
	0x7f,0xe3,0xff,0xb7,0xdf,0x7f,0xbe,0xfd,
	0xe7,0x5a,0x77,0xaf,0xf5,0xd6,0xdb,0x77,
	0xfe,0xe5,0x9f,0x7f,0xfa,0xfd,0xf7,0x9f,
	0x7f,0xbf,0xf9,0xf7,0x9b,0x7f,0xff,0xfd,
	0xf7,0xfd,0x95,0xfd,0xff,0xfd,0xf7,0xbf,
	0xfd,0xff,0xff,0x7f,0xff,0xfd,0xf3,0xfb,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xf7,
	0xff,0xff,0xff,0xfc,0xff,0xcf,0x7f,0xff,
	0xfd,0xff,0xff,0x9b,0xfc,0xff,0xff,0xdf,
	0x7f,0xfb,0xff,0xf7,0x7f,0xff,0xfd,0xdf,
	0xff,0xff,0x7f,0xff,0xff,0xde,0x7f,0xff,
	0xdd,0xff,0xdf,0xff,0xf7,0xfd,0x7f,0xdf,
	0xfd,0xf7,0xff,0xff,0xff,0xe4,0x7f,0xdd,
	0xad,0xff,0xf1,0xe3,0x1f,0x3f,0xbe,0xf8,
	0xe3,0x8f,0x2f,0xf7,0xf1,0xe3,0x3f,0xfc,
	0xf9,0xff,0x8f,0x7d,0xfe,0xdf,0xf3,0xeb,
	0x3f,0xff,0x7f,0xfa,0xff,0xff,0x4f,0xff,
	0xff,0x75,0xf1,0xdf,0x57,0xfb,0x7d,0xf7,
	0xd7,0x77,0x7f,0xfd,0xf5,0xdf,0x5f,0xef,
	0xf1,0xdf,0x7f,0xdf,0xfd,0xf7,0xdf,0x37,
	0x5f,0x7e,0xf3,0xee,0x97,0xff,0xfe,0x7f,
	0xfa,0xff,0xff,0xeb,0xfe,0xfe,0xfe,0x3f,
	0xef,0x2f,0xff,0xfa,0xf3,0xef,0xff,0xfe,
	0xda,0xcf,0xfc,0xfb,0xff,0xf3,0xdf,0xfe,
	0xbf,0xee,0xff,0xeb,0xfe,0x9f,0xfe,0xff,
	0xff,0xc5,0xff,0xff,0x47,0x9f,0x7f,0xe4,
	0xb9,0xc7,0x9e,0x7d,0xfc,0xd9,0xe7,0x9f,
	0x7f,0xf6,0x69,0x9f,0x7b,0xfe,0xf1,0xe7,
	0x9f,0x7f,0xf4,0xf9,0x47,0x9f,0x7f,0xf4,
	0xf9,0xe7,0x27,0xfe,0xff,0xfe,0xd2,0xdf,
	0xff,0xff,0xfd,0xfa,0xef,0xbf,0xff,0xfe,
	0xfa,0xdf,0xaf,0xff,0xe3,0xdf,0xff,0xff,
	0xfe,0xff,0xff,0x3f,0xbf,0xfe,0xf3,0xdf,
	0xbf,0xff,0xff,0x7f,0xf5,0xff,0xd7,0xd7,
	0xff,0xfe,0xed,0xef,0xd7,0xfe,0xfb,0xed,
	0xbd,0xdf,0xfe,0x7e,0xed,0xdb,0xff,0xff,
	0xff,0xb7,0xff,0xff,0x7f,0xfb,0xf5,0xb7,
	0xdf,0x7e,0xff,0xff,0xff,0x8f,0xff,0xbf,
	0xbf,0xfc,0xef,0xff,0xff,0xfe,0xff,0xfb,
	0xef,0xbf,0xff,0xfe,0xef,0xff,0xff,0xfa,
	0xf7,0xff,0xbf,0xff,0xff,0xff,0xeb,0xaf,
	0xbf,0xfe,0xff,0xef,0xbf,0xff,0x4f,0xbc,
	0xff,0xfd,0xf5,0xff,0x7f,0xff,0xff,0xff,
	0xff,0x7f,0x7f,0xff,0xff,0xff,0xff,0xff,
	0xf7,0xff,0xfe,0xfb,0xfd,0xbf,0xff,0x5e,
	0x7b,0xfd,0xb5,0xbf,0xdf,0xff,0xfd,0xef,
	0xe7,0xfd,0xbb,0xff,0xfe,0xff,0xff,0xff,
	0xff,0xff,0xbf,0xff,0xff,0xff,0xff,0xfb,
	0xff,0xff,0xbe,0xfb,0xff,0xff,0xff,0xff,
	0xfb,0xf2,0xff,0xff,0x3e,0xf9,0xff,0xcb,
	0xff,0x27,0xf7,0xbf,0xf5,0x57,0x5e,0x7d,
	0xed,0xd5,0xd7,0x5e,0x79,0xfd,0xf4,0xd7,
	0x4f,0x3f,0xfd,0x07,0x5f,0x3f,0xf5,0xd4,
	0xd7,0x5f,0x6f,0xfd,0xe5,0x97,0x5b,0x7f,
	0xbd,0xf5,0xff,0xf9,0xff,0xbf,0xcf,0xff,
	0xff,0xff,0x7f,0xcf,0x3f,0xff,0xfb,0xf3,
	0xcf,0xff,0xff,0xff,0xad,0x1f,0xf7,0xfb,
	0xf7,0xff,0xf7,0xfb,0xf8,0xbf,0xcf,0x37,
	0xf7,0x3f,0xff,0xbd,0x8d,0xff,0xff,0xb7,
	0xfd,0xef,0xff,0xfd,0x2f,0xff,0xdc,0xff,
	0xcb,0x5f,0xbf,0xff,0xff,0x7d,0x7f,0xff,
	0xff,0xf7,0xef,0xef,0xff,0xdc,0x7d,0x5b,
	0xfe,0x7f,0xff,0xff,0xff,0xf6,0xfe,0xff,
	0xff,0xdd,0xdf,0x7f,0xff,0x7f,0xff,0xfd,
	0xff,0xff,0xf7,0xff,0xff,0xff,0x3f,0xff,
	0xef,0xff,0xff,0xff,0xff,0xff,0xdf,0xff,
	0xff,0xff,0xeb,0xff,0xff,0xff,0xff,0xe1,
	0xbf,0xbd,0xff,0xfd,0x7f,0xbc,0xff,0xd7,
	0x7f,0xbb,0xfd,0xfd,0xff,0xff,0xeb,0xfd,
	0xfd,0xff,0xff,0xff,0xf1,0xfe,0xff,0xf9,
	0xef,0xf5,0xff,0x7b,0x7e,0xf1,0xfd,0xdf,
	0x9f,0xff,0xbb,0xef,0xfb,0xff,0xf7,0xfe,
	0x7b,0xfe,0x9e,0xfb,0xbf,0xff,0xef,0xff,
	0xf7,0xff,0xff,0xff,0xff,0xff,0xff,0xbf,
	0xef,0xff,0x6f,0xff,0xff,0xdd,0xef,0x7f,
	0xff,0x29,0xf9,0xff,0xef,0x7f,0xff,0xfd,
	0xdf,0xf7,0x7f,0xff,0xff,0xf7,0xff,0xff,
	0xff,0xff,0x6f,0xff,0xff,0xff,0x7f,0xff,
	0xff,0xcf,0xfe,0xef,0x3f,0xf7,0xff,0xdf,
	0xff,0xad,0x7f,0xe6,0xff,0xeb,0xff,0xff,
	0xfd,0xbf,0x7f,0xff,0xff,0xfd,0xf7,0xff,
	0xff,0xfc,0xff,0xf7,0xfb,0xf7,0xf9,0xdf,
	0xdf,0xff,0xdf,0x7f,0xf6,0xdf,0x67,0xff,
	0xfd,0xfa,0xff,0xff,0x73,0xff,0xfb,0xef,
	0xff,0xff,0xff,0xee,0xb5,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xbf,0xff,0xff,0xf7,0xff,0xff,
	0xf9,0xfe,0xff,0xff,0xbe,0x3f,0xf9,0xe7,
	0xcc,0xeb,0xff,0x7f,0xff,0xff,0xfe,0xcf,
	0xff,0xff,0xff,0xff,0xff,0xfb,0xff,0xfd,
	0xff,0x3f,0x7f,0xf8,0xff,0xff,0xff,0xff,
	0xf4,0xbf,0xfe,0xff,0xff,0xef,0xff,0xd2,
	0x7f,0x6d,0xff,0xff,0xff,0xef,0xff,0xff,
	0x7f,0xfe,0xff,0xff,0xfb,0xff,0xff,0xff,
	0xff,0xff,0xff,0xf9,0xdf,0xff,0xff,0xef,
	0xff,0xb6,0xe3,0xff,0xef,0xff,0xff,0xff,
	0x4f,0xfe,0x7f,0xf9,0xff,0xeb,0xbf,0xf7,
	0xfe,0xf3,0xdf,0xaf,0xbb,0xfc,0xf7,0xef,
	0x7e,0x7d,0xff,0xff,0x7f,0xff,0xff,0xff,
	0xeb,0xa2,0xfb,0xed,0xeb,0xeb,0xbf,0xff,
	0xff,0xff,0xf4,0xff,0x7f,0xff,0xff,0xbf,
	0xfd,0xf6,0xcf,0xff,0xff,0xfd,0xe7,0xfb,
	0x1f,0xaf,0xbb,0xfb,0x5e,0xbf,0xff,0xff,
	0xbe,0x5d,0x7f,0xfd,0xff,0x2f,0x5b,0xbf,
	0xed,0xff,0xfe,0x83,0xff,0xff,0xff,0xdb,
	0xef,0xf3,0xcf,0xff,0xfe,0xff,0xf7,0xbf,
	0xff,0xdf,0xff,0xfd,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xbf,0xff,0xf7,0xff,
	0xef,0xeb,0xff,0xdf,0xbf,0xf1,0x7f,0xff,
	0xff,0xff,0xff,0xfb,0xff,0xcf,0x3f,0xff,
	0xff,0x33,0xbf,0xff,0xeb,0xaf,0x3e,0xfe,
	0xff,0xff,0xff,0xff,0xff,0xdf,0xff,0xff,
	0xfd,0xfa,0xff,0x7f,0xff,0x3f,0x61,0xfd,
	0xff,0xf9,0xff,0xff,0xff,0xff,0xff,0xfe,
	0xfb,0xff,0xbf,0xf9,0xff,0x9f,0x7f,0xfe,
	0xf3,0xff,0xff,0xff,0xff,0xf3,0xf7,0x5c,
	0xfd,0x8f,0xe7,0xff,0xff,0xfb,0xbf,0x0e,
	0xef,0xdf,0xff,0xfb,0xef,0xbf,0xff,0xfe,
	0xfb,0xef,0xbf,0xff,0xfe,0xff,0xef,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0x3f,0xf5,
	0xd4,0xff,0xaf,0xff,0xff,0xff,0xf3,0xff,
	0xfd,0xda,0xff,0xff,0xdf,0xdf,0x7e,0xfb,
	0xed,0xb7,0xdf,0x7e,0xfb,0xed,0xff,0xdf,
	0xfe,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xa9,0xb7,0xfe,0x7f,0xfe,0xff,0xf7,0xdf,
	0xff,0xbf,0xcd,0xff,0xdf,0xec,0xfb,0xfb,
	0xaf,0xbf,0xfe,0xfe,0xfb,0xaf,0xbf,0xff,
	0xfb,0xeb,0xbf,0xef,0xfb,0xef,0xbf,0xff,
	0xff,0x7b,0x6d,0xb5,0xfb,0x5e,0xfa,0x6f,
	0xff,0xf9,0xff,0xff,0xf6,0xbf,0x56,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xab,0xaf,0xc6,0xff,0xea,
	0xff,0xff,0xce,0xfb,0x6f,0xe2,0xff,0xf7,
	0x7f,0xfe,0xef,0xff,0xff,0xfa,0xeb,0xff,
	0xff,0xff,0xfa,0xff,0xff,0xff,0xff,0xff,
	0xfd,0xbf,0xfe,0x7f,0x5f,0x7f,0xfd,0xdf,
	0x57,0xfb,0xff,0x7f,0xff,0xff,0x99,0xff,
	0xaf,0xff,0xff,0x7f,0xbe,0xff,0xf6,0xdf,
	0xff,0xbf,0xff,0xf7,0xfb,0xef,0xbf,0xdf,
	0xfb,0x3f,0xbf,0xfd,0xff,0x5f,0xea,0xab,
	0xfc,0xbf,0xbc,0xef,0xef,0xff,0xff,0x7f,
	0xfd,0xff,0xcf,0xbe,0xbf,0xe5,0xb6,0x7b,
	0xef,0xbd,0xed,0xfe,0x7b,0xef,0xbf,0xff,
	0xfe,0xef,0xbf,0xff,0xde,0xff,0xef,0xbf,
	0xf9,0xfe,0xfb,0xef,0xbf,0xff,0xf7,0xff,
	0xff,0xe4,0xff,0xff,0xfd,0xfe,0xd9,0x77,
	0xdf,0x7f,0xf6,0x99,0xf7,0x9f,0x7f,0xfe,
	0xfd,0xe7,0x7f,0xfe,0xf9,0xe7,0xbf,0x7f,
	0xfe,0x99,0xe7,0x9f,0x7f,0xff,0xf9,0x7f,
	0xfc,0xff,0x57,0xfe,0x7f,0xf7,0xfd,0xd7,
	0xdd,0x77,0xff,0xd5,0xd7,0xdf,0x7f,0xfd,
	0xf9,0xf7,0x9f,0xbf,0xf9,0xc7,0x9f,0x7f,
	0xff,0xf1,0xe7,0x9f,0x7f,0xac,0xf9,0xe7,
	0xdb,0x7f,0xfb,0x9f,0xda,0xff,0xdf,0x7f,
	0xff,0xff,0x7f,0xff,0xff,0xf7,0xff,0x7f,
	0xff,0xff,0xf7,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xef,0x7f,0xff,0xbf,0xc3,0xff,0xff,
	0xff,0xfd,0xff,0x7f,0xff,0xff,0xdf,0xff,
	0x7f,0xff,0xff,0xf5,0xff,0xdf,0xef,0xfd,
	0xff,0xdf,0x7f,0xff,0xff,0xf7,0xdf,0xff,
	0xff,0xff,0xf7,0xff,0xfd,0xff,0xbf,0xfa,
	0xbf,0xde,0x7f,0x8f,0x2f,0xf6,0xf8,0xe2,
	0x8b,0x3d,0xb6,0xf8,0x77,0x8b,0x7f,0xfe,
	0xe7,0x8d,0x7d,0xfe,0xff,0x67,0x8f,0x3f,
	0xff,0xf9,0xe7,0x8f,0xff,0xdf,0xfe,0xdf,
	0xf1,0xbf,0xd5,0x7f,0x7f,0x7d,0xf5,0xdd,
	0x57,0x5f,0x7d,0xf5,0xdd,0x7f,0xdf,0xfd,
	0xf7,0x7b,0x7b,0xfc,0xf7,0xdf,0x7f,0x7f,
	0xec,0xf1,0xdf,0x7f,0x6b,0xbc,0xfb,0xef,
	0xff,0xab,0xff,0xff,0xf7,0xef,0xeb,0xdf,
	0xbf,0xf6,0xdb,0xf6,0xcf,0xae,0xfe,0xff,
	0x72,0xff,0xfc,0xff,0xeb,0xdf,0xfd,0xff,
	0xeb,0xb2,0xef,0x7e,0xfb,0xcb,0xbb,0xff,
	0xbf,0xff,0x9f,0xfc,0xff,0x7f,0xfe,0xd9,
	0xe7,0x1f,0x7f,0xfe,0xd9,0x47,0x1f,0x7f,
	0xfe,0xf1,0xe7,0x9e,0xef,0xd9,0xe7,0x9f,
	0x7f,0xfe,0xd1,0x67,0x9f,0x7f,0xfe,0x99,
	0xe7,0x9f,0x7f,0xfe,0xec,0xff,0xff,0x7f,
	0xbd,0xfe,0xfb,0xcf,0xbf,0xff,0xfe,0xf7,
	0xef,0xff,0xff,0xfc,0xf7,0x7f,0xff,0xfe,
	0xe7,0xdf,0x7f,0xff,0xfe,0xe2,0xff,0x7f,
	0xff,0xfc,0xff,0xff,0xff,0x1f,0xff,0xff,
	0xfb,0xfb,0xf5,0xdf,0x7f,0xff,0xfd,0xfd,
	0xf7,0xff,0xff,0xff,0xff,0xff,0xbf,0xff,
	0xf7,0xff,0xff,0xfe,0xff,0xf7,0xd7,0xff,
	0xff,0xff,0xf7,0xff,0xff,0xff,0xbf,0xf9,
	0xff,0xfa,0xff,0xff,0xfe,0xfa,0xe3,0xaf,
	0xff,0xfe,0xf9,0xeb,0xff,0x3f,0xfe,0xfd,
	0xdf,0xbf,0xff,0xff,0xff,0xdf,0xaf,0xbf,
	0xfe,0xff,0xdb,0x3f,0xff,0xfe,0xff,0xff,
	0xd9,0xff,0xd3,0xff,0xff,0xff,0xf7,0x9f,
	0x7f,0xff,0xff,0xed,0xff,0xff,0xfb,0xfb,
	0xbf,0xff,0xfb,0xed,0xbf,0xff,0xfe,0x7b,
	0xed,0xb5,0xff,0xfe,0xfb,0xed,0xff,0xff,
	0xff,0xdf,0xfe,0xff,0xfd,0xef,0xfe,0xfb,
	0xff,0xff,0xff,0xbf,0xff,0xff,0xff,0xff,
	0xfe,0xff,0xff,0xdf,0xfe,0xfb,0xff,0xff,
	0xef,0xfe,0x2f,0xff,0xff,0xff,0xbe,0xef,
	0xbf,0xfc,0xf7,0xd5,0xff,0x5f,0x7f,0xf1,
	0xc5,0xd7,0x5f,0x7d,0xfd,0xf5,0xd7,0x5f,
	0x7d,0xfd,0xd5,0xd3,0x7f,0xf5,0xf5,0x17,
	0x5f,0x7e,0xfd,0xf5,0xd7,0x5f,0x7f,0xfd,
	0xd5,0xd6,0x5b,0xff,0x86,0xff,0xe7,0xfb,
	0xff,0xf3,0xcf,0x3d,0xff,0xfc,0xf3,0xcf,
	0x3f,0xdf,0xf8,0xd3,0x8f,0xdf,0xf9,0xef,
	0xff,0xff,0xff,0x7c,0xff,0x87,0xff,0xff,
	0x7c,0xf7,0xff,0xfd,0xff,0x7f,0xf8,0xff,
	0xfb,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xfd,0xff,0x3f,0xff,0xfd,
	0xff,0xfd,0xf7,0xdf,0x7b,0x7f,0xdf,0xff,
	0xff,0xf5,0xf7,0xfe,0xff,0xff,0x7f,0x65,
	0xff,0xff,0xff,0x3f,0xff,0xfc,0xf3,0xcf,
	0x3f,0xff,0xfc,0x71,0xd7,0x1f,0xf7,0xff,
	0xfb,0xf7,0xff,0xff,0xff,0xbf,0xff,0xff,
	0xdd,0xff,0xff,0xff,0xdf,0xff,0xff,0xff,
	0x07,0xff,0x7a,0xf7,0xef,0xe5,0xf7,0x7f,
	0x7f,0xfc,0xf7,0xff,0x1f,0xff,0xef,0xf7,
	0xff,0xfe,0x7b,0xbf,0xff,0xfd,0xff,0xff,
	0xff,0xd7,0xff,0x7f,0xfe,0xff,0xf7,0xdf,
	0xff,0xf7,0xfb,0xff,0xfe,0xbf,0xbf,0xfb,
	0xee,0xfe,0xf7,0xef,0xff,0xff,0xdd,0x7f,
	0xef,0xed,0x67,0xfc,0xff,0xf5,0xff,0xff,
	0xff,0xff,0x7f,0xfd,0xff,0xff,0xfb,0xff,
	0xfe,0xf7,0xfe,0xcf,0xff,0xff,0xfd,0x7f,
	0xfd,0xff,0xbf,0xff,0xfe,0xff,0xf7,0xff,
	0xff,0xff,0xf7,0xff,0xe7,0xfb,0xfd,0xff,
	0xff,0xff,0xf7,0xdc,0xf3,0xff,0x7f,0xf7,
	0xff,0xb7,0xff,0xfe,0x6f,0xfe,0x9b,0xfe,
	0xdf,0xf7,0x7f,0xff,0xdf,0xff,0xdf,0xff,
	0xff,0xff,0xff,0xde,0xff,0xbd,0xbf,0xd5,
	0xff,0xbd,0xfd,0xf7,0x8b,0xff,0xff,0xf7,
	0xf5,0xdf,0x8f,0xbf,0xfe,0x7f,0xf2,0xff,
	0xff,0x7f,0xff,0xff,0x7f,0xff,0xff,0xff,
	0xff,0xfb,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xe7,0xfd,0xff,0xff,0xff,0x5f,
	0xff,0xfe,0xff,0xff,0xff,0xff,0xff,0xaf,
	0x7f,0xff,0xbf,0x76,0xff,0xff,0xfd,0xff,
	0xff,0xff,0xfd,0xcf,0xff,0xff,0xff,0xff,
	0xf7,0xff,0xff,0xff,0xff,0x3f,0xef,0xff,
	0xff,0xff,0xff,0xfb,0xff,0xff,0xff,0xff,
	0x7f,0xfd,0xff,0xef,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0x7f,0xfb,0xbf,0xff,
	0xff,0xff,0xbf,0xff,0xff,0xff,0xff,0xf9,
	0xff,0xff,0xfd,0x7f,0xff,0xff,0xff,0xff,
	0xff,0xff,0xec,0xff,0xff,0x7f,0xff,0xfe,
	0xff,0xff,0xff,0xff,0xfd,0xff,0xde,0xf7,
	0xff,0x7e,0xbf,0x7f,0xff,0x6e,0xff,0xff,
	0xff,0xff,0x7f,0xfb,0xff,0xfe,0xef,0xfb,
	0xff,0xff,0xff,0x27,0xfe,0xff,0xff,0xf7,
	0xf7,0xff,0xff,0xef,0xff,0xff,0xdf,0xff,
	0xff,0xff,0xb3,0x97,0xff,0x7f,0xb3,0xbf,
	0x7e,0xff,0xff,0xff,0xdf,0xff,0xff,0xef,
	0xff,0xff,0xff,0xef,0x5f,0xf8,0xff,0xef,
	0xbf,0xae,0xdf,0xff,0xff,0xff,0xff,0xff,
	0xfe,0xff,0xbf,0xdf,0xff,0xff,0xff,0xbf,
	0xff,0xff,0xfb,0xbf,0xff,0xbf,0xfe,0xef,
	0xff,0xff,0xff,0xff,0xff,0xff,0x1f,0xff,
	0xff,0xdf,0xff,0x37,0xbf,0xff,0xeb,0xad,
	0xef,0x5d,0xfe,0xde,0xfd,0xf7,0xdf,0xfa,
	0xff,0xb7,0xbc,0x7f,0xff,0xff,0xf7,0xdf,
	0xfc,0xff,0xad,0xff,0xdf,0x7f,0xff,0xa3,
	0xd6,0xff,0x5f,0xfe,0xbf,0xf5,0xf3,0xff,
	0x6f,0xfe,0xef,0xf2,0xff,0xee,0xbb,0xdf,
	0x56,0xff,0xb7,0xe9,0xff,0xfb,0xff,0xbf,
	0xdf,0xe6,0xff,0x6f,0xfe,0xff,0x7e,0xfb,
	0xeb,0xf5,0xfe,0xff,0xb7,0xcb,0xff,0xff,
	0xff,0xff,0xff,0xaf,0xff,0xff,0xdc,0x7f,
	0x4f,0xfd,0xff,0x7f,0x7f,0xbf,0xfc,0xff,
	0x73,0xff,0xfd,0x7f,0xf6,0xff,0xff,0xfd,
	0xf7,0xdf,0xa7,0xfd,0xff,0xbf,0xdd,0xfe,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xbf,
	0xdf,0x3a,0xfb,0xff,0xff,0xff,0xfb,0xe9,
	0xff,0xcf,0xfb,0xff,0xff,0xbf,0xff,0x7f,
	0xee,0xff,0xff,0x7f,0xfd,0xff,0xef,0xbd,
	0xdd,0xfe,0xfb,0xff,0xbf,0xff,0xde,0xff,
	0xef,0x7d,0xf9,0x7f,0xff,0xbe,0xd7,0xdf,
	0xfb,0xed,0xbf,0xf7,0x5f,0xbb,0xef,0xb7,
	0xff,0xde,0x7b,0xfd,0xfb,0x6a,0xff,0x6f,
	0xef,0xfe,0xf6,0xff,0xff,0xff,0xff,0xff,
	0xfe,0xff,0xef,0xbf,0x5f,0x7f,0xfd,0xbf,
	0xff,0xee,0x6f,0xff,0xbf,0xff,0x9e,0xfc,
	0xbf,0xfd,0xff,0xde,0xfb,0xf5,0x0f,0xfe,
	0xff,0x7f,0xe7,0xf5,0xff,0xeb,0xff,0x9f,
	0xff,0xd7,0xff,0xfd,0xfd,0xf7,0xd7,0xff,
	0xff,0xf5,0xf7,0x5e,0xff,0xff,0xf5,0xd7,
	0xff,0x7f,0xfd,0xff,0xd7,0x5f,0xff,0x3f,
	0xf9,0xff,0xfb,0xfb,0xaf,0xfd,0xdb,0xff,
	0x3f,0xfc,0xbe,0xf6,0xff,0x7b,0xaf,0xbd,
	0xf7,0xbd,0xef,0xbd,0xfb,0xda,0xbf,0xaf,
	0xbd,0xf7,0xef,0x6b,0xff,0xbe,0xf6,0xde,
	0xff,0xdb,0xff,0xff,0xed,0xfb,0xdf,0xef,
	0xfd,0xff,0xfe,0xfb,0xff,0xbf,0xff,0xb7,
	0xff,0xff,0xff,0xfe,0xff,0xef,0xff,0xff,
	0xfe,0xdf,0xff,0xbf,0xff,0xff,0xfb,0x7f,
	0xfb,0xff,0x4f,0xfe,0xff,0xdf,0xef,0xff,
	0x77,0xfe,0xfb,0xe7,0x9f,0xff,0xfe,0xf9,
	0x6f,0xbf,0xff,0xf6,0xe7,0xbf,0x7d,0xfe,
	0xff,0xe7,0xbf,0xf5,0xff,0xf9,0xff,0x9f,
	0xff,0xf7,0xdb,0x7f,0xe3,0xff,0xf7,0xdd,
	0x7f,0xff,0xf9,0xf7,0x9f,0x7f,0xfe,0xfd,
	0xc7,0xdb,0x6d,0xff,0xdd,0x9b,0x7f,0xdf,
	0xe9,0xf6,0x9f,0x7f,0xf7,0xfd,0xe7,0xdb,
	0x7f,0xae,0x7d,0xf7,0xf9,0x81,0xfd,0xff,
	0xff,0xf7,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfd,0xff,0xcf,0xbf,0xff,0xff,0xff,0xff,
	0xff,0xff,0xfe,0xff,0x7f,0xff,0xfd,0xff,
	0xcf,0xff,0xff,0xfd,0xf7,0xef,0xdb,0xfc,
	0xff,0xff,0xdf,0xff,0xff,0xfd,0xf7,0xdf,
	0x7f,0xfd,0xff,0xdf,0xff,0xff,0xff,0xff,
	0xde,0xff,0xff,0xfd,0xff,0xdf,0xfd,0xf7,
	0xff,0xd7,0xff,0x7f,0xf7,0xef,0xff,0xff,
	0xa2,0xff,0xeb,0xfd,0xbf,0xf6,0xe7,0xff,
	0x7f,0xfc,0xf9,0xdb,0x9f,0xaf,0xfd,0xf6,
	0xdb,0x7f,0xbf,0xf6,0xe7,0x6f,0x7f,0xff,
	0xf6,0xdb,0xdf,0xbf,0xfd,0xf9,0xdb,0x6f,
	0xff,0x47,0xff,0x5f,0xfc,0xf7,0xd5,0x7f,
	0xff,0xfd,0xf7,0xdf,0x57,0xf9,0x7d,0xd5,
	0x57,0x5f,0xfd,0xf7,0x55,0x7f,0x7f,0xed,
	0xf7,0x5d,0x77,0xff,0x7d,0xf5,0xdf,0x77,
	0xdf,0xf5,0x3f,0xf9,0xff,0x9f,0xff,0xfe,
	0xff,0xdb,0xff,0xbf,0xfc,0xff,0xfe,0xf7,
	0xfb,0xbf,0xff,0xfe,0xff,0xff,0xf7,0xff,
	0xfb,0xff,0xbf,0xff,0xde,0xbf,0xef,0xfd,
	0xfb,0xfe,0xfb,0xff,0xd7,0xff,0xff,0xe7,
	0x9f,0x1f,0xfe,0xf9,0xe7,0x9f,0x7b,0x74,
	0xf8,0x67,0x87,0x1d,0x76,0xf8,0x9f,0x1f,
	0xee,0xd9,0xe1,0x1f,0x1f,0x7c,0xf8,0x47,
	0x87,0x7b,0x7c,0xf0,0xe1,0x27,0xfe,0xff,
	0xff,0xf7,0xef,0xf7,0xff,0xff,0xff,0xdf,
	0xff,0xdf,0xff,0xfb,0xed,0xa7,0x5f,0xff,
	0xef,0x77,0xff,0x7e,0xf7,0xeb,0xb7,0xdf,
	0xff,0xff,0xdd,0xaf,0x9f,0x7e,0xff,0xf4,
	0xff,0xff,0xbf,0x7f,0x93,0xef,0xbf,0xff,
	0xfe,0xfb,0x4d,0xbe,0xf7,0xc4,0x13,0x4d,
	0xfa,0x7e,0x93,0xef,0x3d,0xb9,0x5e,0x93,
	0x4d,0xbe,0xdf,0xe4,0x7b,0x4d,0x34,0xf9,
	0xaf,0xff,0xff,0xff,0xf7,0x7b,0xfe,0xff,
	0xff,0xff,0xf7,0xff,0xf9,0xff,0x9f,0x7b,
	0xfe,0xf9,0xff,0x7b,0xde,0xbf,0xe7,0xff,
	0x7f,0xee,0xf9,0xff,0x9f,0xf7,0xff,0xf9,
	0xe7,0x6f,0xfd,0xff,0xef,0xff,0xdf,0xfd,
	0xff,0xff,0xff,0xff,0x7f,0xf7,0xff,0x7f,
	0xf7,0xfd,0xf7,0xff,0xdf,0xfd,0x7f,0xdf,
	0xff,0xff,0x7d,0xf7,0xff,0x77,0xff,0xff,
	0xf7,0xdf,0xff,0x68,0xff,0xfb,0xef,0xbe,
	0xfc,0xff,0xbf,0xff,0xff,0xff,0xfe,0xbf,
	0x2f,0xbf,0xfc,0xe2,0xff,0xbf,0xfc,0xf6,
	0xcb,0xff,0xbf,0xfc,0xf2,0xff,0x2f,0xff,
	0xff,0xf2,0xcb,0xff,0x57,0xfd,0xbf,0xc5,
	0xd7,0x5b,0x7d,0xfd,0xf5,0xd7,0x5f,0x73,
	0xfd,0xd5,0xd6,0x4d,0x76,0xb9,0xd7,0x4d,
	0x5f,0xdd,0xb5,0xd7,0x5d,0x77,0xf9,0x75,
	0xd7,0x5f,0x73,0xdd,0xf5,0xff,0xf8,0xff,
	0xff,0xf7,0x3f,0xff,0xff,0x7f,0xcd,0x3f,
	0xff,0xfc,0xff,0xcf,0x3f,0xfb,0xfc,0x8f,
	0x3c,0xff,0xfe,0xf3,0x8f,0x3f,0xff,0xfc,
	0xe3,0xcd,0x3f,0xfe,0xfc,0xf3,0xff,0x87,
	0xff,0xfd,0xff,0xd7,0xff,0xff,0xff,0xff,
	0xff,0xfe,0xff,0xff,0xff,0xdf,0xff,0xff,
	0xdb,0xff,0xff,0xff,0xff,0xdf,0xff,0xff,
	0xff,0xf7,0xff,0xf7,0xfe,0xff,0xff,0xff,
	0x37,0xfe,0xff,0xfd,0xff,0xf0,0xff,0xff,
	0xdf,0xfc,0xf7,0xcf,0xff,0xef,0xfc,0xf3,
	0xcf,0x3f,0xfd,0xf3,0xbf,0x3f,0xbf,0xfd,
	0xf3,0xcd,0x7d,0xf7,0xfc,0xef,0xcf,0x3f,
	0xff,0x7d,0xf5,0xfd,0xfd,0xff,0x5f,0xff,
	0xdf,0xff,0xff,0xff,0x7e,0xfb,0xff,0xc7,
	0x5b,0x3f,0xbd,0xff,0x5f,0xef,0xbf,0xf5,
	0xff,0x5f,0x7f,0xfd,0xff,0xc7,0xff,0x7f,
	0xfd,0xf5,0xff,0xb9,0xef,0xbf,0xdf,0xff,
	0xb9,0xff,0xff,0xfb,0xff,0xff,0xdf,0xff,
	0x7f,0xff,0xf9,0xe7,0xce,0xff,0xf9,0xff,
	0x9f,0xff,0xff,0xf9,0xe7,0xff,0x7b,0xff,
	0xff,0xe7,0x9f,0xff,0xdf,0xfc,0xff,0xef,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x7b,
	0xfb,0xff,0xff,0xf9,0xff,0x7f,0xff,0xbf,
	0xfe,0xfb,0x7f,0xef,0x7f,0xff,0xf7,0xff,
	0xef,0xff,0xfd,0xfb,0xff,0x7f,0xec,0xfb,
	0xff,0xff,0xff,0xfd,0xff,0xff,0xff,0xff,
	0xed,0xf8,0xdf,0x7f,0xff,0xfd,0xff,0x7b,
	0xfe,0xff,0xff,0xdf,0xff,0xff,0xfd,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x67,
	0xff,0xff,0xfd,0xbf,0xff,0xfe,0xff,0xef,
	0xff,0xff,0x7f,0xff,0xbb,0xef,0xbf,0xff,
	0xbf,0xff,0xbf,0xff,0xff,0xff,0xff,0xfe,
	0xfb,0xff,0xdd,0xff,0x9f,0xff,0xef,0xff,
	0xff,0xfb,0xf7,0xff,0xff,0xc7,0xdd,0xff,
	0xfc,0xf3,0xff,0x3f,0x7d,0xfe,0xf3,0xff,
	0xff,0xff,0xfe,0xf7,0x1f,0x7f,0xf8,0xfd,
	0xce,0x7b,0xff,0xfc,0xf3,0xff,0xdf,0xff,
	0xff,0xff,0xcb,0xff,0x7f,0xf7,0x7b,0xfe,
	0xff,0xd7,0x5f,0xff,0xef,0xed,0xf7,0xdf,
	0xbe,0xff,0xff,0xff,0xff,0xff,0xfb,0xdf,
	0xff,0x7f,0xfb,0xfb,0xb7,0x9f,0xff,0xff,
	0xff,0xff,0xff,0xef,0xfe,0xbf,0xfe,0xff,
	0xff,0x7f,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xf7,0xff,0xfa,0x7b,0x97,0xff,0xde,
	0xff,0xff,0xff,0xff,0xff,0xff,0xfd,0xff,
	0xdf,0xff,0xff,0xff,0x7f,0xe1,0xff,0xf7,
	0xff,0xff,0xff,0x5b,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xfb,0xff,0xbf,0x7f,0xff,
	0xfb,0xb3,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xaf,0xff,
	0xbf,0xff,0xfb,0xff,0xff,0xff,0xff,0xfb,
	0xef,0xff,0xff,0xff,0xff,0xff,0xff,0xdf,
	0xeb,0xff,0xff,0xfe,0xff,0xff,0xff,0xff,
	0xf7,0xfe,0xff,0xff,0xff,0xff,0xff,0xdf,
	0xf5,0xff,0x9f,0xff,0xff,0xf9,0xfb,0xff,
	0xff,0xff,0xff,0xaf,0xff,0xbf,0x7f,0xf9,
	0xfb,0xbe,0x7f,0xdf,0xcb,0x37,0xff,0x7f,
	0xff,0xfd,0xff,0xdf,0xfa,0xff,0xfd,0xf7,
	0x3f,0x61,0xfd,0xff,0xc5,0x3f,0xcf,0xfc,
	0xf3,0xff,0xff,0xff,0x5f,0xfe,0xff,0xff,
	0xcb,0xff,0xdf,0xf3,0xfb,0x3e,0x3f,0xff,
	0xe7,0xfb,0xef,0xff,0xff,0xe6,0x5f,0x6f,
	0xbd,0xbf,0x1e,0xef,0xff,0x6f,0xbb,0xfd,
	0xb6,0xdb,0xff,0xff,0xff,0xdf,0xff,0xff,
	0xff,0xff,0xff,0x2b,0xfd,0xff,0xf7,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0x7d,0xd8,0xff,0x7f,0xcd,
	0xb5,0xd7,0x5c,0xfb,0xff,0xff,0xa7,0xff,
	0xff,0xff,0xff,0xff,0xff,0xfb,0xff,0xbf,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xbf,0x9f,0xff,0xff,
	0x7f,0xbf,0xcd,0x36,0xdb,0xed,0xff,0xff,
	0x35,0xfb,0xfe,0xfb,0xef,0xbf,0x7f,0xab,
	0xef,0xfd,0xff,0xfe,0xff,0xef,0xbf,0xff,
	0xfe,0xfb,0xef,0xbf,0xff,0xfe,0xdf,0xf6,
	0xdb,0xff,0xfe,0x7a,0xe5,0xbd,0x7f,0xff,
	0xff,0xaf,0xd6,0xff,0xff,0xff,0xff,0xff,
	0xaf,0xfd,0xef,0xfe,0xfd,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xef,
	0xe7,0xff,0xff,0xf7,0xde,0x7f,0xff,0xbd,
	0xfd,0xff,0xff,0xfd,0xbf,0x7f,0xfb,0xfd,
	0xff,0xd7,0xfb,0x2f,0xff,0xcf,0xff,0xff,
	0xff,0xf7,0x9f,0x7f,0xeb,0xff,0xf7,0xff,
	0xff,0xa5,0xff,0xff,0xff,0xfe,0xfb,0xef,
	0xbf,0xff,0xff,0xff,0xbb,0xff,0xfe,0x8f,
	0xdf,0xbf,0x9f,0x8b,0x4f,0xfb,0xff,0xfd,
	0xff,0xcf,0x7f,0xf5,0xf3,0xd7,0xef,0x7f,
	0xfc,0xf0,0x5f,0xf9,0xff,0xdf,0xbd,0xf7,
	0xdc,0x73,0xcf,0x3f,0xff,0xff,0xb6,0xfb,
	0xef,0xff,0xff,0xfe,0x6e,0xbe,0xf7,0xe3,
	0xfb,0xff,0xbf,0xff,0xfe,0xfb,0xef,0xbd,
	0xff,0xfe,0xfb,0xff,0xe5,0xff,0xff,0xf7,
	0xfe,0xcf,0x3f,0xbf,0xbc,0xf3,0xfb,0x67,
	0x9f,0x7f,0xfe,0xff,0xe7,0x7f,0xe6,0xf9,
	0xff,0xff,0xff,0xfe,0xf9,0xe7,0x9f,0x7f,
	0xfe,0xf9,0xe7,0x9f,0xff,0x4f,0xfe,0x7f,
	0x7f,0xfd,0xf5,0xd7,0x5f,0xff,0xfd,0xf7,
	0x9a,0x7d,0xfe,0xf1,0xf7,0x9f,0xef,0xf9,
	0xe7,0xdf,0x6b,0xff,0xfd,0xe7,0x1f,0x7f,
	0xfc,0xf1,0xe7,0x1f,0x7f,0xfe,0x9f,0xd9,
	0x7f,0xff,0x6b,0xaf,0xbd,0xf6,0xda,0x7f,
	0xff,0xdf,0xff,0xff,0xff,0xff,0xff,0x6f,
	0xff,0xff,0xdf,0x7f,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xbf,
	0xcb,0xff,0xfe,0xff,0xff,0xfb,0xef,0xbf,
	0xff,0xff,0xd7,0xdf,0x7f,0xff,0xff,0xff,
	0x5f,0xef,0xfd,0xd7,0xff,0xfd,0xff,0xfd,
	0xf7,0xff,0xff,0xff,0xff,0xf7,0xff,0x7f,
	0xff,0x9f,0xfe,0xff,0xfe,0x7f,0xff,0xfd,
	0xf7,0xff,0xff,0xff,0x2f,0xff,0xd9,0xe7,
	0xff,0x7f,0xff,0xf2,0x1f,0xef,0xff,0xf9,
	0xff,0x1f,0x7f,0xfc,0xf9,0xc7,0x9f,0x7f,
	0xfc,0xf1,0xdf,0xf5,0xff,0xdf,0x7e,0xfd,
	0xfd,0xf7,0x5f,0x7f,0xff,0x7d,0xb1,0xdf,
	0x7f,0xff,0xed,0xf7,0x17,0xff,0xfd,0xf5,
	0xcf,0x7f,0xff,0xfd,0xf7,0xdf,0x7f,0xff,
	0xfd,0xf7,0xdf,0xff,0xa3,0xff,0xff,0xfd,
	0xe7,0x9a,0x6b,0xaf,0xb5,0xe6,0xff,0xed,
	0xfd,0xff,0xfd,0xff,0xff,0xff,0xfe,0xff,
	0xef,0xf7,0xff,0xff,0xf7,0xdf,0x7d,0xff,
	0xdd,0xff,0xff,0xff,0xfe,0xbf,0xfc,0xff,
	0x7f,0xbe,0xd1,0x46,0x1b,0x6d,0xb4,0xf9,
	0xc7,0x9f,0x7b,0xfe,0xf9,0xe5,0x9f,0xfd,
	0xb9,0xc1,0x9f,0x1b,0xfe,0xf9,0xe1,0x87,
	0x7f,0x7e,0xb8,0xe7,0x86,0x1f,0xfe,0xe4,
	0xff,0xff,0xaf,0xbf,0xfe,0xfa,0xeb,0xaf,
	0xff,0xfd,0xf6,0xdf,0x7f,0xfe,0xfd,0xff,
	0x67,0xfe,0xf9,0xfa,0xcf,0xfd,0xff,0xff,
	0xe7,0x9f,0x7f,0xfe,0xfd,0xe7,0xff,0xff,
	0x0f,0xff,0xff,0x7f,0xfd,0xf5,0xd3,0x5f,
	0x7f,0xfd,0xbf,0x9f,0xee,0xfa,0xef,0xae,
	0xfb,0xae,0xed,0x3f,0xd3,0xdf,0xba,0xef,
	0x3f,0xfb,0xec,0xfb,0xcf,0xbe,0xff,0xec,
	0xb3,0xbf,0xfa,0xff,0xff,0xeb,0xff,0xef,
	0xff,0xff,0xff,0x7f,0xff,0xbf,0xff,0xff,
	0xfb,0xeb,0xbf,0xff,0xff,0xe7,0xfb,0xbb,
	0xfe,0xff,0xe7,0x9f,0xff,0xff,0x79,0xff,
	0x9f,0x7f,0xfe,0xdf,0xff,0xfb,0x5e,0xfb,
	0x6f,0xbe,0xff,0xfe,0xfb,0xff,0xff,0xf9,
	0xff,0x9f,0x7f,0xfe,0xf9,0xff,0x5f,0xfe,
	0xf7,0xe3,0xff,0x5f,0x7e,0xf9,0xff,0x97,
	0xff,0x7f,0xf9,0xe5,0x2f,0xf6,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xef,0xff,0xff,
	0xff,0xf7,0xff,0xff,0xff,0xff,0xe3,0xff,
	0xef,0xff,0xff,0xff,0xff,0xff,0xfe,0xff,
	0xff,0xbf,0xff,0xff,0xff,0xff,0xf3,0xdf,
	0x5f,0x7f,0xed,0xf5,0xd7,0x1e,0x7f,0xf5,
	0xf5,0x57,0x5f,0x7f,0xfd,0xe5,0x97,0x7b,
	0xfd,0x75,0xd7,0x5f,0x7f,0xfd,0xf5,0x57,
	0x5f,0x7f,0xf5,0xf4,0xd7,0x5f,0xff,0x86,
	0xff,0xf7,0xff,0xff,0xff,0xbd,0xff,0xfe,
	0xfb,0xf7,0x8f,0xff,0xde,0xff,0xb3,0x8f,
	0xff,0xf8,0xf3,0x7f,0x3f,0xff,0xfb,0xf3,
	0xcf,0xff,0xff,0xfc,0xf7,0xcf,0x3f,0xff,
	0xdf,0xf9,0xff,0x9f,0x3f,0xfd,0xfd,0xff,
	0xef,0xff,0xef,0xcf,0xff,0xff,0xff,0xfe,
	0xdb,0xff,0xfd,0xff,0xff,0xff,0xff,0xff,
	0xff,0xf7,0xff,0xbf,0xff,0xf7,0xfb,0xef,
	0xff,0x7f,0xed,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xdf,0xff,0xfd,0xff,0xff,
	0x3f,0xff,0xfd,0xdb,0x3f,0xff,0xff,0xf0,
	0xff,0x3f,0xff,0xdc,0xfe,0xcf,0xff,0xff,
	0xfc,0xf3,0xff,0x3f,0xff,0xdf,0xff,0xfd,
	0xf9,0xf7,0x7d,0xff,0xfd,0xff,0xdf,0xff,
	0xff,0xff,0xff,0xff,0x9f,0xdd,0xff,0xd6,
	0x7f,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0x3f,0xf3,0xff,0xf7,
	0xf6,0xff,0xff,0xfd,0xfd,0xfd,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xfe,0xff,
	0xdf,0xfe,0xff,0xff,0xff,0xff,0xfb,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0x93,0xfb,
	0xff,0xfd,0xff,0xff,0xff,0xff,0xdf,0xfe,
	0xff,0xfd,0xff,0xff,0xff,0xff,0xf7,0xdf,
	0xfe,0xff,0x7d,0xf7,0xff,0xff,0xdf,0xbf,
	0xfd,0xfe,0xfe,0xff,0xff,0xff,0xff,0x1f,
	0xe8,0xbd,0xff,0xf7,0xfd,0x7f,0xff,0xff,
	0xff,0xff,0x7f,0xfa,0xff,0xf7,0x7f,0x7f,
	0xb7,0xff,0xdf,0x76,0xff,0xfd,0xf7,0xdf,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0x7f,0xf7,0xff,0xf7,0xff,0xff,0xef,0xff,
	0xff,0xff,0xff,0xff,0xff,0xbf,0xff,0xff,
	0xdf,0xeb,0x7f,0xff,0xeb,0xff,0xfe,0xff,
	0xff,0xff,0x7f,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0x2b,0xfe,0xff,0x37,0xfe,0xfc,
	0xff,0xfe,0xff,0xff,0xff,0xff,0xff,0x3f,
	0xff,0x7c,0xf3,0xff,0xff,0xff,0xf3,0xcf,
	0x3f,0x67,0xbf,0xfd,0xff,0xdd,0xff,0xff,
	0xff,0xff,0xff,0x7f,0xfc,0xff,0xff,0xf7,
	0xb7,0xfd,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xef,0xd7,0xdf,0xfe,0xdf,0xff,0xbf,
	0x73,0xfb,0xf9,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xe8,0x7f,0xed,
	0xbf,0xff,0xfd,0xff,0xff,0xff,0xfb,0xff,
	0xff,0xff,0xdf,0xdf,0xbf,0xf7,0xff,0xff,
	0xff,0xf5,0x9f,0xff,0xfe,0xff,0xe7,0xdf,
	0xff,0xff,0xff,0xff,0xff,0xff,0x7f,0xff,
	0x7d,0xfd,0xf5,0xef,0xbf,0xff,0xff,0xff,
	0xef,0xff,0xff,0xff,0xff,0xff,0xdf,0xff,
	0xdb,0xff,0xff,0xff,0xb7,0xf7,0xff,0xff,
	0xfe,0xef,0xff,0xff,0xff,0xff,0xff,0x5f,
	0xf9,0xef,0xfc,0xf3,0xff,0xff,0xff,0xef,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfe,0x7d,0xff,0xff,0xff,0xff,0xff,0xfe,
	0xff,0x7f,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfd,0x1d,0xff,0xff,0xfe,0xff,0xff,0xff,
	0xff,0xf3,0xe3,0xff,0xff,0xff,0xff,0xff,
	0x8f,0xfd,0xea,0xff,0xff,0xff,0xfc,0xd3,
	0xff,0xbb,0xbe,0xff,0xf8,0xcf,0xff,0xff,
	0xff,0xff,0x43,0xd6,0xff,0xff,0xfb,0xff,
	0xff,0xff,0xff,0x7f,0xfd,0xf5,0xff,0xff,
	0xff,0xff,0xec,0xe7,0xff,0xff,0xff,0xf3,
	0xff,0xfe,0xff,0xf9,0xff,0xcf,0xff,0xff,
	0xff,0xff,0xff,0xeb,0xf5,0x3e,0xf7,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xe7,0xff,0xff,0xff,0xff,
	0xff,0xff,0xf7,0xff,0xff,0xff,0xff,0xbf,
	0xff,0xff,0xff,0xff,0xdf,0x93,0xfd,0xfd,
	0xff,0xfe,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0x7f,0xfe,0xff,0xff,
	0xff,0xff,0xff,0xbf,0xff,0xff,0xff,0xff,
	0xff,0xfd,0xff,0xff,0xff,0xff,0xdb,0xfd,
	0xdf,0xab,0xee,0xfa,0xfe,0xff,0xff,0xbf,
	0xff,0xfe,0xfb,0xef,0xff,0xff,0xde,0xfb,
	0xbe,0xff,0xff,0xfb,0xef,0xbd,0xff,0xff,
	0xfb,0xff,0xff,0xff,0xfe,0xff,0xff,0xff,
	0x67,0xff,0xff,0xb6,0xda,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xee,0xff,0xff,0xff,0xff,0xff,0xef,0xfe,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0x0e,0xfe,0xff,0xff,0xff,0xdf,0xff,
	0xff,0xff,0xff,0xdf,0x7f,0xff,0xfd,0xff,
	0xdf,0x57,0xbf,0xff,0xff,0x7a,0xf9,0x7d,
	0xff,0xbf,0xff,0xff,0xff,0xff,0xff,0xfa,
	0xff,0xff,0x5f,0xfb,0xff,0xef,0xef,0xff,
	0xff,0xff,0xff,0xbf,0xff,0xff,0x5f,0xff,
	0xff,0xff,0xbf,0xfe,0xbf,0xff,0xf7,0xff,
	0xff,0xbb,0xff,0xd7,0xfb,0xff,0xff,0xff,
	0xf6,0xff,0xff,0xff,0xd9,0xff,0xff,0xed,
	0xb7,0xfb,0xff,0xff,0xff,0xfe,0xff,0xef,
	0xff,0xff,0xff,0xff,0x6f,0xee,0xfe,0x7f,
	0xbf,0xbf,0xff,0xfe,0xff,0xef,0xff,0xfd,
	0xff,0x7b,0xff,0xff,0xff,0x0f,0xfe,0xff,
	0xdf,0x7f,0x9e,0xff,0xfe,0xfb,0xe7,0xff,
	0x7f,0xfe,0xff,0xef,0xff,0x7f,0xfe,0xe7,
	0xbf,0xff,0xff,0xf9,0xe7,0xbf,0x7f,0xfe,
	0x7f,0xef,0x9f,0xff,0xfe,0xfb,0xff,0xe3,
	0xbf,0x76,0xdf,0x7f,0xfc,0xfd,0xf7,0x9f,
	0x7f,0xff,0xf1,0xf7,0xdf,0x7f,0xff,0xe1,
	0x9f,0x7f,0xff,0xfd,0xc7,0x9f,0x7f,0xff,
	0xf9,0xf7,0xdb,0x7f,0xfe,0xfd,0xf7,0xff,
	0xad,0xfd,0xf7,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0x7f,0xff,0xff,0xff,
	0xff,0xff,0xff,0xcf,0xf7,0xff,0xff,0xff,
	0xff,0xfb,0xfd,0xf7,0xff,0xff,0xff,0xff,
	0xfd,0xf7,0xdf,0xff,0xff,0xff,0xff,0xdf,
	0xff,0xff,0xff,0xdf,0x7f,0xff,0xff,0xff,
	0xdf,0x7f,0xff,0xfd,0x7f,0xdf,0x7f,0xff,
	0xfd,0xf7,0xff,0xae,0xff,0xdf,0xff,0xfe,
	0xf9,0xff,0xff,0x7f,0xfe,0xff,0xe7,0xff,
	0xff,0xff,0xff,0xe2,0x77,0xf6,0xdf,0xfe,
	0x9f,0x7d,0xf6,0xff,0xe7,0xeb,0xaf,0xf7,
	0xf9,0xff,0xff,0xff,0x5d,0xfb,0xef,0xfc,
	0xbf,0xdf,0x7e,0xff,0xed,0xf7,0xdf,0x7f,
	0xff,0xfd,0xf7,0xdf,0x16,0xff,0xf7,0xdf,
	0x76,0xff,0xfd,0xf7,0xdf,0x7e,0x5f,0x7c,
	0xf1,0xdf,0x7f,0xff,0xfd,0x37,0xfa,0xff,
	0xdb,0xdd,0x7f,0xff,0xff,0xff,0xdf,0xff,
	0xf3,0xeb,0xbf,0xff,0xff,0xff,0xdd,0xff,
	0xfe,0xbf,0xdf,0xf7,0xdf,0xfe,0xf7,0xff,
	0x5f,0xeb,0xff,0xf2,0xff,0xff,0xff,0xd7,
	0xff,0xff,0xe7,0x9f,0x7f,0xfe,0xf9,0xe7,
	0x9f,0x7f,0xfe,0xf9,0xe7,0x9f,0x7f,0xfe,
	0xe9,0x9e,0x7f,0xfc,0xf9,0xe7,0x9f,0x7f,
	0xee,0xf9,0x47,0x9f,0x7f,0xfe,0xf9,0xe7,
	0xbf,0xfe,0xff,0xff,0xff,0xdf,0x7f,0xff,
	0xff,0xff,0xff,0xff,0xff,0xfd,0xf7,0xdf,
	0xbf,0xff,0xf7,0x9f,0x3f,0xfe,0xff,0xff,
	0xdf,0x7f,0xff,0xff,0xf7,0xff,0x7f,0xff,
	0xfd,0xff,0xf2,0xfe,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xbf,
	0xff,0xfe,0xfd,0xfb,0xff,0xff,0xf5,0xff,
	0xff,0xff,0xff,0xff,0xff,0xf7,0xff,0xff,
	0xff,0xff,0xdf,0xaf,0xff,0xf7,0xfe,0xff,
	0xf7,0xdf,0xff,0xff,0xff,0xff,0xff,0x7f,
	0xff,0xff,0xff,0xef,0xff,0xfd,0xff,0xef,
	0xff,0xff,0xff,0xf7,0xdf,0xff,0xfe,0xf5,
	0xff,0xdf,0x7f,0xff,0xbf,0xfd,0xff,0xff,
	0xfd,0xff,0xfe,0xfb,0xef,0xbf,0xff,0xfe,
	0xfb,0xef,0xbf,0xff,0x7e,0xfb,0xbf,0xff,
	0x7e,0xff,0xef,0xbf,0xff,0xfe,0xfb,0xff,
	0xfd,0xff,0xfe,0xfb,0xef,0xff,0xec,0xff,
	0xdf,0x4f,0xff,0xff,0xef,0xff,0xff,0xff,
	0xff,0xff,0x3f,0xff,0xfd,0xbf,0xfe,0xff,
	0xff,0xff,0xff,0xff,0xef,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x3f,
	0xfd,0xff,0x55,0xd7,0x5f,0x7d,0xfd,0xd5,
	0xd7,0x5f,0x7b,0xf5,0xe5,0x97,0x5f,0x7d,
	0xf9,0xd7,0x5f,0x7f,0xfc,0xd5,0xd2,0x5f,
	0x7d,0xfd,0xd5,0xd7,0x5f,0x7d,0xfd,0xf5,
	0xef,0xba,0xff,0xfe,0xff,0xbf,0xff,0xfc,
	0xef,0xbe,0xff,0xff,0xff,0xf3,0xff,0xff,
	0x7b,0xff,0xcd,0xff,0xff,0x7b,0xef,0xff,
	0x3f,0xff,0xfc,0xff,0xcf,0xf7,0xff,0xfc,
	0xfb,0xdf,0x91,0xf7,0xfb,0xff,0xbf,0xff,
	0x7d,0xff,0xff,0xd7,0x5f,0x7f,0xf7,0xf7,
	0xef,0xff,0xdf,0x3f,0xff,0x7f,0xff,0xff,
	0xfd,0xb7,0xdc,0xfd,0xff,0x3f,0xff,0xff,
	0xfd,0xfe,0xfb,0xe7,0xf6,0xff,0xbf,0xff,
	0xef,0xbb,0xfd,0xff,0xff,0xff,0xfe,0xff,
	0xfc,0xff,0xff,0xff,0x3f,0xff,0xff,0xff,
	0xff,0xff,0xff,0x7e,0xdb,0xff,0xdf,0xff,
	0xff,0xbf,0x7f,0xbf,0xff,0xf2,0xff,0x51,
	0xdf,0xf7,0xff,0xff,0xff,0xbf,0xff,0xff,
	0xbf,0xfe,0xff,0xff,0x7f,0xff,0xff,0xff,
	0xff,0xfd,0xff,0xff,0xff,0xdf,0x7f,0xf5,
	0xc7,0xff,0xff,0xff,0xff,0xff,0xaf,0xff,
	0xdf,0xbf,0xf7,0xbf,0x7f,0xff,0xff,0xeb,
	0xff,0xff,0xff,0xfd,0xff,0x7f,0xdf,0xee,
	0xdb,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0x5f,0xff,0xff,0xff,0x7f,0xff,0xff,0x7f,
	0xfd,0xff,0xff,0xff,0xdf,0xf7,0xff,0xef,
	0xff,0xff,0x7d,0xff,0xef,0xf7,0xdf,0x7f,
	0x7f,0xff,0x9f,0xdf,0xfe,0xfd,0x9f,0xdf,
	0x7e,0xff,0xff,0xf5,0xff,0x7f,0xfb,0x7d,
	0xff,0xeb,0xff,0xdf,0x7f,0x7f,0x77,0xf7,
	0xff,0xfe,0xff,0xff,0xfd,0xff,0xdf,0x7d,
	0xff,0x5d,0x73,0x77,0xef,0xff,0xd7,0xff,
	0x7f,0xff,0xfd,0xff,0x5f,0xff,0xf7,0xfd,
	0xf5,0xff,0x3f,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xfe,0xff,0xf9,
	0xff,0xff,0xfd,0xff,0xff,0xff,0xff,0xe7,
	0xbf,0xff,0xff,0xff,0xfe,0xff,0xff,0xff,
	0xff,0xfb,0xff,0xef,0xdb,0xef,0xfc,0xe3,
	0xcf,0x3f,0xef,0xff,0xeb,0xff,0x3f,0x7f,
	0xff,0xff,0xff,0xf7,0xff,0xff,0x4f,0x7f,
	0xff,0xff,0xff,0xf7,0x39,0xff,0xef,0xf1,
	0xff,0x3f,0xff,0xfc,0xff,0xcd,0xff,0xef,
	0x7f,0xff,0xfe,0xf9,0xff,0xff,0xfb,0xff,
	0x7b,0xff,0xef,0xff,0xff,0xff,0xff,0xff,
	0xbe,0xfb,0xff,0xff,0xff,0xff,0xfb,0xff,
	0xbf,0xfd,0xff,0xfb,0xef,0xff,0x6f,0xfe,
	0x7b,0xaf,0xbf,0xe7,0xff,0xff,0xff,0xff,
	0xff,0x9f,0xff,0xeb,0xfb,0xed,0xbf,0xff,
	0xf7,0xf7,0xff,0xfd,0xfe,0xff,0xef,0x9f,
	0xff,0xfb,0xf9,0xff,0x9f,0x7f,0xfe,0x7f,
	0xe1,0xff,0xff,0xff,0x3f,0xf3,0xff,0xff,
	0xff,0xff,0xbb,0xee,0xff,0x8f,0x3f,0xff,
	0xbc,0x7f,0x5f,0xbf,0xf7,0xb2,0xbb,0x36,
	0xff,0xfc,0xfe,0xcf,0xff,0xef,0xfc,0xf3,
	0xff,0xa5,0xff,0xff,0xff,0xff,0xfd,0x7f,
	0xff,0xff,0xff,0xff,0xbd,0xff,0x7f,0xff,
	0xfd,0xb5,0xdf,0xfb,0xff,0xfd,0xdf,0xff,
	0xff,0xfd,0xf7,0xff,0x7f,0xff,0xff,0xf7,
	0xdf,0xfe,0x7f,0xf0,0x2f,0xe7,0xcf,0x7f,
	0xff,0xaf,0xff,0xff,0xff,0xff,0xfd,0xff,
	0xdf,0x77,0xff,0xad,0xfd,0x7f,0xff,0x7f,
	0xf7,0xbf,0x77,0xdf,0xfd,0xff,0xdd,0xfc,
	0xff,0xfd,0xf7,0x3f,0x6d,0xfd,0xc3,0xf9,
	0xff,0xfd,0x7f,0xfd,0xff,0xff,0xff,0xf4,
	0xff,0xff,0xbe,0xfb,0x6f,0xee,0x3f,0xfd,
	0xfc,0xbb,0xff,0xbf,0xfb,0xf6,0xff,0x6f,
	0xff,0xff,0xf4,0xd3,0xbf,0x76,0xef,0x5f,
	0xff,0x93,0xf5,0xff,0xff,0xff,0xff,0x7f,
	0xde,0xff,0x67,0xfd,0xfd,0xd2,0x7f,0xff,
	0xf7,0xdf,0xd9,0xff,0xff,0xfd,0xf7,0xff,
	0x4f,0xff,0xff,0xda,0x7f,0xff,0x7d,0xd8,
	0xff,0xfa,0xdf,0xa9,0xff,0xff,0xff,0xff,
	0xff,0xbf,0xff,0xff,0xeb,0xef,0xa7,0xfe,
	0xfb,0xed,0xbf,0x9f,0xfe,0xff,0xed,0xb3,
	0xff,0x7f,0xfb,0xff,0xa7,0xff,0xfe,0xbf,
	0xd1,0xff,0xd7,0xfe,0x7f,0xef,0xbf,0xff,
	0xfe,0xff,0xff,0xff,0xfb,0x7e,0xbf,0xff,
	0xfd,0xdf,0xdb,0xdf,0xf6,0xfa,0xfe,0xdb,
	0xff,0xbf,0xff,0xf7,0xfb,0xef,0xbf,0xf7,
	0xeb,0x5f,0xf6,0xbf,0xfa,0xff,0x7b,0xff,
	0xff,0xff,0xff,0xff,0xff,0xcf,0xff,0x7e,
	0xfc,0xef,0xff,0xee,0xff,0xe5,0xda,0xf7,
	0xff,0xff,0xff,0xfd,0xff,0xdf,0xff,0xff,
	0x99,0x6f,0x6f,0xe6,0xff,0x75,0xfe,0xdf,
	0xff,0xff,0xff,0xff,0xff,0x7f,0x7d,0xdf,
	0xdf,0x5f,0x7f,0xff,0xf7,0x5f,0xff,0xfd,
	0xf5,0xff,0x5b,0x7f,0xfd,0xff,0xd7,0xff,
	0x61,0xfd,0xfd,0xff,0xad,0xff,0xaf,0xf0,
	0xff,0xde,0xbf,0xff,0xfe,0xff,0xff,0x6b,
	0xff,0xb3,0xf7,0xda,0x7b,0x9f,0xfb,0xda,
	0xfb,0xaf,0xfd,0x8b,0xda,0x6b,0xff,0xbf,
	0xf6,0x7f,0x6b,0xef,0xfd,0x1f,0xfd,0xf7,
	0xfb,0xff,0xff,0xed,0xfe,0xfb,0xff,0xff,
	0xff,0xff,0xfb,0x7f,0xfb,0xed,0xff,0xef,
	0xff,0xed,0xff,0xff,0xef,0xff,0xff,0xff,
	0xff,0x7f,0xbb,0xf7,0xff,0xff,0xff,0xef,
	0x7f,0x9d,0xff,0xfe,0xff,0xe7,0x9f,0xff,
	0xfe,0xfb,0xef,0x9f,0xff,0xe6,0xdb,0xef,
	0x7f,0xfe,0xdb,0xef,0xff,0x7d,0xfe,0xfb,
	0x7f,0xbf,0xff,0xf7,0xf9,0x7f,0xbe,0xff,
	0x0f,0xfe,0x77,0xfe,0xfd,0xf7,0x9f,0x7f,
	0xfe,0xfd,0xf7,0xdf,0x7f,0xbc,0xfd,0xf7,
	0xd8,0xbf,0xb9,0x76,0xdf,0x6f,0xf7,0xf9,
	0xf6,0xdb,0x77,0xff,0x7d,0xc7,0xdf,0x6f,
	0xff,0xdf,0xdb,0xff,0xff,0x7f,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xbf,0xfd,
	0xff,0xdf,0xff,0xff,0xf7,0xff,0x3f,0xff,
	0xff,0xf6,0xdb,0xff,0xff,0xfd,0xff,0xdf,
	0xbf,0xff,0xbf,0xcd,0xff,0x7f,0xff,0xfd,
	0xff,0xdf,0x7f,0xff,0xfd,0xf7,0xff,0xff,
	0xfd,0xff,0xdf,0xff,0xff,0xf5,0xff,0x5f,
	0xff,0xff,0xdd,0x7f,0xff,0x7f,0xfb,0xff,
	0xdf,0xff,0xff,0xff,0x3f,0xfa,0xff,0xf1,
	0xff,0x6b,0x7f,0xff,0xfd,0xff,0xff,0xaf,
	0xfd,0xf9,0xdb,0x6b,0xbf,0x7d,0xe7,0x6f,
	0xfd,0xff,0xf6,0xf7,0x6f,0xbf,0xbd,0xfe,
	0x5b,0x9f,0xaf,0xbd,0xf6,0x5f,0xf3,0xff,
	0xdf,0x7e,0x5f,0xed,0xf7,0xdf,0x7e,0xff,
	0xfd,0x95,0xdf,0x57,0x5d,0x75,0xd5,0x7f,
	0x7f,0xf5,0xf7,0xd5,0x7f,0x5f,0x75,0xf5,
	0xc4,0x57,0xff,0x7d,0xf7,0x5d,0xff,0x87,
	0x7f,0xfd,0xfe,0xef,0xfb,0xff,0xfe,0xfb,
	0xdf,0xff,0xeb,0xff,0xfe,0xff,0xfb,0xef,
	0xfd,0xef,0xfb,0xff,0xff,0xff,0xff,0xfb,
	0xff,0xaf,0xff,0xff,0xef,0xef,0xaf,0xff,
	0x7f,0xfd,0xff,0x7f,0xfe,0xd9,0xe1,0x9f,
	0x7f,0xfe,0xf9,0x67,0x87,0x7f,0x7c,0xd8,
	0x41,0x87,0xee,0xf1,0xe1,0x1f,0x1f,0xfe,
	0xd1,0xc1,0x07,0x7d,0x7c,0xf8,0xe7,0x07,
	0x1f,0xfe,0xee,0xff,0xff,0x7f,0xbd,0xff,
	0xf7,0xff,0x7f,0xff,0xfd,0xfb,0xdf,0xbf,
	0xbf,0xff,0xff,0x7f,0xff,0xfe,0xf7,0xef,
	0x7f,0xfe,0xff,0xfb,0xcb,0xbf,0xff,0xff,
	0xfa,0xef,0xff,0x77,0xff,0xff,0xff,0xfb,
	0x7d,0xbb,0xff,0xff,0xfb,0xff,0xff,0xed,
	0x7f,0xdd,0x7c,0xf3,0xad,0xff,0x77,0xbb,
	0x5f,0xb7,0xff,0x7d,0xd3,0x4d,0x7f,0xdd,
	0xfe,0xf7,0xed,0xb7,0x3f,0xfa,0xff,0xff,
	0xff,0xff,0xe7,0xff,0xff,0xff,0x7f,0xff,
	0x9e,0xf7,0xff,0xf9,0xe7,0x9f,0xdf,0xbf,
	0xe7,0xff,0x7f,0xfe,0xff,0xe7,0x9f,0xf2,
	0xff,0xf9,0xff,0x9e,0x7b,0xfe,0xc9,0xff,
	0xff,0xfe,0xff,0x4f,0xbe,0xff,0xfe,0xfb,
	0xef,0x3f,0xf9,0xde,0x93,0x4f,0x3c,0xf9,
	0x7b,0x4f,0xbe,0xf7,0xe4,0x7b,0x4f,0x3c,
	0xd1,0xde,0x93,0xef,0x3f,0xf9,0xe4,0xaf,
	0xfc,0xff,0xff,0xff,0xcf,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xcb,0xef,
	0xff,0xf7,0x7f,0xff,0xbf,0xff,0xff,0xfb,
	0x7f,0xff,0xfb,0xfe,0xff,0xff,0xbf,0xff,
	0xff,0xf2,0xff,0x5f,0x7f,0xfd,0xd5,0xd7,
	0x5f,0x7d,0xfd,0xb5,0x57,0x5f,0x7e,0xf1,
	0xe1,0x96,0x7f,0xf9,0xc5,0x97,0x5f,0x7f,
	0xdd,0xe5,0x97,0x5f,0x7e,0xfd,0xf5,0xd3,
	0x5d,0xff,0x96,0xff,0xff,0xff,0xff,0xff,
	0x8f,0x3d,0xfe,0xbf,0xf3,0x7f,0xbb,0xfe,
	0xfc,0xf1,0xcd,0x9f,0xf8,0xff,0xff,0x1f,
	0xff,0xf8,0xf3,0x7b,0x3f,0x7e,0xff,0xff,
	0xff,0xff,0xff,0xbf,0xf8,0xff,0x7f,0xff,
	0xfb,0xcf,0x5f,0xff,0xfe,0xce,0xff,0xff,
	0x7d,0xff,0xfb,0xfb,0xdf,0x7d,0xfd,0xef,
	0xef,0xff,0xfc,0xfe,0xef,0x7f,0xff,0xff,
	0xfb,0xee,0xfb,0xbf,0x7f,0xe0,0xff,0xff,
	0xff,0xfd,0xff,0xfd,0xf6,0xfd,0xfd,0xbf,
	0xff,0xdf,0xcf,0x3b,0xff,0xfc,0xbb,0xfd,
	0xff,0xff,0xf3,0xdf,0x3f,0xf7,0xff,0xef,
	0xff,0xff,0xff,0xff,0xff,0xff,0x4f,0xfe,
	0xff,0xf7,0xff,0xff,0xfd,0xff,0xdf,0xff,
	0xcf,0xdf,0xf3,0xfe,0xbf,0xf5,0xd7,0xbb,
	0xbf,0xff,0xfe,0xff,0xff,0xcf,0xf7,0xff,
	0xff,0xff,0xff,0xff,0xff,0xdb,0xff,0x3f,
	0xfb,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0x7f,0xff,0xff,0xff,0xbf,0xeb,
	0xfe,0xfe,0xff,0xff,0xff,0xff,0xff,0xdf,
	0xfe,0xff,0xff,0xff,0xff,0xfb,0xff,0xbb,
	0xff,0x87,0xff,0xff,0xff,0xff,0xbc,0xff,
	0xff,0xfe,0xff,0xff,0xa7,0xdf,0x7f,0x7f,
	0xfc,0xfd,0xf7,0xff,0xf5,0xff,0xff,0xff,
	0xe7,0xfd,0xff,0xff,0x7d,0xff,0xf9,0xf7,
	0xdf,0xff,0xcf,0xfe,0xff,0xff,0xdf,0xff,
	0xff,0xfe,0xff,0xff,0xff,0xcf,0x3f,0x66,
	0x75,0xeb,0xff,0x3d,0xfa,0xf7,0xff,0xbf,
	0xfe,0xff,0xd7,0xef,0xff,0xef,0xfd,0xf3,
	0xc7,0x7d,0xff,0xef,0x76,0x7f,0xff,0xff,
	0xff,0xff,0xff,0xff,0x7f,0xff,0x6f,0xff,
	0xdf,0xfe,0xff,0xff,0xff,0xfd,0xf7,0xef,
	0xef,0xfe,0xff,0xff,0x7f,0xff,0xff,0xfb,
	0xdf,0x6f,0xbf,0xff,0xff,0x97,0xff,0xff,
	0xff,0xff,0xfc,0xff,0xc7,0xdf,0x5f,0xfc,
	0xf3,0xce,0x3f,0x77,0xff,0xf3,0xa7,0xff,
	0xbf,0xf3,0xff,0xff,0xef,0xff,0xdf,0xff,
	0x3f,0xff,0xfc,0xdd,0xc7,0xff,0xbf,0xfd,
	0xff,0xff,0xff,0xef,0xff,0xff,0xfe,0xff,
	0xef,0xbf,0x7f,0x6b,0xfb,0xff,0xbf,0xff,
	0xfd,0xfb,0xdf,0xfe,0xff,0xbf,0xff,0xef,
	0xff,0xff,0xfb,0xb7,0xff,0xff,0xfe,0xff,
	0xc4,0xff,0xff,0xff,0x7f,0xff,0xff,0xf7,
	0xff,0x7f,0x7f,0xf9,0xdf,0x9d,0x9f,0x7e,
	0xf9,0xff,0xff,0x7e,0xff,0xff,0xf7,0xbf,
	0xfa,0xfb,0xdf,0x8f,0xdf,0xfd,0xeb,0xe7,
	0xdf,0x77,0xff,0xff,0xff,0xff,0xbe,0xfb,
	0xff,0xef,0xff,0xff,0xcf,0xff,0xff,0xbd,
	0xf7,0xcf,0xbf,0xbf,0xf3,0xff,0x7f,0xff,
	0xff,0xf7,0xcf,0xff,0xff,0xfd,0xff,0xcf,
	0x3b,0x7f,0x5f,0xfa,0xff,0xff,0xff,0xfd,
	0xff,0xf3,0xff,0xff,0xff,0x7a,0xff,0xff,
	0xf3,0xd7,0x7f,0xff,0xff,0xdf,0xff,0xff,
	0xfb,0xbd,0xbf,0x5f,0x7f,0xff,0xf5,0xff,
	0x5f,0xff,0xfd,0xf7,0x0b,0xff,0xff,0xff,
	0xff,0xb7,0xde,0x7a,0xf3,0xab,0xe7,0xdf,
	0xff,0xdf,0xfd,0xf7,0xdd,0xfa,0xff,0xf7,
	0xbf,0x7f,0xff,0x7f,0xf7,0xdd,0xfa,0xde,
	0xfd,0xff,0xdf,0x7f,0xff,0x83,0xd6,0xff,
	0xff,0xff,0xdb,0xff,0xfe,0xcd,0xff,0xdf,
	0x5f,0xf3,0xf7,0xf6,0xbf,0x6f,0xe7,0xff,
	0xb3,0xff,0xbf,0xfb,0xff,0xbb,0xff,0xe6,
	0xff,0xf6,0x7f,0xcf,0xfe,0xfd,0xeb,0xf7,
	0xfe,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0x53,0xff,0xfd,0xff,0xf7,0xdf,0xcf,0xff,
	0xff,0x7e,0xff,0xff,0xff,0xff,0xd9,0x7f,
	0xff,0xff,0xf7,0xd3,0xcf,0x3f,0xf5,0xdf,
	0x8f,0xfd,0xff,0xff,0xff,0xfe,0xff,0xff,
	0xff,0xcf,0xfa,0xff,0xff,0xbf,0xcf,0x7b,
	0xfe,0xff,0xdf,0xfe,0xff,0xef,0xff,0xcf,
	0xfe,0xff,0xff,0xb3,0xcf,0x7b,0xfe,0xad,
	0xff,0x1b,0xfd,0xfe,0xef,0xff,0xdf,0xfe,
	0xfb,0xef,0xbf,0xff,0x5e,0x37,0xed,0xf7,
	0xdf,0xdf,0xff,0xb5,0xfd,0xff,0x7b,0xef,
	0xff,0xff,0x5f,0xfb,0xef,0xbf,0xdf,0x6c,
	0x9b,0xdd,0xfe,0x69,0xff,0xff,0xff,0xbf,
	0x76,0xff,0xff,0xff,0xff,0xfb,0x7e,0xa9,
	0xef,0x9f,0x5f,0xfe,0xaf,0xfe,0x7f,0xff,
	0x6b,0xf5,0xff,0xff,0x5e,0xfd,0xff,0xbd,
	0xfb,0x9e,0xf8,0xf5,0x66,0x9e,0xff,0xfd,
	0xff,0xf7,0xfd,0xf9,0xe3,0xff,0x75,0xd6,
	0x5f,0xfd,0x7d,0xf5,0xd7,0x7f,0xed,0xf5,
	0xfd,0xde,0xff,0xff,0xf5,0xd7,0xff,0x7f,
	0xfd,0xd5,0xd7,0x7f,0xff,0x9f,0xd9,0xfd,
	0xff,0xff,0xaf,0xfd,0xc3,0x0f,0xbf,0xaf,
	0xb0,0xf7,0xfa,0x7b,0xaf,0xbd,0xf6,0xa9,
	0xaf,0xfd,0xfb,0xda,0xff,0xaf,0xbd,0xf7,
	0xef,0x6b,0xaf,0xbf,0xf6,0xde,0xff,0xd9,
	0xfe,0xff,0xff,0xff,0xff,0xef,0xbf,0xff,
	0xfe,0xfb,0xff,0xbf,0xff,0xb7,0xdf,0x7e,
	0xff,0xfe,0xff,0xef,0xff,0xed,0xff,0xdf,
	0xfe,0xbf,0xff,0xff,0x9b,0xff,0xff,0xed,
	0x6f,0xfe,0xff,0xff,0xef,0xff,0x7f,0xfe,
	0xf9,0xe7,0x9f,0xff,0xe6,0xf9,0x6f,0xbf,
	0xfd,0xd6,0xe7,0xbf,0x7f,0xfe,0xdf,0xef,
	0xbf,0xfd,0xff,0xf9,0xff,0x9f,0xff,0xff,
	0xfb,0xff,0xe2,0xff,0xf7,0xdf,0x6f,0xff,
	0xf9,0xe7,0x9f,0x7f,0xfe,0xfd,0xc7,0xdf,
	0x6d,0xf7,0x9d,0x9a,0x6b,0xff,0xf9,0x77,
	0xdf,0x7f,0xb7,0xbd,0xe6,0xdf,0x7f,0xbc,
	0xfd,0xf7,0xff,0x85,0xfd,0xff,0xff,0xef,
	0xff,0xff,0xff,0xff,0xff,0xff,0xfd,0xff,
	0xff,0xef,0xff,0xff,0xf7,0xff,0xff,0xff,
	0xff,0xff,0x7f,0xbf,0xfd,0xf7,0xff,0xff,
	0xff,0xfc,0xff,0xff,0xdb,0xfc,0xff,0xff,
	0xdf,0xff,0xff,0xfd,0xf7,0xdf,0x7f,0xf7,
	0xff,0xff,0xff,0xff,0xff,0xdf,0xdf,0xff,
	0xff,0xfd,0xff,0xdf,0xfe,0xfb,0xf7,0xf7,
	0xff,0xff,0xf7,0xff,0xff,0xff,0xaf,0xff,
	0xff,0xff,0xff,0xf6,0xc7,0x1f,0x7d,0xfe,
	0xf1,0xda,0x9f,0xaf,0xfd,0xf6,0x5a,0x37,
	0xfe,0xf6,0xf7,0x6f,0xff,0xff,0xf6,0x5b,
	0x8f,0xaf,0xb5,0xf8,0xdb,0x6b,0xff,0x4d,
	0xff,0xff,0xed,0xf7,0xdd,0x7f,0xff,0xfd,
	0xb7,0xdf,0x77,0xfd,0x7d,0xd5,0x5d,0x57,
	0x7d,0xf3,0x55,0x7e,0x7f,0xfd,0xf7,0x57,
	0x57,0xcf,0xfc,0xb5,0xc5,0x57,0x5f,0xf5,
	0x7f,0xf8,0xff,0xbf,0xff,0xbf,0xff,0xeb,
	0xef,0xff,0xff,0xff,0xde,0xf7,0xef,0xbf,
	0xff,0x5e,0xef,0xfe,0xff,0xef,0xfa,0xff,
	0xbf,0xff,0xde,0xfb,0xeb,0x3f,0xff,0xfe,
	0xfe,0xff,0xc3,0xfb,0xff,0xe7,0x9f,0x1d,
	0xfe,0xf9,0xe7,0x9e,0x7b,0x7e,0xf8,0x67,
	0x87,0x1d,0x76,0xc8,0x1e,0x1f,0xfe,0xd1,
	0xe1,0x1f,0x1f,0x74,0xb0,0x67,0x87,0x7d,
	0x74,0xf0,0xe1,0xdf,0xfe,0xff,0xfd,0xf7,
	0xef,0xf7,0xff,0xff,0xf7,0x9f,0xaf,0xdf,
	0xff,0xff,0xed,0xe7,0xdf,0xf3,0xef,0x77,
	0xff,0x7e,0xf7,0xef,0xf7,0xdf,0xfc,0xfb,
	0xdd,0xff,0xdf,0x7e,0xff,0xf7,0xff,0xff,
	0xff,0xdf,0x97,0xff,0xff,0xff,0xff,0x7f,
	0x5f,0xfe,0xff,0xc5,0x17,0x5f,0xb2,0x5f,
	0x97,0xfb,0x77,0xf9,0x7f,0x17,0x5f,0xfc,
	0xff,0xe5,0x7f,0x5f,0x76,0xf9,0xb7,0xff,
	0x7f,0xff,0xfd,0x7b,0xfe,0xff,0xff,0xfd,
	0xff,0xef,0xf9,0xff,0x9e,0x7b,0xfe,0xf9,
	0xfe,0x7f,0xfe,0xff,0xe7,0xfd,0x7b,0xfe,
	0x39,0xff,0x9e,0xf3,0xff,0xf9,0xe7,0x5f,
	0xfc,0xff,0xef,0xbf,0xdf,0xfc,0xfb,0xef,
	0xbf,0xff,0xfe,0xf3,0xef,0x37,0xdf,0xfc,
	0xf3,0xbd,0xf7,0xfc,0xfb,0xcf,0xbf,0xf7,
	0xdc,0x73,0xef,0x3f,0xdf,0xde,0xf3,0xcf,
	0xff,0xe1,0xff,0xbf,0xff,0xbe,0xfc,0xff,
	0xff,0xff,0xfe,0xff,0xf2,0xff,0x3b,0xff,
	0xff,0xfe,0xff,0xff,0xfb,0xff,0xfb,0xfb,
	0xbf,0xff,0xe3,0xff,0xeb,0xff,0xff,0xff,
	0xff,0xff,0x6f,0xfd,0xff,0xf5,0xd7,0x5e,
	0x7f,0xfd,0xd5,0xd7,0x5e,0x7d,0xf5,0xa5,
	0x57,0x5f,0x7e,0xed,0x17,0x5f,0x7b,0xf1,
	0xd5,0x57,0x5d,0x7d,0xfd,0xd5,0x96,0x5f,
	0x7b,0xfd,0xf5,0xaf,0xfb,0xff,0xf3,0xdf,
	0xdf,0xff,0xfc,0xff,0xdf,0xff,0x7f,0xef,
	0xff,0xff,0x3f,0x7f,0xff,0xfd,0x3f,0xf3,
	0xf7,0xf1,0x8f,0x3d,0xff,0xec,0xf3,0x7f,
	0x3f,0xfe,0xfc,0xff,0xff,0x8f,0xff,0xf3,
	0xdf,0xff,0x7f,0xff,0xbd,0xaf,0xdf,0xbf,
	0xff,0xff,0xff,0xdf,0x7f,0xff,0xff,0xff,
	0x7f,0xbf,0xff,0x5d,0xff,0xff,0xff,0xcb,
	0xff,0x7f,0xfd,0xff,0xfe,0xbf,0x57,0xfe,
	0xff,0xfe,0xff,0xff,0xcf,0xff,0xff,0xdf,
	0xfd,0xff,0xf5,0xef,0xbf,0xf3,0xff,0x3f,
	0xff,0xf2,0xfb,0x3f,0xff,0xdf,0xf1,0xc7,
	0xff,0xdf,0xff,0xff,0xcf,0xff,0xff,0xff,
	0xf4,0xff,0xff,0xfd,0xff,0xef,0xff,0xff,
	0xbe,0xff,0xb6,0xfb,0xff,0xd6,0xff,0x7f,
	0xfd,0xff,0xff,0xdb,0xff,0xb5,0xff,0xff,
	0xff,0xff,0x9f,0xd7,0x5f,0x7f,0xfd,0xf5,
	0xff,0x97,0xff,0xff,0xff,0xfb,0xbf,0xff,
	0xff,0xff,0xbf,0xdf,0xff,0xff,0xff,0xf6,
	0xdf,0xef,0xcf,0xff,0xff,0x7d,0xaf,0xfd,
	0xff,0xdf,0xff,0xfd,0xfd,0xb6,0xfb,0xdf,
	0x7d,0xfd,0x9f,0xfc,0xff,0xff,0x7f,0x7f,
	0xff,0xf7,0xff,0xff,0xff,0xfd,0xff,0xdf,
	0x7d,0xff,0xdd,0x77,0x9f,0xff,0xf9,0xff,
	0x9f,0xef,0xff,0xef,0xff,0xff,0xfd,0xf7,
	0xdd,0x77,0xfd,0xff,0xe2,0xff,0xff,0xff,
	0xff,0xff,0xff,0xdf,0xff,0xfd,0xfd,0x77,
	0xdf,0xff,0xff,0xed,0xf7,0xfb,0xff,0xdd,
	0xff,0xdf,0xfb,0xff,0xfd,0xb6,0xdf,0xff,
	0xff,0xfb,0xff,0xf7,0xbf,0x1f,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xbf,0xff,0xff,
	0xfb,0xff,0xef,0xb7,0xdf,0xff,0x9f,0xf7,
	0xff,0xfd,0xff,0xbf,0xff,0xfe,0xdf,0xef,
	0xff,0xf7,0xbf,0xff,0xff,0xfd,0x7f,0xfb,
	0xff,0xf8,0xff,0xcf,0x3f,0xff,0xff,0xf3,
	0xff,0x37,0xdf,0xbc,0xfd,0xff,0x3f,0xfd,
	0xfe,0xfe,0xff,0x7f,0xfc,0xff,0xcf,0xdf,
	0xff,0xff,0xfd,0xcf,0x3f,0xff,0xfc,0xff,
	0xd1,0xff,0xef,0xff,0xff,0xfe,0x6d,0xfb,
	0xdf,0xfa,0xff,0xfa,0xb7,0xef,0xff,0xff,
	0xfb,0xfb,0xff,0xff,0xff,0xef,0xff,0xff,
	0xbe,0xff,0xff,0xff,0x7f,0xf5,0xfb,0xb7,
	0xff,0x3f,0xfe,0xff,0xff,0xff,0xf5,0x77,
	0xf7,0xdd,0xf7,0x9c,0xdf,0xff,0xdd,0x5f,
	0xe9,0xd7,0xf7,0x7f,0xfd,0x77,0x7d,0xff,
	0xf7,0xf7,0xf7,0xff,0xfd,0xfa,0xdd,0xdf,
	0xff,0xfc,0x7f,0xa3,0xff,0xff,0xff,0xfb,
	0xff,0xff,0xef,0x7f,0xff,0xff,0xef,0xbf,
	0xff,0x7f,0xfb,0xff,0xff,0xff,0xff,0xfb,
	0xbf,0xff,0xfe,0xfb,0xf7,0xff,0xdf,0xff,
	0xff,0xff,0xf1,0xff,0xb1,0xff,0xff,0xff,
	0xff,0xff,0xbf,0xfe,0xff,0xff,0xff,0x7f,
	0xff,0xfa,0xff,0xbd,0xff,0xde,0xff,0xff,
	0xff,0xff,0xf7,0xff,0xff,0xff,0xff,0xff,
	0xfe,0xff,0xff,0xff,0xff,0x7f,0xf0,0xff,
	0xff,0x3f,0x7e,0xff,0xfb,0xff,0xff,0xfa,
	0xfe,0xad,0x3f,0xbd,0xf7,0xfe,0xad,0xbf,
	0xfc,0xeb,0xab,0x37,0xff,0x7f,0xff,0x2f,
	0xbf,0xfe,0xff,0xeb,0xfd,0xff,0xef,0x63,
	0xfd,0xff,0xfc,0xe7,0xfd,0xff,0xff,0xcf,
	0xe7,0xff,0x77,0xfe,0xe9,0xa7,0xff,0x77,
	0xfe,0xd7,0x9f,0x5f,0xdd,0xf5,0xff,0xfd,
	0x7f,0xfa,0xf9,0xff,0x5f,0x77,0xfe,0xff,
	0x2f,0xef,0xff,0xff,0xfb,0xff,0xff,0xff,
	0xff,0xfd,0xff,0xff,0xff,0x7f,0xff,0xfd,
	0xff,0xff,0xff,0xf5,0xd7,0xff,0xff,0xff,
	0xff,0xf7,0xdf,0x7f,0xfd,0xff,0xff,0xff,
	0xff,0x7f,0xdb,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xfb,0xff,0xff,0xff,0xff,0xfb,
	0xef,0xff,0xff,0xff,0xa9,0xa7,0xfe,0xff,
	0xff,0xff,0xb3,0xff,0x7e,0xeb,0xff,0xff,
	0xff,0xff,0xb7,0xd8,0xff,0xff,0xfe,0xf9,
	0xef,0xbf,0xff,0x7e,0xb3,0xef,0xbf,0xff,
	0xde,0x7b,0xef,0xbf,0xff,0xfb,0xef,0xbf,
	0xff,0xfe,0xfb,0xef,0xb5,0xf7,0xf6,0xfb,
	0xef,0xbf,0xd7,0xfe,0xa7,0xf6,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xee,0xff,0xff,
	0xff,0xff,0xee,0xbb,0xff,0xff,0xff,0xbf,
	0xff,0xfe,0xff,0xff,0xff,0xaf,0xbe,0xfb,
	0xef,0xff,0xff,0xbf,0xfa,0x2f,0xe6,0xff,
	0xdf,0x7f,0xfe,0xfd,0xf7,0xdf,0x5e,0xff,
	0xff,0xbf,0xde,0xf7,0xdf,0xfd,0xbf,0xfe,
	0x5f,0x6f,0xfd,0xdf,0x7f,0xff,0x2f,0xff,
	0xfd,0xd7,0xff,0xff,0xf7,0xd5,0xff,0x89,
	0xfc,0xff,0xf3,0x5f,0xdf,0xff,0xff,0xa3,
	0xdf,0xef,0x3f,0xff,0xbf,0xfb,0xfe,0x3f,
	0xff,0xcb,0x6a,0xaa,0xff,0xfd,0x5f,0x4f,
	0xb9,0xed,0xbe,0xfa,0xef,0x7f,0xaf,0xff,
	0xef,0xfc,0xff,0xfb,0xbf,0xbf,0xff,0xfe,
	0xff,0xef,0xbd,0xff,0xfe,0xfb,0xef,0xbf,
	0xff,0xfe,0x6f,0xbb,0xf1,0xfe,0xfb,0xef,
	0xbf,0xff,0x96,0xfb,0xef,0xbf,0xff,0xfe,
	0xfb,0xff,0xeb,0xff,0x9f,0xff,0xfe,0xf9,
	0xe7,0xff,0x7f,0xfe,0xf9,0xe7,0x9f,0x7f,
	0xfe,0xf9,0xe7,0x7f,0xfe,0xf9,0xe7,0x9f,
	0x7f,0xfe,0xf9,0x67,0x9f,0x7f,0xfe,0xf9,
	0xe7,0x9f,0xff,0x67,0xfe,0x7f,0xfc,0xfd,
	0xc7,0x1f,0x7f,0xff,0xf1,0xe7,0x9f,0x7f,
	0xfc,0xe9,0xc7,0x9f,0xff,0xf9,0xe7,0x9f,
	0x7f,0xfc,0xf1,0xe7,0x9f,0x77,0xfe,0xf9,
	0xe7,0x1f,0x7f,0xfc,0x1f,0xd8,0xff,0xff,
	0x7f,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xdd,0xff,
	0xff,0xff,0xfd,0xff,0xff,0xff,0xff,0xff,
	0xf7,0xdf,0xff,0xff,0xfd,0xff,0xdf,0xff,
	0xfd,0xf7,0xdf,0xff,0xff,0xff,0xf7,0xdf,
	0x7f,0xff,0xfd,0xf7,0xff,0xff,0xff,0x8f,
	0xfa,0xff,0xd9,0xff,0x1f,0x7d,0xfe,0xff,
	0x67,0x9f,0x7f,0xfc,0xf9,0xf7,0x8d,0x2f,
	0x7c,0xe2,0x89,0x2f,0xfe,0xd1,0xe7,0x5f,
	0x2d,0xf7,0xd9,0xe3,0xdf,0x7f,0xf4,0xf8,
	0x7f,0xf3,0xbf,0xdf,0x7f,0xff,0xfd,0xf7,
	0xdf,0x7e,0xff,0xfd,0xb7,0xdf,0x7f,0x7f,
	0x7c,0xf1,0x17,0x5b,0x2c,0xf1,0xdf,0x7f,
	0xff,0x6d,0xf1,0xdf,0x1f,0xff,0xfd,0xf7,
	0xc7,0xff,0xb3,0xff,0xff,0xfa,0xff,0xff,
	0xdf,0xff,0xff,0xdd,0xff,0xbf,0x7f,0xff,
	0xdf,0xeb,0xab,0xf5,0xff,0x3f,0xeb,0xff,
	0xff,0xeb,0xff,0xff,0xfc,0xf7,0xde,0xbf,
	0xff,0x3f,0xff,0x1f,0xfd,0xff,0x7f,0xfe,
	0xb9,0xe1,0x9f,0x7f,0xfe,0xb9,0xe7,0x87,
	0x7f,0x7e,0xd8,0x45,0x87,0xee,0xb9,0x45,
	0x9e,0x1b,0xfe,0xf9,0xe1,0x97,0x7b,0x66,
	0xf8,0xe7,0x86,0x1d,0x7e,0xe5,0xff,0xdf,
	0xff,0xfd,0xfd,0xf7,0x9f,0x7f,0xfe,0xfd,
	0xf7,0xff,0xff,0xff,0xee,0xf7,0x2f,0xbe,
	0xe8,0xf7,0xdf,0xff,0xff,0xf9,0xe3,0xdf,
	0x3f,0xf3,0xfd,0xf7,0xef,0xfc,0x07,0xff,
	0xff,0xfe,0xff,0x7f,0xfb,0xff,0xff,0xff,
	0xff,0xff,0xed,0xff,0xff,0xf6,0xfb,0x2d,
	0xfd,0xf5,0xfb,0xff,0xb7,0xff,0x7f,0xf3,
	0xef,0xff,0xfd,0xee,0xff,0x6d,0xbf,0xff,
	0xf8,0xff,0xff,0xff,0x7f,0xe7,0xfd,0xff,
	0xff,0x7f,0xff,0x9d,0xff,0xff,0xb8,0xef,
	0x9d,0xff,0xff,0xef,0xfd,0x77,0xfe,0xff,
	0xe7,0xae,0xf7,0xcf,0xfb,0xff,0x9d,0xfb,
	0xfe,0xd9,0xff,0xff,0xfe,0xff,0x4f,0xbe,
	0xff,0xfe,0xfb,0xef,0x3f,0xf9,0xfe,0x83,
	0x2d,0x3e,0xf9,0xfb,0x2f,0xbe,0xff,0xe4,
	0xfb,0x4f,0x3c,0xf8,0x7e,0x8b,0xef,0x3f,
	0xd9,0x62,0x3f,0xfe,0xff,0xff,0xff,0xff,
	0xff,0xff,0xfb,0xdf,0xff,0xff,0xff,0xd7,
	0xf3,0xcb,0xff,0xff,0x7f,0xff,0xff,0xf7,
	0xff,0xff,0xbf,0xef,0xbf,0xdf,0xff,0xff,
	0xfd,0xef,0xfb,0x3b,0xd7,0xff,0x5f,0x7f,
	0xe9,0xf5,0x97,0x5f,0x7a,0xf1,0xe5,0x53,
	0x4f,0x7d,0xf5,0xd5,0x57,0x7f,0xf5,0xd5,
	0xc7,0x4f,0x3f,0xf5,0x55,0x97,0x5a,0x7f,
	0xec,0xb5,0xd7,0x5f,0xb3,0x87,0xff,0xff,
	0xff,0xff,0xf3,0xcf,0xff,0xff,0xfb,0xf3,
	0xff,0xfa,0xfe,0xf8,0x5d,0xff,0xdf,0xdf,
	0xaf,0xdf,0xf5,0xf5,0xfb,0xff,0x0e,0x3f,
	0xfe,0xbc,0x63,0xff,0xff,0xff,0x9f,0xf8,
	0xdf,0xff,0xff,0xed,0xbf,0xbf,0xbf,0xff,
	0xef,0xf9,0xf7,0xfb,0xfd,0xff,0xef,0xf5,
	0x7f,0xff,0xdf,0xef,0xff,0xfe,0xfb,0xdf,
	0xdf,0xfc,0xf6,0xf3,0xff,0xbf,0xff,0x7f,
	0xe3,0xff,0xff,0xff,0xff,0xfe,0xbd,0xff,
	0x7f,0x7d,0xf7,0xdf,0xff,0xdd,0xff,0xb7,
	0xff,0xf1,0xff,0x77,0xdf,0xff,0xff,0xfd,
	0xd7,0xfd,0xfd,0xef,0x5d,0xef,0xff,0x7f,
	0xff,0x67,0xff,0xff,0xff,0x9f,0xff,0xfe,
	0xff,0xed,0xff,0xef,0x3f,0xff,0xff,0xbb,
	0xff,0x87,0xbf,0xfb,0xfe,0xc6,0xef,0xff,
	0xff,0xff,0xd7,0xff,0xbf,0xff,0xff,0xfb,
	0xfb,0xfe,0xff,0xfa,0xbf,0xff,0xff,0xff,
	0xff,0xff,0xff,0xfe,0xff,0xef,0xbf,0xbf,
	0xff,0xff,0x7f,0xff,0xfe,0xff,0x7f,0xff,
	0x7f,0xfe,0xff,0x9f,0xfd,0xff,0xfd,0xfb,
	0xfe,0xff,0xff,0xff,0xc3,0xff,0xff,0xef,
	0xf7,0xff,0xef,0xdf,0x7d,0xff,0xff,0xff,
	0xff,0xff,0xff,0xed,0x7f,0xf6,0xdf,0xff,
	0x76,0xcf,0xfe,0xf7,0xff,0xef,0xfb,0xff,
	0xbb,0xfd,0xff,0xff,0xfd,0x4f,0xfe,0xfd,
	0xbf,0xff,0xdf,0xa7,0xfe,0xfd,0xf7,0xdf,
	0x7f,0xff,0xef,0xf7,0xe3,0xff,0xbf,0xdf,
	0x7d,0xcf,0xfe,0xff,0xf7,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xf3,
	0xff,0xff,0xff,0xff,0xff,0xe7,0xef,0xbf,
	0xee,0xbf,0xef,0x3f,0x7f,0xf7,0xbd,0xff,
	0xfd,0xfa,0xff,0xfd,0x9f,0xff,0xfe,0xdb,
	0xff,0xff,0xff,0xeb,0xfa,0xff,0xff,0xff,
	0x8b,0xff,0xff,0x3b,0xfe,0xff,0xfe,0xff,
	0x3f,0xff,0xf7,0xf3,0xf7,0xf7,0x7f,0xfc,
	0xff,0xeb,0x6f,0xdf,0x73,0xcf,0x3f,0xff,
	0xf8,0x73,0xff,0x3f,0xfe,0xfc,0xff,0xff,
	0xff,0xff,0xfc,0xff,0xff,0xfb,0xfd,0xe7,
	0xff,0xff,0x6c,0xff,0xdf,0xfe,0x7f,0xff,
	0xee,0xff,0xcf,0xbf,0xff,0x8a,0x7f,0xfb,
	0xed,0xdf,0xdf,0xdc,0xff,0xef,0xb7,0xff,
	0xff,0xff,0xff,0xe8,0xff,0xff,0x2f,0xff,
	0xff,0xf7,0xfd,0x7e,0xff,0xf7,0xbf,0xdf,
	0xff,0x5d,0xd2,0x8f,0xf7,0xde,0xbf,0xbf,
	0xcf,0x7f,0xff,0x4f,0xb7,0xdf,0x3f,0xff,
	0xed,0xff,0xff,0xff,0x77,0xfc,0xff,0xbf,
	0xfd,0xff,0x97,0xf6,0xfb,0xe9,0xee,0xfb,
	0x5e,0xfb,0xff,0xe1,0xfd,0xbf,0xff,0xff,
	0x7f,0x7f,0xff,0xff,0xff,0xff,0xff,0xbf,
	0xfd,0xff,0xff,0xff,0xff,0x9f,0xfa,0xff,
	0xef,0xf7,0xfd,0xfb,0xff,0xff,0xef,0xff,
	0xdf,0xff,0xff,0xbf,0xff,0xff,0xff,0xff,
	0xfd,0xff,0xff,0xfa,0xff,0xff,0xf7,0xff,
	0xff,0xf3,0xff,0xff,0xff,0xff,0xff,0x5d,
	0xff,0xfc,0xeb,0xff,0xbf,0xfe,0xfa,0xfe,
	0x2f,0xef,0x7f,0xf6,0xf9,0xff,0xb7,0xfe,
	0xff,0xff,0xff,0xbf,0xff,0xff,0xff,0xff,
	0xff,0xff,0xf3,0xff,0xff,0xff,0xfc,0xff,
	0x43,0xd6,0xd7,0x9f,0xfc,0xff,0xc9,0x67,
	0xff,0x7e,0xfa,0xcf,0xbf,0x5f,0xde,0xd3,
	0xf9,0xff,0xff,0xff,0xdf,0xff,0xf7,0xfd,
	0xff,0xff,0xff,0x5f,0xff,0xff,0xff,0xff,
	0xff,0x6b,0xf3,0xfe,0xff,0xdf,0xff,0xff,
	0xfd,0xf7,0xff,0x7f,0xcf,0xff,0xfd,0xdf,
	0x7f,0xff,0xff,0xff,0xff,0xff,0xfd,0xff,
	0xd4,0xff,0xff,0xff,0xff,0xff,0x7f,0xff,
	0xff,0xff,0xdf,0x87,0xfd,0xff,0xbf,0xff,
	0xff,0xfb,0xff,0xbf,0x9f,0x7e,0xff,0xef,
	0xb7,0xfe,0xfe,0xff,0xff,0xff,0xff,0xff,
	0xff,0xb3,0xff,0xff,0xff,0xff,0xff,0xdf,
	0xfb,0xff,0xff,0xff,0xdb,0xf9,0xff,0xff,
	0xb7,0xfb,0xdf,0x7b,0xcd,0xb6,0xf7,0x76,
	0x3b,0x6f,0x3f,0xf3,0xfe,0xfb,0xbf,0xff,
	0x5e,0x33,0xed,0x37,0xfb,0xfe,0xfb,0xef,
	0xbf,0xf5,0xfe,0xff,0xef,0xff,0x6a,0xff,
	0xfd,0xef,0xca,0xff,0xae,0xb9,0xa5,0x9e,
	0xfb,0xae,0xb8,0xef,0x9e,0x57,0xff,0xff,
	0xff,0xff,0x6e,0xa9,0xff,0xd6,0xff,0xff,
	0xff,0xff,0xbf,0xfb,0xff,0xff,0xff,0x76,
	0xfe,0xf7,0xff,0xfd,0xff,0xf7,0x4b,0xff,
	0x75,0xf7,0xd7,0xde,0xff,0x6d,0xfd,0xff,
	0xeb,0xff,0x7f,0xd6,0x5b,0x7d,0xe5,0xdf,
	0x7e,0xff,0xef,0xff,0xd5,0xfa,0xff,0xfd,
	0x9f,0xfa,0xff,0xff,0xbb,0xff,0xbf,0xcb,
	0xee,0xeb,0xec,0xbe,0xfa,0xee,0xab,0xef,
	0xfd,0x5b,0xbf,0xff,0xb6,0x9b,0xfa,0xfb,
	0xff,0xe7,0xcf,0xef,0xbe,0xaf,0xf7,0xff,
	0x7f,0xfd,0x89,0x7f,0xfe,0xff,0xfe,0xff,
	0xef,0xbf,0xfd,0xfe,0xdb,0xef,0xbf,0xff,
	0xfe,0xdf,0x6f,0xff,0xb6,0x7b,0x6f,0xbc,
	0xff,0xfe,0x7b,0xef,0xbf,0xff,0xfe,0x5b,
	0xfe,0xbf,0xff,0xcf,0xfe,0xf7,0xfb,0xe7,
	0xbf,0x7f,0xfe,0x59,0xe7,0x9f,0x75,0xf6,
	0xf9,0x67,0xff,0x75,0xd6,0xe7,0x9f,0x7f,
	0xfe,0xf9,0xe7,0x9f,0x7d,0xf6,0xf9,0xe7,
	0x9f,0xf9,0xfe,0xf9,0x7f,0xe4,0x7f,0xf7,
	0x9f,0x7f,0xbf,0xf9,0x66,0x1f,0x6b,0x8e,
	0xd9,0xe7,0x9a,0x6d,0xdf,0x79,0x9f,0x7f,
	0xae,0xf9,0xc7,0x1a,0x7f,0xf4,0x71,0xe7,
	0x9f,0x7f,0xfc,0xfd,0xc7,0xff,0xb9,0xfd,
	0xff,0xff,0xff,0xff,0xfc,0xf3,0xdf,0x7f,
	0xff,0xff,0xff,0xff,0x3f,0xff,0xff,0xff,
	0xff,0xff,0xfd,0xff,0xdf,0xff,0xff,0xff,
	0xff,0xff,0x7f,0xff,0xff,0xff,0xff,0xdb,
	0xfc,0xff,0xf7,0xdf,0x7f,0xf7,0xdd,0x77,
	0x7f,0x7f,0xff,0xfd,0xf7,0xdf,0xfd,0xff,
	0xfd,0xdf,0x7f,0xf7,0xfd,0xdf,0xff,0xff,
	0xff,0xff,0xf7,0xdf,0xfe,0xff,0xfd,0xff,
	0xff,0xe2,0xff,0xff,0x3f,0xfe,0xff,0xe3,
	0x8f,0x3f,0xbe,0x78,0xf3,0xcf,0x25,0xff,
	0xf6,0xe2,0x27,0xff,0xdd,0x62,0x8d,0x3d,
	0xb6,0xd8,0xe2,0xdf,0x37,0xb6,0xf8,0x7f,
	0x9f,0xff,0x35,0xff,0xff,0xfd,0xf1,0xdf,
	0x37,0xdf,0x7c,0xf3,0xc5,0x37,0x6f,0x7c,
	0xb1,0xd5,0x17,0x7b,0xb1,0xdf,0x17,0x7f,
	0x7c,0xf3,0xc5,0x17,0xff,0xed,0xf1,0xc5,
	0x7f,0xff,0xfd,0xbf,0xfb,0xff,0x7f,0xef,
	0xfe,0xf7,0xce,0xfb,0xcf,0xbd,0xb7,0xde,
	0xfb,0xff,0xff,0xf7,0x6f,0xff,0xfe,0xfb,
	0xfd,0xf3,0xcf,0x77,0xdf,0xdd,0xff,0xef,
	0x7f,0xf7,0xff,0xaf,0xff,0xdd,0xff,0xff,
	0x63,0x86,0x7f,0x6c,0xb0,0xc5,0x07,0x1b,
	0xf4,0xd9,0xe3,0x1f,0x7f,0xee,0xe9,0x87,
	0x7f,0x7e,0xd8,0xc7,0x8f,0x7f,0xfe,0xf9,
	0x65,0x9e,0x1f,0xfe,0xf9,0xe7,0x8f,0xfe,
	0xff,0xff,0x33,0xdf,0x3f,0xb2,0xec,0x23,
	0xdb,0x3c,0xff,0xde,0xfb,0xef,0x2f,0xfe,
	0xe2,0x9f,0x3f,0xfa,0xfc,0x73,0x8b,0x2f,
	0xfe,0xf9,0xf3,0x8f,0x7c,0xff,0xff,0x7f,
	0xf0,0xff,0xff,0x9b,0xef,0xff,0xfd,0xe4,
	0xdb,0x6f,0x3f,0xfb,0xf7,0xf3,0x5f,0x7f,
	0xfd,0xd3,0xef,0x7f,0xff,0xf6,0xdf,0x4f,
	0x7f,0xfd,0xff,0x9b,0xdf,0xbf,0xff,0xff,
	0xff,0xbb,0xff,0xff,0xe7,0xbe,0xf7,0xef,
	0xfb,0xef,0xbe,0xf7,0xfe,0xbf,0xef,0xfe,
	0xff,0xff,0xff,0x8f,0xff,0xef,0x39,0xff,
	0xbc,0xff,0xff,0xff,0xeb,0xfe,0xfb,0xde,
	0xff,0xff,0x1f,0xfd,0xff,0x0f,0x36,0xf9,
	0x7e,0x93,0x0f,0x36,0xf9,0xe2,0xfb,0x4d,
	0xbc,0xf7,0xfe,0xfb,0x3f,0xf8,0xde,0x83,
	0xed,0x37,0xf9,0xfe,0xfb,0x0f,0xb6,0xf7,
	0xe2,0xfb,0xef,0xff,0xe8,0xff,0xff,0xff,
	0xff,0xff,0xf3,0xd3,0x4f,0xaf,0xff,0xff,
	0x5f,0x4f,0xff,0xff,0xff,0xff,0xff,0xff,
	0xf4,0xff,0x4f,0xff,0xff,0xff,0xbf,0xff,
	0x3f,0xf5,0xff,0xff,0xff,0x37,0xff,0xeb,
	0xf5,0x57,0x5f,0x7b,0xdd,0xd5,0xd7,0x5f,
	0x7d,0xfd,0xc5,0x43,0x0f,0x3d,0xf5,0x97,
	0x1f,0x7d,0xf5,0xc5,0xd7,0x4e,0x3b,0xe9,
	0xd5,0xd7,0x4f,0x7d,0xfd,0xf5,0x27,0xf8,
	0xff,0xef,0xff,0x7f,0xf6,0xbf,0xff,0xf7,
	0xd6,0xfa,0xf8,0xe2,0x87,0xfd,0xff,0xfb,
	0x8d,0xf7,0x77,0xff,0xa3,0xe7,0xff,0xfd,
	0xef,0x7f,0xff,0xdf,0xfe,0xfc,0xef,0xfe,
	0x83,0xff,0xfd,0xf7,0xff,0xbe,0xff,0xfe,
	0xff,0xff,0xff,0xfd,0xf5,0xcf,0xbb,0xff,
	0xde,0xdb,0xef,0xff,0x6f,0xd7,0xed,0xdd,
	0xff,0x7f,0xff,0xfd,0xb7,0xff,0x75,0xfb,
	0xfe,0x37,0xfe,0xff,0xff,0xff,0xff,0xff,
	0xfd,0xff,0xdf,0x7f,0xdf,0x7d,0xff,0xfd,
	0x7f,0xf7,0x3b,0xfd,0xff,0xff,0xd7,0x7f,
	0xfb,0x7f,0xff,0xff,0xff,0xdf,0xff,0xff,
	0xff,0xff,0x7f,0xe6,0xff,0xff,0xff,0xf9,
	0x7f,0xf4,0xb1,0x87,0x7f,0xef,0x7d,0xbf,
	0xc7,0xff,0xff,0x7f,0xff,0xff,0x7f,0xfc,
	0xff,0x86,0xef,0xde,0xfb,0xff,0xff,0x17,
	0xff,0xbf,0xff,0xff,0x93,0xfc,0xff,0xfb,
	0x7f,0xff,0xff,0xff,0xef,0xff,0xbf,0xfd,
	0xff,0xff,0xaf,0xff,0xbf,0xcf,0xbf,0xff,
	0xfc,0xfb,0xef,0xff,0xff,0xff,0xff,0xff,
	0xef,0xff,0xff,0xff,0xff,0x9f,0xe9,0xff,
	0xff,0x7f,0xff,0x1f,0x76,0xff,0xf5,0xd6,
	0x5f,0xfb,0xdd,0xfd,0xcf,0x35,0x77,0xff,
	0xff,0xdf,0xff,0xdf,0xfd,0xcf,0xff,0xff,
	0xff,0xdf,0xf7,0x37,0xfe,0xed,0xff,0xee,
	0xff,0xff,0xff,0xdf,0xe7,0xfe,0xdf,0x0f,
	0x37,0xfc,0xfe,0xf3,0xff,0xbc,0xfe,0xf2,
	0xf3,0xff,0xfd,0xdf,0xdb,0x7e,0xbf,0xbd,
	0xf7,0xdf,0xf7,0xff,0xfd,0xfa,0xf7,0xff,
	0x06,0xff,0xff,0xef,0xef,0xdf,0xdf,0xff,
	0xff,0xff,0xff,0xb6,0x7e,0xfb,0xff,0xff,
	0xff,0xd7,0xf7,0xff,0xfd,0xdf,0xff,0xef,
	0xfd,0xff,0xff,0xff,0xbf,0xfe,0xff,0xff,
	0xff,0xbf,0xfa,0xff,0xbc,0xff,0xcf,0xff,
	0xe7,0xdf,0x7f,0x46,0x1f,0x7f,0xff,0xf3,
	0xfb,0x3b,0xff,0xfe,0xcb,0x1f,0xff,0xf7,
	0xf3,0xff,0x3f,0xbf,0xff,0xf0,0xce,0x3f,
	0xff,0xfc,0xff,0xc9,0xff,0xe7,0xff,0x7f,
	0x7e,0xbf,0xff,0xe4,0x7b,0xfb,0xfb,0xff,
	0xbf,0xff,0xff,0xfb,0xff,0xdd,0xfe,0xed,
	0xff,0xd1,0xfe,0x7f,0xfb,0xfb,0xbf,0x7f,
	0xfb,0xfa,0xef,0xff,0xdf,0xf6,0xfb,0x7f,
	0xbf,0xcf,0x3e,0xdf,0xfd,0xbf,0x7d,0x9e,
	0xd6,0xdd,0xb9,0xec,0x8c,0xff,0xdf,0xe7,
	0x37,0xaa,0xfe,0x33,0xff,0x9c,0xff,0xed,
	0xb9,0xdf,0x9f,0x3f,0xfe,0x7f,0xf3,0xff,
	0xff,0xbf,0x2f,0x7f,0xfc,0xff,0x3f,0xff,
	0xff,0x7d,0xff,0xce,0xbf,0xff,0xf9,0xff,
	0x7d,0xfa,0xfd,0x7f,0xdf,0x7e,0xff,0xfa,
	0xef,0xef,0xff,0xdf,0xbe,0xff,0xff,0xad,
	0xff,0xff,0xff,0xff,0xfd,0xff,0xff,0xff,
	0xff,0xff,0xeb,0xff,0xff,0xfe,0xff,0xf7,
	0xfd,0xff,0xfc,0xf3,0xff,0x3f,0xfb,0xdb,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0x7f,0xf0,0xff,0xff,0xff,0xff,0xea,0xaf,
	0xff,0xff,0xfc,0xfe,0xfd,0xff,0xdf,0x7f,
	0xff,0xe5,0xff,0x7f,0xd3,0xff,0x37,0xff,
	0x7f,0xff,0xfd,0xff,0xdf,0xff,0xff,0xfd,
	0xf7,0x3f,0x69,0xfd,0xff,0xf5,0xff,0xcf,
	0x7c,0xfe,0xcf,0xd7,0xf7,0xf4,0x73,0x4f,
	0xff,0xfb,0x37,0xff,0x7f,0xfd,0xfe,0xbf,
	0xff,0xbf,0xfb,0xf6,0xff,0x6f,0xff,0xff,
	0xf7,0xdf,0xbf,0x7e,0xef,0xff,0xff,0xbf,
	0xf4,0xfa,0xff,0xff,0xfd,0x75,0xfe,0x7f,
	0x7f,0x3f,0xf7,0xda,0xff,0xbf,0xf4,0xf7,
	0x4a,0xff,0xff,0xff,0xff,0xff,0xaf,0xfd,
	0x7f,0xf6,0xcb,0xff,0x3d,0xd8,0xff,0xff,
	0xff,0xa9,0xff,0xff,0x3f,0xfb,0xaf,0xff,
	0x9f,0x3e,0xfb,0xbc,0xff,0xfe,0xff,0xed,
	0xbf,0xff,0xfa,0xff,0xef,0xbf,0xff,0x7f,
	0xeb,0xff,0xb3,0xff,0xfe,0xbf,0xcb,0xff,
	0xff,0xfe,0xff,0xef,0x37,0xfb,0xfe,0x7b,
	0xef,0xf7,0xff,0x7e,0x7f,0xff,0xfd,0xff,
	0xdb,0xdd,0xbc,0xf7,0xfe,0xfb,0xfd,0x37,
	0xfb,0xf6,0xfb,0xef,0xb7,0xdf,0xff,0x6f,
	0xf6,0xff,0xff,0xff,0xbf,0xef,0xd6,0xff,
	0x6f,0xb9,0xef,0xdb,0x7b,0xfe,0xf9,0xef,
	0xff,0xff,0xff,0xe5,0xbe,0xf7,0xff,0xeb,
	0xa6,0xdd,0xff,0x9e,0xfd,0xbf,0xfd,0xfe,
	0xef,0xe1,0xff,0xdf,0xfe,0x5f,0x7f,0xed,
	0xbf,0xd7,0xca,0x7f,0xfd,0x75,0xf4,0x57,
	0x7f,0xfd,0xff,0x5f,0x7f,0xf7,0xfd,0x7f,
	0x7f,0xff,0x7d,0xff,0xdf,0xff,0x7d,0xfd,
	0xf5,0xff,0xa9,0xff,0xff,0xc7,0xff,0xea,
	0xab,0xff,0xbe,0xca,0xee,0x6b,0xaf,0xb6,
	0xf7,0xda,0x6b,0xdf,0xfb,0xde,0xfb,0xef,
	0xfd,0xff,0xda,0x6b,0xff,0xb2,0xf6,0xff,
	0x6b,0xaf,0xfd,0xd3,0xfc,0xff,0xfb,0xff,
	0xbf,0xff,0xb6,0xfb,0xef,0xbf,0xff,0xb7,
	0x5b,0xfe,0xff,0xff,0xb7,0xef,0xff,0xed,
	0xfe,0xdf,0xef,0xff,0xff,0xf7,0xfb,0xff,
	0xbf,0xff,0xb7,0xdf,0xff,0xe1,0xff,0x9f,
	0xff,0xfe,0xf9,0x67,0x9e,0x7f,0xf6,0xd9,
	0x6f,0x9e,0xff,0xfe,0x9b,0x6f,0x7e,0xfe,
	0xfb,0xe7,0xff,0x77,0xfe,0xfb,0x7f,0x9d,
	0xff,0xe7,0xf9,0xff,0xbf,0xf5,0x0f,0xfe,
	0x7f,0xfc,0xfd,0xe7,0x9f,0x7f,0xbe,0x99,
	0xe6,0xdd,0x6f,0xfe,0xfd,0xf7,0xdb,0xbf,
	0xf9,0xf7,0x1f,0x7f,0xff,0xb1,0xf6,0xdb,
	0x77,0xfe,0xfd,0xc7,0xdf,0x7f,0x8f,0x1f,
	0xd8,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfc,0xf7,0xff,0x3f,0xff,0xff,0xff,0xfb,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xf7,
	0xfb,0xff,0xff,0xfd,0xff,0xff,0x7f,0xff,
	0xbd,0xdd,0xff,0xff,0xff,0xfd,0xf7,0xdf,
	0x7f,0xf7,0xdd,0xf7,0xff,0x7d,0xff,0xff,
	0xff,0xff,0xff,0xfd,0xff,0xff,0xff,0xff,
	0xdf,0xff,0xff,0x7f,0xf7,0xff,0xff,0xff,
	0xfd,0xfb,0xff,0xfa,0xff,0xf9,0xff,0x8f,
	0x25,0xff,0xf9,0xf2,0x8b,0xbf,0xbd,0xfc,
	0xdb,0x6f,0xaf,0x7d,0xf7,0x6b,0x37,0xbe,
	0xf6,0xe3,0x6b,0xaf,0xfd,0xfd,0xdb,0x9f,
	0xbf,0xbd,0x76,0x5f,0xf4,0xff,0xdf,0x7e,
	0x7b,0x7c,0xb1,0xdf,0x17,0x5b,0xfc,0xd5,
	0xc5,0x77,0x7d,0xf5,0xd5,0x7f,0x5f,0xe5,
	0xf1,0xd5,0x1e,0x7f,0xf5,0xf5,0xdf,0x5f,
	0xfb,0xfd,0xf5,0x5d,0xff,0xab,0xff,0xff,
	0xfa,0xdf,0xf3,0xfb,0xfd,0xd7,0x7f,0xff,
	0xef,0xfd,0xff,0xfe,0xfa,0xef,0xff,0xef,
	0xff,0xcf,0xbe,0xff,0xfc,0xfa,0xef,0xfd,
	0xfb,0xfe,0xb7,0xef,0xbf,0xff,0x1f,0xfc,
	0xff,0x7f,0xfe,0xd9,0xc1,0x9f,0x7b,0xfe,
	0xb9,0x67,0x87,0x7f,0x7c,0xd0,0xe1,0x87,
	0xfe,0xf9,0x61,0x9f,0x1d,0xf6,0xd9,0xe1,
	0x87,0x7f,0x7c,0xf8,0x67,0x87,0x1f,0xfe,
	0xea,0xff,0xdf,0xff,0xff,0xcc,0xfb,0xdf,
	0x3f,0xfe,0xfc,0xfa,0x8f,0xbf,0xff,0xfe,
	0xfa,0xff,0xff,0xfe,0xfb,0xfb,0xbf,0xff,
	0xfe,0xfa,0x9f,0xbf,0xff,0xfd,0xfa,0xeb,
	0xff,0x5f,0xff,0xff,0xfe,0xff,0xf7,0xdb,
	0xff,0x7f,0xff,0xed,0xd7,0xcd,0xff,0xdd,
	0x76,0xdb,0xad,0xff,0x77,0xdb,0xdf,0xb7,
	0xfd,0x7f,0xdb,0xed,0xff,0xdd,0xfe,0xd7,
	0xcd,0xb7,0x3f,0xfb,0xff,0xff,0xff,0x3f,
	0xef,0xff,0xf7,0xef,0xbf,0xff,0x9f,0xfb,
	0xef,0xf9,0xe7,0x9e,0xff,0xbf,0xe7,0xfe,
	0x7f,0xee,0xbf,0xe7,0x9e,0xff,0xef,0x79,
	0xff,0x9f,0x7b,0xfe,0xcb,0xff,0xff,0xfe,
	0xfb,0x4d,0xbe,0xff,0xde,0x7b,0xef,0x3f,
	0xf1,0x7e,0x93,0x4f,0x3e,0xf9,0xfb,0x4d,
	0xb6,0xff,0x64,0xfb,0x4f,0x3e,0xf9,0xde,
	0x93,0xef,0x3f,0xf9,0xe4,0xaf,0xfe,0xff,
	0xfb,0xef,0xff,0xff,0xff,0xbf,0xff,0xdb,
	0xff,0xef,0xff,0xf6,0xd3,0x4f,0xff,0xef,
	0xd3,0x6f,0xef,0xfd,0xdf,0xd3,0x4f,0xff,
	0xff,0xf4,0xbf,0x4f,0x3f,0xfd,0xbf,0xd2,
	0xff,0x5f,0x7d,0xe5,0xf5,0x17,0x5f,0x7c,
	0xf5,0xd5,0xd7,0x5f,0x6b,0xf5,0xc5,0x97,
	0x7e,0xf9,0x75,0xd7,0x1f,0x7d,0xfd,0xd1,
	0xd7,0x5e,0x7b,0xfd,0xb5,0xd7,0x5f,0xff,
	0x93,0xff,0xff,0xff,0xff,0xf3,0x0d,0x3f,
	0xfe,0xdd,0xac,0xff,0x77,0xbf,0xff,0xf1,
	0xf7,0xdf,0xf8,0x7d,0xf9,0x26,0xff,0xbb,
	0xf1,0xf7,0x3f,0x7e,0xff,0xf3,0xf7,0xdf,
	0xff,0xdf,0xf8,0xff,0xff,0xff,0xf7,0xdf,
	0xbf,0xff,0xfc,0xfb,0xef,0x7f,0x7f,0xff,
	0xff,0xef,0xf7,0xfd,0xff,0xfd,0xf7,0xff,
	0xfe,0xff,0xff,0x7f,0xdf,0xff,0xf7,0xff,
	0xff,0xff,0x7f,0xee,0xff,0xff,0xff,0xff,
	0xfe,0xfd,0xeb,0xff,0xff,0xff,0xdf,0x9d,
	0xff,0x1f,0x7f,0xff,0xd1,0xdf,0xff,0xff,
	0xf1,0xff,0x3f,0x7f,0xff,0xf7,0xf7,0xff,
	0xfc,0xff,0xff,0xff,0x07,0xff,0xff,0xff,
	0xff,0xff,0xdf,0xf6,0x7e,0xfd,0xb5,0x84,
	0x5d,0x6f,0xfc,0xf5,0xff,0xbf,0xff,0xf1,
	0xff,0x9e,0xdf,0xef,0xfd,0xff,0xff,0xff,
	0x7f,0xff,0xc7,0xff,0xef,0x9f,0xf8,0xff,
	0xff,0xff,0xff,0xdf,0xff,0xff,0xee,0xbe,
	0xdb,0xef,0xfb,0xfe,0x7e,0xff,0xff,0xfe,
	0xfb,0xdf,0xff,0xff,0xbf,0x7b,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xc9,
	0x7f,0xdf,0xff,0xff,0xfd,0x7d,0xcf,0xdf,
	0x7f,0xff,0xf5,0xf7,0x1f,0x7f,0xdc,0xf3,
	0xf7,0x7f,0xfc,0xf7,0xff,0xff,0xfe,0xe8,
	0xff,0x7f,0x3c,0xfa,0xfc,0xfd,0xdf,0xff,
	0xb7,0xfe,0xbf,0xff,0xdf,0xf7,0x2f,0xff,
	0xff,0xf2,0xfb,0xdf,0xbb,0xd9,0xfe,0xdb,
	0xaf,0x1f,0xf7,0xc8,0xcf,0xff,0xdd,0xfa,
	0xea,0x7f,0xff,0x7f,0xfa,0xfb,0xff,0xff,
	0xef,0x7f,0xf4,0xfb,0xed,0x7f,0xfa,0xff,
	0xd7,0x7f,0xff,0x77,0xf7,0xdf,0xfd,0xdf,
	0xdf,0xff,0xaf,0xfd,0xfe,0xff,0xfd,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0x75,0xff,
	0xff,0xef,0xff,0xbf,0xff,0xff,0xeb,0x7f,
	0xfc,0x7d,0xf7,0xdf,0x67,0x7d,0xfd,0xf7,
	0x3f,0xff,0xfc,0xf3,0xef,0xff,0xfc,0xf3,
	0xff,0xff,0xff,0xff,0xff,0x8b,0x3f,0xff,
	0xff,0xf3,0xc7,0xff,0xdf,0xfd,0xff,0xff,
	0xff,0xef,0xfe,0xeb,0xff,0xff,0xbf,0xfe,
	0xff,0xf7,0xfb,0xef,0xbb,0xff,0xff,0xef,
	0xd7,0xfe,0xff,0xdf,0xff,0xfb,0xdf,0xbe,
	0xfb,0xff,0xb7,0xff,0xfe,0xff,0x65,0xff,
	0x7f,0x27,0x5e,0xdf,0xfb,0xfd,0xb7,0xda,
	0xfe,0x7f,0xfd,0x9e,0x7f,0xfe,0x7d,0xff,
	0x77,0xff,0xf7,0xff,0xff,0xff,0xff,0xfb,
	0xcf,0xdf,0xff,0xff,0xf9,0xe7,0xdf,0x6b,
	0xfe,0xff,0xff,0xfd,0xe7,0xdd,0x7f,0xfb,
	0xfd,0xfb,0xfd,0xf6,0xff,0xfd,0x73,0xff,
	0xbf,0xbf,0xff,0xbf,0xff,0xf7,0xff,0xff,
	0xef,0x3f,0xf7,0xff,0xff,0xdf,0x3f,0x7f,
	0x2f,0xfb,0xff,0xef,0xef,0xdf,0xbf,0xde,
	0xfd,0xab,0xff,0xf7,0xff,0xef,0xf3,0xbf,
	0xff,0xff,0xbf,0xff,0xf7,0xff,0xbf,0xff,
	0xff,0xff,0xff,0xfb,0xff,0xfd,0x3e,0xff,
	0xfd,0xf7,0x0f,0xff,0xfc,0xfe,0xfb,0xb7,
	0xfc,0xfa,0xea,0xcb,0xff,0xdd,0xfa,0xff,
	0xfd,0xf7,0xdf,0xfc,0xfb,0x97,0xff,0x7f,
	0xeb,0xfb,0xf7,0xdf,0xf7,0xff,0xfd,0xff,
	0xdf,0x7f,0xff,0x33,0xd6,0xa7,0xff,0xfc,
	0xd3,0xe9,0xe7,0x97,0x7c,0xfd,0xef,0xe6,
	0xff,0xf7,0xdf,0x7f,0xd7,0xff,0xd7,0xfc,
	0x7f,0x9b,0xff,0xbb,0xdf,0xbe,0xff,0x77,
	0xfd,0x6f,0x7f,0xfd,0xeb,0xf6,0xfe,0xff,
	0x9f,0x7f,0x7f,0x3f,0xf5,0xdc,0x5f,0x7f,
	0xff,0xff,0xf2,0xdf,0x67,0xff,0xff,0x7f,
	0xff,0xff,0xf5,0xff,0xff,0x7f,0xfd,0xfd,
	0xdc,0xff,0xff,0xfd,0xf7,0xdf,0x97,0xfd,
	0xef,0xb7,0xff,0xfe,0xfb,0xac,0xa7,0x9f,
	0xfa,0xfb,0xff,0xb7,0xdf,0x7e,0xfb,0xff,
	0xff,0xff,0xff,0xe9,0xff,0xff,0x7e,0xfb,
	0xef,0xf7,0xfe,0x7f,0xeb,0xff,0xff,0x5b,
	0x7c,0x7f,0x6f,0x9f,0xf3,0x6a,0x7b,0xef,
	0xbf,0xda,0x6a,0x7f,0x6d,0xf7,0xdd,0x77,
	0xff,0xbf,0xd7,0x4f,0xfb,0xcd,0xb4,0xdf,
	0x65,0x3b,0x6f,0xbd,0xd7,0xf6,0x33,0xbd,
	0xfe,0x65,0xff,0xab,0xf5,0x8a,0x57,0xee,
	0xb9,0xff,0xbe,0x6b,0xfe,0xab,0xbf,0xff,
	0xfe,0xfb,0xff,0xbf,0xbf,0xea,0x7f,0xa7,
	0xbe,0x5e,0x9f,0xb8,0xef,0xbd,0xfa,0xde,
	0xfb,0xf6,0x16,0xfe,0xdf,0x6f,0xf5,0xfd,
	0xdf,0xd6,0x7f,0xbd,0xb7,0xdf,0x5f,0x6f,
	0xfd,0xf5,0xd7,0xff,0xff,0xf5,0x57,0x5e,
	0x5f,0xfd,0xf7,0xd7,0xdf,0x7f,0xfd,0xf5,
	0xdf,0x5f,0xff,0x9f,0xf9,0xff,0xea,0xea,
	0xec,0xbd,0xfb,0xee,0xab,0xeb,0xbe,0xf7,
	0x2a,0x6a,0xaf,0xbd,0xf6,0xbd,0xec,0xbd,
	0xfe,0xde,0xab,0xaf,0xbd,0xf7,0xee,0x6b,
	0xaf,0xbe,0xf6,0xde,0xff,0xd7,0xff,0xbf,
	0xfd,0xb3,0xdf,0xee,0xbf,0xff,0xfe,0xfb,
	0x7f,0xbb,0xff,0xff,0xff,0xff,0xff,0xb6,
	0xdf,0xee,0xff,0xf9,0xfe,0xdf,0xff,0xbf,
	0xff,0xff,0xfb,0xff,0xff,0xff,0xbf,0xfe,
	0xff,0x79,0x6f,0xff,0x7d,0xfe,0xf9,0xe7,
	0x9f,0xff,0xe6,0xf9,0xef,0xbf,0xfd,0xfe,
	0xe7,0xbf,0x7d,0xfe,0xff,0xe7,0xbf,0xf7,
	0xff,0xf9,0x7f,0x9f,0xff,0xff,0xfb,0xff,
	0xe5,0xbf,0xe6,0xdb,0x77,0xb7,0xb9,0xe6,
	0x9f,0x7f,0xbe,0xfd,0xe6,0xda,0x7f,0xa7,
	0xfd,0x9b,0x7f,0xf7,0xf1,0xf7,0x9f,0x7f,
	0xbf,0xfd,0xe7,0xda,0x77,0xae,0xfd,0xf6,
	0xff,0xad,0xfd,0xf7,0xcf,0xf7,0xff,0xfc,
	0xf7,0xdf,0xff,0xbf,0xfd,0xfe,0xdf,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xbf,0xfd,0xff,0xff,0xff,0xff,0xff,
	0xf6,0xff,0x3b,0xfc,0xf7,0x77,0xdf,0xff,
	0xf7,0xdd,0xd7,0xdf,0x7f,0xf7,0xff,0xb7,
	0xff,0xff,0xff,0xff,0xdf,0xff,0xff,0xff,
	0xff,0xdf,0xff,0xfd,0xff,0xf7,0xff,0x7f,
	0xff,0xf7,0xff,0xff,0xe8,0x7f,0x8f,0xad,
	0xbf,0xf6,0xe3,0x8f,0x2f,0xfe,0xf8,0xdb,
	0x8d,0xbf,0xfd,0xf6,0xdb,0x2f,0xbe,0xd6,
	0xe2,0x6f,0x7f,0xfe,0xf6,0xdb,0xcf,0xaf,
	0xfd,0xfd,0xdb,0x6f,0xff,0x0d,0xff,0xdf,
	0xec,0xf1,0xd5,0x17,0xdf,0x7c,0xb3,0xc7,
	0x77,0xd9,0xfc,0xd5,0x57,0x77,0x79,0xf1,
	0x55,0x17,0xdf,0xfd,0xf7,0x5d,0x77,0x7f,
	0x7c,0xf5,0xdf,0x5f,0x5f,0xf5,0x3f,0xf2,
	0xff,0x7a,0xef,0xfe,0xf7,0xde,0x7b,0xeb,
	0xbf,0xff,0xde,0xbb,0xef,0xbf,0xff,0xfe,
	0xff,0xff,0xbf,0xfc,0x7b,0xff,0xbf,0xff,
	0xde,0xbb,0xff,0xff,0xff,0xfe,0xfb,0xff,
	0xcb,0xff,0x97,0xe1,0x87,0x1f,0x64,0xb0,
	0xc3,0x86,0x19,0x7c,0xb0,0x63,0x87,0x1d,
	0x7c,0xf8,0x86,0x1f,0x74,0xf0,0xe1,0x06,
	0x1f,0x7c,0xd8,0xe1,0x87,0x5f,0x76,0xd0,
	0xe1,0x8f,0xfe,0xff,0x48,0x12,0xed,0x7f,
	0xd3,0x5c,0x23,0xcd,0xa4,0xff,0x78,0xfb,
	0xeb,0xaf,0xff,0x62,0xed,0xff,0x93,0xfe,
	0xb7,0xe9,0xbf,0xff,0x4c,0xfb,0xff,0xb6,
	0xff,0xff,0x7f,0xf5,0xff,0xfd,0xd9,0x67,
	0x37,0x7f,0xe6,0xf9,0xc7,0x1e,0xdd,0xfc,
	0xf1,0x4d,0x37,0xdd,0xd2,0xc7,0xb7,0x7f,
	0x74,0xfb,0x47,0xb7,0xdd,0xe6,0xf1,0xed,
	0x1f,0xdf,0x7c,0xfb,0xbf,0xff,0xff,0xef,
	0xbc,0x7b,0xde,0xbb,0xef,0xbe,0xfb,0xfe,
	0xb9,0xe7,0x9e,0x7f,0xfe,0xf9,0xaf,0x7b,
	0xfe,0xfb,0xe7,0x9d,0x7f,0xee,0xb9,0xef,
	0x9e,0xff,0xee,0xf9,0xe7,0xdf,0xfd,0xff,
	0xcf,0x7f,0xdf,0xc4,0xf3,0x8d,0x3f,0xf7,
	0xfa,0x13,0x8f,0x3d,0xf9,0xe4,0x93,0x3f,
	0xf6,0xe4,0xeb,0x4f,0xbe,0xfe,0x64,0x93,
	0xcd,0x3d,0xf9,0xda,0x13,0x4f,0xfe,0xe8,
	0xff,0xd3,0xff,0xff,0xff,0xf6,0xdb,0xff,
	0xff,0xff,0xf6,0xff,0x5f,0x7f,0xfd,0xf5,
	0xff,0x7f,0xfd,0xff,0xd7,0xff,0x7f,0xfd,
	0xf5,0xfb,0x5f,0xff,0xff,0xf5,0xd7,0xff,
	0x27,0xff,0xd7,0xf5,0x57,0x5f,0x7f,0xdc,
	0xe5,0x97,0x1f,0x6d,0xf5,0xd5,0x57,0x1f,
	0x7d,0xf1,0x57,0x4f,0x7d,0xfd,0xd0,0xd7,
	0x5e,0x7d,0xbd,0xf1,0xd7,0x1f,0x3f,0xfd,
	0xf5,0x3f,0xf8,0xff,0xef,0xdf,0xff,0x3f,
	0xfc,0xfe,0xbd,0xf7,0xb6,0xfc,0x6f,0xff,
	0xff,0xff,0xf7,0xbd,0xfd,0xff,0x7f,0xf3,
	0xcf,0x3f,0xff,0xfc,0xff,0xff,0x3f,0xfe,
	0xff,0xff,0xff,0x95,0xff,0xff,0xbf,0xff,
	0xff,0xf3,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfd,0xf7,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xdf,0x7f,0xff,0xef,0x3f,
	0xff,0xff,0xff,0xff,0xf7,0xfe,0xef,0xff,
	0xef,0xff,0xff,0xff,0xff,0xff,0xfe,0xcf,
	0xff,0xff,0xff,0xff,0xff,0xbf,0xff,0xfd,
	0x77,0x3f,0x7f,0xf2,0xf3,0xcf,0xff,0xff,
	0xff,0xf7,0xfe,0xff,0xff,0xff,0xb0,0xff,
	0xf7,0xfb,0x1f,0x7f,0xfd,0xf1,0xdf,0x1f,
	0x5f,0x7c,0xff,0xc7,0x1f,0x7f,0xfc,0xfb,
	0x1f,0x7f,0x7f,0xf1,0xff,0x1d,0x7f,0xfc,
	0xff,0xc7,0xff,0x7f,0xfc,0xf7,0xff,0x8b,
	0xff,0xff,0xff,0xff,0xbe,0xeb,0xdf,0xff,
	0xef,0xfd,0xfb,0xff,0xab,0xff,0xfe,0xfb,
	0xef,0xeb,0xbe,0xef,0xef,0xff,0xff,0xfe,
	0xfb,0xff,0xbb,0xef,0xff,0xfb,0xef,0xff,
	0x5f,0xfc,0x7f,0xff,0xff,0xcf,0xff,0xff,
	0xff,0xfd,0xbf,0xff,0xff,0xff,0xff,0x9f,
	0xff,0x5e,0xff,0xff,0xff,0xff,0xe3,0xf3,
	0xff,0xfd,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xe9,0xff,0xff,0xff,0xaf,0xfd,
	0xff,0xdf,0xff,0xff,0xff,0xef,0xff,0xff,
	0xbf,0xff,0xff,0xe3,0xff,0xff,0xff,0xfe,
	0xef,0xff,0xfd,0xbf,0xff,0xff,0xff,0xff,
	0xf7,0xfb,0xff,0x2f,0xff,0xff,0xff,0xff,
	0x7d,0xff,0xff,0xb7,0xff,0xff,0xff,0xff,
	0xff,0xff,0xed,0xff,0x1c,0xff,0xff,0xff,
	0xff,0x77,0xff,0xff,0xff,0xff,0xdf,0xff,
	0xff,0xff,0xdf,0xff,0x7f,0x98,0xff,0xbf,
	0xff,0xff,0xef,0x7f,0xf7,0xfe,0xff,0xff,
	0xff,0xff,0xff,0xff,0xef,0xdf,0xfe,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0x7f,0xff,0xff,0xff,0xff,0xff,0xc7,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x3f,
	0xfe,0xff,0xff,0xff,0xff,0xff,0xf2,0xff,
	0xbf,0xff,0xfe,0xfb,0xff,0xff,0xff,0xbf,
	0xff,0xbf,0xff,0xfc,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xfd,0xff,
	0xfd,0xe2,0xff,0xff,0xff,0x7f,0xff,0xef,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xfb,0x7e,0xfb,0xeb,0xff,0xdf,
	0xff,0xff,0xff,0xef,0xff,0xff,0xff,0xff,
	0xff,0xf7,0x81,0xff,0xff,0xff,0xdb,0xfb,
	0xff,0xff,0xfe,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xfb,0xff,0xff,0xfd,0xbf,0xfd,
	0x7d,0xdf,0xff,0xff,0xff,0xfe,0xf7,0xef,
	0xff,0xff,0x7f,0x3f,0xdd,0xff,0xf1,0xe7,
	0x9f,0x7f,0x7e,0xf8,0xe7,0x9f,0x7f,0xfe,
	0xf9,0xe7,0x9f,0x7f,0xfe,0xa7,0x9f,0x7f,
	0xfe,0xf9,0xe7,0x9f,0x7f,0xfe,0xf9,0xe7,
	0x9f,0x7f,0xfe,0xf9,0x7f,0xed,0xff,0xdf,
	0x7f,0xff,0xfd,0xf7,0xd5,0x7f,0xff,0xfd,
	0xf7,0xcf,0x7f,0xff,0x7d,0xd7,0x3f,0xff,
	0xfd,0xf7,0xdf,0x7f,0xff,0xfd,0xf7,0xdf,
	0x7f,0xff,0xfd,0xf7,0xdf,0xff,0x27,0xff,
	0x7f,0xff,0xfd,0xf7,0xdf,0x7f,0xff,0xfd,
	0xb7,0xdd,0x7e,0xff,0xfd,0xf7,0xdf,0xff,
	0xfd,0xf7,0xdf,0x7f,0xfb,0xfd,0xf7,0xdf,
	0x7f,0xdf,0xfd,0xf7,0xdf,0x7f,0xff,0xbf,
	0xb8,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xef,0xff,0xf7,0x7f,0xff,0xff,
	0xff,0xdf,0xff,0xff,0xf7,0xff,0x7f,0xff,
	0xff,0xff,0xff,0xfe,0xff,0xfe,0xff,0xff,
	0xff,0xc9,0xff,0xdf,0x79,0xff,0xfd,0xf7,
	0xdf,0x7d,0xff,0xfd,0xf7,0xdf,0x7f,0xff,
	0xfd,0xb7,0x6f,0xff,0xfd,0xf7,0xdf,0x7f,
	0xff,0xfd,0xf7,0xdf,0x7f,0xff,0xdd,0xf7,
	0xdf,0xff,0xbd,0xfe,0x3f,0xff,0xfc,0xf3,
	0xcf,0x3f,0xff,0xdc,0xf1,0xcb,0x3f,0x7f,
	0xfc,0xf3,0xcf,0xff,0xfc,0xf3,0xcf,0x1f,
	0xff,0xfc,0xf3,0xcf,0x3f,0xbf,0xfc,0xf3,
	0xcf,0x3f,0xff,0xff,0x77,0xff,0xfd,0xf6,
	0xdf,0x7f,0xff,0xfd,0xf6,0xdf,0x7f,0xff,
	0x7d,0xf7,0xdf,0x7f,0xf7,0xf3,0xdf,0x7f,
	0xff,0xfd,0xf7,0xdf,0x7f,0x7f,0xfd,0xf7,
	0xdf,0x6f,0xf7,0xfd,0xdf,0xbb,0xff,0x9f,
	0x7e,0xfa,0xe9,0xa7,0x9f,0x7e,0xfa,0xe9,
	0xa7,0x9f,0x7e,0xfa,0xe9,0xa7,0x7f,0xfa,
	0xe9,0xa7,0x9f,0x7e,0xfa,0xe9,0xa7,0x9f,
	0x7e,0xfa,0xe9,0xa7,0x9f,0xfe,0xdf,0xfc,
	0xff,0xfe,0xfb,0xef,0xbf,0xff,0xfe,0xfb,
	0xef,0xbf,0xff,0xee,0xfb,0xef,0xbe,0xff,
	0xfb,0xee,0xbf,0xff,0xfe,0xfb,0xef,0xbf,
	0xff,0xfe,0xfb,0xef,0xbf,0xff,0xfe,0xff,
	0xe2,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xdf,0xff,0xff,0xff,0xff,0xbf,0xff,0xff,
	0xdf,0xff,0xfc,0xff,0xff,0xff,0xff,0xff,
	0xde,0x1f,0xff,0xbf,0xef,0xfe,0xfb,0xef,
	0xbf,0xef,0xfe,0xbb,0xef,0xbf,0xff,0xfe,
	0xfb,0xef,0xdf,0xfe,0xfb,0xef,0xbf,0xff,
	0xfe,0xfb,0xef,0xbe,0xff,0xfe,0xfb,0xee,
	0xbf,0xff,0x3f,0xfa,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xed,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xf7,0xdf,0xff,0xbf,0xff,
	0xfe,0xfb,0xef,0xbf,0xff,0xfe,0xfb,0xef,
	0xbf,0xfb,0xfe,0xfb,0xef,0xef,0xfe,0xfb,
	0xef,0xbf,0xff,0xfe,0x7b,0xef,0xbf,0xff,
	0xfa,0xfb,0xef,0xbf,0xff,0x2f,0xde,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xfb,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0x7f,0xf2,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xdf,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfd,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xfb,0xef,0xff,0xff,0xff,0xff,
	0xbf,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xfe,0xff,0xff,0xff,0xff,
	0xff,0xcf,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xfb,0x5d,0xfd,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xee,0xff,0xff,0xff,0xfb,
	0xff,0xbf,0xff,0xfe,0xfb,0xef,0xbf,0xff,
	0xfe,0xfb,0xff,0xbf,0xff,0xff,0xef,0xbf,
	0xff,0xfe,0xff,0xef,0xbf,0xff,0xfe,0xfb,
	0xef,0xbf,0xff,0xff,0x0f,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xdf,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0x7f,0xff,0xf8,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xfe,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xfb,0xc5,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0x9f,0xfe,0xff,0xdf,0x7f,0xff,0xfd,0xf7,
	0xdf,0x7f,0xff,0xfd,0xf7,0xdf,0x7f,0xff,
	0xfd,0xf7,0x7f,0xff,0xfd,0xf7,0xdf,0x7f,
	0xff,0xfd,0xf7,0xdf,0x7f,0xff,0xfd,0xf7,
	0x3f,0xff,0xf0,0xfe,0xf7,0xdf,0x3f,0xff,
	0xfd,0xf7,0xdf,0x7f,0xff,0xfd,0xf7,0xdf,
	0x7f,0xff,0xfc,0xdd,0x3f,0xff,0xfd,0xf7,
	0xcf,0x7f,0xff,0xfd,0xf7,0xdf,0x7f,0xff,
	0xfd,0xf7,0xf7,0x83,0xff,0xff,0xfd,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xef,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	0xff,0xff,0xff,0xff,0x9f,0xfc,0x7f,0xff,
	0xdd,0xff,0xd7,0x7f,0xff,0xfd,0xf7,0xdf,
	0x7f,0xff,0xfd,0xf7,0xf7,0xff,0xfd,0xfd,
	0xdf,0x7f,0xff,0xff,0xf5,0xdf,0x7f,0xff,
	0xfd,0xf7,0xdf,0x7f,0x7f,0xff,0xe9,0xff,
	0xfe,0xd7,0xdf,0xde,0xfd,0xf5,0xd7,0x5f,
	0x7f,0xfd,0xf5,0xd7,0x5f,0xdf,0xfb,0xdd,
	0xf7,0x7e,0xfd,0xf5,0xb7,0x77,0x7f,0xfd,
	0xf5,0xd7,0x5f,0x7f,0xfd,0xf5,0xed,0x2d,
	0xff,0xff
};
